/*
* Descent 3 
* Copyright (C) 2024 Parallax Software
*
* This program is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef DIFFICULTY_EXTERNAL_H_
#define DIFFICULTY_EXTERNAL_H_

#define MAX_DIFFICULTY_LEVELS 5 //	Number of difficulty levels.

// skill difficulties
#define DIFFICULTY_TRAINEE 0
#define DIFFICULTY_ROOKIE 1
#define DIFFICULTY_HOTSHOT 2
#define DIFFICULTY_ACE 3
#define DIFFICULTY_INSANE 4

#endif
