/*
* Descent 3 
* Copyright (C) 2024 Parallax Software
*
* This program is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program.  If not, see <http://www.gnu.org/licenses/>.

--- HISTORICAL COMMENTS FOLLOW ---

 * $Logfile: /DescentIII/main/help.h $
 * $Revision: 2 $
 * $Date: 3/02/98 5:14p $
 * $Author: Jeff $
 *
 * Help functions (when user presses F1 in game)
 *
 * $Log: /DescentIII/main/help.h $
 *
 * 2     3/02/98 5:14p Jeff
 * initial help system created
 *
 * $NoKeywords: $
 */

#ifndef __D3HELP_H_
#define __D3HELP_H_

void HelpDisplay(void);

#endif
