LIBRARY
EXPORTS	
	_CreateMenuItemWArgs@20
	CreateMenuItemWArgs@20=_CreateMenuItemWArgs@20
	_CreateDMFC@0
	CreateDMFC@0=_CreateDMFC@0
	_CreateDmfcStats@0
	CreateDmfcStats@0=_CreateDmfcStats@0
	_CreateMenuItem@0
	CreateMenuItem@0=_CreateMenuItem@0
	DLLAddBlinkingHUDMessage
	DLLAddColoredHUDMessage
	DLLAddHUDItem
	DLLAddHUDMessage
	DLLApplyDamageToPlayer
	DLLAttachObject
	DLLAttachObjectRadius
	DLLAttachRandomNapalmEffectsToObject
	DLLButtonCreate
	DLLCheckBoxCreate
	DLLCheckBoxIsChecked
	DLLCheckBoxSetCheck
	DLLComputeRoomCenter
	DLLConvertAxisAmountToEuler
	DLLConvertEulerToAxisAmount
	DLLCreateAndFireWeapon
	DLLCreateNewUIBmpItem
	DLLCreateNewUITextItem
	DLLCreateRandomLineSparks
	DLLCreateRandomParticles
	DLLCreateRandomSparks
	DLLCreateStringTable
	DLLDebugBreak_callback_resume
	DLLDebugBreak_callback_stop
	DLLDebug_ConsolePrintf
	DLLDeleteUIItem
	DLLDescentDefer
	DLLDestroyStringTable
	DLLDoMessageBox
	DLLDoUI
	DLLEditCreate
	DLLEditGetText
	DLLEditSetText
	DLLEndFrame
	DLLFindArg
	DLLFindObjectIDName
	DLLFindShipName
	DLLFindSoundName
	DLLFindTextureName
	DLLFindWeaponName
	DLLFireWeaponFromObject
	DLLGameFrame
	DLLGameRenderWorld
	DLLGetFrameParameters
	DLLGetGameAPI
	DLLGetGoalRoomForTeam
	DLLGetUICallback
	DLLGetUIItemHeight
	DLLGetUIItemPosition
	DLLGetUIItemWidth
	DLLGetUltimateParentForObject
	DLLHideCursor
	DLLHotSpotCreate
	DLLHotSpotSetStates
	DLLIncTeamScore
	DLLInitObjectScripts
	DLLInitPlayerNewShip
	DLLInt3MessageBox
	DLLInvAddTypeID
	DLLInvCheckItem
	DLLInvGetTypeIDCount
	DLLInvRemove
	DLLInvReset
	DLLListAddItem
	DLLListCreate
	DLLListGetItem
	DLLListGetSelectedIndex
	DLLListRemoveAll
	DLLListRemoveItem
	DLLListSelectItem
	DLLListSetSelectedIndex
	DLLMultiClientSendSpecialPacket
	DLLMultiDisconnectPlayer
	DLLMultiEndLevel
	DLLMultiGetMatchChecksum
	DLLMultiMatchGeneric
	DLLMultiMatchWeapon
	DLLMultiPaintGoalRooms
	DLLMultiSendClientExecuteDLL
	DLLMultiSendObject
	DLLMultiSendRequestToObserve
	DLLMultiSendSpecialPacket
	DLLMultiSetLogoState
	DLLNewUIGameWindowClose
	DLLNewUIGameWindowCreate
	DLLNewUIGameWindowDestroy
	DLLNewUIGameWindowOpen
	DLLNewUIWindowClose
	DLLNewUIWindowCreate
	DLLNewUIWindowDestroy
	DLLNewUIWindowLoadBackgroundImage
	DLLNewUIWindowOpen
	DLLNewUIWindowSetFocusOnEditGadget
	DLLObjCreate
	DLLObjGet
	DLLObjSetPos
	DLLObjSetPosNoMark
	DLLOldEditCreate
	DLLOldEditGetText
	DLLOldEditSetText
	DLLOldListAddItem
	DLLOldListCreate
	DLLOldListGetItem
	DLLOldListGetSelectedIndex
	DLLOldListRemoveAll
	DLLOldListRemoveItem
	DLLOldListSelectItem
	DLLOpenCFILE
	DLLPPic_GetBitmapHandle
	DLLPPic_GetPilot
	DLLPlay2dSound
	DLLPlay3dSound
	DLLPlayerChangeShip
	DLLPlayerGetRandomStartPosition
	DLLPlayerSetHUDNameFOV
	DLLPlayerSetLighting
	DLLPlayerSetRotatingBall
	DLLPlayerSpewInventory
	DLLPlayerStopSounds
	DLLPollUI
	DLLRemoveUIBmpItem
	DLLRemoveUITextItem
	DLLRenderHUDGetTextHeight
	DLLRenderHUDGetTextLineWidth
	DLLRenderHUDQuad
	DLLRenderHUDText
	DLLRenderHUDTextFlags
	DLLResetFacings
	DLLResumeControls
	DLLSelectNextCameraView
	DLLSetMaxTeams
	DLLSetObjectDeadFlag
	DLLSetObjectDeadFlagRaw
	DLLSetOldEditBufferLen
	DLLSetUICallback
	DLLSetUITextItemText
	DLLShowCursor
	DLLSliderCreate
	DLLSliderGetPos
	DLLSliderGetRange
	DLLSliderSetPos
	DLLSliderSetRange
	DLLSliderSetSelectChangeCallback
	DLLSliderSetSelectChangeCallbackWData
	DLLSpewClearEvent
	DLLSpewCreate
	DLLStartFrame
	DLLSuspendControls
	DLLTextCreate
	DLLTextSetTitle
	DLLToggleUICallback
	DLLTouchSound
	DLLUIConsoleGadgetCreate
	DLLUIConsoleGadgetputs
	DLLUnattachChild
	DLLUnattachChildren
	DLLUnattachFromParent
	DLLVisEffectAllocate
	DLLVisEffectCreate
	DLLVisEffectCreateControlled
	DLLVisEffectDelete
	DLLVisEffectFree
	DLLVisEffectInitType
	DLLVisEffectLink
	DLLVisEffectRelink
	DLLVisEffectUnlink
	DLLassert
	DLLbm_AllocBitmap
	DLLbm_AllocLoadFileBitmap
	DLLbm_CreateChunkedBitmap
	DLLbm_DestroyChunkedBitmap
	DLLbm_FreeBitmap
	DLLbm_data
	DLLbm_h
	DLLbm_w
	DLLcf_CloseLibrary
	DLLcf_CopyFile
	DLLcf_Diff
	DLLcf_OpenLibrary
	DLLcf_ReadByte
	DLLcf_ReadBytes
	DLLcf_ReadDouble
	DLLcf_ReadFloat
	DLLcf_ReadInt
	DLLcf_ReadShort
	DLLcf_ReadString
	DLLcf_WriteByte
	DLLcf_WriteBytes
	DLLcf_WriteDouble
	DLLcf_WriteFloat
	DLLcf_WriteInt
	DLLcf_WriteShort
	DLLcf_WriteString
	DLLcfclose
	DLLcfeof
	DLLcfexist
	DLLddio_MakePath
	DLLddio_SplitPath
	DLLfvi_FindIntersection
	DLLfvi_QuickDistCellList
	DLLfvi_QuickDistFaceList
	DLLfvi_QuickDistObjectList
	DLLfvi_QuickRoomCheck
	DLLg3_AddDeltaVec
	DLLg3_CalcPointDepth
	DLLg3_CheckAndDrawPoly
	DLLg3_CheckNormalFacing
	DLLg3_ClipPolygon
	DLLg3_CodePoint
	DLLg3_DoneInstance
	DLLg3_DrawBitmap
	DLLg3_DrawBox
	DLLg3_DrawLine
	DLLg3_DrawPlanarRotatedBitmap
	DLLg3_DrawPoly
	DLLg3_DrawRotatedBitmap
	DLLg3_DrawSpecialLine
	DLLg3_DrawSphere
	DLLg3_EndFrame
	DLLg3_FreeTempPoints
	DLLg3_GetMatrixScale
	DLLg3_GetUnscaledMatrix
	DLLg3_GetViewMatrix
	DLLg3_GetViewPosition
	DLLg3_Point2Vec
	DLLg3_ProjectPoint
	DLLg3_RotateDeltaVec
	DLLg3_RotateDeltaX
	DLLg3_RotateDeltaY
	DLLg3_RotateDeltaZ
	DLLg3_RotatePoint
	DLLg3_SetCustomClipPlane
	DLLg3_SetFarClipZ
	DLLg3_SetTriangulationTest
	DLLg3_StartFrame
	DLLg3_StartInstanceAngles
	DLLg3_StartInstanceMatrix
	DLLgrfont_GetHeight
	DLLgrtext_CenteredPrintf
	DLLgrtext_Flush
	DLLgrtext_GetAlpha
	DLLgrtext_GetFont
	DLLgrtext_GetTextLineWidth
	DLLgrtext_Printf
	DLLgrtext_SetAlpha
	DLLgrtext_SetColor
	DLLgrtext_SetFancyColor
	DLLgrtext_SetFont
	DLLnw_GetHostAddressFromNumbers
	DLLnw_GetNumbersFromHostAddress
	DLLnw_GetThisIP
	DLLphys_apply_force
	DLLphys_apply_rot
	DLLrend_ClearScreen
	DLLrend_DrawChunkedBitmap
	DLLrend_DrawCircle
	DLLrend_DrawLFBBitmap
	DLLrend_DrawLine
	DLLrend_DrawPolygon
	DLLrend_DrawScaledBitmap
	DLLrend_DrawScaledChunkedBitmap
	DLLrend_DrawSimpleBitmap
	DLLrend_DrawSpecialLine
	DLLrend_FillCircle
	DLLrend_FillRect
	DLLrend_GetLFBLock
	DLLrend_GetPixel
	DLLrend_ReleaseLFBLock
	DLLrend_SetAlphaType
	DLLrend_SetAlphaValue
	DLLrend_SetColorModel
	DLLrend_SetFiltering
	DLLrend_SetFlatColor
	DLLrend_SetFogState
	DLLrend_SetLighting
	DLLrend_SetMipState
	DLLrend_SetOverlayMap
	DLLrend_SetOverlayType
	DLLrend_SetPixel
	DLLrend_SetTextureType
	DLLrend_SetWrapType
	DLLrend_SetZBias
	DLLrend_SetZBufferState
	DLLrend_SetZBufferWriteMask
	DLLtaunt_AreEnabled
	DLLtaunt_Enable
	DLLvm_AddVectors
	DLLvm_AnglesToMatrix
	DLLvm_AverageVector
	DLLvm_CalcDetValue
	DLLvm_ClearMatrix
	DLLvm_ComputeBoundingSphere
	DLLvm_CrossProduct
	DLLvm_DeltaAngVec
	DLLvm_DeltaAngVecNorm
	DLLvm_DistToPlane
	DLLvm_DivVector
	DLLvm_DotProduct
	DLLvm_ExtractAnglesFromMatrix
	DLLvm_GetCentroid
	DLLvm_GetCentroidFast
	DLLvm_GetMagnitude
	DLLvm_GetMagnitudeFast
	DLLvm_GetNormal
	DLLvm_GetNormalizedDir
	DLLvm_GetNormalizedDirFast
	DLLvm_GetPerp
	DLLvm_GetSlope
	DLLvm_MakeAngleZero
	DLLvm_MakeIdentity
	DLLvm_MakeInverseMatrix
	DLLvm_MakeRandomVector
	DLLvm_MakeVectorZero
	DLLvm_MatrixMul
	DLLvm_MatrixMulTMatrix
	DLLvm_MatrixMulVector
	DLLvm_NormalizeVector
	DLLvm_NormalizeVectorFast
	DLLvm_Orthogonalize
	DLLvm_ScaleAddVector
	DLLvm_ScaleVector
	DLLvm_SinCos
	DLLvm_SinCosToMatrix
	DLLvm_SubVectors
	DLLvm_TransposeMatrix
	DLLvm_VectorAngleToMatrix
	DLLvm_VectorDistance
	DLLvm_VectorDistanceQuick
	DLLvm_VectorMulTMatrix
	DLLvm_VectorToMatrix
	DPrintf
	DatabaseRead1
	DatabaseRead2
	DatabaseRead3
	DatabaseWrite1
	DatabaseWrite2
	FatalError
	GetPlayerRankIndex
	Inven_Add
	Inven_AddCounterMeasure
	Inven_AddObject
	Inven_AtBeginning
	Inven_AtEnd
	Inven_CheckItem
	Inven_FindPos
	Inven_GetAuxPosTypeID
	Inven_GetInventoryItemList
	Inven_GetPos
	Inven_GetPosCount
	Inven_GetPosDescription
	Inven_GetPosIconName
	Inven_GetPosInfo
	Inven_GetPosName
	Inven_GetPosTypeID
	Inven_GetTypeIDCount
	Inven_GotoPos
	Inven_GotoPosTypeID
	Inven_IsSelectable
	Inven_IsUsable
	Inven_NextPos
	Inven_PrevPos
	Inven_Remove
	Inven_Reset
	Inven_ResetPos
	Inven_Size
	Inven_Use
	Inven_UseObjHandle
	Inven_UsePos
	Inven_ValidatePos
	TableFileAdd
	TableFilesClear
	_IDMFC_AddDeathMessage@12
	IDMFC_AddDeathMessage@12=_IDMFC_AddDeathMessage@12
	_IDMFC_AddHUDItemCallback@12
	IDMFC_AddHUDItemCallback@12=_IDMFC_AddHUDItemCallback@12
	_IDMFC_AddInputCommand@16
	IDMFC_AddInputCommand@16=_IDMFC_AddInputCommand@16
	_IDMFC_AddSuicideMessage@8
	IDMFC_AddSuicideMessage@8=_IDMFC_AddSuicideMessage@8
	_IDMFC_AddWeaponHash@16
	IDMFC_AddWeaponHash@16=_IDMFC_AddWeaponHash@16
	_IDMFC_AllowTeamChange@4
	IDMFC_AllowTeamChange@4=_IDMFC_AllowTeamChange@4
	_IDMFC_AnnounceTeamChangeDeny@8
	IDMFC_AnnounceTeamChangeDeny@8=_IDMFC_AnnounceTeamChangeDeny@8
	_IDMFC_AreLogosEnabled@4
	IDMFC_AreLogosEnabled@4=_IDMFC_AreLogosEnabled@4
	_IDMFC_AreTauntsEnabled@4
	IDMFC_AreTauntsEnabled@4=_IDMFC_AreTauntsEnabled@4
	_IDMFC_AutoDeathMessage@8
	IDMFC_AutoDeathMessage@8=_IDMFC_AutoDeathMessage@8
	_IDMFC_AutoTeamSelect@8
	IDMFC_AutoTeamSelect@8=_IDMFC_AutoTeamSelect@8
	_IDMFC_AutoTimeLimit@8
	IDMFC_AutoTimeLimit@8=_IDMFC_AutoTimeLimit@8
	_IDMFC_BanPlayerFromGame@8
	IDMFC_BanPlayerFromGame@8=_IDMFC_BanPlayerFromGame@8
	_IDMFC_CFGClose@4
	IDMFC_CFGClose@4=_IDMFC_CFGClose@4
	_IDMFC_CFGCreateKey@8
	IDMFC_CFGCreateKey@8=_IDMFC_CFGCreateKey@8
	_IDMFC_CFGCreateRecord@16
	IDMFC_CFGCreateRecord@16=_IDMFC_CFGCreateRecord@16
	_IDMFC_CFGFlush@4
	IDMFC_CFGFlush@4=_IDMFC_CFGFlush@4
	_IDMFC_CFGLookupKey@8
	IDMFC_CFGLookupKey@8=_IDMFC_CFGLookupKey@8
	_IDMFC_CFGLookupRecord@12
	IDMFC_CFGLookupRecord@12=_IDMFC_CFGLookupRecord@12
	_IDMFC_CFGOpen@8
	IDMFC_CFGOpen@8=_IDMFC_CFGOpen@8
	_IDMFC_CallClientEvent@24
	IDMFC_CallClientEvent@24=_IDMFC_CallClientEvent@24
	_IDMFC_CallOnAllowObserverChange@8
	IDMFC_CallOnAllowObserverChange@8=_IDMFC_CallOnAllowObserverChange@8
	_IDMFC_CallOnCanChangeTeam@12
	IDMFC_CallOnCanChangeTeam@12=_IDMFC_CallOnCanChangeTeam@12
	_IDMFC_CallOnClientCollideA@12
	IDMFC_CallOnClientCollideA@12=_IDMFC_CallOnClientCollideA@12
	_IDMFC_CallOnClientCollideB@20
	IDMFC_CallOnClientCollideB@20=_IDMFC_CallOnClientCollideB@20
	_IDMFC_CallOnClientGameCreated@4
	IDMFC_CallOnClientGameCreated@4=_IDMFC_CallOnClientGameCreated@4
	_IDMFC_CallOnClientLevelChange@4
	IDMFC_CallOnClientLevelChange@4=_IDMFC_CallOnClientLevelChange@4
	_IDMFC_CallOnClientLevelEnd@4
	IDMFC_CallOnClientLevelEnd@4=_IDMFC_CallOnClientLevelEnd@4
	_IDMFC_CallOnClientLevelStart@4
	IDMFC_CallOnClientLevelStart@4=_IDMFC_CallOnClientLevelStart@4
	_IDMFC_CallOnClientObjectChangeSegment@16
	IDMFC_CallOnClientObjectChangeSegment@16=_IDMFC_CallOnClientObjectChangeSegment@16
	_IDMFC_CallOnClientObjectDestroyed@8
	IDMFC_CallOnClientObjectDestroyed@8=_IDMFC_CallOnClientObjectDestroyed@8
	_IDMFC_CallOnClientObjectKilled@12
	IDMFC_CallOnClientObjectKilled@12=_IDMFC_CallOnClientObjectKilled@12
	_IDMFC_CallOnClientObjectShieldsChanged@12
	IDMFC_CallOnClientObjectShieldsChanged@12=_IDMFC_CallOnClientObjectShieldsChanged@12
	_IDMFC_CallOnClientPlayerChangeSegment@16
	IDMFC_CallOnClientPlayerChangeSegment@16=_IDMFC_CallOnClientPlayerChangeSegment@16
	_IDMFC_CallOnClientPlayerDisconnect@8
	IDMFC_CallOnClientPlayerDisconnect@8=_IDMFC_CallOnClientPlayerDisconnect@8
	_IDMFC_CallOnClientPlayerEntersGame@8
	IDMFC_CallOnClientPlayerEntersGame@8=_IDMFC_CallOnClientPlayerEntersGame@8
	_IDMFC_CallOnClientPlayerExploded@8
	IDMFC_CallOnClientPlayerExploded@8=_IDMFC_CallOnClientPlayerExploded@8
	_IDMFC_CallOnClientPlayerKilled@12
	IDMFC_CallOnClientPlayerKilled@12=_IDMFC_CallOnClientPlayerKilled@12
	_IDMFC_CallOnClientShowUI@12
	IDMFC_CallOnClientShowUI@12=_IDMFC_CallOnClientShowUI@12
	_IDMFC_CallOnClientWallCollide@32
	IDMFC_CallOnClientWallCollide@32=_IDMFC_CallOnClientWallCollide@32
	_IDMFC_CallOnControlMessage@12
	IDMFC_CallOnControlMessage@12=_IDMFC_CallOnControlMessage@12
	_IDMFC_CallOnDisconnectSaveStatsToFile@4
	IDMFC_CallOnDisconnectSaveStatsToFile@4=_IDMFC_CallOnDisconnectSaveStatsToFile@4
	_IDMFC_CallOnDoControls@8
	IDMFC_CallOnDoControls@8=_IDMFC_CallOnDoControls@8
	_IDMFC_CallOnGameStateRequest@8
	IDMFC_CallOnGameStateRequest@8=_IDMFC_CallOnGameStateRequest@8
	_IDMFC_CallOnGetHudCallSignColor@8
	IDMFC_CallOnGetHudCallSignColor@8=_IDMFC_CallOnGetHudCallSignColor@8
	_IDMFC_CallOnHUDInterval@4
	IDMFC_CallOnHUDInterval@4=_IDMFC_CallOnHUDInterval@4
	_IDMFC_CallOnInputString@8
	IDMFC_CallOnInputString@8=_IDMFC_CallOnInputString@8
	_IDMFC_CallOnInterval@4
	IDMFC_CallOnInterval@4=_IDMFC_CallOnInterval@4
	_IDMFC_CallOnKeypress@8
	IDMFC_CallOnKeypress@8=_IDMFC_CallOnKeypress@8
	_IDMFC_CallOnLevelEndSaveStatsToFile@4
	IDMFC_CallOnLevelEndSaveStatsToFile@4=_IDMFC_CallOnLevelEndSaveStatsToFile@4
	_IDMFC_CallOnMeDisconnectFromServer@4
	IDMFC_CallOnMeDisconnectFromServer@4=_IDMFC_CallOnMeDisconnectFromServer@4
	_IDMFC_CallOnPLRInit@4
	IDMFC_CallOnPLRInit@4=_IDMFC_CallOnPLRInit@4
	_IDMFC_CallOnPLRInterval@4
	IDMFC_CallOnPLRInterval@4=_IDMFC_CallOnPLRInterval@4
	_IDMFC_CallOnPlayAudioTaunt@8
	IDMFC_CallOnPlayAudioTaunt@8=_IDMFC_CallOnPlayAudioTaunt@8
	_IDMFC_CallOnPlayerChangeTeam@20
	IDMFC_CallOnPlayerChangeTeam@20=_IDMFC_CallOnPlayerChangeTeam@20
	_IDMFC_CallOnPlayerConnect@8
	IDMFC_CallOnPlayerConnect@8=_IDMFC_CallOnPlayerConnect@8
	_IDMFC_CallOnPlayerEntersObserver@12
	IDMFC_CallOnPlayerEntersObserver@12=_IDMFC_CallOnPlayerEntersObserver@12
	_IDMFC_CallOnPlayerExitsObserver@8
	IDMFC_CallOnPlayerExitsObserver@8=_IDMFC_CallOnPlayerExitsObserver@8
	_IDMFC_CallOnPlayerReconnect@8
	IDMFC_CallOnPlayerReconnect@8=_IDMFC_CallOnPlayerReconnect@8
	_IDMFC_CallOnPrintScores@8
	IDMFC_CallOnPrintScores@8=_IDMFC_CallOnPrintScores@8
	_IDMFC_CallOnSaveStatsToFile@4
	IDMFC_CallOnSaveStatsToFile@4=_IDMFC_CallOnSaveStatsToFile@4
	_IDMFC_CallOnServerCollideA@12
	IDMFC_CallOnServerCollideA@12=_IDMFC_CallOnServerCollideA@12
	_IDMFC_CallOnServerCollideB@20
	IDMFC_CallOnServerCollideB@20=_IDMFC_CallOnServerCollideB@20
	_IDMFC_CallOnServerGameCreated@4
	IDMFC_CallOnServerGameCreated@4=_IDMFC_CallOnServerGameCreated@4
	_IDMFC_CallOnServerIsAddressBanned@12
	IDMFC_CallOnServerIsAddressBanned@12=_IDMFC_CallOnServerIsAddressBanned@12
	_IDMFC_CallOnServerLevelChange@4
	IDMFC_CallOnServerLevelChange@4=_IDMFC_CallOnServerLevelChange@4
	_IDMFC_CallOnServerLevelEnd@4
	IDMFC_CallOnServerLevelEnd@4=_IDMFC_CallOnServerLevelEnd@4
	_IDMFC_CallOnServerLevelStart@4
	IDMFC_CallOnServerLevelStart@4=_IDMFC_CallOnServerLevelStart@4
	_IDMFC_CallOnServerObjectChangeSegment@16
	IDMFC_CallOnServerObjectChangeSegment@16=_IDMFC_CallOnServerObjectChangeSegment@16
	_IDMFC_CallOnServerObjectDestroyed@8
	IDMFC_CallOnServerObjectDestroyed@8=_IDMFC_CallOnServerObjectDestroyed@8
	_IDMFC_CallOnServerObjectKilled@12
	IDMFC_CallOnServerObjectKilled@12=_IDMFC_CallOnServerObjectKilled@12
	_IDMFC_CallOnServerObjectShieldsChanged@12
	IDMFC_CallOnServerObjectShieldsChanged@12=_IDMFC_CallOnServerObjectShieldsChanged@12
	_IDMFC_CallOnServerPlayerChangeSegment@16
	IDMFC_CallOnServerPlayerChangeSegment@16=_IDMFC_CallOnServerPlayerChangeSegment@16
	_IDMFC_CallOnServerPlayerDisconnect@8
	IDMFC_CallOnServerPlayerDisconnect@8=_IDMFC_CallOnServerPlayerDisconnect@8
	_IDMFC_CallOnServerPlayerEntersGame@8
	IDMFC_CallOnServerPlayerEntersGame@8=_IDMFC_CallOnServerPlayerEntersGame@8
	_IDMFC_CallOnServerPlayerExploded@8
	IDMFC_CallOnServerPlayerExploded@8=_IDMFC_CallOnServerPlayerExploded@8
	_IDMFC_CallOnServerPlayerKilled@12
	IDMFC_CallOnServerPlayerKilled@12=_IDMFC_CallOnServerPlayerKilled@12
	_IDMFC_CallOnServerWallCollide@32
	IDMFC_CallOnServerWallCollide@32=_IDMFC_CallOnServerWallCollide@32
	_IDMFC_CallOnSpecialPacket@4
	IDMFC_CallOnSpecialPacket@4=_IDMFC_CallOnSpecialPacket@4
	_IDMFC_CallOnTeamChangeName@16
	IDMFC_CallOnTeamChangeName@16=_IDMFC_CallOnTeamChangeName@16
	_IDMFC_CallOnWeaponFired@12
	IDMFC_CallOnWeaponFired@12=_IDMFC_CallOnWeaponFired@12
	_IDMFC_CheckPInfo@4
	IDMFC_CheckPInfo@4=_IDMFC_CheckPInfo@4
	_IDMFC_CheckPlayerNum@8
	IDMFC_CheckPlayerNum@8=_IDMFC_CheckPlayerNum@8
	_IDMFC_ClipString@16
	IDMFC_ClipString@16=_IDMFC_ClipString@16
	_IDMFC_CompareNetworkAddress@16
	IDMFC_CompareNetworkAddress@16=_IDMFC_CompareNetworkAddress@16
	_IDMFC_ConvertHUDAlphaByte@8
	IDMFC_ConvertHUDAlphaByte@8=_IDMFC_ConvertHUDAlphaByte@8
	_IDMFC_ConvertHUDAlphaFloat@8
	IDMFC_ConvertHUDAlphaFloat@8=_IDMFC_ConvertHUDAlphaFloat@8
	_IDMFC_ConvertHUDCoord@20
	IDMFC_ConvertHUDCoord@20=_IDMFC_ConvertHUDCoord@20
	_IDMFC_ConvertLocalToServerObjnum@8
	IDMFC_ConvertLocalToServerObjnum@8=_IDMFC_ConvertLocalToServerObjnum@8
	_IDMFC_ConvertServerToLocalObjnum@8
	IDMFC_ConvertServerToLocalObjnum@8=_IDMFC_ConvertServerToLocalObjnum@8
	_IDMFC_DatabaseReadA@16
	IDMFC_DatabaseReadA@16=_IDMFC_DatabaseReadA@16
	_IDMFC_DatabaseReadB@16
	IDMFC_DatabaseReadB@16=_IDMFC_DatabaseReadB@16
	_IDMFC_DatabaseReadC@12
	IDMFC_DatabaseReadC@12=_IDMFC_DatabaseReadC@12
	_IDMFC_DatabaseRegister@8
	IDMFC_DatabaseRegister@8=_IDMFC_DatabaseRegister@8
	_IDMFC_DatabaseWriteA@16
	IDMFC_DatabaseWriteA@16=_IDMFC_DatabaseWriteA@16
	_IDMFC_DatabaseWriteB@12
	IDMFC_DatabaseWriteB@12=_IDMFC_DatabaseWriteB@12
	_IDMFC_DecryptData@12
	IDMFC_DecryptData@12=_IDMFC_DecryptData@12
	_IDMFC_Delete@4
	IDMFC_Delete@4=_IDMFC_Delete@4
	_IDMFC_DestroyPointer@4
	IDMFC_DestroyPointer@4=_IDMFC_DestroyPointer@4
	_IDMFC_DisconnectMe@4
	IDMFC_DisconnectMe@4=_IDMFC_DisconnectMe@4
	_IDMFC_DisplayInputCommandHelp@8
	IDMFC_DisplayInputCommandHelp@8=_IDMFC_DisplayInputCommandHelp@8
	_IDMFC_DisplayNetGameInfo@12
	IDMFC_DisplayNetGameInfo@12=_IDMFC_DisplayNetGameInfo@12
	_IDMFC_DisplayOutrageLogo@4
	IDMFC_DisplayOutrageLogo@4=_IDMFC_DisplayOutrageLogo@4
	_IDMFC_DisplayPlayerInfo@12
	IDMFC_DisplayPlayerInfo@12=_IDMFC_DisplayPlayerInfo@12
	_IDMFC_DisplayingPlayerInfo@4
	IDMFC_DisplayingPlayerInfo@4=_IDMFC_DisplayingPlayerInfo@4
	_IDMFC_DoDamageToPlayer@20
	IDMFC_DoDamageToPlayer@20=_IDMFC_DoDamageToPlayer@20
	_IDMFC_DoRandomDeathMessage@16
	IDMFC_DoRandomDeathMessage@16=_IDMFC_DoRandomDeathMessage@16
	_IDMFC_DrawMenu@4
	IDMFC_DrawMenu@4=_IDMFC_DrawMenu@4
	_IDMFC_DuplicatePointer@4
	IDMFC_DuplicatePointer@4=_IDMFC_DuplicatePointer@4
	_IDMFC_Dynamic_Cast@8
	IDMFC_Dynamic_Cast@8=_IDMFC_Dynamic_Cast@8
	_IDMFC_EnableAudioTaunts@8
	IDMFC_EnableAudioTaunts@8=_IDMFC_EnableAudioTaunts@8
	_IDMFC_EnableAutoSaveDisconnect@8
	IDMFC_EnableAutoSaveDisconnect@8=_IDMFC_EnableAutoSaveDisconnect@8
	_IDMFC_EnableAutoSaveLevelEnd@8
	IDMFC_EnableAutoSaveLevelEnd@8=_IDMFC_EnableAutoSaveLevelEnd@8
	_IDMFC_EnableOnScreenMenu@8
	IDMFC_EnableOnScreenMenu@8=_IDMFC_EnableOnScreenMenu@8
	_IDMFC_EnableOnScreenMenuBackground@8
	IDMFC_EnableOnScreenMenuBackground@8=_IDMFC_EnableOnScreenMenuBackground@8
	_IDMFC_EnableShipLogos@8
	IDMFC_EnableShipLogos@8=_IDMFC_EnableShipLogos@8
	_IDMFC_EnableStatisticalMessages@8
	IDMFC_EnableStatisticalMessages@8=_IDMFC_EnableStatisticalMessages@8
	_IDMFC_EncryptData@12
	IDMFC_EncryptData@12=_IDMFC_EncryptData@12
	_IDMFC_EndLevel@4
	IDMFC_EndLevel@4=_IDMFC_EndLevel@4
	_IDMFC_FindPInfoStatClose@4
	IDMFC_FindPInfoStatClose@4=_IDMFC_FindPInfoStatClose@4
	_IDMFC_FindPInfoStatFirst@12
	IDMFC_FindPInfoStatFirst@12=_IDMFC_FindPInfoStatFirst@12
	_IDMFC_FindPInfoStatNext@8
	IDMFC_FindPInfoStatNext@8=_IDMFC_FindPInfoStatNext@8
	_IDMFC_FreeHostsLists@4
	IDMFC_FreeHostsLists@4=_IDMFC_FreeHostsLists@4
	_IDMFC_GameClose@4
	IDMFC_GameClose@4=_IDMFC_GameClose@4
	_IDMFC_GameInit@8
	IDMFC_GameInit@8=_IDMFC_GameInit@8
	_IDMFC_GenerateStatFilename@16
	IDMFC_GenerateStatFilename@16=_IDMFC_GenerateStatFilename@16
	_IDMFC_GetBannedPlayerCallsign@8
	IDMFC_GetBannedPlayerCallsign@8=_IDMFC_GetBannedPlayerCallsign@8
	_IDMFC_GetCameraViewType@8
	IDMFC_GetCameraViewType@8=_IDMFC_GetCameraViewType@8
	_IDMFC_GetChangeTeamPacket@8
	IDMFC_GetChangeTeamPacket@8=_IDMFC_GetChangeTeamPacket@8
	_IDMFC_GetCounterMeasureOwner@8
	IDMFC_GetCounterMeasureOwner@8=_IDMFC_GetCounterMeasureOwner@8
	_IDMFC_GetCurrentMission@4
	IDMFC_GetCurrentMission@4=_IDMFC_GetCurrentMission@4
	_IDMFC_GetDLLInfoCallData@4
	IDMFC_GetDLLInfoCallData@4=_IDMFC_GetDLLInfoCallData@4
	_IDMFC_GetDMFCGameInfo@8
	IDMFC_GetDMFCGameInfo@8=_IDMFC_GetDMFCGameInfo@8
	_IDMFC_GetFrametime@4
	IDMFC_GetFrametime@4=_IDMFC_GetFrametime@4
	_IDMFC_GetGameFontTranslateArray@4
	IDMFC_GetGameFontTranslateArray@4=_IDMFC_GetGameFontTranslateArray@4
	_IDMFC_GetGamePolyModels@4
	IDMFC_GetGamePolyModels@4=_IDMFC_GetGamePolyModels@4
	_IDMFC_GetGameStateRequest@8
	IDMFC_GetGameStateRequest@8=_IDMFC_GetGameStateRequest@8
	_IDMFC_GetGameTextures@4
	IDMFC_GetGameTextures@4=_IDMFC_GetGameTextures@4
	_IDMFC_GetGameVClips@4
	IDMFC_GetGameVClips@4=_IDMFC_GetGameVClips@4
	_IDMFC_GetGameWindowH@4
	IDMFC_GetGameWindowH@4=_IDMFC_GetGameWindowH@4
	_IDMFC_GetGameWindowW@4
	IDMFC_GetGameWindowW@4=_IDMFC_GetGameWindowW@4
	_IDMFC_GetGameWindowX@4
	IDMFC_GetGameWindowX@4=_IDMFC_GetGameWindowX@4
	_IDMFC_GetGameWindowY@4
	IDMFC_GetGameWindowY@4=_IDMFC_GetGameWindowY@4
	_IDMFC_GetGametime@4
	IDMFC_GetGametime@4=_IDMFC_GetGametime@4
	_IDMFC_GetHighestRoomIndex@4
	IDMFC_GetHighestRoomIndex@4=_IDMFC_GetHighestRoomIndex@4
	_IDMFC_GetHudAspectX@4
	IDMFC_GetHudAspectX@4=_IDMFC_GetHudAspectX@4
	_IDMFC_GetHudAspectY@4
	IDMFC_GetHudAspectY@4=_IDMFC_GetHudAspectY@4
	_IDMFC_GetItObjNum@4
	IDMFC_GetItObjNum@4=_IDMFC_GetItObjNum@4
	_IDMFC_GetLastGameControls@4
	IDMFC_GetLastGameControls@4=_IDMFC_GetLastGameControls@4
	_IDMFC_GetLevelInfo@4
	IDMFC_GetLevelInfo@4=_IDMFC_GetLevelInfo@4
	_IDMFC_GetLocalD3Dir@4
	IDMFC_GetLocalD3Dir@4=_IDMFC_GetLocalD3Dir@4
	_IDMFC_GetLocalRole@4
	IDMFC_GetLocalRole@4=_IDMFC_GetLocalRole@4
	_IDMFC_GetMeObjNum@4
	IDMFC_GetMeObjNum@4=_IDMFC_GetMeObjNum@4
	_IDMFC_GetMyTeam@4
	IDMFC_GetMyTeam@4=_IDMFC_GetMyTeam@4
	_IDMFC_GetNetPlayers@4
	IDMFC_GetNetPlayers@4=_IDMFC_GetNetPlayers@4
	_IDMFC_GetNetgameInfo@4
	IDMFC_GetNetgameInfo@4=_IDMFC_GetNetgameInfo@4
	_IDMFC_GetNumBannedPlayers@4
	IDMFC_GetNumBannedPlayers@4=_IDMFC_GetNumBannedPlayers@4
	_IDMFC_GetNumTeams@4
	IDMFC_GetNumTeams@4=_IDMFC_GetNumTeams@4
	_IDMFC_GetObjectInfo@8
	IDMFC_GetObjectInfo@8=_IDMFC_GetObjectInfo@8
	_IDMFC_GetObjects@4
	IDMFC_GetObjects@4=_IDMFC_GetObjects@4
	_IDMFC_GetObserverModeBitmap@4
	IDMFC_GetObserverModeBitmap@4=_IDMFC_GetObserverModeBitmap@4
	_IDMFC_GetOnScreenMenu@4
	IDMFC_GetOnScreenMenu@4=_IDMFC_GetOnScreenMenu@4
	_IDMFC_GetOsirisModuleData@4
	IDMFC_GetOsirisModuleData@4=_IDMFC_GetOsirisModuleData@4
	_IDMFC_GetPilotPicBitmapHandles@4
	IDMFC_GetPilotPicBitmapHandles@4=_IDMFC_GetPilotPicBitmapHandles@4
	_IDMFC_GetPlayerColors@4
	IDMFC_GetPlayerColors@4=_IDMFC_GetPlayerColors@4
	_IDMFC_GetPlayerLogoBmp@12
	IDMFC_GetPlayerLogoBmp@12=_IDMFC_GetPlayerLogoBmp@12
	_IDMFC_GetPlayerNum@4
	IDMFC_GetPlayerNum@4=_IDMFC_GetPlayerNum@4
	_IDMFC_GetPlayerRecord@8
	IDMFC_GetPlayerRecord@8=_IDMFC_GetPlayerRecord@8
	_IDMFC_GetPlayerRecordByPnum@8
	IDMFC_GetPlayerRecordByPnum@8=_IDMFC_GetPlayerRecordByPnum@8
	_IDMFC_GetPlayerRecordData@8
	IDMFC_GetPlayerRecordData@8=_IDMFC_GetPlayerRecordData@8
	_IDMFC_GetPlayerTeam@8
	IDMFC_GetPlayerTeam@8=_IDMFC_GetPlayerTeam@8
	_IDMFC_GetPlayers@4
	IDMFC_GetPlayers@4=_IDMFC_GetPlayers@4
	_IDMFC_GetRenderZoom@4
	IDMFC_GetRenderZoom@4=_IDMFC_GetRenderZoom@4
	_IDMFC_GetRooms@4
	IDMFC_GetRooms@4=_IDMFC_GetRooms@4
	_IDMFC_GetScoreLimit@8
	IDMFC_GetScoreLimit@8=_IDMFC_GetScoreLimit@8
	_IDMFC_GetShieldDeltaArray@4
	IDMFC_GetShieldDeltaArray@4=_IDMFC_GetShieldDeltaArray@4
	_IDMFC_GetShips@4
	IDMFC_GetShips@4=_IDMFC_GetShips@4
	_IDMFC_GetSortedPlayerSlots@12
	IDMFC_GetSortedPlayerSlots@12=_IDMFC_GetSortedPlayerSlots@12
	_IDMFC_GetSortedPlayerSlotsByEfficiency@12
	IDMFC_GetSortedPlayerSlotsByEfficiency@12=_IDMFC_GetSortedPlayerSlotsByEfficiency@12
	_IDMFC_GetTeamAssignmentPacket@8
	IDMFC_GetTeamAssignmentPacket@8=_IDMFC_GetTeamAssignmentPacket@8
	_IDMFC_GetTeamColor@8
	IDMFC_GetTeamColor@8=_IDMFC_GetTeamColor@8
	_IDMFC_GetTeamForNewPlayer@12
	IDMFC_GetTeamForNewPlayer@12=_IDMFC_GetTeamForNewPlayer@12
	_IDMFC_GetTeamFromString@8
	IDMFC_GetTeamFromString@8=_IDMFC_GetTeamFromString@8
	_IDMFC_GetTeamString@8
	IDMFC_GetTeamString@8=_IDMFC_GetTeamString@8
	_IDMFC_GetTerrainSegs@4
	IDMFC_GetTerrainSegs@4=_IDMFC_GetTerrainSegs@4
	_IDMFC_GetTimeInGame@8
	IDMFC_GetTimeInGame@8=_IDMFC_GetTimeInGame@8
	_IDMFC_GetTimeLeft@8
	IDMFC_GetTimeLeft@8=_IDMFC_GetTimeLeft@8
	_IDMFC_GetTimeString@8
	IDMFC_GetTimeString@8=_IDMFC_GetTimeString@8
	_IDMFC_GetViewerObjectPtr@8
	IDMFC_GetViewerObjectPtr@8=_IDMFC_GetViewerObjectPtr@8
	_IDMFC_GetVisEffectArray@8
	IDMFC_GetVisEffectArray@8=_IDMFC_GetVisEffectArray@8
	_IDMFC_GetWeaponDeathMessage@12
	IDMFC_GetWeaponDeathMessage@12=_IDMFC_GetWeaponDeathMessage@12
	_IDMFC_GetWeapons@4
	IDMFC_GetWeapons@4=_IDMFC_GetWeapons@4
	_IDMFC_IAmDedicatedServer@4
	IDMFC_IAmDedicatedServer@4=_IDMFC_IAmDedicatedServer@4
	_IDMFC_IsAddressBanned@12
	IDMFC_IsAddressBanned@12=_IDMFC_IsAddressBanned@12
	_IDMFC_IsDisplayingNetGameInfo@4
	IDMFC_IsDisplayingNetGameInfo@4=_IDMFC_IsDisplayingNetGameInfo@4
	_IDMFC_IsMasterTrackerGame@4
	IDMFC_IsMasterTrackerGame@4=_IDMFC_IsMasterTrackerGame@4
	_IDMFC_IsMenuUp@4
	IDMFC_IsMenuUp@4=_IDMFC_IsMenuUp@4
	_IDMFC_IsPlayerAlive@8
	IDMFC_IsPlayerAlive@8=_IDMFC_IsPlayerAlive@8
	_IDMFC_IsPlayerBanned@8
	IDMFC_IsPlayerBanned@8=_IDMFC_IsPlayerBanned@8
	_IDMFC_IsPlayerDedicatedServer@8
	IDMFC_IsPlayerDedicatedServer@8=_IDMFC_IsPlayerDedicatedServer@8
	_IDMFC_IsPlayerDedicatedServerPnum@8
	IDMFC_IsPlayerDedicatedServerPnum@8=_IDMFC_IsPlayerDedicatedServerPnum@8
	_IDMFC_IsPlayerObserver@8
	IDMFC_IsPlayerObserver@8=_IDMFC_IsPlayerObserver@8
	_IDMFC_KillAllTimers@4
	IDMFC_KillAllTimers@4=_IDMFC_KillAllTimers@4
	_IDMFC_KillTimer@8
	IDMFC_KillTimer@8=_IDMFC_KillTimer@8
	_IDMFC_LoadFunctions@8
	IDMFC_LoadFunctions@8=_IDMFC_LoadFunctions@8
	_IDMFC_OnAllowObserverChange@8
	IDMFC_OnAllowObserverChange@8=_IDMFC_OnAllowObserverChange@8
	_IDMFC_OnCanChangeTeam@12
	IDMFC_OnCanChangeTeam@12=_IDMFC_OnCanChangeTeam@12
	_IDMFC_OnClientGameCreated@4
	IDMFC_OnClientGameCreated@4=_IDMFC_OnClientGameCreated@4
	_IDMFC_OnClientLevelChange@4
	IDMFC_OnClientLevelChange@4=_IDMFC_OnClientLevelChange@4
	_IDMFC_OnClientLevelEnd@4
	IDMFC_OnClientLevelEnd@4=_IDMFC_OnClientLevelEnd@4
	_IDMFC_OnClientLevelStart@4
	IDMFC_OnClientLevelStart@4=_IDMFC_OnClientLevelStart@4
	_IDMFC_OnClientObjectChangeSegment@16
	IDMFC_OnClientObjectChangeSegment@16=_IDMFC_OnClientObjectChangeSegment@16
	_IDMFC_OnClientObjectDestroyed@8
	IDMFC_OnClientObjectDestroyed@8=_IDMFC_OnClientObjectDestroyed@8
	_IDMFC_OnClientObjectKilled@12
	IDMFC_OnClientObjectKilled@12=_IDMFC_OnClientObjectKilled@12
	_IDMFC_OnClientObjectShieldsChanged@12
	IDMFC_OnClientObjectShieldsChanged@12=_IDMFC_OnClientObjectShieldsChanged@12
	_IDMFC_OnClientPlayerChangeSegment@16
	IDMFC_OnClientPlayerChangeSegment@16=_IDMFC_OnClientPlayerChangeSegment@16
	_IDMFC_OnClientPlayerDisconnect@8
	IDMFC_OnClientPlayerDisconnect@8=_IDMFC_OnClientPlayerDisconnect@8
	_IDMFC_OnClientPlayerEntersGame@8
	IDMFC_OnClientPlayerEntersGame@8=_IDMFC_OnClientPlayerEntersGame@8
	_IDMFC_OnClientPlayerExploded@8
	IDMFC_OnClientPlayerExploded@8=_IDMFC_OnClientPlayerExploded@8
	_IDMFC_OnClientPlayerKilled@12
	IDMFC_OnClientPlayerKilled@12=_IDMFC_OnClientPlayerKilled@12
	_IDMFC_OnClientShowUI@12
	IDMFC_OnClientShowUI@12=_IDMFC_OnClientShowUI@12
	_IDMFC_OnClientWallCollide@32
	IDMFC_OnClientWallCollide@32=_IDMFC_OnClientWallCollide@32
	_IDMFC_OnControlMessage@12
	IDMFC_OnControlMessage@12=_IDMFC_OnControlMessage@12
	_IDMFC_OnDisconnectSaveStatsToFile@4
	IDMFC_OnDisconnectSaveStatsToFile@4=_IDMFC_OnDisconnectSaveStatsToFile@4
	_IDMFC_OnDoControls@8
	IDMFC_OnDoControls@8=_IDMFC_OnDoControls@8
	_IDMFC_OnGameStateRequest@8
	IDMFC_OnGameStateRequest@8=_IDMFC_OnGameStateRequest@8
	_IDMFC_OnGetHudCallSignColor@8
	IDMFC_OnGetHudCallSignColor@8=_IDMFC_OnGetHudCallSignColor@8
	_IDMFC_OnHUDInterval@4
	IDMFC_OnHUDInterval@4=_IDMFC_OnHUDInterval@4
	_IDMFC_OnInputString@8
	IDMFC_OnInputString@8=_IDMFC_OnInputString@8
	_IDMFC_OnInterval@4
	IDMFC_OnInterval@4=_IDMFC_OnInterval@4
	_IDMFC_OnKeypress@8
	IDMFC_OnKeypress@8=_IDMFC_OnKeypress@8
	_IDMFC_OnLevelEndSaveStatsToFile@4
	IDMFC_OnLevelEndSaveStatsToFile@4=_IDMFC_OnLevelEndSaveStatsToFile@4
	_IDMFC_OnMeDisconnectFromServer@4
	IDMFC_OnMeDisconnectFromServer@4=_IDMFC_OnMeDisconnectFromServer@4
	_IDMFC_OnPLRInit@4
	IDMFC_OnPLRInit@4=_IDMFC_OnPLRInit@4
	_IDMFC_OnPLRInterval@4
	IDMFC_OnPLRInterval@4=_IDMFC_OnPLRInterval@4
	_IDMFC_OnPlayAudioTaunt@8
	IDMFC_OnPlayAudioTaunt@8=_IDMFC_OnPlayAudioTaunt@8
	_IDMFC_OnPlayerChangeTeam@20
	IDMFC_OnPlayerChangeTeam@20=_IDMFC_OnPlayerChangeTeam@20
	_IDMFC_OnPlayerConnect@8
	IDMFC_OnPlayerConnect@8=_IDMFC_OnPlayerConnect@8
	_IDMFC_OnPlayerEntersObserver@12
	IDMFC_OnPlayerEntersObserver@12=_IDMFC_OnPlayerEntersObserver@12
	_IDMFC_OnPlayerExitsObserver@8
	IDMFC_OnPlayerExitsObserver@8=_IDMFC_OnPlayerExitsObserver@8
	_IDMFC_OnPlayerReconnect@8
	IDMFC_OnPlayerReconnect@8=_IDMFC_OnPlayerReconnect@8
	_IDMFC_OnPrintScores@8
	IDMFC_OnPrintScores@8=_IDMFC_OnPrintScores@8
	_IDMFC_OnSaveStatsToFile@4
	IDMFC_OnSaveStatsToFile@4=_IDMFC_OnSaveStatsToFile@4
	_IDMFC_OnServerGameCreated@4
	IDMFC_OnServerGameCreated@4=_IDMFC_OnServerGameCreated@4
	_IDMFC_OnServerIsAddressBanned@12
	IDMFC_OnServerIsAddressBanned@12=_IDMFC_OnServerIsAddressBanned@12
	_IDMFC_OnServerLevelChange@4
	IDMFC_OnServerLevelChange@4=_IDMFC_OnServerLevelChange@4
	_IDMFC_OnServerLevelEnd@4
	IDMFC_OnServerLevelEnd@4=_IDMFC_OnServerLevelEnd@4
	_IDMFC_OnServerLevelStart@4
	IDMFC_OnServerLevelStart@4=_IDMFC_OnServerLevelStart@4
	_IDMFC_OnServerObjectChangeSegment@16
	IDMFC_OnServerObjectChangeSegment@16=_IDMFC_OnServerObjectChangeSegment@16
	_IDMFC_OnServerObjectDestroyed@8
	IDMFC_OnServerObjectDestroyed@8=_IDMFC_OnServerObjectDestroyed@8
	_IDMFC_OnServerObjectKilled@12
	IDMFC_OnServerObjectKilled@12=_IDMFC_OnServerObjectKilled@12
	_IDMFC_OnServerObjectShieldsChanged@12
	IDMFC_OnServerObjectShieldsChanged@12=_IDMFC_OnServerObjectShieldsChanged@12
	_IDMFC_OnServerPlayerChangeSegment@16
	IDMFC_OnServerPlayerChangeSegment@16=_IDMFC_OnServerPlayerChangeSegment@16
	_IDMFC_OnServerPlayerDisconnect@8
	IDMFC_OnServerPlayerDisconnect@8=_IDMFC_OnServerPlayerDisconnect@8
	_IDMFC_OnServerPlayerEntersGame@8
	IDMFC_OnServerPlayerEntersGame@8=_IDMFC_OnServerPlayerEntersGame@8
	_IDMFC_OnServerPlayerExploded@8
	IDMFC_OnServerPlayerExploded@8=_IDMFC_OnServerPlayerExploded@8
	_IDMFC_OnServerPlayerKilled@12
	IDMFC_OnServerPlayerKilled@12=_IDMFC_OnServerPlayerKilled@12
	_IDMFC_OnServerWallCollide@32
	IDMFC_OnServerWallCollide@32=_IDMFC_OnServerWallCollide@32
	_IDMFC_OnSpecialPacket@4
	IDMFC_OnSpecialPacket@4=_IDMFC_OnSpecialPacket@4
	_IDMFC_OnTeamChangeName@16
	IDMFC_OnTeamChangeName@16=_IDMFC_OnTeamChangeName@16
	_IDMFC_OnWeaponFired@12
	IDMFC_OnWeaponFired@12=_IDMFC_OnWeaponFired@12
	_IDMFC_PacketCheckPlayerNum@8
	IDMFC_PacketCheckPlayerNum@8=_IDMFC_PacketCheckPlayerNum@8
	_IDMFC_ReadInHostsAllowDeny@4
	IDMFC_ReadInHostsAllowDeny@4=_IDMFC_ReadInHostsAllowDeny@4
	_IDMFC_ReceiveControlMessage@8
	IDMFC_ReceiveControlMessage@8=_IDMFC_ReceiveControlMessage@8
	_IDMFC_ReceiveRequestForPlayerRecords@8
	IDMFC_ReceiveRequestForPlayerRecords@8=_IDMFC_ReceiveRequestForPlayerRecords@8
	_IDMFC_RegisterPacketReceiver@12
	IDMFC_RegisterPacketReceiver@12=_IDMFC_RegisterPacketReceiver@12
	_IDMFC_RehashAllowDeny@4
	IDMFC_RehashAllowDeny@4=_IDMFC_RehashAllowDeny@4
	_IDMFC_RemoveAllBans@4
	IDMFC_RemoveAllBans@4=_IDMFC_RemoveAllBans@4
	_IDMFC_RemoveBan@8
	IDMFC_RemoveBan@8=_IDMFC_RemoveBan@8
	_IDMFC_RequestGameState@4
	IDMFC_RequestGameState@4=_IDMFC_RequestGameState@4
	_IDMFC_RequestTeamChange@16
	IDMFC_RequestTeamChange@16=_IDMFC_RequestTeamChange@16
	_IDMFC_ResetPInfo@4
	IDMFC_ResetPInfo@4=_IDMFC_ResetPInfo@4
	_IDMFC_RespawnPlayer@16
	IDMFC_RespawnPlayer@16=_IDMFC_RespawnPlayer@16
	_IDMFC_SelectNextCameraView@8
	IDMFC_SelectNextCameraView@8=_IDMFC_SelectNextCameraView@8
	_IDMFC_SendChangeTeamRequest@12
	IDMFC_SendChangeTeamRequest@12=_IDMFC_SendChangeTeamRequest@12
	_IDMFC_SendControlMessageToPlayer@12
	IDMFC_SendControlMessageToPlayer@12=_IDMFC_SendControlMessageToPlayer@12
	_IDMFC_SendDMFCGameInfo@8
	IDMFC_SendDMFCGameInfo@8=_IDMFC_SendDMFCGameInfo@8
	_IDMFC_SendNetGameInfoSync@8
	IDMFC_SendNetGameInfoSync@8=_IDMFC_SendNetGameInfoSync@8
	_IDMFC_SendPacket@16
	IDMFC_SendPacket@16=_IDMFC_SendPacket@16
	_IDMFC_SendRequestForPlayerRecords@4
	IDMFC_SendRequestForPlayerRecords@4=_IDMFC_SendRequestForPlayerRecords@4
	_IDMFC_SendTeamAssignment@16
	IDMFC_SendTeamAssignment@16=_IDMFC_SendTeamAssignment@16
	_IDMFC_SetDeathMessageFilter@8
	IDMFC_SetDeathMessageFilter@8=_IDMFC_SetDeathMessageFilter@8
	_IDMFC_SetMaxPlayerHardLimit@8
	IDMFC_SetMaxPlayerHardLimit@8=_IDMFC_SetMaxPlayerHardLimit@8
	_IDMFC_SetNumberOfTeams@8
	IDMFC_SetNumberOfTeams@8=_IDMFC_SetNumberOfTeams@8
	_IDMFC_SetTeamName@16
	IDMFC_SetTeamName@16=_IDMFC_SetTeamName@16
	_IDMFC_SetTimerInterval@20
	IDMFC_SetTimerInterval@20=_IDMFC_SetTimerInterval@20
	_IDMFC_SetViewerObjectPtr@8
	IDMFC_SetViewerObjectPtr@8=_IDMFC_SetViewerObjectPtr@8
	_IDMFC_SetWeaponDeathMessage@16
	IDMFC_SetWeaponDeathMessage@16=_IDMFC_SetWeaponDeathMessage@16
	_IDMFC_Set_OnAllowObserverChange@8
	IDMFC_Set_OnAllowObserverChange@8=_IDMFC_Set_OnAllowObserverChange@8
	_IDMFC_Set_OnCanChangeTeam@8
	IDMFC_Set_OnCanChangeTeam@8=_IDMFC_Set_OnCanChangeTeam@8
	_IDMFC_Set_OnClientCollideA@8
	IDMFC_Set_OnClientCollideA@8=_IDMFC_Set_OnClientCollideA@8
	_IDMFC_Set_OnClientCollideB@8
	IDMFC_Set_OnClientCollideB@8=_IDMFC_Set_OnClientCollideB@8
	_IDMFC_Set_OnClientGameCreated@8
	IDMFC_Set_OnClientGameCreated@8=_IDMFC_Set_OnClientGameCreated@8
	_IDMFC_Set_OnClientLevelChange@8
	IDMFC_Set_OnClientLevelChange@8=_IDMFC_Set_OnClientLevelChange@8
	_IDMFC_Set_OnClientLevelEnd@8
	IDMFC_Set_OnClientLevelEnd@8=_IDMFC_Set_OnClientLevelEnd@8
	_IDMFC_Set_OnClientLevelStart@8
	IDMFC_Set_OnClientLevelStart@8=_IDMFC_Set_OnClientLevelStart@8
	_IDMFC_Set_OnClientObjectChangeSegment@8
	IDMFC_Set_OnClientObjectChangeSegment@8=_IDMFC_Set_OnClientObjectChangeSegment@8
	_IDMFC_Set_OnClientObjectDestroyed@8
	IDMFC_Set_OnClientObjectDestroyed@8=_IDMFC_Set_OnClientObjectDestroyed@8
	_IDMFC_Set_OnClientObjectKilled@8
	IDMFC_Set_OnClientObjectKilled@8=_IDMFC_Set_OnClientObjectKilled@8
	_IDMFC_Set_OnClientObjectShieldsChanged@8
	IDMFC_Set_OnClientObjectShieldsChanged@8=_IDMFC_Set_OnClientObjectShieldsChanged@8
	_IDMFC_Set_OnClientPlayerChangeSegment@8
	IDMFC_Set_OnClientPlayerChangeSegment@8=_IDMFC_Set_OnClientPlayerChangeSegment@8
	_IDMFC_Set_OnClientPlayerDisconnect@8
	IDMFC_Set_OnClientPlayerDisconnect@8=_IDMFC_Set_OnClientPlayerDisconnect@8
	_IDMFC_Set_OnClientPlayerEntersGame@8
	IDMFC_Set_OnClientPlayerEntersGame@8=_IDMFC_Set_OnClientPlayerEntersGame@8
	_IDMFC_Set_OnClientPlayerExploded@8
	IDMFC_Set_OnClientPlayerExploded@8=_IDMFC_Set_OnClientPlayerExploded@8
	_IDMFC_Set_OnClientPlayerKilled@8
	IDMFC_Set_OnClientPlayerKilled@8=_IDMFC_Set_OnClientPlayerKilled@8
	_IDMFC_Set_OnClientShowUI@8
	IDMFC_Set_OnClientShowUI@8=_IDMFC_Set_OnClientShowUI@8
	_IDMFC_Set_OnClientWallCollide@8
	IDMFC_Set_OnClientWallCollide@8=_IDMFC_Set_OnClientWallCollide@8
	_IDMFC_Set_OnControlMessage@8
	IDMFC_Set_OnControlMessage@8=_IDMFC_Set_OnControlMessage@8
	_IDMFC_Set_OnDisconnectSaveStatsToFile@8
	IDMFC_Set_OnDisconnectSaveStatsToFile@8=_IDMFC_Set_OnDisconnectSaveStatsToFile@8
	_IDMFC_Set_OnDoControls@8
	IDMFC_Set_OnDoControls@8=_IDMFC_Set_OnDoControls@8
	_IDMFC_Set_OnGameStateRequest@8
	IDMFC_Set_OnGameStateRequest@8=_IDMFC_Set_OnGameStateRequest@8
	_IDMFC_Set_OnGetHudCallSignColor@8
	IDMFC_Set_OnGetHudCallSignColor@8=_IDMFC_Set_OnGetHudCallSignColor@8
	_IDMFC_Set_OnHUDInterval@8
	IDMFC_Set_OnHUDInterval@8=_IDMFC_Set_OnHUDInterval@8
	_IDMFC_Set_OnInputString@8
	IDMFC_Set_OnInputString@8=_IDMFC_Set_OnInputString@8
	_IDMFC_Set_OnInterval@8
	IDMFC_Set_OnInterval@8=_IDMFC_Set_OnInterval@8
	_IDMFC_Set_OnKeypress@8
	IDMFC_Set_OnKeypress@8=_IDMFC_Set_OnKeypress@8
	_IDMFC_Set_OnLevelEndSaveStatsToFile@8
	IDMFC_Set_OnLevelEndSaveStatsToFile@8=_IDMFC_Set_OnLevelEndSaveStatsToFile@8
	_IDMFC_Set_OnMeDisconnectFromServer@8
	IDMFC_Set_OnMeDisconnectFromServer@8=_IDMFC_Set_OnMeDisconnectFromServer@8
	_IDMFC_Set_OnPLRInit@8
	IDMFC_Set_OnPLRInit@8=_IDMFC_Set_OnPLRInit@8
	_IDMFC_Set_OnPLRInterval@8
	IDMFC_Set_OnPLRInterval@8=_IDMFC_Set_OnPLRInterval@8
	_IDMFC_Set_OnPlayAudioTaunt@8
	IDMFC_Set_OnPlayAudioTaunt@8=_IDMFC_Set_OnPlayAudioTaunt@8
	_IDMFC_Set_OnPlayerChangeTeam@8
	IDMFC_Set_OnPlayerChangeTeam@8=_IDMFC_Set_OnPlayerChangeTeam@8
	_IDMFC_Set_OnPlayerConnect@8
	IDMFC_Set_OnPlayerConnect@8=_IDMFC_Set_OnPlayerConnect@8
	_IDMFC_Set_OnPlayerEntersObserver@8
	IDMFC_Set_OnPlayerEntersObserver@8=_IDMFC_Set_OnPlayerEntersObserver@8
	_IDMFC_Set_OnPlayerExitsObserver@8
	IDMFC_Set_OnPlayerExitsObserver@8=_IDMFC_Set_OnPlayerExitsObserver@8
	_IDMFC_Set_OnPlayerReconnect@8
	IDMFC_Set_OnPlayerReconnect@8=_IDMFC_Set_OnPlayerReconnect@8
	_IDMFC_Set_OnPrintScores@8
	IDMFC_Set_OnPrintScores@8=_IDMFC_Set_OnPrintScores@8
	_IDMFC_Set_OnSaveStatsToFile@8
	IDMFC_Set_OnSaveStatsToFile@8=_IDMFC_Set_OnSaveStatsToFile@8
	_IDMFC_Set_OnServerCollideA@8
	IDMFC_Set_OnServerCollideA@8=_IDMFC_Set_OnServerCollideA@8
	_IDMFC_Set_OnServerCollideB@8
	IDMFC_Set_OnServerCollideB@8=_IDMFC_Set_OnServerCollideB@8
	_IDMFC_Set_OnServerGameCreated@8
	IDMFC_Set_OnServerGameCreated@8=_IDMFC_Set_OnServerGameCreated@8
	_IDMFC_Set_OnServerIsAddressBanned@8
	IDMFC_Set_OnServerIsAddressBanned@8=_IDMFC_Set_OnServerIsAddressBanned@8
	_IDMFC_Set_OnServerLevelChange@8
	IDMFC_Set_OnServerLevelChange@8=_IDMFC_Set_OnServerLevelChange@8
	_IDMFC_Set_OnServerLevelEnd@8
	IDMFC_Set_OnServerLevelEnd@8=_IDMFC_Set_OnServerLevelEnd@8
	_IDMFC_Set_OnServerLevelStart@8
	IDMFC_Set_OnServerLevelStart@8=_IDMFC_Set_OnServerLevelStart@8
	_IDMFC_Set_OnServerObjectChangeSegment@8
	IDMFC_Set_OnServerObjectChangeSegment@8=_IDMFC_Set_OnServerObjectChangeSegment@8
	_IDMFC_Set_OnServerObjectDestroyed@8
	IDMFC_Set_OnServerObjectDestroyed@8=_IDMFC_Set_OnServerObjectDestroyed@8
	_IDMFC_Set_OnServerObjectKilled@8
	IDMFC_Set_OnServerObjectKilled@8=_IDMFC_Set_OnServerObjectKilled@8
	_IDMFC_Set_OnServerObjectShieldsChanged@8
	IDMFC_Set_OnServerObjectShieldsChanged@8=_IDMFC_Set_OnServerObjectShieldsChanged@8
	_IDMFC_Set_OnServerPlayerChangeSegment@8
	IDMFC_Set_OnServerPlayerChangeSegment@8=_IDMFC_Set_OnServerPlayerChangeSegment@8
	_IDMFC_Set_OnServerPlayerDisconnect@8
	IDMFC_Set_OnServerPlayerDisconnect@8=_IDMFC_Set_OnServerPlayerDisconnect@8
	_IDMFC_Set_OnServerPlayerEntersGame@8
	IDMFC_Set_OnServerPlayerEntersGame@8=_IDMFC_Set_OnServerPlayerEntersGame@8
	_IDMFC_Set_OnServerPlayerExploded@8
	IDMFC_Set_OnServerPlayerExploded@8=_IDMFC_Set_OnServerPlayerExploded@8
	_IDMFC_Set_OnServerPlayerKilled@8
	IDMFC_Set_OnServerPlayerKilled@8=_IDMFC_Set_OnServerPlayerKilled@8
	_IDMFC_Set_OnServerWallCollide@8
	IDMFC_Set_OnServerWallCollide@8=_IDMFC_Set_OnServerWallCollide@8
	_IDMFC_Set_OnSpecialPacket@8
	IDMFC_Set_OnSpecialPacket@8=_IDMFC_Set_OnSpecialPacket@8
	_IDMFC_Set_OnTeamChangeName@8
	IDMFC_Set_OnTeamChangeName@8=_IDMFC_Set_OnTeamChangeName@8
	_IDMFC_Set_OnWeaponFired@8
	IDMFC_Set_OnWeaponFired@8=_IDMFC_Set_OnWeaponFired@8
	_IDMFC_SetupPlayerRecord@16
	IDMFC_SetupPlayerRecord@16=_IDMFC_SetupPlayerRecord@16
	_IDMFC_ShouldIDisplayHUDName@8
	IDMFC_ShouldIDisplayHUDName@8=_IDMFC_ShouldIDisplayHUDName@8
	_IDMFC_StartPacket@16
	IDMFC_StartPacket@16=_IDMFC_StartPacket@16
	_IDMFC_StartUIWindow@12
	IDMFC_StartUIWindow@12=_IDMFC_StartUIWindow@12
	_IDMFC_SwitchAllowTeamChange@8
	IDMFC_SwitchAllowTeamChange@8=_IDMFC_SwitchAllowTeamChange@8
	_IDMFC_SwitchNetGameInfoDisplay@8
	IDMFC_SwitchNetGameInfoDisplay@8=_IDMFC_SwitchNetGameInfoDisplay@8
	_IDMFC_SwitchPlayerInfoDisplay@8
	IDMFC_SwitchPlayerInfoDisplay@8=_IDMFC_SwitchPlayerInfoDisplay@8
	_IDMFC_SwitchServerHudCallsignLevel@8
	IDMFC_SwitchServerHudCallsignLevel@8=_IDMFC_SwitchServerHudCallsignLevel@8
	_IDMFC_SwitchShowHudCallsignLevel@12
	IDMFC_SwitchShowHudCallsignLevel@12=_IDMFC_SwitchShowHudCallsignLevel@12
	_IDMFC_TranslateEvent@12
	IDMFC_TranslateEvent@12=_IDMFC_TranslateEvent@12
	_IDMFC_UpdatePInfo@16
	IDMFC_UpdatePInfo@16=_IDMFC_UpdatePInfo@16
	_IDMFC_VersionCheck@8
	IDMFC_VersionCheck@8=_IDMFC_VersionCheck@8
	_IDMFC_WarpToLevel@8
	IDMFC_WarpToLevel@8=_IDMFC_WarpToLevel@8
	_IDMFC_WasPlayerInGameAtLevelEnd@8
	IDMFC_WasPlayerInGameAtLevelEnd@8=_IDMFC_WasPlayerInGameAtLevelEnd@8
	_IDMFC_WriteDMFCStatsToFile@8
	IDMFC_WriteDMFCStatsToFile@8=_IDMFC_WriteDMFCStatsToFile@8
	_IDmfcStats_CanScrollDown@4
	IDmfcStats_CanScrollDown@4=_IDmfcStats_CanScrollDown@4
	_IDmfcStats_CanScrollUp@4
	IDmfcStats_CanScrollUp@4=_IDmfcStats_CanScrollUp@4
	_IDmfcStats_Delete@4
	IDmfcStats_Delete@4=_IDmfcStats_Delete@4
	_IDmfcStats_DestroyPointer@4
	IDmfcStats_DestroyPointer@4=_IDmfcStats_DestroyPointer@4
	_IDmfcStats_DoFrame@4
	IDmfcStats_DoFrame@4=_IDmfcStats_DoFrame@4
	_IDmfcStats_DuplicatePointer@4
	IDmfcStats_DuplicatePointer@4=_IDmfcStats_DuplicatePointer@4
	_IDmfcStats_Dynamic_Cast@8
	IDmfcStats_Dynamic_Cast@8=_IDmfcStats_Dynamic_Cast@8
	_IDmfcStats_Enable@8
	IDmfcStats_Enable@8=_IDmfcStats_Enable@8
	_IDmfcStats_Initialize@8
	IDmfcStats_Initialize@8=_IDmfcStats_Initialize@8
	_IDmfcStats_IsEnabled@4
	IDmfcStats_IsEnabled@4=_IDmfcStats_IsEnabled@4
	_IDmfcStats_ScrollDown@4
	IDmfcStats_ScrollDown@4=_IDmfcStats_ScrollDown@4
	_IDmfcStats_ScrollUp@4
	IDmfcStats_ScrollUp@4=_IDmfcStats_ScrollUp@4
	_IMenuItem_AddSubMenu@8
	IMenuItem_AddSubMenu@8=_IMenuItem_AddSubMenu@8
	_IMenuItem_Back@4
	IMenuItem_Back@4=_IMenuItem_Back@4
	_IMenuItem_CallFunc@8
	IMenuItem_CallFunc@8=_IMenuItem_CallFunc@8
	_IMenuItem_Delete@4
	IMenuItem_Delete@4=_IMenuItem_Delete@4
	_IMenuItem_DestroyPointer@4
	IMenuItem_DestroyPointer@4=_IMenuItem_DestroyPointer@4
	_IMenuItem_DetachSubMenu@8
	IMenuItem_DetachSubMenu@8=_IMenuItem_DetachSubMenu@8
	_IMenuItem_Down@4
	IMenuItem_Down@4=_IMenuItem_Down@4
	_IMenuItem_Draw@24
	IMenuItem_Draw@24=_IMenuItem_Draw@24
	_IMenuItem_DuplicatePointer@4
	IMenuItem_DuplicatePointer@4=_IMenuItem_DuplicatePointer@4
	_IMenuItem_Dynamic_Cast@8
	IMenuItem_Dynamic_Cast@8=_IMenuItem_Dynamic_Cast@8
	_IMenuItem_Execute@4
	IMenuItem_Execute@4=_IMenuItem_Execute@4
	_IMenuItem_Forward@4
	IMenuItem_Forward@4=_IMenuItem_Forward@4
	_IMenuItem_GetCustomSubMenuCount@4
	IMenuItem_GetCustomSubMenuCount@4=_IMenuItem_GetCustomSubMenuCount@4
	_IMenuItem_GetFocus@4
	IMenuItem_GetFocus@4=_IMenuItem_GetFocus@4
	_IMenuItem_GetTitle@4
	IMenuItem_GetTitle@4=_IMenuItem_GetTitle@4
	_IMenuItem_GetType@4
	IMenuItem_GetType@4=_IMenuItem_GetType@4
	_IMenuItem_HasSubMenus@4
	IMenuItem_HasSubMenus@4=_IMenuItem_HasSubMenus@4
	_IMenuItem_LoseInputFocus@4
	IMenuItem_LoseInputFocus@4=_IMenuItem_LoseInputFocus@4
	_IMenuItem_SetInputFocus@4
	IMenuItem_SetInputFocus@4=_IMenuItem_SetInputFocus@4
	_IMenuItem_SetState@8
	IMenuItem_SetState@8=_IMenuItem_SetState@8
	_IMenuItem_SetStateItemList@12
	IMenuItem_SetStateItemList@12=_IMenuItem_SetStateItemList@12
	_IMenuItem_Up@4
	IMenuItem_Up@4=_IMenuItem_Up@4
