/*
* Descent 3 
* Copyright (C) 2024 Parallax Software
*
* This program is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef FIREBALL_EXTERNAL_H_
#define FIREBALL_EXTERNAL_H_

#define MED_EXPLOSION_INDEX2 0
#define SMALL_EXPLOSION_INDEX2 1
#define MED_EXPLOSION_INDEX 2
#define MED_EXPLOSION_INDEX3 3
#define BIG_EXPLOSION_INDEX 4
#define BILLOWING_INDEX 5
#define SMALL_EXPLOSION_INDEX 6
#define MED_SMOKE_INDEX 7
#define BLACK_SMOKE_INDEX 8
#define BLAST_RING_INDEX 9
#define SMOKE_TRAIL_INDEX 10
#define CUSTOM_EXPLOSION_INDEX 11
#define SHRINKING_BLAST_INDEX 12
#define SMOLDERING_INDEX 13
#define SHRINKING_BLAST_INDEX2 14
#define HOT_SPARK_INDEX 15
#define COOL_SPARK_INDEX 16
#define GRADIENT_BALL_INDEX 17
#define SPRAY_INDEX 18
#define FADING_LINE_INDEX 19
#define MUZZLE_FLASH_INDEX 20
#define SHIP_HIT_INDEX 21
#define BLUE_BLAST_RING_INDEX 22
#define PARTICLE_INDEX 23
#define AFTERBURNER_INDEX 24
#define NAPALM_BALL_INDEX 25
#define LIGHTNING_ORIGIN_INDEXA 26
#define LIGHTNING_ORIGIN_INDEXB 27
#define RAINDROP_INDEX 28
#define PUDDLEDROP_INDEX 29
#define GRAVITY_FIELD_INDEX 30
#define LIGHTNING_BOLT_INDEX 31
#define INVUL_HIT_INDEX 32
#define SINE_WAVE_INDEX 33
#define AXIS_BILLBOARD_INDEX 34
#define DEFAULT_CORONA_INDEX 35
#define HEADLIGHT_CORONA_INDEX 36
#define STAR_CORONA_INDEX 37
#define SUN_CORONA_INDEX 38
#define SNOWFLAKE_INDEX 39
#define THICK_LIGHTNING_INDEX 40
#define BLUE_FIRE_INDEX 41
#define RUBBLE1_INDEX 42
#define RUBBLE2_INDEX 43
#define WATER_SPLASH_INDEX 44
#define SHATTER_INDEX 45
#define SHATTER_INDEX2 46
#define BILLBOARD_SMOKETRAIL_INDEX 47
#define MASSDRIVER_EFFECT_INDEX 48
#define BLUE_EXPLOSION_INDEX 49
#define GRAY_SPARK_INDEX 50
#define GRAY_LIGHTNING_BOLT_INDEX 51
#define MERCBOSS_MASSDRIVER_EFFECT_INDEX 52

// Fireball types
#define FT_EXPLOSION 0
#define FT_SMOKE 1
#define FT_EFFECT 2
#define FT_BILLOW 3
#define FT_SPARK 4

#endif
