/*
* Descent 3 
* Copyright (C) 2024 Parallax Software
*
* This program is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef WEAPON_EXTERNAL_H_
#define WEAPON_EXTERNAL_H_

#define LASER_INDEX 0
#define VAUSS_INDEX 1
#define MICROWAVE_INDEX 2
#define PLASMA_INDEX 3
#define FUSION_INDEX 4
#define SUPER_LASER_INDEX 5
#define MASSDRIVER_INDEX 6
#define NAPALM_INDEX 7
#define EMD_INDEX 8
#define OMEGA_INDEX 9
#define CONCUSSION_INDEX 10
#define HOMING_INDEX 11
#define IMPACTMORTAR_INDEX 12
#define SMART_INDEX 13
#define MEGA_INDEX 14
#define FRAG_INDEX 15
#define GUIDED_INDEX 16
#define NAPALMROCKET_INDEX 17
#define CYCLONE_INDEX 18
#define BLACKSHARK_INDEX 19

#define FLARE_INDEX 20

#endif
