/*
* Descent 3 
* Copyright (C) 2024 Parallax Software
*
* This program is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef _LANSTRINGS_HEADER
#define _LANSTRINGS_HEADER

#define TXT(d) GetString(d)

#define TXT_LC_HELP1 TXT(0)
#define TXT_LC_HELP2 TXT(1)
#define TXT_LC_RETURNMAIN TXT(2)
#define TXT_LC_GAMELISTHDR TXT(3)
#define TXT_LC_HELP3 TXT(4)
#define TXT_LC_HELP4 TXT(4)
#define TXT_LC_BLANK TXT(4)
#define TXT_LC_EXIT TXT(5)
#define TXT_LC_JOINSEL TXT(6)
#define TXT_LC_STARTNEW TXT(7)
#define TXT_LC_SRCHADDR TXT(8)
#define TXT_LC_SCANLOCAL TXT(9)
#define TXT_LC_GAMENAME TXT(10)
#define TXT_LC_MSNNAME TXT(11)
#define TXT_LC_SCRIPTNAME TXT(12)
#define TXT_LC_STARTGAME TXT(13)
#define TXT_LC_MPLYROPTIONS TXT(14)
#define TXT_LC_PREVMENU TXT(15)
#define TXT_LC_TIMELIMIT TXT(16)
#define TXT_LC_KILLGOAL TXT(17)
#define TXT_LC_PPS TXT(18)
#define TXT_LC_CFGALLOWEDSHIP TXT(19)
#define TXT_LC_SERVERMODE TXT(20)
#define TXT_LC_CLIENTSERVER TXT(21)
#define TXT_LC_PEERPEER TXT(22)
#define TXT_LC_SHIPSALLOWED TXT(23)
#define TXT_LC_RESPAWNRATE TXT(24)
#define TXT_LC_ERROR TXT(25)
#define TXT_LC_NO_TCPIP TXT(26)

#define TXT_LC_USEROTVEL TXT(29)
#define TXT_LC_USESMOOTHING TXT(30)
#define TXT_LC_NO_GAMES TXT(31)
#define TXT_LC_GAMEHEADER TXT(32)
#define TXT_LC_CANCEL TXT(33)
#define TXT_LC_MAXPLAYERS TXT(34)

#define TXT_LC_ACC_WEAP_COLL TXT(35)
#define TXT_LC_BRIGHT_PLAYERS TXT(36)
#endif
