EXPORTS
InitializeDLL@4
_InitializeDLL@4=InitializeDLL@4
ShutdownDLL@0
_ShutdownDLL@0=ShutdownDLL@0
GetGOScriptID@8
_GetGOScriptID@8=GetGOScriptID@8
CreateInstance@4
_CreateInstance@4=CreateInstance@4
DestroyInstance@8
_DestroyInstance@8=DestroyInstance@8
CallInstanceEvent@16
_CallInstanceEvent@16=CallInstanceEvent@16
SaveRestoreState@8
_SaveRestoreState@8=SaveRestoreState@8
