/*
 * Descent 3
 * Copyright (C) 2024 Parallax Software
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef _PXOSTRINGS_HEADER
#define _PXOSTRINGS_HEADER

#define TXT(d) GetString(d)

#define TXT_PXO_HELP1 TXT(0)
#define TXT_PXO_HELP2 TXT(1)
#define TXT_PXO_RETURNMAIN TXT(2)
#define TXT_PXO_GAMELISTHDR TXT(3)
#define TXT_PXO_BLANK TXT(4)
#define TXT_PXO_EXIT TXT(5)
#define TXT_PXO_JOINSEL TXT(6)
#define TXT_PXO_STARTNEW TXT(7)
#define TXT_PXO_SRCHADDR TXT(8)
#define TXT_PXO_SCANLOCAL TXT(9)
#define TXT_PXO_GAMENAME TXT(10)
#define TXT_PXO_MSNNAME TXT(11)
#define TXT_PXO_SCRIPTNAME TXT(12)
#define TXT_PXO_STARTGAME TXT(13)
#define TXT_PXO_MPLYROPTIONS TXT(14)
#define TXT_PXO_PREVMENU TXT(15)
#define TXT_PXO_TIMELIMIT TXT(16)
#define TXT_PXO_KILLGOAL TXT(17)
#define TXT_PXO_PPS TXT(18)
#define TXT_PXO_CFGALLOWEDSHIP TXT(19)
#define TXT_PXO_SERVERMODE TXT(20)
#define TXT_PXO_CLIENTSERVER TXT(21)
#define TXT_PXO_PEERPEER TXT(22)
#define TXT_PXO_SHIPSALLOWED TXT(23)
#define TXT_PXO_RESPAWNRATE TXT(24)
#define TXT_PXO_REGURL TXT(25)
#define TXT_PXO_HELP3 TXT(26)
#define TXT_PXO_HELP4 TXT(27)
#define TXT_PXO_CONNECTING TXT(28)
#define TXT_PXO_CANTCONNECT TXT(29)
#define TXT_PXO_MASTERTRACKER TXT(30)
#define TXT_PXO_CANTAUTOVALIDATE TXT(31)
#define TXT_PXO_LOGINMASTERTRKR TXT(32)
#define TXT_PXO_CANCEL TXT(33)
#define TXT_PXO_LOGIN TXT(34)
#define TXT_PXO_LOGINID TXT(35)
#define TXT_PXO_TRACKERID TXT(36)
#define TXT_PXO_PASSWORD TXT(37)
#define TXT_PXO_CREATENEWPROFILE TXT(38)
#define TXT_PXO_BADLOGIN TXT(39)
#define TXT_PXO_TIMEOUTMT TXT(40)
#define TXT_PXO_SENDPRIV TXT(41)
#define TXT_PXO_JOINPRIV TXT(42)
#define TXT_PXO_JOINCHAN TXT(43)
#define TXT_PXO_FINDPILOT TXT(44)
#define TXT_PXO_GETPILOTINFO TXT(45)
#define TXT_PXO_JOINSTARTGAME TXT(46)
#define TXT_PXO_INNEWLOBBY TXT(47)
#define TXT_PXO_WRITINGGAMESTATS TXT(48)
#define TXT_PXO_SENDINGGAMEOVER TXT(49)
#define TXT_PXO_ENTERINGLOBBY TXT(50)
#define TXT_PXO_CANTJOINLOBBY TXT(51)
#define TXT_PXO_PRIVATEMESSAGE TXT(52)
#define TXT_PXO_SEND TXT(53)
#define TXT_PXO_PILOTNAME TXT(54)
#define TXT_PXO_MESSAGE TXT(55)
#define TXT_PXO_JOIN TXT(56)
#define TXT_PXO_CHANNELNAME TXT(57)
#define TXT_PXO_SEARCH TXT(58)
#define TXT_PXO_SEARCHINGPILOT TXT(59)
#define TXT_PXO_CANBEFOUNDIN TXT(60)
#define TXT_PXO_GETTINGPILOTSTAT TXT(61)
#define TXT_PXO_PILOTSTATS TXT(62)
#define TXT_PXO_CLOSE TXT(63)
#define TXT_PXO_TOTALKILLS TXT(64)
#define TXT_PXO_TOTALDEATHS TXT(65)
#define TXT_PXO_SUICIDES TXT(66)
#define TXT_PXO_EFFECIENCY TXT(67)
#define TXT_PXO_TIMEINGAME TXT(68)
#define TXT_PXO_ERRORNOTONLINE TXT(69)
#define TXT_PXO_BADNICK TXT(70)
#define TXT_PXO_PRIVMSGTO TXT(71)
#define TXT_PXO_PRIVMSGFROM TXT(72)
#define TXT_PXO_CHATDISCONNECTED TXT(73)
#define TXT_PXO_YOUAREINLOBBY TXT(74)
#define TXT_PXO_MAKEGAMEDEFAULT TXT(75)
#define TXT_PXO_RETURNTOCHAT TXT(76)
#define TXT_PXO_NO_TCPIP TXT(77)
#define TXT_PXO_ERROR TXT(78)

#define TXT_PXO_NOTFOUND TXT(81)
#define TXT_PXO_USEROTVEL TXT(82)
#define TXT_PXO_USESMOOTHING TXT(83)
#define TXT_PXO_GAME_HDR TXT(84)
#define TXT_PXO_NO_GAMES TXT(85)
#define TXT_PXO_MAXPLAYERS TXT(86)
#define TXT_PXO_ACC_WEAP_COLL TXT(87)
#define TXT_PXO_BRIGHT_PLAYERS TXT(88)
#define TXT_RANKING0 89
// strings 89-98 are used for rankings
#define TXT_PXO_GAMENAME2 TXT(99)
#define TXT_PXO_GAMETYPE TXT(100)
#define TXT_PXO_MISSION TXT(101)
#define TXT_PXO_LEVEL TXT(102)
#define TXT_PXO_PLAYERS TXT(103)
#define TXT_PXO_PING TXT(104)
#endif
