//Microsoft Developer Studio generated resource script.
//
#include "resource.h"

// Generated Help ID header file
#define APSTUDIO_HIDDEN_SYMBOLS
#include "resource.hm"
#undef APSTUDIO_HIDDEN_SYMBOLS

#define APSTUDIO_READONLY_SYMBOLS
/////////////////////////////////////////////////////////////////////////////
//
// Generated from the TEXTINCLUDE 2 resource.
//
#include "afxres.h"

/////////////////////////////////////////////////////////////////////////////
#undef APSTUDIO_READONLY_SYMBOLS

/////////////////////////////////////////////////////////////////////////////
// English (U.S.) resources

#if !defined(AFX_RESOURCE_DLL) || defined(AFX_TARG_ENU)
#ifdef _WIN32
LANGUAGE LANG_ENGLISH, SUBLANG_ENGLISH_US
#pragma code_page(1252)
#endif //_WIN32

#ifdef APSTUDIO_INVOKED
/////////////////////////////////////////////////////////////////////////////
//
// TEXTINCLUDE
//

1 TEXTINCLUDE DISCARDABLE 
BEGIN
    "resource.h\0"
END

2 TEXTINCLUDE DISCARDABLE 
BEGIN
    "#include ""afxres.h""\r\n"
    "\0"
END

3 TEXTINCLUDE DISCARDABLE 
BEGIN
    "#define _AFX_NO_SPLITTER_RESOURCES\r\n"
    "#define _AFX_NO_OLE_RESOURCES\r\n"
    "#define _AFX_NO_TRACKER_RESOURCES\r\n"
    "#define _AFX_NO_PROPERTY_RESOURCES\r\n"
    "\r\n"
    "#if !defined(AFX_RESOURCE_DLL) || defined(AFX_TARG_ENU)\r\n"
    "#ifdef _WIN32\r\n"
    "LANGUAGE 9, 1\r\n"
    "#pragma code_page(1252)\r\n"
    "#endif\r\n"
    "#include ""res\\editor.rc2""  // non-Microsoft Visual C++ edited resources\r\n"
    "#include ""afxres.rc""         // Standard components\r\n"
    "#include ""afxprint.rc""      // printing/print preview resources\r\n"
    "#include ""afxolesv.rc""      // OLE server resources\r\n"
    "#endif\0"
END

#endif    // APSTUDIO_INVOKED


/////////////////////////////////////////////////////////////////////////////
//
// Icon
//

// Icon with lowest ID value placed first to ensure application icon
// remains consistent on all systems.
IDR_MAINFRAME           ICON    DISCARDABLE     "res\\editor.ico"
IDR_EDITORTYPE          ICON    DISCARDABLE     "res\\editorDoc.ico"

/////////////////////////////////////////////////////////////////////////////
//
// Bitmap
//

IDR_MAINFRAME           BITMAP  MOVEABLE PURE   "res\\Toolbar.bmp"
IDR_SRVR_INPLACE        BITMAP  MOVEABLE PURE   "res\\IToolbar.bmp"
IDB_MATTHEAD            BITMAP  DISCARDABLE     "res\\Matt2.bmp"
IDB_D3SPLASH            BITMAP  DISCARDABLE     "res\\D3.bmp"
IDB_NORTH               BITMAP  DISCARDABLE     "res\\north.bmp"
IDB_EAST                BITMAP  DISCARDABLE     "res\\east.bmp"
IDB_SOUTH               BITMAP  DISCARDABLE     "res\\south.bmp"
IDB_WEST                BITMAP  DISCARDABLE     "res\\west.bmp"
IDB_ROTATELEFT          BITMAP  DISCARDABLE     "res\\rotleft.bmp"
IDB_ROTATERIGHT         BITMAP  DISCARDABLE     "res\\rotright.bmp"
IDB_OUTARROW            BITMAP  DISCARDABLE     "res\\outarrow.bmp"
IDB_INARROW             BITMAP  DISCARDABLE     "res\\inarrow.bmp"
IDB_DALLAS_TREE_ICONS_BITMAP BITMAP  DISCARDABLE     "res\\DallasTreeIcons.bmp"
IDB_ORPHREMOVE          BITMAP  DISCARDABLE     "res\\orphremo.bmp"

/////////////////////////////////////////////////////////////////////////////
//
// Toolbar
//

IDR_MAINFRAME TOOLBAR DISCARDABLE  16, 16
BEGIN
    BUTTON      ID_FILE_NEW
    BUTTON      ID_FILE_OPEN
    BUTTON      ID_FILE_SAVE
    SEPARATOR
    BUTTON      ID_EDIT_CUT
    BUTTON      ID_EDIT_COPY
    BUTTON      ID_EDIT_PASTE
    SEPARATOR
    BUTTON      ID_APP_ABOUT
    SEPARATOR
    BUTTON      ID_FILE_PLAY640X480
    SEPARATOR
    BUTTON      ID_BUTTON_OUTLINE
    BUTTON      ID_BUTTON_WINDOWSELECTION
    BUTTON      ID_BUTTON_LIGHTING
    BUTTON      ID_ZBUTTON
    SEPARATOR
    BUTTON      ID_OBJBUTTON
    BUTTON      ID_REINIT_OPENGL
    BUTTON      ID_OSIRISCOMPILE
END

IDR_SRVR_INPLACE TOOLBAR DISCARDABLE  16, 15
BEGIN
    BUTTON      ID_EDIT_CUT
    BUTTON      ID_EDIT_COPY
    BUTTON      ID_EDIT_PASTE
    SEPARATOR
    BUTTON      ID_APP_ABOUT
END


/////////////////////////////////////////////////////////////////////////////
//
// Menu
//

IDR_MAINFRAME MENU PRELOAD DISCARDABLE 
BEGIN
    POPUP "&File"
    BEGIN
        MENUITEM "&New\tCtrl+N",                ID_FILE_NEW
        MENUITEM "&Open...\tCtrl+O",            ID_FILE_OPEN
        MENUITEM "&Save\tCtrl+S",               ID_FILE_SAVE
        MENUITEM "Save &As...",                 ID_FILE_SAVE_AS
        MENUITEM SEPARATOR
        MENUITEM "Edit Level Info...",          ID_FILE_LEVEL_INFO, HELP
        MENUITEM "Save Goal Text to File",      ID_FILE_SAVEGOALTEXT
        MENUITEM "Show Level Stats",            ID_FILE_STATS
        MENUITEM "Verify Level",                ID_FILE_VERIFY_LEVEL
        MENUITEM "Fix Level Cracks",            ID_FILE_FIXCRACKS
        MENUITEM "Remove Extra Points",         ID_FILE_REMOVEEXTRAPOINTS
        MENUITEM "Fix/Remove Degenerate Faces", ID_FILE_FIXDEGENERATEFACES
        MENUITEM "Remove Duplicate Faces from Current Room", 
                                                ID_FILE_REMOVEDUPLICATEFACESFROMCURRENTROOM

        MENUITEM "Set Selected Room(s) Properties", ID_FILE_LEVELPROPS
        MENUITEM SEPARATOR
        MENUITEM "Import 3DSMax Room",          ID_FILE_IMPORT_ROOM
        MENUITEM SEPARATOR
        MENUITEM "&Play in 640 x 480\tAlt+P",   ID_FILE_PLAY640X480
        MENUITEM "Go to main menu...",          ID_FILE_LEAVE_EDITOR
        MENUITEM SEPARATOR
        MENUITEM "Autosave",                    ID_FILE_AUTOSAVE, CHECKED
        MENUITEM "Restore Gamestate",           ID_FILE_RESTOREGAMESTATE
        MENUITEM "&Information\tAlt+I",         ID_FILE_INFORMATION
        MENUITEM "&Check Mine\tAlt+C",          ID_FILE_CHECKMINE
        MENUITEM SEPARATOR
        MENUITEM "&Print...\tCtrl+P",           ID_FILE_PRINT
        MENUITEM "Print Pre&view",              ID_FILE_PRINT_PREVIEW
        MENUITEM "P&rint Setup...",             ID_FILE_PRINT_SETUP
        MENUITEM SEPARATOR
        MENUITEM "Preferences",                 ID_FILE_PREFERENCES
        MENUITEM SEPARATOR
        MENUITEM "Recent File",                 ID_FILE_MRU_FILE1, GRAYED
        MENUITEM SEPARATOR
        MENUITEM "E&xit\tAlt+X",                ID_APP_EXIT
    END
    POPUP "&Edit"
    BEGIN
        MENUITEM "&Undo\tCtrl+Z",               ID_EDIT_UNDO
        MENUITEM SEPARATOR
        MENUITEM "Cu&t\tCtrl+X",                ID_EDIT_CUT
        MENUITEM "&Copy\tCtrl+C",               ID_EDIT_COPY
        MENUITEM "&Place in Mine\tCtrl+V",      ID_EDIT_PLACE
        MENUITEM "&Place on Terrain",           ID_EDIT_PLACE_TERRAIN
        MENUITEM "&Attach",                     ID_EDIT_ATTACH
        MENUITEM "&Delete",                     ID_EDIT_DELETE
        MENUITEM SEPARATOR
        MENUITEM "Load Scrap from Disk",        ID_EDIT_LOADSCRAP
        MENUITEM "Save Scrap to Disk",          ID_EDIT_SAVESCRAP
        MENUITEM SEPARATOR
        MENUITEM "Add Current Room to Selected List", ID_EDIT_ADDSELECT
        MENUITEM "Remove Current Room from Selected List", 
                                                ID_EDIT_REMOVESELECT
        MENUITEM "Select All Rooms Attached to Current Room", 
                                                ID_EDIT_SELECTATTACHED
        MENUITEM "Clear Selected List",         ID_EDIT_CLEARSELECTED
    END
    POPUP "&View"
    BEGIN
        MENUITEM "&Toolbar",                    ID_VIEW_TOOLBAR
        MENUITEM "Keypad\tF10",                 ID_VIEW_KEYPAD_TOGGLE
        MENUITEM "Te&xtured view",              ID_VIEW_TEXTUREMINE
        MENUITEM "&Wireframe view",             ID_VIEW_WIREFRAMEMINE
        MENUITEM "Show objects in Wireframe view", 
                                                ID_VIEW_SHOWOBJECTSINWIREFRAMEVIEW

        MENUITEM SEPARATOR
        MENUITEM "Center on Current &Room\tHome", ID_VIEW_CENTERONCUBE
        MENUITEM "Center on Current &Object",   ID_VIEW_CENTERONOBJECT
        MENUITEM "Center on &Mine Origin",      ID_VIEW_CENTERONMINE
        MENUITEM "Reset View Radius",           ID_VIEW_RESETVIEWRADIUS
        MENUITEM SEPARATOR
        MENUITEM "Move camera to &Current Room\tC", 
                                                ID_VIEW_MOVECAMERATOSELECTEDROOM

        MENUITEM "Move camera to &Current Face\tShift-C", 
                                                ID_VIEW_MOVECAMERATOSELECTEDFACE

        MENUITEM "Move camera to Current Object", 
                                                ID_VIEW_MOVECAMERATOCURRENTOBJECT

        MENUITEM "Flip Viewer\tKeypad-/",       ID_VIEW_FLIP
        MENUITEM SEPARATOR
        MENUITEM "Show Viewer Forward Vector",  ID_VIEW_SHOWVIEWERFORWARDVECTOR

        MENUITEM SEPARATOR
        MENUITEM "Mine View\tF2",               ID_MINE_VIEW
        MENUITEM "Terrain View\tF3",            ID_TERRAIN_VIEW
        MENUITEM "Room View\tF4",               ID_ROOM_VIEW
        MENUITEM SEPARATOR
        MENUITEM "Next Viewer\tK",              ID_VIEW_NEXTVIEWER
        MENUITEM "New Viewer",                  ID_VIEW_NEWVIEWER
        MENUITEM "Delete Viewer",               ID_VIEW_DELETEVIEWER
        MENUITEM SEPARATOR
        MENUITEM "Display Viewer Properties",   ID_VIEW_VIEWPROP
    END
    POPUP "&Room"
    BEGIN
        MENUITEM "&Add Room at Current Face\tA", ID_ROOM_ADD
        MENUITEM SEPARATOR
        MENUITEM "Delete Current Room",         ID_ROOM_DELETE
        MENUITEM "Delete Current Face",         ID_ROOM_DELETEFACE
        MENUITEM "Delete Current Portal",       ID_ROOM_DELETEPORTAL
        MENUITEM "Delete Current Vert From Face", ID_ROOM_DELETEVERT
        MENUITEM "Delete Connected Faces",      ID_ROOM_DELETECONNECTEDFACES
        MENUITEM SEPARATOR
        MENUITEM "Place Room at Current Room",  ID_ROOM_PLACEROOM
        MENUITEM "Place Room at Current Terrain Cell", 
                                                ID_ROOM_PLACETERRAINROOM
        MENUITEM "Rotate Placed Room 45 Degrees", 
                                                ID_ROOM_ROTATEPLACEDROOM45DEGREES

        MENUITEM "Snap Placed Room",            ID_ROOM_SNAPPLACEDROOM
        MENUITEM "Attach Room",                 ID_ROOM_ATTACHROOM
        MENUITEM "Un-place Room/Group/Door",    ID_ROOM_UNPLACEROOM
        MENUITEM "Drop Room",                   ID_ROOM_DROPROOM
        MENUITEM SEPARATOR
        MENUITEM "&Mark Room/Face\tCtrl+M",     ID_ROOM_MARK
        MENUITEM "Swap Marked and Current Room/Face", 
                                                ID_ROOM_SWAPMAKEDANDCURRENTROOMFACE

        MENUITEM SEPARATOR
        MENUITEM "Select Room by Number",       ID_ROOM_SELECTBYNUMBER
        MENUITEM SEPARATOR
        MENUITEM "Save Current Room",           ID_ROOM_SAVECURRENTROOM
        MENUITEM "Rename Room\tCtrl+R",         ID_ROOM_RENAMEROOM
        MENUITEM SEPARATOR
        MENUITEM "&Build Bridge\tCtrl+B",       ID_ROOM_BUILDBRIDGE
        MENUITEM "Build Smooth Bridge",         ID_ROOM_BUILDSMOOTHBRIDGE
        MENUITEM "Join Rooms",                  ID_ROOM_JOINROOMS
        MENUITEM "Join Rooms Exact",            ID_ROOM_JOINROOMSEXACT
        MENUITEM "Join Adjacent Faces",         ID_ROOM_JOIN_ADJACENT_FACES
        MENUITEM "Combine Rooms",               ID_ROOM_COMBINE
        MENUITEM "Link Room to New External Room", ID_ROOM_LINKTONEWEXTERNAL
        MENUITEM "Merge Object into Room",      ID_ROOM_MERGEOBJECTINTOROOM
        MENUITEM SEPARATOR
        MENUITEM "Add First Point to New Face (Start Face)", 
                                                ID_ROOM_STARTNEWFACE
        MENUITEM "Add Vert to New Face\tShift+V", ID_ROOM_ADDVERTTONEWFACE
        MENUITEM "Add Last Point to New Face (Finish Face)", 
                                                ID_ROOM_FINISHNEWFACE
        MENUITEM SEPARATOR
        MENUITEM "Snap Point to Edge",          ID_ROOM_SNAPPOINTTOEDGE
        MENUITEM "Snap Point to Point",         ID_ROOM_SNAPPOINTTOPOINT
        MENUITEM "Snap Point to Face",          ID_ROOM_SNAPPOINTTOFACE
        MENUITEM "Undo Snap",                   ID_ROOM_UNDOSNAP
        MENUITEM SEPARATOR
        MENUITEM "Set Room Center from Viewer", ID_ROOM_SETCENTERFROMVIEWER
        MENUITEM "Clear Center Point",          ID_ROOM_CLEARCENTERPOINT
        MENUITEM SEPARATOR
        MENUITEM "Next Face\tF",                ID_ROOM_NEXTFACE
        MENUITEM "Previous Face\tShift+F",      ID_ROOM_PREVIOUSFACE
        MENUITEM "Select Face by Number",       ID_ROOM_SELECTFACEBYNUMBER
        POPUP "Face Editing"
        BEGIN
            MENUITEM "Add Vert To Edge",            ID_ROOM_FACE_ADDVERTTOEDGE

            MENUITEM "Delete Vert on Edge",         ID_ROOM_FACE_DELETEVERTONEDGE

            MENUITEM "Move Vert on Edge",           ID_ROOM_FACE_MOVEVERTONEDGE

            MENUITEM "Split Face",                  ID_ROOM_FACE_SPLITFACE
            MENUITEM "Triangulate face",            ID_ROOM_SPLITFACE
        END
        MENUITEM SEPARATOR
        MENUITEM "Grab Texture\tG",             ID_ROOM_GRABTEXTURE
        MENUITEM "Propagate to All Faces in Room", ID_ROOM_PROPAGATETOALL
        MENUITEM "Propagate to Adjacent Coplanar Faces\tCtrl+F", 
                                                ID_ROOM_PROPAGATETOADJACENTCOPLANARFACES

    END
    POPUP "&Object"
    BEGIN
        MENUITEM "Place Object\tInsert",        ID_OBJECT_PLACEOBJECT
        MENUITEM "Delete Object\tDelete",       ID_OBJECT_DELETEOBJECT
        MENUITEM "Move Player 0 to Viewer",     ID_OBJECT_MOVEPLAYER
        MENUITEM "Create New Camera at Viewer", ID_OBJECT_PLACECAMERAATVIEWER

        MENUITEM "Create New Camera at Current Face", 
                                                ID_OBJECT_PLACECAMERAATCURRENTFACE

        MENUITEM "Move Camera to Viewer",       ID_OBJECT_SETCAMERAFROMVIEWER

        MENUITEM "Move Viewer to Camera",       ID_OBJECT_SETVIEWERFROMCAMERA

        MENUITEM "Select Object By Number",     ID_OBJECT_SELECTBYNUMBER
        MENUITEM "Create SoundSource at Viewer", 
                                                ID_OBJECT_PLACESOUNDSOURCEATVIEWER

        MENUITEM "Place Waypoint at Viewer",    ID_OBJECT_PLACEWAYPOINTATVIEWER

    END
    POPUP "Edi&tors"
    BEGIN
        MENUITEM "Textures...",                 ID_TOOLS_WORLD_TEXTURES
        MENUITEM "Megacells...",                ID_EDITORS_MEGACELLS
        MENUITEM "Robots...",                   ID_TOOLS_WORLD_OBJECTS_ROBOTS

        MENUITEM "Powerups...",                 ID_TOOLS_WORLD_OBJECTS_POWERUPS

        MENUITEM "Buildings...",                ID_TOOLS_WORLD_OBJECTS_BUILDINGS

        MENUITEM "Clutter...",                  ID_TOOLS_WORLD_OBJECTS_CLUTTER

        MENUITEM "Player Ships...",             ID_TOOLS_WORLD_OBJECTS_PLAYER

        MENUITEM "Weapons...",                  ID_TOOLS_WORLD_WEAPONS
        MENUITEM "Doors...",                    ID_TOOLS_WORLD_OBJECTS_DOOR
        MENUITEM "Sounds...",                   ID_TOOLS_WORLD_OBJECTS_SOUND
        MENUITEM "Lights...",                   ID_TOOLS_WORLD_OBJECTS_LIGHTS

        MENUITEM "Ambient Sounds...",           ID_EDITORS_AMBIENTSOUNDS
        MENUITEM "Levels and Scripts...",       ID_SCRIPT_LEVEL_INTERFACE
        MENUITEM "Files...",                    ID_EDITORS_FILES
        MENUITEM "AI Properties...",            ID_EDITORS_AIPROPERTIES
        , GRAYED
        MENUITEM SEPARATOR
        MENUITEM "Reorder Net Pages",           IDD_REORDER_PAGES
        MENUITEM "Show All Pages Checked Out",  IDD_SHOW_ALL_CHECKED_OUT
        MENUITEM "Orphan Hunter",               IDD_ORPHANHUNTER
        MENUITEM SEPARATOR
        MENUITEM "Bitmap/Animation Converter...", IDM_IMPORT_BITMAP
        MENUITEM "HotSpot TGA Extractor...",    ID_HOTSPOT_TGA
        MENUITEM SEPARATOR
        MENUITEM "Briefing Editor...",          ID_BRIEFING_EDITOR
        MENUITEM "Font Editor...",              ID_SUBEDITORS_FONT
        MENUITEM "Hog Maker...",                ID_SUBEDITORS_HOGMAKER
        MENUITEM "Table File Text Editor...",   ID_SUBEDITORS_TABLEFILEEDIT
        MENUITEM "Table File Filter...",        ID_SUBEDITORS_TABLEFILEFILTER

        MENUITEM "Dallas Graphical Script Editor...", ID_EDITORS_DALLAS
    END
    POPUP "&Window"
    BEGIN
        MENUITEM "&Tile",                       ID_WINDOW_TILE
        MENUITEM "&Cascade",                    ID_WINDOW_CASCADE
    END
    POPUP "&Help"
    BEGIN
        MENUITEM "Help on D3 Editor",           ID_D3HELP
        MENUITEM "&About D3 Editor...",         ID_APP_ABOUT
    END
    POPUP "Test"
    BEGIN
        MENUITEM "Test function 1",             ID_TEST_TEST1
        MENUITEM "Test function 2",             ID_TEST_TEST2
        MENUITEM "Test function 3",             ID_TEST_TEST3
    END
END

IDR_SRVR_INPLACE MENU PRELOAD DISCARDABLE 
BEGIN
    POPUP "&Edit"
    BEGIN
        MENUITEM "&Undo\tCtrl+Z",               ID_EDIT_UNDO
        MENUITEM SEPARATOR
        MENUITEM "Cu&t\tCtrl+X",                ID_EDIT_CUT
        MENUITEM "&Copy\tCtrl+C",               ID_EDIT_COPY
        MENUITEM "&Paste\tCtrl+V",              ID_EDIT_PASTE
    END
    POPUP "&View"
    BEGIN
        MENUITEM "&Toolbar",                    ID_VIEW_TOOLBAR
    END
    MENUITEM SEPARATOR
    MENUITEM SEPARATOR
    POPUP "&Help"
    BEGIN
        MENUITEM "Help on D3 Editor",           ID_D3HELP
        MENUITEM "&About D3 Editor...",         ID_APP_ABOUT
    END
END

IDR_SRVR_EMBEDDED MENU PRELOAD DISCARDABLE 
BEGIN
    POPUP "&File"
    BEGIN
        MENUITEM "&Update\tCtrl+S",             ID_FILE_UPDATE
        MENUITEM "Save Copy &As...",            ID_FILE_SAVE_COPY_AS
        MENUITEM SEPARATOR
        MENUITEM "&Print...\tCtrl+P",           ID_FILE_PRINT
        MENUITEM "Print Pre&view",              ID_FILE_PRINT_PREVIEW
        MENUITEM "P&rint Setup...",             ID_FILE_PRINT_SETUP
        MENUITEM SEPARATOR
        MENUITEM "E&xit",                       ID_APP_EXIT
    END
    POPUP "&Edit"
    BEGIN
        MENUITEM "&Undo\tCtrl+Z",               ID_EDIT_UNDO
        MENUITEM SEPARATOR
        MENUITEM "Cu&t\tCtrl+X",                ID_EDIT_CUT
        MENUITEM "&Copy\tCtrl+C",               ID_EDIT_COPY
        MENUITEM "&Paste\tCtrl+V",              ID_EDIT_PASTE
    END
    POPUP "&View"
    BEGIN
        MENUITEM "&Toolbar",                    ID_VIEW_TOOLBAR
        MENUITEM "&Status Bar",                 ID_VIEW_STATUS_BAR
    END
    POPUP "&Help"
    BEGIN
        MENUITEM "&About editor...",            ID_APP_ABOUT
    END
END

IDR_TABLE_FILE_EDIT_MENU MENU DISCARDABLE 
BEGIN
    POPUP "&File"
    BEGIN
        MENUITEM "&Load...",                    ID_TABLEEDIT_FILE_LOAD
        MENUITEM "&Save",                       ID_TABLEEDIT_FILE_SAVE
        MENUITEM "S&ave as...",                 ID_TABLEEDIT_FILE_SAVE_AS
        MENUITEM SEPARATOR
        MENUITEM "&Exit",                       ID_TABLEEDIT_FILE_EXIT
    END
END

IDR_TABLE_FILE_FILTER_MENU MENU DISCARDABLE 
BEGIN
    POPUP "File"
    BEGIN
        MENUITEM "New",                         ID_TABLEFILTER_FILE_NEW
        MENUITEM "Load...",                     ID_TABLEFILTER_FILE_LOAD
        MENUITEM "Save",                        ID_TABLEFILTER_FILE_SAVE
        MENUITEM "Save as...",                  ID_TABLEFILTER_FILE_SAVEAS
        MENUITEM SEPARATOR
        MENUITEM "Exit",                        ID_TABLEFILTER_FILE_EXIT
    END
END


/////////////////////////////////////////////////////////////////////////////
//
// Accelerator
//

IDR_MAINFRAME ACCELERATORS PRELOAD MOVEABLE PURE 
BEGIN
    "A",            ID_TOOLS_AI,            VIRTKEY, ALT, NOINVERT
    "A",            ID_ROOM_ADD,            VIRTKEY, SHIFT, NOINVERT
    "B",            ID_ROOM_BUILDSMOOTHBRIDGE, VIRTKEY, CONTROL, NOINVERT
    "C",            ID_VIEW_MOVECAMERATOSELECTEDROOM, VIRTKEY, NOINVERT
    "C",            ID_EDIT_COPY,           VIRTKEY, CONTROL, NOINVERT
    "C",            ID_VIEW_MOVECAMERATOSELECTEDFACE, VIRTKEY, SHIFT, 
                                                    NOINVERT
    "F",            ID_ROOM_NEXTFACE,       VIRTKEY, NOINVERT
    "F",            ID_ROOM_PROPAGATETOADJACENTCOPLANARFACES, VIRTKEY, 
                                                    CONTROL, NOINVERT
    "F",            ID_ROOM_PREVIOUSFACE,   VIRTKEY, SHIFT, NOINVERT
    "G",            ID_ROOM_GRABTEXTURE,    VIRTKEY, NOINVERT
    "G",            ID_TOOLS_CURVEGENERATOR, VIRTKEY, ALT, NOINVERT
    "I",            ID_FILE_INFORMATION,    VIRTKEY, ALT, NOINVERT
    "K",            ID_VIEWER_NEXT,         VIRTKEY, NOINVERT
    "L",            ID_TOOLS_LIGHTING,      VIRTKEY, ALT, NOINVERT
    "M",            ID_ROOM_MARK,           VIRTKEY, CONTROL, NOINVERT
    "M",            ID_ACCEL_COPY_LIGHT_MULTIPLE, VIRTKEY, SHIFT, NOINVERT
    "N",            ID_FILE_NEW,            VIRTKEY, CONTROL, NOINVERT
    "O",            ID_FILE_OPEN,           VIRTKEY, CONTROL, NOINVERT
    "O",            ID_TOOLS_OBJECT,        VIRTKEY, ALT, NOINVERT
    "P",            ID_FILE_PLAY640X480,    VIRTKEY, ALT, NOINVERT
    "R",            ID_ROOM_RENAMEROOM,     VIRTKEY, CONTROL, NOINVERT
    "R",            ID_TOOLS_TRIGGER,       VIRTKEY, ALT, NOINVERT
    "R",            ID_ROOM_PROPAGATETOALL, VIRTKEY, SHIFT, NOINVERT
    "S",            ID_FILE_SAVE,           VIRTKEY, CONTROL, NOINVERT
    "T",            ID_TOOLS_TEXTURE,       VIRTKEY, ALT, NOINVERT
    "V",            ID_ACCEL_NEXTVERTEX,    VIRTKEY, NOINVERT
    "V",            ID_EDIT_PASTE,          VIRTKEY, CONTROL, NOINVERT
    "V",            ID_ROOM_ADDVERTTONEWFACE, VIRTKEY, SHIFT, NOINVERT
    VK_ADD,         ID_NUMPADADD,           VIRTKEY, CONTROL, NOINVERT
    VK_BACK,        ID_EDIT_UNDO,           VIRTKEY, ALT, NOINVERT
    VK_DECIMAL,     ID_NUMPADDEL,           VIRTKEY, CONTROL, NOINVERT
    VK_DELETE,      ID_OBJECT_DELETEOBJECT, VIRTKEY, NOINVERT
    VK_DELETE,      ID_EDIT_CUT,            VIRTKEY, SHIFT, NOINVERT
    VK_DIVIDE,      ID_VIEW_FLIP,           VIRTKEY, NOINVERT
    VK_DIVIDE,      ID_NUMPADDIV,           VIRTKEY, CONTROL, NOINVERT
    VK_DOWN,        ID_ACCEL_SLIDEDOWN,     VIRTKEY, NOINVERT
    VK_F10,         ID_VIEW_KEYPAD_TOGGLE,  VIRTKEY, NOINVERT
    VK_F2,          ID_MINE_VIEW,           VIRTKEY, NOINVERT
    VK_F3,          ID_TERRAIN_VIEW,        VIRTKEY, NOINVERT
    VK_F4,          ID_ROOM_VIEW,           VIRTKEY, NOINVERT
    VK_F6,          ID_NEXT_PANE,           VIRTKEY, NOINVERT
    VK_F6,          ID_PREV_PANE,           VIRTKEY, SHIFT, NOINVERT
    VK_HOME,        ID_VIEW_CENTERONCUBE,   VIRTKEY, NOINVERT
    VK_INSERT,      ID_OBJECT_PLACEOBJECT,  VIRTKEY, NOINVERT
    VK_INSERT,      ID_EDIT_COPY,           VIRTKEY, CONTROL, NOINVERT
    VK_INSERT,      ID_EDIT_PASTE,          VIRTKEY, SHIFT, NOINVERT
    VK_LEFT,        ID_ACCEL_SLIDELEFT,     VIRTKEY, NOINVERT
    VK_MULTIPLY,    ID_NUMPADMUL,           VIRTKEY, CONTROL, NOINVERT
    VK_NUMPAD0,     ID_NUMPAD0,             VIRTKEY, CONTROL, NOINVERT
    VK_NUMPAD1,     ID_NUMPAD1,             VIRTKEY, CONTROL, NOINVERT
    VK_NUMPAD2,     ID_NUMPAD2,             VIRTKEY, CONTROL, NOINVERT
    VK_NUMPAD3,     ID_NUMPAD3,             VIRTKEY, CONTROL, NOINVERT
    VK_NUMPAD4,     ID_NUMPAD4,             VIRTKEY, CONTROL, NOINVERT
    VK_NUMPAD5,     ID_NUMPAD5,             VIRTKEY, CONTROL, NOINVERT
    VK_NUMPAD6,     ID_NUMPAD6,             VIRTKEY, CONTROL, NOINVERT
    VK_NUMPAD7,     ID_NUMPAD7,             VIRTKEY, CONTROL, NOINVERT
    VK_NUMPAD8,     ID_NUMPAD8,             VIRTKEY, CONTROL, NOINVERT
    VK_NUMPAD9,     ID_NUMPAD9,             VIRTKEY, CONTROL, NOINVERT
    VK_PAUSE,       ID_NUMPADLOCK,          VIRTKEY, CONTROL, NOINVERT
    VK_RETURN,      ID_NUMPADENTER,         VIRTKEY, CONTROL, NOINVERT
    VK_RIGHT,       ID_ACCEL_SLIDERIGHT,    VIRTKEY, NOINVERT
    VK_SUBTRACT,    ID_NUMPADMINUS,         VIRTKEY, CONTROL, NOINVERT
    VK_UP,          ID_ACCEL_SLIDEUP,       VIRTKEY, NOINVERT
    "W",            ID_TOOLS_WALL,          VIRTKEY, ALT, NOINVERT
    "X",            ID_EDIT_CUT,            VIRTKEY, CONTROL, NOINVERT
    "X",            ID_APP_EXIT,            VIRTKEY, ALT, NOINVERT
    "Z",            ID_EDIT_UNDO,           VIRTKEY, CONTROL, NOINVERT
END

IDR_SRVR_INPLACE ACCELERATORS PRELOAD MOVEABLE PURE 
BEGIN
    "Z",            ID_EDIT_UNDO,           VIRTKEY, CONTROL
    "X",            ID_EDIT_CUT,            VIRTKEY, CONTROL
    "C",            ID_EDIT_COPY,           VIRTKEY, CONTROL
    "V",            ID_EDIT_PASTE,          VIRTKEY, CONTROL
    VK_BACK,        ID_EDIT_UNDO,           VIRTKEY, ALT
    VK_DELETE,      ID_EDIT_CUT,            VIRTKEY, SHIFT
    VK_INSERT,      ID_EDIT_COPY,           VIRTKEY, CONTROL
    VK_INSERT,      ID_EDIT_PASTE,          VIRTKEY, SHIFT
    VK_ESCAPE,      ID_CANCEL_EDIT_SRVR,    VIRTKEY, NOINVERT
END

IDR_SRVR_EMBEDDED ACCELERATORS PRELOAD MOVEABLE PURE 
BEGIN
    "S",            ID_FILE_UPDATE,         VIRTKEY, CONTROL
    "P",            ID_FILE_PRINT,          VIRTKEY, CONTROL
    "Z",            ID_EDIT_UNDO,           VIRTKEY, CONTROL
    "X",            ID_EDIT_CUT,            VIRTKEY, CONTROL
    "C",            ID_EDIT_COPY,           VIRTKEY, CONTROL
    "V",            ID_EDIT_PASTE,          VIRTKEY, CONTROL
    VK_BACK,        ID_EDIT_UNDO,           VIRTKEY, ALT
    VK_DELETE,      ID_EDIT_CUT,            VIRTKEY, SHIFT
    VK_INSERT,      ID_EDIT_COPY,           VIRTKEY, CONTROL
    VK_INSERT,      ID_EDIT_PASTE,          VIRTKEY, SHIFT
    VK_F6,          ID_NEXT_PANE,           VIRTKEY 
    VK_F6,          ID_PREV_PANE,           VIRTKEY, SHIFT
END


/////////////////////////////////////////////////////////////////////////////
//
// Dialog
//

IDD_ABOUTBOX DIALOG DISCARDABLE  0, 0, 199, 99
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "About Descent 3 Editor"
FONT 8, "MS Sans Serif"
BEGIN
    LTEXT           "D3Edit  version 0.0",IDC_STATIC,20,20,56,8,SS_NOPREFIX
    LTEXT           "Copyright 1996-1998 Outrage Entertainment, Inc.  All rights reserved.",
                    IDC_STATIC,20,35,145,25
    DEFPUSHBUTTON   "OK",IDOK,140,68,32,14,WS_GROUP
END

IDD_KEYPAD_DLGBAR DIALOG DISCARDABLE  0, 0, 221, 518
STYLE WS_CHILD | WS_BORDER
FONT 7, "Small Fonts"
BEGIN
    CONTROL         "Tab1",IDC_KEYPADS,"SysTabControl32",0x0,0,5,220,510
END

IDD_TEXTUREKEYPAD DIALOG DISCARDABLE  0, 0, 210, 532
STYLE DS_CONTROL | WS_CHILD
FONT 7, "Small Fonts"
BEGIN
    PUSHBUTTON      "Slide Left",IDC_TEXPAD_SLIDELEFT,10,85,30,25,BS_BITMAP | 
                    BS_MULTILINE | BS_FLAT
    PUSHBUTTON      "Default UVs",IDC_TEXPAD_SETDEFAULT,45,85,30,25,
                    BS_MULTILINE | BS_FLAT
    PUSHBUTTON      "Rotate Left",IDC_TEXPAD_ROTLEFT,10,115,30,25,BS_BITMAP | 
                    BS_MULTILINE | BS_FLAT
    PUSHBUTTON      "Slide Down",IDC_TEXPAD_SLIDEDOWN,45,115,30,25,BS_BITMAP | 
                    BS_MULTILINE | BS_FLAT
    PUSHBUTTON      "Rotate Right",IDC_TEXPAD_ROTRIGHT,80,115,30,25,
                    BS_BITMAP | BS_MULTILINE | BS_FLAT
    PUSHBUTTON      "Stretch Less",IDC_TEXPAD_STRETCHLESS,10,55,30,25,
                    BS_BITMAP | BS_MULTILINE | BS_FLAT
    PUSHBUTTON      "Slide Up",IDC_TEXPAD_SLIDEUP,45,55,30,25,BS_BITMAP | 
                    BS_MULTILINE | BS_FLAT
    PUSHBUTTON      "Stretch More",IDC_TEXPAD_STRETCHMORE,80,55,30,25,
                    BS_BITMAP | BS_MULTILINE | BS_FLAT
    PUSHBUTTON      "Next Edge",IDC_TEXPAD_NEXTEDGE,45,145,30,25,
                    BS_MULTILINE | BS_FLAT
    PUSHBUTTON      "Rot 90",IDC_TEXPAD_ROTATE90,10,145,30,25,BS_MULTILINE | 
                    BS_FLAT
    PUSHBUTTON      "Flip X",IDC_TEXPAD_FLIPX,10,25,30,25,BS_MULTILINE | 
                    BS_FLAT
    PUSHBUTTON      "Flip Y",IDC_TEXPAD_FLIPY,45,25,30,25,BS_MULTILINE | 
                    BS_FLAT
    PUSHBUTTON      "Slide Right",IDC_TEXPAD_SLIDERIGHT,80,85,30,25,
                    BS_BITMAP | BS_MULTILINE | BS_FLAT
    GROUPBOX        "Pad",IDC_TEXTUREPAD_GROUP,5,10,200,170
    GROUPBOX        "Palette",IDC_TEXTURE_GROUP,5,205,200,320
    PUSHBUTTON      "Grab",IDC_TEXPAD_GRAB,80,145,30,25,BS_FLAT
    CONTROL         "",IDC_TEXTURE_LISTBOX,"Static",SS_BLACKFRAME,15,220,175,
                    275
    LTEXT           "Show texture types:",IDC_STATIC,16,500,61,8
    CONTROL         "Terrain",IDC_SHOW_TERRAIN_TEXTURES,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,85,510,40,11
    CONTROL         "Light",IDC_SHOW_LIGHT_TEXTURES,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,140,510,35,10
    CONTROL         "Object",IDC_SHOW_OBJ_TEXTURES,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,140,500,40,11
    CONTROL         "Mine",IDC_SHOW_MINE_TEXTURES2,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,85,500,40,11
    PUSHBUTTON      "Reset All",IDC_RESET_ALL_TEXTURES,150,85,30,25,
                    BS_MULTILINE | BS_FLAT
    LTEXT           "Texture:",IDC_CURRENT_TEXTURE_NAME,5,190,140,10
    PUSHBUTTON      "Replace face texture with current texture",
                    IDC_REPLACE_TEXTURE,150,115,50,40,BS_MULTILINE | BS_FLAT
    CONTROL         "",IDC_TEXTURE_SWATCH,"Static",SS_BLACKFRAME,150,185,45,
                    20
    LTEXT           "Scale:",IDC_STATIC,140,160,20,10
    EDITTEXT        IDC_TEXSCALE_EDIT,170,160,25,12,ES_AUTOHSCROLL
    PUSHBUTTON      "Expand U",IDC_TEXPAD_EXPAND_U,115,25,30,25,BS_MULTILINE | 
                    BS_FLAT
    PUSHBUTTON      "Expand V",IDC_TEXPAD_EXPAND_V,115,55,30,25,BS_MULTILINE | 
                    BS_FLAT
    PUSHBUTTON      "Shrink U",IDC_TEXPAD_CONTRACT_U,115,85,30,25,
                    BS_MULTILINE | BS_FLAT
    PUSHBUTTON      "Shrink V",IDC_TEXPAD_CONTRACT_V,115,115,30,25,
                    BS_MULTILINE | BS_FLAT
    PUSHBUTTON      "Face map",IDC_FACE_MAP,80,25,30,25,BS_MULTILINE | 
                    BS_FLAT
    CONTROL         "Currently used",IDC_CURRENT_USED,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,15,510,65,10
    PUSHBUTTON      "Find next face with current texture",IDC_FIND_TEXTURE,
                    150,25,45,55,BS_MULTILINE
END

IDD_OBJECTKEYPAD DIALOG DISCARDABLE  0, 0, 210, 553
STYLE DS_CONTROL | WS_CHILD
FONT 7, "Small Fonts"
BEGIN
    PUSHBUTTON      "Set Default",IDC_OBJPAD_SETDEFAULT,150,70,40,20,
                    BS_MULTILINE | BS_FLAT
    PUSHBUTTON      "Delete Object",IDC_OBJ_DELOBJ,60,70,40,20,BS_MULTILINE | 
                    BS_FLAT
    PUSHBUTTON      "Place Object",IDC_OBJPAD_PLACEOBJ,15,70,40,20,
                    BS_MULTILINE | BS_FLAT
    PUSHBUTTON      "Next Object",IDC_OBJPAD_NEXTOBJ,105,70,40,20,
                    BS_MULTILINE | BS_FLAT
    PUSHBUTTON      "Flip",IDC_OBJPAD_FLIPOBJ,20,30,35,15,BS_MULTILINE | 
                    BS_FLAT
    GROUPBOX        "Object placement",IDC_OBJECTPAD_GROUP,5,10,200,140
    COMBOBOX        IDC_COMBO_OBJTYPE,69,288,120,200,CBS_DROPDOWNLIST | 
                    WS_VSCROLL | WS_TABSTOP
    LTEXT           "Type",IDC_STATIC,14,287,21,9
    LTEXT           "Specific name",IDC_STATIC,14,303,50,9
    COMBOBOX        IDC_COMBO_OBJID,69,304,120,180,CBS_DROPDOWNLIST | 
                    CBS_SORT | WS_VSCROLL | WS_TABSTOP
    GROUPBOX        "Selection",IDC_OBJECT_GROUP,5,274,200,198
    PUSHBUTTON      "Props.",IDC_OBJPAD_PROPERTIES,55,30,35,15,BS_FLAT
    PUSHBUTTON      "Flag Object as Multiplay Only",IDC_OBJPAD_FLAGOBJMULTI,
                    105,95,85,20,BS_MULTILINE | BS_FLAT
    CONTROL         "",IDC_OBJ_PALETTE,"Static",SS_BLACKFRAME,10,322,188,145
    PUSHBUTTON      "Reset Objects",IDC_RESET_OBJECTS,15,121,40,20,
                    BS_MULTILINE | BS_FLAT
    CONTROL         "",IDC_STATIC,"Static",SS_ETCHEDFRAME | SS_SUNKEN,10,25,
                    190,40
    CONTROL         "X",IDC_OBJMOVEX,"Button",BS_AUTORADIOBUTTON | 
                    BS_PUSHLIKE | BS_FLAT,90,30,25,15
    CONTROL         "Y",IDC_OBJMOVEY,"Button",BS_AUTORADIOBUTTON | 
                    BS_PUSHLIKE | BS_FLAT,115,30,25,15
    CONTROL         "Z",IDC_OBJMOVEZ,"Button",BS_AUTORADIOBUTTON | 
                    BS_PUSHLIKE | BS_FLAT,140,30,25,15
    CONTROL         "XY",IDC_OBJMOVEXY,"Button",BS_AUTORADIOBUTTON | 
                    BS_PUSHLIKE | BS_FLAT,165,30,25,15
    CONTROL         "P",IDC_OBJMOVEP,"Button",BS_AUTORADIOBUTTON | 
                    BS_PUSHLIKE | BS_FLAT,90,45,25,15
    CONTROL         "H",IDC_OBJMOVEH,"Button",BS_AUTORADIOBUTTON | 
                    BS_PUSHLIKE | BS_FLAT,115,45,25,15
    CONTROL         "B",IDC_OBJMOVEB,"Button",BS_AUTORADIOBUTTON | 
                    BS_PUSHLIKE | BS_FLAT,140,45,25,15
    CONTROL         "PH",IDC_OBJMOVEPH,"Button",BS_AUTORADIOBUTTON | 
                    BS_PUSHLIKE | BS_FLAT,165,45,25,15
    COMBOBOX        IDC_COORDSYS_SELECT,20,45,70,45,CBS_DROPDOWNLIST | 
                    WS_VSCROLL | WS_TABSTOP
    PUSHBUTTON      "Reset Current Object Height",
                    IDC_OBJPAD_RESET_CUROBJ_HEIGHT,61,121,60,20,BS_MULTILINE | 
                    BS_FLAT
    PUSHBUTTON      "Reset All\nObject Heights",
                    IDC_OBJPAD_RESET_ALL_OBJ_HEIGHTS,129,121,60,20,
                    BS_MULTILINE | BS_FLAT
    GROUPBOX        "Player start flags",IDC_STATIC,7,483,200,63
    LTEXT           "Current start position:",IDC_START_POS_TEXT,15,495,95,
                    10
    PUSHBUTTON      "<< Prev",IDC_PREV_START_POSITON,15,511,30,15,BS_FLAT
    PUSHBUTTON      "Next >>",IDC_NEXT_START_POS,55,511,30,15,BS_FLAT
    PUSHBUTTON      "Move viewer to start position",IDC_JUMP_TO_START_POS,
                    155,490,50,25,BS_MULTILINE | BS_FLAT
    CONTROL         "Red",IDC_RED_CHECK,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,15,530,30,10
    CONTROL         "Blue",IDC_BLUE_CHECK,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,61,530,30,10
    CONTROL         "Green",IDC_GREEN_CHECK,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,105,530,40,10
    CONTROL         "Yellow",IDC_YELLOW_CHECK,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,146,530,40,10
    GROUPBOX        "Object Type Swapping",IDC_STATIC,5,175,200,50
    LTEXT           "Convert this type:",IDC_STATIC,10,191,65,10
    LTEXT           "Into this type:",IDC_STATIC,10,205,50,10
    COMBOBOX        IDC_SWAP_SOURCE_COMBO,75,191,80,65,CBS_DROPDOWN | 
                    CBS_SORT | WS_VSCROLL | WS_TABSTOP
    COMBOBOX        IDC_SWAP_DEST_COMBO,75,205,80,65,CBS_DROPDOWN | CBS_SORT | 
                    WS_VSCROLL | WS_TABSTOP
    PUSHBUTTON      "Swap object types",IDC_OBJECT_SWAP_BUTTON,160,191,40,30,
                    BS_MULTILINE | BS_FLAT
    GROUPBOX        "Re-ground Object",IDC_STATIC,7,230,196,39
    LTEXT           "Object Type:",IDC_STATIC,16,244,49,13
    COMBOBOX        IDC_REGROUND_COMBO,75,242,80,65,CBS_DROPDOWN | CBS_SORT | 
                    WS_VSCROLL | WS_TABSTOP
    PUSHBUTTON      "Go!",IDC_REGROUND_BUTTON,160,238,40,25,BS_FLAT
    PUSHBUTTON      "Rot 90",IDC_OBJ_ROT90,15,97,38,19,BS_FLAT
    CONTROL         "Allow objects to be pushed into walls",
                    IDC_OBJECT_PUSHTHROUGHWALLS,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,10,155,190,15
    PUSHBUTTON      "Delete All",IDC_OBJPAD_DELETEALL,60,97,40,20,
                    BS_MULTILINE | BS_FLAT
END

IDD_LIGHTINGKEYPAD DIALOG DISCARDABLE  0, 0, 210, 481
STYLE DS_CONTROL | WS_CHILD
FONT 7, "Small Fonts"
BEGIN
    PUSHBUTTON      "Set Default",IDC_LITEPAD_SETDEFAULT,70,210,30,20,
                    BS_MULTILINE | BS_FLAT
    PUSHBUTTON      "Count lights in mine",IDC_LITEPAD_COUNT_LIGHTS,10,10,40,
                    25,BS_MULTILINE | BS_FLAT
    PUSHBUTTON      "Do terrain radiosity",IDC_LITEPAD_TERRAIN_RADIOSITY,10,
                    70,40,25,BS_MULTILINE | BS_FLAT
    PUSHBUTTON      "Next Vertex",IDC_LITEPAD_NEXTVERTEX,105,210,30,20,
                    BS_MULTILINE | BS_FLAT
    PUSHBUTTON      "Do mine radiosity",IDC_LITEPAD_SMOOTHALL,10,40,40,25,
                    BS_MULTILINE | BS_FLAT
    EDITTEXT        IDC_LIGHT_ITERATION_EDIT,155,135,45,15,ES_AUTOHSCROLL
    LTEXT           "Iterations",IDC_STATIC,115,135,35,10
    GROUPBOX        "Radiosity",IDC_STATIC,5,0,200,195
    GROUPBOX        "Volumetrics and pulse lighting",IDC_STATIC,5,200,200,
                    175
    LTEXT           "Alpha",IDC_STATIC,10,215,25,10
    EDITTEXT        IDC_ALPHA_EDIT,35,215,30,12,ES_AUTOHSCROLL
    CONTROL         "Show lightmaps",IDC_SHOW_LIGHTMAPS,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,130,5,70,10
    EDITTEXT        IDC_HEMICUBE_RES_EDIT,155,119,45,15,ES_AUTOHSCROLL
    LTEXT           "Hemicube res",IDC_STATIC,105,119,50,10
    EDITTEXT        IDC_LIGHT_SPACING_EDIT,155,105,45,15,ES_AUTOHSCROLL
    LTEXT           "Light spacing",IDC_STATIC,105,105,50,10
    CONTROL         "Best fit",IDC_BESTFIT_CHECK,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,130,15,70,8
    PUSHBUTTON      "Create volumetric cone from this face",
                    IDC_CREATE_VOLUME_CONE,10,315,60,25,BS_MULTILINE | 
                    BS_FLAT
    PUSHBUTTON      "Make volumetric have selected texture",
                    IDC_FILL_VOLUME_TEXTURE,80,345,65,25,BS_MULTILINE | 
                    BS_FLAT
    PUSHBUTTON      "Clear all volumetrics in room",
                    IDC_CLEAR_VOLUMETRIC_ROOM,10,345,60,25,BS_MULTILINE | 
                    BS_FLAT
    EDITTEXT        IDC_ANGLE_EDIT,50,265,40,12,ES_AUTOHSCROLL
    LTEXT           "Angle",IDC_STATIC,10,265,25,10
    EDITTEXT        IDC_START_ALPHA_EDIT,50,280,40,12,ES_AUTOHSCROLL
    LTEXT           "Start alpha",IDC_STATIC,10,280,40,10
    EDITTEXT        IDC_RESOLUTION_EDIT,50,295,40,12,ES_AUTOHSCROLL
    LTEXT           "Resolution",IDC_STATIC,10,295,40,10
    PUSHBUTTON      "Create edged volumetric from this face",
                    IDC_EDGE_VOLUMETRIC,80,315,65,25,BS_MULTILINE | BS_FLAT
    PUSHBUTTON      "Get farthest point from this face",IDC_FARTHEST_POINT,
                    150,315,50,25,BS_MULTILINE | BS_FLAT
    EDITTEXT        IDC_FALLOFF_EDIT,50,250,40,12,ES_AUTOHSCROLL
    LTEXT           "Falloff",IDC_STATIC,10,250,35,10
    EDITTEXT        IDC_PITCH_ANGLE_EDIT,155,250,40,12,ES_AUTOHSCROLL
    LTEXT           "Pitch angle",IDC_STATIC,100,250,55,10
    EDITTEXT        IDC_HEADING_ANGLE_EDIT,155,265,40,12,ES_AUTOHSCROLL
    LTEXT           "Heading angle",IDC_STATIC,100,265,55,10
    CONTROL         "Use hemicubes",IDC_HEMICUBE_CHECK,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,130,25,70,8
    PUSHBUTTON      "Count lights in terrain",IDC_COUNT_TERRAIN_LIGHTS,55,10,
                    40,25,BS_MULTILINE | BS_FLAT
    PUSHBUTTON      "Radiosity current room",IDC_CURROOM_RADIOSITY,55,40,40,
                    25,BS_MULTILINE | BS_FLAT
    PUSHBUTTON      "Dynamic terrain lighting",IDC_DYNAMIC_TERRAIN,55,70,40,
                    25,BS_MULTILINE | BS_FLAT
    EDITTEXT        IDC_IGNORE_LIMIT_EDIT,155,150,45,15,ES_AUTOHSCROLL
    LTEXT           "Ignore Limit",IDC_STATIC,115,150,40,10
    PUSHBUTTON      "Blur lightmaps",IDC_BLUR_LIGHTMAPS,10,150,40,21,
                    BS_MULTILINE | BS_FLAT
    LTEXT           "Light ambience:",IDC_STATIC,10,175,60,10
    LTEXT           "R",IDC_STATIC,75,175,10,10
    EDITTEXT        IDC_RED_AMBIENT_EDIT,85,175,20,12,ES_AUTOHSCROLL
    LTEXT           "G",IDC_STATIC,115,175,10,10
    EDITTEXT        IDC_GREEN_AMBIENT_EDIT,125,175,20,12,ES_AUTOHSCROLL
    LTEXT           "B",IDC_STATIC,155,175,10,10
    EDITTEXT        IDC_BLUE_AMBIENT_EDIT,165,175,20,12,ES_AUTOHSCROLL
    CONTROL         "Super Detail",IDC_ELEMENT_CHECK,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,130,35,70,8
    CONTROL         "Ignore terrain",IDC_IGNORE_TERRAIN,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,130,45,70,8
    CONTROL         "Ignore satellites",IDC_IGNORE_SATELLITES,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,130,55,70,8
    CONTROL         "Use BSP tree",IDC_USE_BSP_CHECK,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,130,65,70,8
    LTEXT           "Lighting multiplier",IDC_STATIC,10,392,34,17
    EDITTEXT        IDC_LIGHT_MULTIPLY_EDIT,46,394,30,12,ES_AUTOHSCROLL
    EDITTEXT        IDC_PULSE_TIME_EDIT,155,280,40,12,ES_AUTOHSCROLL
    LTEXT           "Pulse time",IDC_STATIC,100,280,55,10
    EDITTEXT        IDC_PULSE_OFFSET_EDIT,155,295,40,12,ES_AUTOHSCROLL
    LTEXT           "Pulse offset ",IDC_STATIC,100,295,55,10
    PUSHBUTTON      "Make room have pulse",IDC_PULSE_ROOM,150,345,45,25,
                    BS_MULTILINE | BS_FLAT
    PUSHBUTTON      "Reset all multiples",IDC_RESET_MULTIPLES,83,391,40,20,
                    BS_MULTILINE | BS_FLAT
    PUSHBUTTON      "Destroy BSP Tree",IDC_DESTROY_BSP,55,150,40,20,
                    BS_MULTILINE | BS_FLAT
    CONTROL         "Combine faces",IDC_COMBINE_FACES,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,130,75,70,10
    CONTROL         "Volume lights",IDC_VOLUME_LIGHTS,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,130,85,70,10
    CONTROL         "Face has corona",IDC_CORONA_CHECK,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,10,415,75,10
    CONTROL         "Fog current room",IDC_FOG_CHECK,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,10,445,80,10
    GROUPBOX        "Miscellaneous",IDC_STATIC,5,380,200,95
    LTEXT           "Fog depth",IDC_STATIC,95,445,45,10
    EDITTEXT        IDC_FOG_DENSITY,140,445,40,12,ES_AUTOHSCROLL
    LTEXT           "Fog color:",IDC_STATIC,10,460,40,10
    LTEXT           "R",IDC_STATIC,75,460,10,10
    EDITTEXT        IDC_FOG_RED,85,460,20,12,ES_AUTOHSCROLL
    LTEXT           "G",IDC_STATIC,115,460,10,10
    EDITTEXT        IDC_FOG_GREEN,125,460,20,12,ES_AUTOHSCROLL
    LTEXT           "B",IDC_STATIC,155,460,10,10
    EDITTEXT        IDC_FOG_BLUE,165,460,20,12,ES_AUTOHSCROLL
    PUSHBUTTON      "Fill Room Coronas",IDC_ROOM_CORONAS,90,415,45,20,
                    BS_MULTILINE | BS_FLAT
    PUSHBUTTON      "Fill Mine Coronas",IDC_MINE_CORONAS,140,415,45,20,
                    BS_MULTILINE | BS_FLAT
    CONTROL         "Reverse operation",IDC_REVERSE_CORONA,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,10,425,75,10
    LTEXT           "Global multiplier",IDC_STATIC,10,135,55,10
    EDITTEXT        IDC_GLOBAL_MULTIPLIER_EDIT,65,135,30,12,ES_AUTOHSCROLL
    CONTROL         "Strobe",IDC_STROBE_CHECK,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,150,210,50,10
    CONTROL         "Flicker",IDC_FLICKER_CHECK,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,150,225,50,10
    LTEXT           "Room multiplier",IDC_STATIC,10,120,55,10
    EDITTEXT        IDC_ROOM_MULTIPLIER_EDIT,65,120,30,12,ES_AUTOHSCROLL
    EDITTEXT        IDC_ROOM_AMGREEN,40,105,20,12,ES_AUTOHSCROLL
    EDITTEXT        IDC_ROOM_AMRED,10,105,20,12,ES_AUTOHSCROLL
    EDITTEXT        IDC_ROOM_AMBLUE,70,105,20,12,ES_AUTOHSCROLL
    LTEXT           "Room ambience",IDC_STATIC,11,97,60,9
    PUSHBUTTON      "Clear Lightmaps",IDC_CLEAR_LIGHTMAPS,127,391,39,20,
                    BS_MULTILINE | BS_FLAT
    PUSHBUTTON      "Copy Multiple",IDC_COPY_MULTIPLIER,170,390,33,20,
                    BS_MULTILINE | BS_FLAT
END

IDD_TRIGGERKEYPAD DIALOG DISCARDABLE  0, 0, 212, 461
STYLE DS_CONTROL | WS_CHILD
FONT 7, "Small Fonts"
BEGIN
    PUSHBUTTON      "Add Trigger to Current Face",IDC_TRIG_ADD_TO_CURFACE,
                    155,90,55,20,BS_MULTILINE | BS_FLAT
    PUSHBUTTON      "Add Trigger to Current Portal",
                    IDC_TRIG_ADD_TO_CURPORTAL,155,119,55,20,BS_MULTILINE | 
                    BS_FLAT
    GROUPBOX        "Current trigger",IDC_STATIC,5,85,135,250
    PUSHBUTTON      "Next in Mine",IDC_TRIG_NEXT_IN_MINE,50,20,40,20,
                    BS_MULTILINE | BS_FLAT
    PUSHBUTTON      "Previous in Mine",IDC_TRIG_PREV_IN_MINE,5,20,40,20,
                    BS_MULTILINE | BS_FLAT
    LTEXT           "Number:",IDC_TRIG_CURRENT_NUM,17,120,50,8
    LTEXT           "Type:",IDC_TRIG_CURRENT_TYPE,70,120,67,10
    LTEXT           "Room:",IDC_TRIG_CURRENT_ROOM,17,130,50,10
    LTEXT           "Face:",IDC_TRIG_CURRENT_FACE,17,140,50,10
    PUSHBUTTON      "Delete Trigger",IDC_TRIG_DELETE,95,155,40,20,
                    BS_MULTILINE | BS_FLAT
    PUSHBUTTON      "Add New Floating Trigger",IDC_TRIG_ADD_FLOATING,155,150,
                    55,20,BS_MULTILINE | BS_FLAT
    GROUPBOX        "Floating trigger size",IDC_TRIG_FLOAT_BORDER,14,266,100,
                    60
    PUSHBUTTON      "Select by Number",IDC_TRIG_GOTO,95,20,40,20,
                    BS_MULTILINE | BS_FLAT
    CONTROL         "Player",IDC_TRIG_ACTIV_PLAYER,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,22,208,65,10
    LTEXT           "Activators:",IDC_TRIG_ACTIVATORS,17,198,40,10
    CONTROL         "Player Weapons",IDC_TRIG_ACTIV_PLAYER_WEAPONS,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,22,219,65,10
    CONTROL         "Robots",IDC_TRIG_ACTIV_ROBOTS,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,22,229,65,10
    CONTROL         "Robot Weapons",IDC_TRIG_ACTIV_ROBOT_WEAPONS,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,22,239,65,10
    CONTROL         "Render Floating Triggers",IDC_TRIG_RENDER_FLOATING,
                    "Button",BS_AUTOCHECKBOX | WS_TABSTOP,7,343,90,10
    CONTROL         "One-shot",IDC_TRIG_ONESHOT,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,17,183,45,10
    PUSHBUTTON      "Wider",IDC_TRIG_FLOAT_WIDER,66,282,30,15,BS_FLAT
    PUSHBUTTON      "Narrower",IDC_TRIG_FLOAT_NARROWER,66,302,30,15,BS_FLAT
    PUSHBUTTON      "Taller",IDC_TRIG_FLOAT_TALLER,22,282,30,15,BS_FLAT
    PUSHBUTTON      "Shorter",IDC_TRIG_FLOAT_SHORTER,22,302,30,15,BS_FLAT
    PUSHBUTTON      "Next Portal",IDC_TRIG_NEXT_PORTAL,155,215,40,20,BS_FLAT
    LTEXT           "Current Portal:",IDC_TRIG_CURRENT_PORTAL,155,194,40,15
    PUSHBUTTON      "Next in Room",IDC_TRIG_NEXT_IN_ROOM,50,50,40,20,
                    BS_MULTILINE | BS_FLAT
    PUSHBUTTON      "Previous in Room",IDC_TRIG_PREV_IN_ROOM,5,50,40,20,
                    BS_MULTILINE | BS_FLAT
    LTEXT           "Name",IDC_TRIG_CURRENT_NAME,15,100,120,8
    PUSHBUTTON      "Edit Script",IDC_TRIG_EDIT_SCRIPT,17,155,35,21,
                    BS_MULTILINE | BS_FLAT
    PUSHBUTTON      "Edit Name",IDC_TRIG_EDIT_NAME,57,155,35,21,BS_MULTILINE | 
                    BS_FLAT
    LTEXT           "Portal:",IDC_TRIG_CURRENT_TRIG_PORTAL,72,140,50,10
    CONTROL         "Clutter",IDC_TRIG_ACTIV_CLUTTER,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,22,249,65,10
END

IDD_STATUS DIALOG DISCARDABLE  0, 0, 295, 182
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "File Management"
FONT 8, "MS Sans Serif"
BEGIN
    DEFPUSHBUTTON   "OK",IDOK,232,33,50,14
    PUSHBUTTON      "Cancel",IDCANCEL,232,54,50,14
    LTEXT           "Data Checked Out By",IDC_STATIC,18,23,69,8
    GROUPBOX        "Descent 3 Data Status",IDC_STATIC,7,7,281,168
    LISTBOX         IDC_LIST1,17,33,212,134,LBS_SORT | LBS_NOINTEGRALHEIGHT | 
                    WS_VSCROLL | WS_TABSTOP
    COMBOBOX        IDC_COMBO1,89,19,138,30,CBS_DROPDOWN | CBS_SORT | 
                    WS_VSCROLL | WS_TABSTOP
END

IDD_WORLDOBJECTSDOOR DIALOG DISCARDABLE  0, 0, 319, 262
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "WorldObjects >> Door"
FONT 8, "MS Sans Serif"
BEGIN
    LTEXT           "Open Sound",IDC_STATIC,10,191,41,8
    LTEXT           "Close Sound",IDC_STATIC,10,205,41,8
    LTEXT           "Open Time",IDC_STATIC,10,120,36,8
    LTEXT           "Close Time",IDC_STATIC,10,153,36,8
    LTEXT           "Energy Hit Sound",IDC_STATIC,10,220,56,8,WS_DISABLED
    LTEXT           "Material Hit Sound",IDC_STATIC,10,235,59,8,WS_DISABLED
    EDITTEXT        IDC_DOOR_OPEN_TIME,65,120,35,14,ES_AUTOHSCROLL | 
                    ES_WANTRETURN
    EDITTEXT        IDC_CLOSE_TIME,65,150,35,15,ES_AUTOHSCROLL | 
                    ES_WANTRETURN
    COMBOBOX        IDC_DOOR_OPEN_SOUND,70,190,95,65,CBS_DROPDOWNLIST | 
                    CBS_SORT | WS_VSCROLL | WS_TABSTOP
    COMBOBOX        IDC_DOOR_CLOSE_SOUND,70,205,95,55,CBS_DROPDOWNLIST | 
                    CBS_SORT | WS_VSCROLL | WS_TABSTOP
    COMBOBOX        IDC_DOOR_MATERIAL_SOUND,70,235,95,30,CBS_DROPDOWNLIST | 
                    CBS_SORT | WS_DISABLED | WS_VSCROLL | WS_TABSTOP
    COMBOBOX        IDC_DOOR_ENERGY_SOUND,70,220,95,30,CBS_DROPDOWNLIST | 
                    CBS_SORT | WS_DISABLED | WS_VSCROLL | WS_TABSTOP
    CONTROL         "",IDC_DOORVIEW,"Static",SS_ETCHEDFRAME,180,40,130,109
    PUSHBUTTON      "Add New Door",IDC_ADD_DOOR,120,15,50,14
    EDITTEXT        IDC_DOOR_MODEL_NAME_EDIT,50,75,120,15,ES_AUTOHSCROLL | 
                    ES_WANTRETURN
    PUSHBUTTON      "Delete Door",IDC_DELETE_DOOR,120,30,50,14
    LTEXT           "Stays Open For",IDC_STATIC,10,136,50,8
    EDITTEXT        IDC_DOOR_STAYS_OPEN,65,135,35,15,ES_AUTOHSCROLL | 
                    ES_WANTRETURN
    COMBOBOX        IDC_DOOR_PULLDOWN,10,20,95,235,CBS_DROPDOWNLIST | 
                    CBS_SORT | WS_VSCROLL | WS_TABSTOP
    PUSHBUTTON      "Load Model",IDC_LOAD_DOOR_MODEL,5,75,42,14
    PUSHBUTTON      "Lock",IDC_LOCK_DOOR,180,15,20,15
    PUSHBUTTON      "Check In",IDC_CHECKIN_DOOR,205,15,35,15
    PUSHBUTTON      "<< Previous",IDC_DOOR_PREV,10,40,45,14
    PUSHBUTTON      "Next >>",IDC_DOOR_NEXT,60,40,30,14
    PUSHBUTTON      "Copy Door Info",IDC_BUTTON14,195,220,56,14
    PUSHBUTTON      "Paste Door Info",IDC_BUTTON28,195,240,56,14
    PUSHBUTTON      "Doors checked out",IDC_DOORS_OUT,245,15,65,15
    GROUPBOX        "Sounds",IDC_STATIC,5,175,165,80
    GROUPBOX        "Timings",IDC_STATIC,5,105,105,65
    GROUPBOX        "Properties",IDC_STATIC,115,105,55,65
    CONTROL         "See-through",IDC_TRANSPARENCY,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,120,120,47,10
    EDITTEXT        IDC_SCRIPTNAME,177,170,90,12,ES_AUTOHSCROLL
    PUSHBUTTON      "Change Name",IDC_DOOR_CHANGE_NAME,120,50,50,15
    GROUPBOX        "Current Door",IDC_STATIC,5,10,105,55
    DEFPUSHBUTTON   "Done",IDOK,270,240,40,14
    PUSHBUTTON      "Browse",IDC_BROWSE,269,169,38,14
    GROUPBOX        "Default Script Module",IDC_STATIC,173,161,137,24
    CONTROL         "Blastable",IDC_DOOR_BLASTABLE,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,120,130,47,9
    LTEXT           "Hit Points:",IDC_DOOR_HITPOINTS_TEXT,120,143,35,8
    EDITTEXT        IDC_DOOR_HITPOINTS_EDIT,120,152,45,12,ES_AUTOHSCROLL
END

IDD_WORLDOBJECTSLIGHT DIALOG DISCARDABLE  0, 0, 334, 297
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "WorldObjects >> Lights"
FONT 8, "MS Sans Serif"
BEGIN
    PUSHBUTTON      "Add New Light",IDC_BUTTON1,15,25,50,14
    PUSHBUTTON      "Delete Light",IDC_BUTTON2,69,25,50,14
    PUSHBUTTON      "Edit Light",IDC_BUTTON3,123,25,50,14
    CONTROL         "Track on Player",IDC_CHECK1,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,17,146,66,10
    CONTROL         "Track on Robots",IDC_CHECK7,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,17,161,69,10
    GROUPBOX        "Light Object Options",IDC_STATIC,4,10,326,283
    PUSHBUTTON      "Lock",IDC_BUTTON5,177,25,21,14
    PUSHBUTTON      "Check In",IDC_BUTTON7,202,25,34,14
    PUSHBUTTON      "<< Previous",IDC_BUTTON8,240,25,42,14
    PUSHBUTTON      "Next >>",IDC_BUTTON10,286,25,30,14
    LTEXT           "New Light ID",IDC_STATIC,15,47,42,8
    EDITTEXT        IDC_EDIT24,73,47,104,12,ES_AUTOHSCROLL
    LTEXT           "H",IDC_STATIC,49,193,8,8
    EDITTEXT        IDC_EDIT33,23,193,20,12,ES_AUTOHSCROLL
    CONTROL         "",IDC_STATIC,"Static",SS_BLACKFRAME,65,193,100,9
    LTEXT           "H",IDC_STATIC,49,207,8,8
    CONTROL         "",IDC_STATIC,"Static",SS_BLACKFRAME,65,207,100,9
    EDITTEXT        IDC_EDIT34,23,207,20,12,ES_AUTOHSCROLL
    LTEXT           "H",IDC_STATIC,49,220,8,8
    EDITTEXT        IDC_EDIT14,23,220,20,12,ES_AUTOHSCROLL
    CONTROL         "",IDC_STATIC,"Static",SS_BLACKFRAME,65,220,100,9
    GROUPBOX        "Cast Light Color",IDC_STATIC,17,177,190,90
    CONTROL         "",IDC_STATIC,"Static",SS_BLACKFRAME,23,244,150,17
    CONTROL         "",IDC_STATIC,"Static",SS_BLACKFRAME,173,193,30,30
    PUSHBUTTON      "Copy Light Info",IDC_BUTTON14,264,269,56,14
    PUSHBUTTON      "Paste Light Info",IDC_BUTTON23,204,269,56,14
    PUSHBUTTON      "Load Model",IDC_BUTTON37,15,65,42,14
    EDITTEXT        IDC_EDIT25,73,65,104,12,ES_AUTOHSCROLL
    LTEXT           "Object",IDC_STATIC,16,89,22,8
    LTEXT           "Model Name",IDC_STATIC,16,108,41,8
    COMBOBOX        IDC_COMBO20,73,108,104,30,CBS_DROPDOWN | CBS_SORT | 
                    WS_VSCROLL | WS_TABSTOP
    COMBOBOX        IDC_COMBO21,73,89,104,30,CBS_DROPDOWN | CBS_SORT | 
                    WS_VSCROLL | WS_TABSTOP
    LTEXT           "Light ID",IDC_STATIC,16,127,26,8
    COMBOBOX        IDC_COMBO22,73,127,104,30,CBS_DROPDOWN | CBS_SORT | 
                    WS_VSCROLL | WS_TABSTOP
END

IDD_WORLDOBJECTSPLAYER DIALOG DISCARDABLE  0, 0, 402, 231
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "WorldObjects >> Player"
FONT 8, "MS Sans Serif"
BEGIN
    GROUPBOX        "Player Ship Options",IDC_STATIC,7,7,383,218
    PUSHBUTTON      "Add New Ship",IDC_ADD_PSHIP,15,20,56,14
    PUSHBUTTON      "Delete Player",IDC_PSHIP_DELETE,76,20,50,14
    EDITTEXT        IDC_PSHIP_NAME_EDIT,70,40,104,12,ES_AUTOHSCROLL | 
                    ES_WANTRETURN
    LTEXT           "Ship name",IDC_STATIC,15,42,46,8
    PUSHBUTTON      "Lock",IDC_PSHIP_LOCK,130,20,21,14
    PUSHBUTTON      "Check In",IDC_PSHIP_CHECKIN,156,20,34,14
    PUSHBUTTON      "<< Previous",IDC_PSHIP_PREV,195,20,42,14
    PUSHBUTTON      "Next >>",IDC_PSHIP_NEXT,242,20,30,14
    PUSHBUTTON      "Player ships checked out",IDC_PSHIPS_OUT,275,20,85,15
    COMBOBOX        IDC_PSHIP_PULLDOWN,71,175,105,55,CBS_DROPDOWN | CBS_SORT | 
                    WS_VSCROLL | WS_TABSTOP
    LTEXT           "Player ships",IDC_STATIC,15,175,45,10
    CONTROL         "",IDC_SHIPVIEW,"Static",SS_ETCHEDFRAME,252,44,108,97
    PUSHBUTTON      "Edit Physics",IDC_PSHIP_EDIT_PHYSICS,15,200,65,15
    EDITTEXT        IDC_PSHIP_DYING_MODEL_NAME_EDIT,70,115,104,12,
                    ES_AUTOHSCROLL | ES_WANTRETURN
    PUSHBUTTON      "Dying Model",IDC_PSHIP_DYING_MODEL,15,115,50,14
    PUSHBUTTON      "Set no dying model",IDC_NULL_DYING,180,115,65,15
    PUSHBUTTON      "Edit Weapon Batteries",IDC_EDIT_WEAPONS,90,200,95,15
    EDITTEXT        IDC_PSHIP_COCKPIT_EDIT,70,135,104,12,ES_AUTOHSCROLL | 
                    ES_WANTRETURN
    PUSHBUTTON      "Cockpit",IDC_PSHIP_COCKPIT,15,135,50,14
    LTEXT           "Armor scalar",IDC_STATIC,315,180,40,10
    EDITTEXT        IDC_SHIP_ARMOR_EDIT,360,180,25,12,ES_AUTOHSCROLL
    CONTROL         "Allowed By Default",IDC_DEFAULTALLOW,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,310,200,78,10
    EDITTEXT        IDC_PSHIP_MODEL_NAME_EDIT,45,86,95,12,ES_AUTOHSCROLL | 
                    ES_READONLY
    PUSHBUTTON      "Change Model",IDC_PSHIP_LOAD_MODEL,155,80,50,10
    LTEXT           "Model",IDC_STATIC,20,86,20,10
    GROUPBOX        "Models",IDC_STATIC,15,60,195,45
    CONTROL         "Hi-res",IDC_HIRES_RADIO,"Button",BS_AUTORADIOBUTTON,20,
                    70,35,10
    CONTROL         "Med-res",IDC_MEDRES_RADIO,"Button",BS_AUTORADIOBUTTON,
                    60,70,40,10
    CONTROL         "Lo-res",IDC_LORES_RADIO,"Button",BS_AUTORADIOBUTTON,105,
                    70,35,10
    PUSHBUTTON      "Set none",IDC_NOLOD,155,90,50,10
    EDITTEXT        IDC_LOD_DISTANCE_EDIT,180,67,25,12,ES_AUTOHSCROLL
    LTEXT           "LOD Dist",IDC_STATIC,147,66,30,10
END

IDD_WORLDTEXTURES DIALOG DISCARDABLE  0, 0, 617, 391
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "World >> Textures"
FONT 8, "MS Sans Serif"
BEGIN
    PUSHBUTTON      "Add 128x128 Texture",IDC_WTEXDLG_ADDNEW,15,30,76,13
    GROUPBOX        "Texture Options",IDC_STATIC,0,7,608,377
    PUSHBUTTON      "Delete Texture",IDC_DELETE,94,22,52,14
    PUSHBUTTON      "Lock",IDC_LOCK,149,22,21,14
    PUSHBUTTON      "Check In",IDC_CHECKIN,173,22,34,14
    PUSHBUTTON      "<< Prev",IDC_PREVIOUS,210,22,42,14
    PUSHBUTTON      "Next >>",IDC_NEXT,255,22,30,14
    LTEXT           "Reflectivity",IDC_STATIC,374,185,36,8
    EDITTEXT        IDC_REFLECT,411,184,30,12,ES_AUTOHSCROLL | ES_WANTRETURN
    LTEXT           "Texture Name:",IDC_STATIC,15,75,50,8
    LTEXT           "Damage",IDC_STATIC,382,201,28,8
    EDITTEXT        IDC_DAMAGE,411,199,30,12,ES_AUTOHSCROLL | ES_WANTRETURN
    LTEXT           "Slide",IDC_STATIC,392,215,16,8
    EDITTEXT        IDC_SLIDEU,411,214,20,12,ES_AUTOHSCROLL | ES_WANTRETURN
    EDITTEXT        IDC_SLIDEV,447,214,19,12,ES_AUTOHSCROLL | ES_WANTRETURN
    LTEXT           "U",IDC_STATIC,416,227,8,8
    LTEXT           "V",IDC_STATIC,452,227,8,8
    CONTROL         "Volatile",IDC_VOLATILE,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,374,256,39,10
    CONTROL         "Water",IDC_WATER,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,
                    374,246,35,10
    CONTROL         "Forcefield",IDC_FORCEFIELD,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,426,275,47,10
    CONTROL         "Object",IDC_OBJECT_TEXTURE,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,380,320,36,10
    CONTROL         "Terrain",IDC_TERRAIN_TEXTURE,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,380,330,65,10
    CONTROL         "Fly Thru",IDC_TEXTURE_FLY_THRU_CHECK,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,426,266,41,10
    CONTROL         "Mine",IDC_MINE_TEXTURE,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,380,310,60,10
    COMBOBOX        IDC_TEX_LIST,70,75,110,143,CBS_DROPDOWN | CBS_SORT | 
                    WS_VSCROLL | WS_TABSTOP
    EDITTEXT        IDC_BITMAP_NAME,283,40,87,12,ES_AUTOHSCROLL | 
                    ES_WANTRETURN
    PUSHBUTTON      "Load Bitmap",IDC_LOAD_BITMAP,200,40,80,14
    CONTROL         "",IDC_CHECK_ANIMATE,"Button",BS_AUTOCHECKBOX | 
                    WS_DISABLED | WS_TABSTOP,200,95,8,10
    CONTROL         "",IDC_TEXTUREVIEW,"Static",SS_ETCHEDFRAME,374,15,102,85
    PUSHBUTTON      "Textures Checked Out",IDC_RCS_STATUS,293,22,75,14
    CONTROL         "Destroyable",IDC_CHECK_DESTROY,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,200,65,50,9
    CONTROL         "",IDC_DESTROYVIEW,"Static",SS_ETCHEDFRAME,485,15,102,85
    LTEXT           "Texture Number",IDC_TEX_STRING,15,90,50,10
    EDITTEXT        IDC_TEX_NUM,70,90,50,12,ES_AUTOHSCROLL | ES_WANTRETURN
    CONTROL         "Mipped",IDC_CHECK_MIPPED,"Button",BS_AUTOCHECKBOX | 
                    WS_DISABLED | WS_TABSTOP,200,85,8,10
    LTEXT           "Animated",IDC_STATIC,209,95,35,10
    LTEXT           "Destroyed Image",IDC_STATIC,510,100,60,10
    LTEXT           "Normal Image",IDC_STATIC,400,100,45,10
    PUSHBUTTON      "Goto texture",IDC_TEXTURE_CURRENT,25,115,50,15
    CONTROL         "",IDC_TEXTURE_LISTBOX,"Static",SS_BLACKFRAME,35,135,335,
                    240
    SCROLLBAR       IDC_TEXLIST_VERT,25,135,10,240,SBS_VERT
    LTEXT           "Mipped",IDC_STATIC,209,85,30,10
    PUSHBUTTON      "Save as TGA",IDC_SAVE_AS_TGA,315,75,55,20,BS_MULTILINE
    EDITTEXT        IDC_RED_LIGHTING,412,116,29,12,ES_AUTOHSCROLL
    EDITTEXT        IDC_GREEN_LIGHTING,412,129,30,12,ES_AUTOHSCROLL
    EDITTEXT        IDC_BLUE_LIGHTING,412,143,30,12,ES_AUTOHSCROLL
    LTEXT           "Red",IDC_STATIC,394,118,17,10
    LTEXT           "Green",IDC_STATIC,388,131,21,10
    LTEXT           "Blue",IDC_STATIC,392,143,17,10
    CONTROL         "",IDC_CHECK_TMAP2,"Button",BS_AUTOCHECKBOX | 
                    WS_DISABLED | WS_TABSTOP,200,105,8,10
    LTEXT           "Tmap2",IDC_STATIC,209,105,35,10
    PUSHBUTTON      "Add 64x64 Texture",IDC_ADD_NEW_SMALL,15,45,75,11
    PUSHBUTTON      "Add 32x32 Texture",IDC_ADD_NEW_TINY,15,60,75,13
    PUSHBUTTON      "Size to 128",IDC_CONVERT_TO_NORMAL,260,75,50,15,
                    BS_MULTILINE
    PUSHBUTTON      "Size to 64",IDC_CONVERT_TO_SMALL,260,95,50,15
    PUSHBUTTON      "Size to 32",IDC_CONVERT_TO_TINY,260,115,50,15
    LTEXT           "Width:",IDC_TEXTURE_WIDTH,95,40,45,10
    LTEXT           "Height:",IDC_TEXTURE_HEIGHT,95,50,45,10
    LTEXT           "Alpha",IDC_STATIC,388,158,21,8
    EDITTEXT        IDC_ALPHA_EDIT,412,158,30,12,ES_AUTOHSCROLL | 
                    ES_WANTRETURN
    CONTROL         "Ping pong",IDC_PING_PONG,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,200,115,50,10
    DEFPUSHBUTTON   "Done",IDOK,560,360,40,14
    PUSHBUTTON      "Change Name",IDC_TEXTURE_CHANGE_NAME,130,90,50,15
    LTEXT           "Destroyed texture:",IDC_STATIC,205,55,60,8
    COMBOBOX        IDC_DESTROY_PULLDOWN,270,55,100,143,CBS_DROPDOWN | 
                    CBS_SORT | WS_VSCROLL | WS_TABSTOP
    CONTROL         "Light",IDC_LIGHT_CHECK,"Button",BS_AUTOCHECKBOX | 
                    WS_DISABLED | WS_TABSTOP,426,246,31,10
    COMBOBOX        IDC_TEXTURE_AMBIENT_SOUND_PULLDOWN,510,305,87,79,
                    CBS_DROPDOWN | CBS_SORT | WS_VSCROLL | WS_TABSTOP
    GROUPBOX        "Ambient Sound Info",IDC_STATIC,505,295,95,43
    CONTROL         "Saturate",IDC_SATURATE,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,426,256,43,10
    PUSHBUTTON      "Emergency override to unlock",IDC_OVERRIDE,315,100,55,
                    25,BS_MULTILINE
    CONTROL         "",IDC_TEXTURE_SWATCH,"Static",SS_BLACKFRAME,446,119,30,
                    35
    LTEXT           "Speed",IDC_STATIC,386,172,25,8
    EDITTEXT        IDC_SPEED_EDIT,412,170,30,12,ES_AUTOHSCROLL | 
                    ES_WANTRETURN
    CONTROL         "",IDC_BUMPVIEW,"Static",SS_ETCHEDFRAME,495,125,97,85
    GROUPBOX        "Bump map/Specular map",IDC_STATIC,485,115,115,145
    PUSHBUTTON      "Import PCX Bump",IDC_IMPORT_BUMP,495,235,40,20,
                    BS_MULTILINE
    PUSHBUTTON      "Generate Omni-bumps",IDC_GENERATE_OMNI,540,235,50,20,
                    BS_MULTILINE
    CONTROL         "Src",IDC_SRC_BUMP,"Button",BS_AUTORADIOBUTTON,495,210,
                    23,10
    CONTROL         "N",IDC_BUMP1,"Button",BS_AUTORADIOBUTTON,534,210,18,10
    CONTROL         "E",IDC_BUMP2,"Button",BS_AUTORADIOBUTTON,574,210,18,10
    CONTROL         "S",IDC_BUMP3,"Button",BS_AUTORADIOBUTTON,534,222,18,10
    CONTROL         "W",IDC_BUMP4,"Button",BS_AUTORADIOBUTTON,574,222,22,10
    CONTROL         "Metal",IDC_METAL_CHECK,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,374,276,33,10
    CONTROL         "Plastic",IDC_PLASTIC_CHECK,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,374,286,37,10
    CONTROL         "Marble",IDC_MARBLE_CHECK,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,374,266,37,10
    GROUPBOX        "Texture Type",IDC_STATIC,375,300,75,75
    CONTROL         "Effect",IDC_EFFECT_TEXTURE,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,380,340,60,10
    CONTROL         "HUD / Cockput",IDC_HUD_COCKPIT_TEXTURE,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,380,350,65,10
    CONTROL         "Light",IDC_LIGHT_TEXTURE,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,380,360,65,10
    CONTROL         "Procedural",IDC_PROCEDURAL_CHECK,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,455,335,50,10
    PUSHBUTTON      "Edit Procedural",IDC_PROCEDURAL_BUTTON,455,360,70,15
    CONTROL         "Water procedural",IDC_WATER_PROCEDURAL_CHECK,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,455,345,70,10
    GROUPBOX        "Corona style",IDC_STATIC,485,265,115,25
    CONTROL         "A",IDC_CORONA_A,"Button",BS_AUTORADIOBUTTON,490,275,20,
                    10
    CONTROL         "B",IDC_CORONA_B,"Button",BS_AUTORADIOBUTTON,515,275,20,
                    10
    CONTROL         "C",IDC_CORONA_C,"Button",BS_AUTORADIOBUTTON,540,275,20,
                    10
    CONTROL         "D",IDC_CORONA_D,"Button",BS_AUTORADIOBUTTON,565,275,20,
                    10
    CONTROL         "Force lightmap",IDC_FORCE_LIGHTMAPS,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,426,286,66,10
    PUSHBUTTON      "Multiple Check Ins",IDC_MULTICHECKTEX,150,40,45,20,
                    BS_MULTILINE
    CONTROL         "Import as 4444",IDC_IMPORT_4444,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,95,60,65,8
    CONTROL         "4444",IDC_4444_CHECK,"Button",BS_AUTOCHECKBOX | 
                    WS_DISABLED | WS_TABSTOP,200,125,49,9
    CONTROL         "Saturate lightmap",IDC_SATURATE_LIGHTMAP,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,426,296,70,8
    PUSHBUTTON      "Add 256x256 Texture",IDC_ADD_NEW_HUGE,15,15,75,13
    CONTROL         "Lava",IDC_LAVA_CHECKBOX,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,374,236,32,10
    CONTROL         "Rubble",IDC_RUBBLE_CHECKBOX,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,426,236,39,10
    LTEXT           "Volume",IDC_STATIC,542,322,30,8
    EDITTEXT        IDC_TEXTURE_AMBIENT_SOUND_VOLUME,572,322,23,10,
                    ES_AUTOHSCROLL
    CONTROL         "Smooth specular",IDC_SMOOTH_SPEC_CHECK,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,123,123,67,10
    PUSHBUTTON      "Select",IDC_AMBIENT_SOUND_SELECT,512,322,27,11
    CONTROL         "Breakable",IDC_CHECK_BREAKABLE,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,200,75,50,9
END

IDD_WORLDWEAPONS DIALOG DISCARDABLE  0, 0, 547, 393
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "World >> Weapons"
FONT 7, "Small Fonts"
BEGIN
    PUSHBUTTON      "Add New",IDC_ADD_WEAPON,94,5,40,15
    PUSHBUTTON      "Delete",IDC_DELETE_WEAPON,139,5,35,15
    PUSHBUTTON      "Lock",IDC_LOCK_WEAPON,179,5,21,15
    PUSHBUTTON      "Check In",IDC_CHECKIN_WEAPON,204,5,34,15
    PUSHBUTTON      "<< Previous",IDC_PREV_WEAPON,5,5,42,15
    PUSHBUTTON      "Next >>",IDC_NEXT_WEAPON,50,5,30,15
    COMBOBOX        IDC_WEAPON_PULLDOWN,65,30,105,72,CBS_DROPDOWN | CBS_SORT | 
                    WS_VSCROLL | WS_TABSTOP
    LTEXT           "Impact Size",IDC_STATIC,185,45,55,8
    EDITTEXT        IDC_WEAPON_BLOB_SIZE_EDIT,62,99,40,12,ES_AUTOHSCROLL | 
                    ES_WANTRETURN
    LTEXT           "Wall Hit",IDC_STATIC,300,35,30,8
    LTEXT           "Fire",IDC_STATIC,300,65,20,8,WS_DISABLED
    LTEXT           "Bounce",IDC_STATIC,300,50,35,8
    COMBOBOX        IDC_WEAPON_BOUNCE_SOUND_COMBO,340,50,104,30,CBS_DROPDOWN | 
                    CBS_SORT | WS_VSCROLL | WS_TABSTOP
    COMBOBOX        IDC_WEAPON_WALL_SOUND_PULLDOWN,340,35,104,78,
                    CBS_DROPDOWN | CBS_SORT | WS_VSCROLL | WS_TABSTOP
    COMBOBOX        IDC_FIRE_SOUND_PULLDOWN,340,65,104,60,CBS_DROPDOWN | 
                    CBS_SORT | WS_DISABLED | WS_VSCROLL | WS_TABSTOP
    RTEXT           "Blob Size",IDC_STATIC,22,99,35,8
    EDITTEXT        IDC_WEAPON_IMPACT_SIZE_EDIT,245,45,40,12,ES_AUTOHSCROLL | 
                    ES_WANTRETURN
    PUSHBUTTON      "Hud Image",IDC_LOAD_WEAPON_ANIM,5,60,55,15
    EDITTEXT        IDC_WEAPON_VCLIP_EDIT,65,60,104,13,ES_AUTOHSCROLL
    CONTROL         "",IDC_WEAPONVIEW,"Static",SS_ETCHEDFRAME,465,150,75,95
    PUSHBUTTON      "Weapons out",IDC_WEAPONS_OUT,250,5,62,15
    CONTROL         "",IDC_WEAPONFIREVIEW,"Static",SS_ETCHEDFRAME,465,35,75,
                    95
    LTEXT           "Hud Image",IDC_WEAPON_HUDVIEW,465,140,50,10
    LTEXT           "Discharge Image",IDC_WEAPON_FIRE_VIEW,465,25,55,10
    PUSHBUTTON      "Discharge File",IDC_LOAD_WEAPON_DISCHARGE,5,45,55,15
    EDITTEXT        IDC_WEAPON_DISCHARGE_EDIT,65,45,104,13,ES_AUTOHSCROLL
    RTEXT           "Thrust Time",IDC_STATIC,17,130,40,8
    EDITTEXT        IDC_WEAPON_THRUST_TIME_EDIT,62,129,40,12,ES_AUTOHSCROLL | 
                    ES_WANTRETURN
    RTEXT           "Life Time",IDC_STATIC,32,114,25,8
    EDITTEXT        IDC_WEAPON_LIFE_TIME_EDIT,62,114,40,12,ES_AUTOHSCROLL | 
                    ES_WANTRETURN
    LTEXT           "Spawns",IDC_STATIC,10,175,35,10
    EDITTEXT        IDC_WEAPON_SPAWN_EDIT,45,175,30,12,ES_AUTOHSCROLL
    LTEXT           "of",IDC_STATIC,16,191,8,10
    COMBOBOX        IDC_WEAPON_SPAWN_PULLDOWN,30,190,120,60,CBS_DROPDOWN | 
                    CBS_SORT | WS_VSCROLL | WS_TABSTOP
    CONTROL         "Homing",IDC_WEAPON_HOMING_CHECK,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,185,190,48,10
    CONTROL         "Use Parent Velocity",
                    IDC_WEAPON_USE_PARENT_VELOCITY_CHECK,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,185,120,80,10
    CONTROL         "Collide With Siblings",
                    IDC_WEAPON_COLLIDE_WITH_SIBLING_CHECK,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,185,130,85,10
    LTEXT           "Flying",IDC_STATIC,300,80,25,8,WS_DISABLED
    COMBOBOX        IDC_FLYING_SOUND_PULLDOWN,340,80,104,60,CBS_DROPDOWN | 
                    CBS_SORT | WS_DISABLED | WS_VSCROLL | WS_TABSTOP
    LTEXT           "Dmg",IDC_STATIC,185,30,17,8
    EDITTEXT        IDC_WEAPON_DAMAGE_EDIT,220,30,30,12,ES_AUTOHSCROLL | 
                    ES_WANTRETURN
    CONTROL         "Smoke trails",IDC_SMOKE_CHECK,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,184,255,55,10
    PUSHBUTTON      "Edit Physics",IDC_EDIT_PHYSICS,110,125,60,15
    GROUPBOX        "Sounds",IDC_STATIC,295,25,160,75
    PUSHBUTTON      "Change Name",IDC_CHANGE_NAME,5,30,55,15
    PUSHBUTTON      "Edit Lighting",IDC_WEAPON_LIGHT_INFO,110,145,60,15
    GROUPBOX        "Flags",IDC_STATIC,180,100,155,140
    CONTROL         "Energy weapon",IDC_ENERGY_RADIO,"Button",
                    BS_AUTORADIOBUTTON,350,120,70,10
    CONTROL         "Matter weapon",IDC_MATTER_RADIO,"Button",
                    BS_AUTORADIOBUTTON,350,130,70,10
    GROUPBOX        "Weapon type",IDC_STATIC,345,110,110,35
    CONTROL         "Electrical stream",IDC_ELECTRICAL_CHECK,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,185,140,70,10
    PUSHBUTTON      "Compute Size",IDC_DEFAULT_SIZE,110,105,60,15
    LTEXT           "Static",IDC_SIZE_MESSAGE,55,75,80,10
    LTEXT           "Impact Force",IDC_STATIC,185,75,60,8
    EDITTEXT        IDC_WEAPON_IMPACT_FORCE_EDIT,245,75,40,12,ES_AUTOHSCROLL | 
                    ES_WANTRETURN
    COMBOBOX        IDC_SMOKE_PULLDOWN,240,275,90,61,CBS_DROPDOWN | CBS_SORT | 
                    WS_VSCROLL | WS_TABSTOP
    LTEXT           "Smoke texture:",IDC_STATIC,184,276,50,10
    COMBOBOX        IDC_EXPLODE_PULLDOWN,350,210,100,61,CBS_DROPDOWN | 
                    CBS_SORT | WS_VSCROLL | WS_TABSTOP
    LTEXT           "Explode texture:",IDC_STATIC,350,200,55,8
    CONTROL         "Spray",IDC_SPRAY_CHECK,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,275,120,45,10
    CONTROL         "Streamers",IDC_INSTANT_CHECK,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,275,150,50,10
    CONTROL         "Invisible",IDC_INVISIBLE,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,275,140,45,8
    CONTROL         "Ring style",IDC_RING,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,275,130,50,10
    CONTROL         "Blast ring",IDC_EXPLODE_RING,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,350,160,50,10
    RTEXT           "Weapon Alpha",IDC_STATIC,7,84,50,8
    EDITTEXT        IDC_WEAPON_ALPHA_EDIT,62,84,40,12,ES_AUTOHSCROLL | 
                    ES_WANTRETURN
    CONTROL         "Saturate weapon",IDC_SATURATE_CHECK,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,185,151,70,10
    CONTROL         "Planar Blast",IDC_PLANAR_BLAST,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,350,170,55,10
    CONTROL         "Planar",IDC_PLANAR_CHECK,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,275,160,55,10
    LTEXT           "Explode Size",IDC_STATIC,350,231,45,10
    EDITTEXT        IDC_EXPLODE_SIZE_EDIT,400,231,30,12,ES_AUTOHSCROLL
    GROUPBOX        "Explosion",IDC_STATIC,345,150,110,170
    LTEXT           "Explode Time",IDC_STATIC,350,245,45,10
    EDITTEXT        IDC_EXPLODE_TIME_EDIT,400,245,30,12,ES_AUTOHSCROLL
    GROUPBOX        "Smoke",IDC_STATIC,180,245,155,45
    CONTROL         "Enable camera",IDC_ENABLE_CAMERA,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,185,161,75,10
    GROUPBOX        "Spawned weapons",IDC_STATIC,5,165,170,90
    CONTROL         "On Impact",IDC_SPAWNS_ON_IMPACT,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,10,225,50,10
    CONTROL         "On Timeout",IDC_SPAWNS_ON_TIMEOUT,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,10,240,55,10
    PUSHBUTTON      "Override to unlocked",IDC_OVERRIDE,320,5,80,15
    GROUPBOX        "Particles",IDC_STATIC,180,290,155,60
    EDITTEXT        IDC_PARTICLE_LIFE_EDIT,295,315,30,12,ES_AUTOHSCROLL
    EDITTEXT        IDC_PARTICLE_SIZE_EDIT,295,329,30,12,ES_AUTOHSCROLL
    EDITTEXT        IDC_PARTICLE_COUNT_EDIT,215,330,30,12,ES_AUTOHSCROLL
    LTEXT           "Count",IDC_STATIC,185,330,25,10
    LTEXT           "Life",IDC_STATIC,265,315,25,10
    LTEXT           "Size",IDC_STATIC,265,330,25,10
    COMBOBOX        IDC_PARTICLE_PULLDOWN,185,315,75,84,CBS_DROPDOWN | 
                    CBS_SORT | WS_VSCROLL | WS_TABSTOP
    LTEXT           "Particle Texture",IDC_STATIC,185,305,65,8
    PUSHBUTTON      "Copy",IDC_WEAPON_COPY_BUTTON,405,5,45,15
    PUSHBUTTON      "Paste",IDC_WEAPON_PASTE_BUTTON,455,5,45,15
    CONTROL         "Expand",IDC_EXPANDING_CHECK,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,405,160,45,8
    CONTROL         "Muzzle flash",IDC_MUZZLE_FLASH,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,275,170,55,10
    LTEXT           "To change explosion lighting, click on the ""Edit Lighting"" button and change the ""End Color"" fields.",
                    IDC_STATIC,355,260,85,45
    CONTROL         "Microwave",IDC_MICROWAVE,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,275,180,55,10
    CONTROL         "Napalm effect",IDC_NAPALM,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,185,171,75,10
    LTEXT           "IDmg",IDC_STATIC,185,90,20,8
    EDITTEXT        IDC_WEAPON_IMPACT_DAMAGE_EDIT,220,90,30,12,
                    ES_AUTOHSCROLL | ES_WANTRETURN
    EDITTEXT        IDC_WEAPON_IMPACT_TIME_EDIT2,245,60,40,12,ES_AUTOHSCROLL | 
                    ES_WANTRETURN
    LTEXT           "Impact Time",IDC_STATIC,185,60,60,8
    COMBOBOX        IDC_SCORCH_PULLDOWN,45,271,120,80,CBS_DROPDOWN | 
                    CBS_SORT | WS_VSCROLL | WS_TABSTOP
    LTEXT           "Texture:",IDC_STATIC,10,271,30,10
    GROUPBOX        "Scorch Mark/Bullet Hole",IDC_STATIC,5,260,170,45
    GROUPBOX        "Icon Image",IDC_STATIC,5,311,170,30
    COMBOBOX        IDC_SMALLIMG_PULLDOWN,45,321,120,70,CBS_DROPDOWN | 
                    CBS_SORT | WS_VSCROLL | WS_TABSTOP
    LTEXT           "Texture",IDC_STATIC,10,321,30,10
    EDITTEXT        IDC_WEAPON_SCORCH_SIZE_EDIT,45,285,40,12,ES_AUTOHSCROLL | 
                    ES_WANTRETURN
    LTEXT           "Size",IDC_STATIC,10,285,30,15
    LTEXT           "or",IDC_STATIC,155,190,15,10
    COMBOBOX        IDC_SPAWN2_PULLDOWN,30,205,120,15,CBS_DROPDOWN | 
                    CBS_SORT | WS_VSCROLL | WS_TABSTOP
    LTEXT           "Alternate chance",IDC_STATIC,70,225,60,10
    EDITTEXT        IDC_ALTERNATE_CHANCE_EDIT,130,225,20,12,ES_AUTOHSCROLL
    LTEXT           "%",IDC_STATIC,152,228,10,13
    CONTROL         "Reverse growth",IDC_REVERSE_SMOKE_CHECK,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,255,255,70,10
    GROUPBOX        "Gravity Field",IDC_STATIC,180,350,155,40
    CONTROL         "Use Gravity Field",IDC_GRAVITY_FIELD_CHECK,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,185,360,70,10
    LTEXT           "Time",IDC_STATIC,265,360,20,10
    LTEXT           "Size",IDC_STATIC,265,370,25,10
    EDITTEXT        IDC_GRAVITY_TIME,295,360,25,12,ES_AUTOHSCROLL
    EDITTEXT        IDC_GRAVITY_SIZE,295,370,25,12,ES_AUTOHSCROLL
    GROUPBOX        "Countermeasures",IDC_STATIC,345,320,110,55
    CONTROL         "Is countermeasure",IDC_COUNTERMEASURE_CHECK,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,350,335,80,8
    CONTROL         "Spawns robot on death",IDC_SPAWNS_ROBOT_CHECK,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,350,345,100,8
    COMBOBOX        IDC_SPAWN_ROBOT_PULLDOWN,350,354,99,84,CBS_DROPDOWN | 
                    CBS_SORT | WS_VSCROLL | WS_TABSTOP
    GROUPBOX        "Terrain damage",IDC_STATIC,5,345,170,45
    CONTROL         "Freeze",IDC_FREEZE_CHECK,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,275,110,55,10
    LTEXT           "Depth",IDC_STATIC,10,375,25,8
    LTEXT           "Size",IDC_STATIC,10,360,25,8
    EDITTEXT        IDC_TERRIAN_DAMAGE_SIZE,34,357,35,12,ES_AUTOHSCROLL
    EDITTEXT        IDC_TERRAIN_DAMAGE_DEPTH,34,374,35,12,ES_AUTOHSCROLL
    CONTROL         "Homing split",IDC_HOMED_SPLIT_CHECK,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,90,240,75,10
    CONTROL         "Silent Homing",IDC_SILENT_HOMING_CHECK,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,185,200,75,10
    LTEXT           "Homing FOV",IDC_STATIC,185,210,50,8
    EDITTEXT        IDC_HOMING_FOV_TEXT,275,210,45,10,ES_AUTOHSCROLL
    CONTROL         "No rotate",IDC_NO_ROTATE_CHECK,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,275,190,55,10
    CONTROL         "Use custom size:",IDC_CUSTOM_SIZE_CHECK,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,185,225,75,10
    EDITTEXT        IDC_CUSTOM_SIZE_EDIT,275,225,45,10,ES_AUTOHSCROLL
    CONTROL         "Elec. Tracer",IDC_ELECTIC_TRACER_CHECK,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,275,200,55,10
    CONTROL         "Timeout like wallhit",IDC_TIMEOUT_WALL_CHECK,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,350,180,80,10
    EDITTEXT        IDC_WEAPON_GENERIC_DAMAGE_EDIT,265,30,30,12,
                    ES_AUTOHSCROLL | ES_WANTRETURN
    LTEXT           "P",IDC_STATIC,210,30,8,10
    LTEXT           "G",IDC_STATIC,255,30,8,10
    LTEXT           "P",IDC_STATIC,210,90,8,10
    EDITTEXT        IDC_WEAPON_IMPACT_GENERIC_DAMAGE_EDIT,266,90,29,12,
                    ES_AUTOHSCROLL | ES_WANTRETURN
    LTEXT           "G",IDC_STATIC,255,90,8,10
    CONTROL         "Planar smoke",IDC_PLANAR_SMOKE_CHECK,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,255,265,67,10
    RTEXT           "Recoil Force",IDC_STATIC,7,145,50,8
    EDITTEXT        IDC_WEAPON_RECOIL_FORCE_EDIT,63,144,40,12,ES_AUTOHSCROLL | 
                    ES_WANTRETURN
END

IDD_FONT_DIALOG DIALOG DISCARDABLE  0, 0, 452, 316
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "Subeditors >> Font"
FONT 8, "MS Sans Serif"
BEGIN
    DEFPUSHBUTTON   "OK",IDOK,395,5,50,15
    PUSHBUTTON      "Cancel",IDCANCEL,395,25,50,15
    CONTROL         "",IDC_STATIC_FONTVIEW,"Static",SS_BLACKFRAME,10,5,365,
                    230
    EDITTEXT        IDC_EDIT_MINASCII,35,260,20,15,ES_AUTOHSCROLL | 
                    ES_NUMBER
    GROUPBOX        "Properties",IDC_STATIC,10,245,195,60
    LTEXT           "Min",IDC_STATIC,15,260,20,10
    GROUPBOX        "Untitled",IDC_STATIC_FONTNAME,210,245,165,60
    LTEXT           "Name",IDC_STATIC,215,260,25,10
    EDITTEXT        IDC_EDIT_FONTNAME,240,260,125,12,ES_AUTOHSCROLL
    PUSHBUTTON      "Open...",IDC_OPEN,255,285,35,15
    PUSHBUTTON      "Save",IDC_SAVE,295,285,35,15
    PUSHBUTTON      "Save as,,,",IDC_SAVEAS,335,285,35,15
    PUSHBUTTON      "New",IDC_NEW,395,70,50,15
    CONTROL         "Monochrome",IDC_PROP_MONO,"Button",BS_AUTORADIOBUTTON,
                    65,260,65,8
    CONTROL         "Bitmapped",IDC_PROP_BITMAPPED,"Button",
                    BS_AUTORADIOBUTTON,65,275,70,8
    LTEXT           "Max",IDC_STATIC,15,280,15,10
    LTEXT           "Static",IDC_STATIC_MAXASCII,35,280,20,8,NOT WS_GROUP
    LTEXT           "Static",IDC_STATIC_PROP,135,260,65,10
    PUSHBUTTON      "Lock",IDC_LOCK,395,90,50,15
    PUSHBUTTON      "Check in",IDC_CHECKIN,395,110,50,15
    PUSHBUTTON      "Delete",IDC_DELETE,395,130,50,15
    CONTROL         "Grayscale",IDC_PROP_GRADIANT,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,65,290,50,10
    EDITTEXT        IDC_BRIGHTNESS,180,285,20,15,ES_AUTOHSCROLL
    LTEXT           "Brighness",IDC_STATIC,135,290,40,10
END

IDD_WORLDSOUNDS DIALOGEX 0, 0, 427, 231
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "World >> Sounds"
FONT 8, "MS Sans Serif", 0, 0, 0x1
BEGIN
    GROUPBOX        "Sound Options",IDC_STATIC,10,10,410,218
    PUSHBUTTON      "Copy Sound Info",IDC_SOUNDCOPY_EDIT,345,210,64,14
    PUSHBUTTON      "Paste Sound Info",IDC_SOUNDPASTE_BUTTON,265,210,66,14
    PUSHBUTTON      "Add New",IDC_ADD_SOUND,135,20,35,15
    PUSHBUTTON      "Delete",IDC_DELETE_SOUND,135,40,35,15
    PUSHBUTTON      "Lock",IDC_LOCK_SOUND,175,20,35,14
    PUSHBUTTON      "Check In",IDC_CHECKIN_SOUND,175,40,34,15
    PUSHBUTTON      "<< Previous",IDC_PREV_SOUND,20,45,42,14
    PUSHBUTTON      "Next >>",IDC_NEXT_SOUND,65,45,30,14
    CONTROL         "Frequency shifts",IDC_SOUNDHALLEFFECT_CHECK,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,25,140,80,10,
                    WS_EX_TRANSPARENT
    CONTROL         "Always Plays",IDC_SOUNDFOREVER_CHECK,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,25,180,68,10
    GROUPBOX        "3D Sound Cone Info",IDC_STATIC,145,85,265,90
    LTEXT           "Max. Distance",IDC_STATIC,25,101,46,8
    LTEXT           "Min. Distance",IDC_STATIC,25,113,44,8
    LTEXT           "Inner Cone Degrees",IDC_STATIC,155,110,64,8
    LTEXT           "Outer Cone Degrees",IDC_STATIC,155,130,66,8
    LTEXT           "Outer Volume %",IDC_STATIC,155,150,62,8
    EDITTEXT        IDC_SOUNDMAXDIST_EDIT,75,100,40,12,ES_AUTOHSCROLL | 
                    ES_WANTRETURN
    EDITTEXT        IDC_SOUNDMINDIST_EDIT,75,115,40,12,ES_AUTOHSCROLL | 
                    ES_WANTRETURN
    EDITTEXT        IDC_SOUNDINNERCONEANGLE_EDIT,225,108,40,14,
                    ES_AUTOHSCROLL | ES_WANTRETURN
    EDITTEXT        IDC_SOUNDOUTERCONEANGLE_EDIT,225,130,40,14,
                    ES_AUTOHSCROLL | ES_WANTRETURN
    EDITTEXT        IDC_SOUNDOUTERCONEVOL_EDIT,225,150,40,14,ES_AUTOHSCROLL | 
                    ES_WANTRETURN
    LTEXT           "Current Sound",IDC_STATIC,20,20,50,8
    COMBOBOX        IDC_SOUND_PULLDOWN,20,30,110,74,CBS_DROPDOWN | CBS_SORT | 
                    WS_VSCROLL | WS_TABSTOP
    PUSHBUTTON      "Play Sound",IDC_PLAYSOUND,385,25,30,20,BS_MULTILINE
    CONTROL         "Looping",IDC_LOOPING_CHECK,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,265,190,41,10
    LTEXT           "Start Sample",IDC_STATIC,330,180,40,10
    LTEXT           "End Sample",IDC_STATIC,375,180,40,8
    EDITTEXT        IDC_SOUNDLOOPSTART_EDIT,330,190,36,15,ES_AUTOHSCROLL | 
                    ES_WANTRETURN
    EDITTEXT        IDC_SOUNDLOOPEND_EDIT,375,190,34,15,ES_AUTOHSCROLL | 
                    ES_WANTRETURN
    PUSHBUTTON      "New",IDC_LOAD_SOUND,340,30,30,14
    EDITTEXT        IDC_RAW_NAME_EDIT,230,30,104,14,ES_AUTOHSCROLL | 
                    ES_WANTRETURN
    GROUPBOX        "Orientation To Object",IDC_STATIC,275,105,125,57
    CONTROL         "Stays At Play Position",IDC_SOUNDPOSATTACH_RADIO,"Button",
                    BS_AUTORADIOBUTTON,25,151,95,10
    CONTROL         "Moves With Linked Object",IDC_SOUNDOBJATTACH_RADIO,
                    "Button",BS_AUTORADIOBUTTON,25,161,105,10
    CONTROL         "New Kills Old",IDC_SOUNDONCE_CHECK,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,25,200,60,10
    CONTROL         "Plays Exclusively",IDC_SOUNDEXCLUSIVE_CHECK,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,25,190,70,10
    CONTROL         "Object",IDC_SOUNDOBJECT_RADIO,"Button",
                    BS_AUTORADIOBUTTON | WS_GROUP,285,115,40,10
    CONTROL         "Turret 1",IDC_SOUNDTURRET1_RADIO,"Button",
                    BS_AUTORADIOBUTTON,285,125,50,10
    CONTROL         "Turret 2",IDC_SOUNDTURRET2_RADIO,"Button",
                    BS_AUTORADIOBUTTON,285,135,40,10
    CONTROL         "Turret 3",IDC_SOUNDTURRET3_RADIO,"Button",
                    BS_AUTORADIOBUTTON,285,145,55,10
    CONTROL         "Forward",IDC_SOUNDFORWARD_RADIO,"Button",
                    BS_AUTORADIOBUTTON | WS_GROUP,340,115,45,10
    CONTROL         "Backward",IDC_SOUNDBACKWARD_RADIO,"Button",
                    BS_AUTORADIOBUTTON,340,125,45,10
    CONTROL         "Upward",IDC_SOUNDUPWARD_RADIO,"Button",
                    BS_AUTORADIOBUTTON,340,135,40,10
    CONTROL         "Downward",IDC_SOUNDDOWNWARD_RADIO,"Button",
                    BS_AUTORADIOBUTTON,340,145,55,10
    LTEXT           "0 K - Total 0 K",IDC_SOUND_MEMORY_STATIC,230,50,135,10
    CONTROL         "Realtime 3D Updating",IDC_SOUND_NO_UPDATE,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,25,130,96,11
    CONTROL         "Sound Starts at Fire",IDC_RADIO3,"Button",
                    BS_AUTORADIOBUTTON,150,195,90,10
    CONTROL         "Sound Starts at Pre-fire",IDC_RADIO4,"Button",
                    BS_AUTORADIOBUTTON,150,205,90,10
    GROUPBOX        "Weapon Fire Info",IDC_STATIC,145,180,110,40
    GROUPBOX        "3D General Info",IDC_STATIC,20,85,120,90
    PUSHBUTTON      "Stop Sounds",IDC_KILLSOUNDS,385,50,30,20,BS_MULTILINE
    CONTROL         "One Instance Per Object",IDC_SOUND_ONCE_PER_OBJ_CHECK,
                    "Button",BS_AUTOCHECKBOX | WS_TABSTOP,25,211,90,10
    EDITTEXT        IDC_SOUND_IMPORT_VOLUME_EDIT,285,65,35,10,ES_AUTOHSCROLL
    LTEXT           "Import Volume %",IDC_STATIC,230,65,55,10
    PUSHBUTTON      "Change Name",IDC_SOUND_CHANGE_NAME,20,65,50,15
    GROUPBOX        "Sample",IDC_STATIC,220,20,160,60
    PUSHBUTTON      "Override to unlock",IDC_OVERRIDE,135,60,70,15
END

IDD_FLOATING_KEYPAD DIALOGEX 80, 0, 258, 214
STYLE WS_CHILD | WS_CLIPSIBLINGS | WS_CAPTION | WS_SYSMENU | WS_THICKFRAME
EXSTYLE WS_EX_TOOLWINDOW
CAPTION "Keypad"
FONT 7, "Small Fonts", 0, 0, 0x1
BEGIN
    CONTROL         "Tab1",IDC_FLOATING_KEYPAD_TAB,"SysTabControl32",0x0,5,5,
                    250,205
END

IDD_PREFERENCES DIALOG DISCARDABLE  0, 0, 307, 191
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "Editor preferences"
FONT 8, "MS Sans Serif"
BEGIN
    DEFPUSHBUTTON   "OK",IDOK,180,170,50,14
    PUSHBUTTON      "Cancel",IDCANCEL,250,170,50,14
    GROUPBOX        "Game Display Mode",IDC_STATIC,7,5,135,70
    GROUPBOX        "Hardware Acceleration",IDC_HARDWARE_ACC,7,78,135,77
    CONTROL         "Use Mip-mapping",IDC_MIPPING_CHECK,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,15,125,81,10
    CONTROL         "Use Bilinear filtering",IDC_BILINEAR_CHECK,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,15,135,75,10
    CONTROL         "Use OpenGL",IDC_USE_OPENGL,"Button",BS_AUTORADIOBUTTON,
                    15,90,86,10
    CONTROL         "Use Glide (3Dfx Voodoo Cards)",IDC_USE_GLIDE,"Button",
                    BS_AUTORADIOBUTTON,15,100,120,10
    CONTROL         "Windowed",IDC_WINDOWED,"Button",BS_AUTORADIOBUTTON | 
                    WS_GROUP,15,20,48,11
    CONTROL         "Full Screen w/ Software",IDC_FULLSCREEN_SW,"Button",
                    BS_AUTORADIOBUTTON,15,30,110,10
    CONTROL         "Full Screen w/ Hardware",IDC_FULLSCREEN_HW,"Button",
                    BS_AUTORADIOBUTTON,15,39,110,13
    CONTROL         "Debug in Fullscreen",IDC_DEBUG_CHECK,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,15,55,115,10
    COMBOBOX        IDC_DEFAULT_PILOT,220,130,65,55,CBS_DROPDOWN | CBS_SORT | 
                    WS_VSCROLL | WS_TABSTOP
    LTEXT           "Default Pilot",IDC_STATIC,165,130,39,8
    CONTROL         "Slider1",IDC_SLEWSLIDER,"msctls_trackbar32",
                    TBS_AUTOTICKS | TBS_BOTH | WS_TABSTOP,220,60,65,10
    LTEXT           "Key slewing",IDC_STATIC,165,60,50,10
    CONTROL         "Use Direct3D",IDC_USE_D3D,"Button",BS_AUTORADIOBUTTON,
                    15,110,110,10
    GROUPBOX        "Static",IDC_STATIC,155,5,135,150
    CONTROL         "Joystick",IDC_JOYENABLE,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,165,25,80,10
    CONTROL         "In Game Cinematics Enabled",IDC_IGC_ENABLED,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,165,111,107,10
END

IDD_HOGDIALOG DIALOG DISCARDABLE  0, 0, 457, 271
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "Hogmaker"
FONT 8, "MS Sans Serif"
BEGIN
    DEFPUSHBUTTON   "OK",IDOK,395,5,50,14
    PUSHBUTTON      "Cancel",IDCANCEL,395,25,50,14
    CONTROL         "List1",IDC_HOGLIST,"SysListView32",LVS_REPORT | 
                    LVS_ALIGNLEFT | LVS_NOSORTHEADER | WS_BORDER | 
                    WS_TABSTOP,10,5,375,250
    CONTROL         "Progress1",IDC_HOGPROGRESS,"msctls_progress32",
                    WS_BORDER,70,255,315,10
    LTEXT           "Loaded.",IDC_STATUSTEXT,10,255,55,10
END

IDD_DOORWAYKEYPAD DIALOG DISCARDABLE  0, 0, 205, 461
STYLE DS_CONTROL | WS_CHILD
FONT 7, "Small Fonts"
BEGIN
    GROUPBOX        "Doors",IDC_DOORWAY_GROUP,5,5,125,152
    CONTROL         "",IDC_DOORPIC,"Static",SS_BLACKFRAME,35,20,69,62
    PUSHBUTTON      "Place Door",IDC_DOORWAY_PLACEDOOR,35,115,70,14,BS_FLAT
    EDITTEXT        IDC_DOORWAYSELEDIT,15,90,60,15,ES_AUTOHSCROLL | 
                    ES_READONLY | ES_WANTRETURN
    PUSHBUTTON      "<<",IDC_PREV_DOOR,80,90,20,15,BS_FLAT
    PUSHBUTTON      ">>",IDC_NEXT_DOOR,105,90,20,15,BS_FLAT
    GROUPBOX        "Current Doorway",IDC_DOORWAY_CURRENT_BOX,7,167,127,168
    LTEXT           "Name:",IDC_DOORWAY_ID,12,182,119,10
    CONTROL         "Locked",IDC_DOORWAY_LOCKED,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,20,205,45,10
    LTEXT           "Keys",IDC_DOORWAY_KEYS_TEXT,86,197,20,10
    CONTROL         "1",IDC_KEY1_CHECK,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,
                    81,212,18,10
    CONTROL         "2",IDC_KEY2_CHECK,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,
                    81,222,18,10
    CONTROL         "3",IDC_KEY3_CHECK,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,
                    81,232,18,10
    CONTROL         "4",IDC_KEY4_CHECK,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,
                    81,242,18,10
    PUSHBUTTON      "Attach Door",IDC_DOORWAY_ATTACHDOOR,35,135,70,14,
                    BS_FLAT
    EDITTEXT        IDC_DOORWAY_POS_EDIT,71,287,56,15,ES_AUTOHSCROLL
    LTEXT           "Initial Position",IDC_DOORWAY_POS_TEXT,20,292,50,10
    CONTROL         "Auto-close",IDC_DOORWAY_AUTO,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,20,220,50,10
    LTEXT           "Hit Points",IDC_DOORWAY_HITPOINT_TEXT,21,318,43,9
    EDITTEXT        IDC_DOORWAY_HITPOINT_EDIT,71,313,56,15,ES_AUTOHSCROLL
    CONTROL         "5",IDC_KEY5_CHECK,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,
                    106,212,18,10
    CONTROL         "6",IDC_KEY6_CHECK,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,
                    106,222,18,10
    CONTROL         "7",IDC_KEY7_CHECK,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,
                    106,232,18,10
    CONTROL         "8",IDC_KEY8_CHECK,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,
                    106,242,20,10
    CONTROL         "Needs all selected keys",IDC_DOORWAY_KEY_ALL,"Button",
                    BS_AUTORADIOBUTTON,20,257,120,10
    CONTROL         "Needs only one selected key",IDC_DOORWAY_KEY_ONLY_ONE,
                    "Button",BS_AUTORADIOBUTTON,20,267,120,10
    CONTROL         "GB Ignores Locked",IDC_DOORWAY_GB_IGNORE_LOCKED,"Button",
                    BS_AUTOCHECKBOX | BS_MULTILINE | WS_TABSTOP,20,233,60,20
END

IDD_TERRAINKEYPAD DIALOG DISCARDABLE  0, 0, 210, 514
STYLE DS_CONTROL | WS_CHILD
FONT 7, "Small Fonts"
BEGIN
    PUSHBUTTON      "Min",IDC_TERRPAD_MAKE_MIN,10,75,40,20,BS_MULTILINE | 
                    BS_FLAT
    PUSHBUTTON      "Max",IDC_TERRPAD_MAKE_MAX,10,25,40,20,BS_MULTILINE | 
                    BS_FLAT
    PUSHBUTTON      "Zero",IDC_TERRPAD_MAKE_ZERO,10,50,40,20,BS_MULTILINE | 
                    BS_FLAT
    PUSHBUTTON      "Fill sel w/texture",IDC_TERRPAD_FILL_AREA,110,20,40,20,
                    BS_MULTILINE | BS_FLAT
    PUSHBUTTON      "Refresh wireframe map",IDC_TERRPAD_REDO_TOPMAP,155,55,
                    45,30,BS_MULTILINE | BS_FLAT
    PUSHBUTTON      "Pyramid from max",IDC_TERRPAD_PYRAMID,55,50,40,20,
                    BS_MULTILINE | BS_FLAT
    PUSHBUTTON      "Hill from max",IDC_TERRPAD_HILL,55,75,40,20,
                    BS_MULTILINE | BS_FLAT
    PUSHBUTTON      "Lower -10",IDC_TERRPAD_LOWER10,55,145,40,20,
                    BS_MULTILINE | BS_FLAT
    PUSHBUTTON      "Lower pts",IDC_TERRPAD_MOVE_DOWN,10,120,40,20,
                    BS_MULTILINE | BS_FLAT
    PUSHBUTTON      "Import PCX Map",IDC_TERRPAD_IMPORT,155,145,45,20,
                    BS_MULTILINE | BS_FLAT
    PUSHBUTTON      "Select None",IDC_TERRPAD_SELECT_NONE,110,120,40,20,
                    BS_MULTILINE | BS_FLAT
    PUSHBUTTON      "Normalize Terrain",IDC_TERRPAD_RENORMALIZE,155,90,45,20,
                    BS_MULTILINE | BS_FLAT
    PUSHBUTTON      "Rotate tex 90*",IDC_TERRPAD_ROT_TEXTURE,110,45,40,20,
                    BS_MULTILINE | BS_FLAT
    PUSHBUTTON      "House o' Pancakes",IDC_TERRPAD_PANCAKES,55,25,40,20,
                    BS_MULTILINE | BS_FLAT
    PUSHBUTTON      "Raise +10",IDC_TERRPAD_RAISE10,10,145,40,20,
                    BS_MULTILINE | BS_FLAT
    PUSHBUTTON      "Raise pts",IDC_TERRPAD_MOVE_UP,55,120,40,20,
                    BS_MULTILINE | BS_FLAT
    PUSHBUTTON      "Select range",IDC_TERRPAD_SELECTRANGE,155,115,45,20,
                    BS_MULTILINE | BS_FLAT
    GROUPBOX        "Simple Geometry",IDC_TEXTUREPAD_GROUP,5,10,95,90
    PUSHBUTTON      "Next 16 textures the sky texture",IDC_SKY_TEXTURE,10,
                    190,60,20,BS_MULTILINE | BS_FLAT
    PUSHBUTTON      "Make sky closer",IDC_SKY_NEARER,10,215,60,20,
                    BS_MULTILINE | BS_FLAT
    PUSHBUTTON      "Make sky farther",IDC_SKY_FARTHER,75,215,60,20,
                    BS_MULTILINE | BS_FLAT
    GROUPBOX        "Sky",IDC_STATIC,5,175,200,180
    CONTROL         "Stars",IDC_STARS_CHECK,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,140,195,34,10
    LTEXT           "Num of moons",IDC_NUM_MOONS_STATIC,10,240,55,10
    LTEXT           "Current moon",IDC_CUR_MOON_STATIC,75,240,50,10
    PUSHBUTTON      "Move sat closer",IDC_TERR_MOVE_MOON,10,270,60,20,
                    BS_MULTILINE | BS_FLAT
    PUSHBUTTON      "Randomize Sky",IDC_TERR_RANDOMIZE_SKY,75,190,60,20,
                    BS_FLAT
    PUSHBUTTON      "<<",IDC_TERR_LESS_MOONS,10,251,20,15
    PUSHBUTTON      ">>",IDC_TERR_MORE_MOONS,40,251,20,15
    PUSHBUTTON      "<<",IDC_TERR_PREV_MOON,75,251,20,15
    PUSHBUTTON      ">>",IDC_TERR_NEXT_MOON,106,251,20,15
    GROUPBOX        "Terrain/Mine integration",IDC_STATIC,5,360,200,40
    CONTROL         "Show terrain pts in mine view",IDC_SHOW_TERRAIN,"Button",
                    BS_AUTOCHECKBOX | BS_MULTILINE | WS_TABSTOP,105,370,85,
                    15
    CONTROL         "Flat shade and Far Fog",IDC_FLAT_SHADE_TERRAIN_CHECK,
                    "Button",BS_AUTOCHECKBOX | WS_TABSTOP,5,405,120,10
    PUSHBUTTON      "Calculate Terrain Occlusion",IDC_TERRAIN_OCCLUSION,155,
                    20,45,30,BS_MULTILINE | BS_FLAT
    PUSHBUTTON      "Select All Cells",IDC_TERR_SELECT_ALL,110,145,40,20,
                    BS_MULTILINE | BS_FLAT
    CONTROL         "Top Down view in texture window",IDC_TERRAIN_2D,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,5,415,120,10
    CONTROL         "Textured sky",IDC_TEXTURE_SKY,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,140,204,60,10
    LTEXT           "Color",IDC_STATIC,10,295,45,10
    LTEXT           "R",IDC_STATIC,10,309,10,10
    LTEXT           "G",IDC_STATIC,10,325,10,10
    LTEXT           "B",IDC_STATIC,10,340,10,10
    EDITTEXT        IDC_SKY_RED_EDIT,25,305,25,12,ES_AUTOHSCROLL
    EDITTEXT        IDC_SKY_GREEN_EDIT,25,321,25,12,ES_AUTOHSCROLL
    EDITTEXT        IDC_SKY_BLUE_EDIT,25,335,25,12,ES_AUTOHSCROLL
    CONTROL         "Don't use LOD engine",IDC_NO_LOD_ENGINE,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,5,425,120,10
    EDITTEXT        IDC_FOG_DISTANCE_EDIT,170,230,25,12,ES_AUTOHSCROLL
    LTEXT           "Visibility",IDC_STATIC,140,230,30,10
    EDITTEXT        IDC_PIXEL_ERROR_EDIT,165,405,35,12,ES_AUTOHSCROLL
    LTEXT           "Pixel Err.",IDC_STATIC,130,405,30,10
    GROUPBOX        "Manual Geometry",IDC_STATIC,5,105,95,65
    GROUPBOX        "Static",IDC_STATIC,105,10,100,160
    CONTROL         "Sky",IDC_SKY_RADIO,"Button",BS_AUTORADIOBUTTON,60,315,
                    45,8
    CONTROL         "Horizon",IDC_HORIZON_RADIO,"Button",BS_AUTORADIOBUTTON,
                    60,326,45,8
    CONTROL         "Fog",IDC_FOG_RADIO,"Button",BS_AUTORADIOBUTTON,60,337,
                    45,8
    CONTROL         "",IDC_SKY_SWATCH,"Static",SS_BLACKFRAME,110,313,25,10
    CONTROL         "",IDC_HORIZON_SWATCH,"Static",SS_BLACKFRAME,110,324,25,
                    10
    CONTROL         "",IDC_FOG_SWATCH,"Static",SS_BLACKFRAME,110,336,25,10
    PUSHBUTTON      "Toggle visibility",IDC_TOGGLE_VISIBILITY,10,370,80,10
    CONTROL         "Show invisible cells",IDC_SHOW_INVISIBLE,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,105,385,80,10
    PUSHBUTTON      "Move sat farther",IDC_TERR_MOVE_MOON_AWAY,75,270,60,20,
                    BS_MULTILINE | BS_FLAT
    PUSHBUTTON      "^",IDC_MOVE_SAT_UP,165,245,15,15
    PUSHBUTTON      "v",IDC_MOVE_SAT_DOWN,165,265,15,15
    PUSHBUTTON      "<",IDC_MOVE_SAT_LEFT,145,255,15,15
    PUSHBUTTON      ">",IDC_MOVE_SAT_RIGHT,185,255,15,15
    CONTROL         "Satellites",IDC_SATELLITE_CHECK,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,140,185,46,10
    PUSHBUTTON      "Smooth Terrain",IDC_SMOOTH_TERRAIN,135,425,65,15
    CONTROL         "Use Fog",IDC_USE_FOG,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,140,215,60,10
    CONTROL         "Draw Halo",IDC_USE_HALO,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,140,280,60,10
    CONTROL         "Draw atmos",IDC_USE_ATMOSPHERE,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,140,290,60,10
    PUSHBUTTON      "Raise/LowerTerrain",IDC_DROP_TERRAIN,5,450,65,15
    PUSHBUTTON      "Save as PCX",IDC_SAVE_AS_PCX,135,450,65,15
    CONTROL         "Skip External Rooms && Objects",IDC_NO_EXT_ROOMS_OBJS,
                    "Button",BS_AUTOCHECKBOX | WS_TABSTOP,5,435,114,10
    CONTROL         "Cur Sat",IDC_SATELLITE_RADIO,"Button",
                    BS_AUTORADIOBUTTON,60,304,45,8
    CONTROL         "",IDC_SATELLITE_SWATCH,"Static",SS_BLACKFRAME,110,301,
                    25,10
    PUSHBUTTON      "Force visibility",IDC_FORCE_VISIBLE,10,385,80,10
    LTEXT           "Damage",IDC_STATIC,140,340,30,10
    EDITTEXT        IDC_DAMAGE_PER_SEC_EDIT,175,340,25,10,ES_AUTOHSCROLL
    LTEXT           "Fog dist",IDC_STATIC,140,330,30,10
    EDITTEXT        IDC_FOG_SCALAR_EDIT,175,330,25,10,ES_AUTOHSCROLL
    CONTROL         "Rotate stars",IDC_ROTATE_STARS,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,140,300,60,10
    CONTROL         "Rotate sky",IDC_ROTATE_SKY,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,140,310,60,10
    LTEXT           "Rot deg",IDC_STATIC,140,320,30,10
    EDITTEXT        IDC_ROTATE_SPEED_EDIT,175,320,25,10,ES_AUTOHSCROLL
    PUSHBUTTON      "Tile more",IDC_TILE_MORE,110,70,40,20,BS_FLAT
    PUSHBUTTON      "Tile less",IDC_TILE_LESS,110,95,40,20,BS_FLAT
    PUSHBUTTON      "Edit Sounds",IDC_TERR_SOUND,7,476,42,17
END

IDD_D3SPLASH DIALOG DISCARDABLE  0, 0, 215, 149
STYLE DS_CENTER | WS_POPUP | WS_BORDER
FONT 8, "MS Sans Serif"
BEGIN
END

IDD_PATHKEYPAD DIALOG DISCARDABLE  0, 0, 207, 538
STYLE DS_CONTROL | WS_CHILD
FONT 6, "Small Fonts"
BEGIN
    PUSHBUTTON      "Create AI Path Nodes",IDC_AIN_CREATE_NODES,5,260,80,15
    PUSHBUTTON      "Verify Level Nodes",IDC_AIN_VERIFY_LEVEL,5,280,80,15
    PUSHBUTTON      "Destroy ALL Nodes",IDC_AIN_VERIFY_ROOM,5,300,80,15
    CONTROL         "Display All Nodes",IDC_AIN_VIEW_ALL_NODES,"Button",
                    BS_AUTORADIOBUTTON,96,265,75,10
    CONTROL         "Cur Room + Next Rooms",IDC_AIN_VIEW_ROOM_AND_NEXT_ROOMS,
                    "Button",BS_AUTORADIOBUTTON,96,274,90,10
    CONTROL         "Only Cur Room",IDC_AIN_VIEW_ROOM,"Button",
                    BS_AUTORADIOBUTTON,96,285,75,10
    CONTROL         "Don't Display Nodes",IDC_AIN_VIEW_NONE,"Button",
                    BS_AUTORADIOBUTTON,96,294,75,10
    COMBOBOX        IDC_AIN_CURRENT_NODE,10,340,85,104,CBS_DROPDOWN | 
                    CBS_SORT | WS_VSCROLL | WS_TABSTOP
    COMBOBOX        IDC_AIN_OTHER_NODE,100,340,85,103,CBS_DROPDOWN | 
                    CBS_SORT | WS_VSCROLL | WS_TABSTOP
    PUSHBUTTON      "Delete Current node",IDC_AIN_DELETE_NODE,10,360,85,10
    PUSHBUTTON      "Auto Make Edges for node",IDC_AIN_AUTO_EDGE_NODE,10,373,
                    85,10
    PUSHBUTTON      "Move NODE To VIEWER",IDC_AIN_MOVE_NODE_TO_VIEWER,10,386,
                    85,10
    CONTROL         "Nodes in this room",IDC_AIN_SEL_ROOM,"Button",
                    BS_AUTORADIOBUTTON,100,360,85,10
    CONTROL         "Cur Room + Next Rooms",IDC_AIN_SEL_ROOM_AND_NEXT_ROOMS,
                    "Button",BS_AUTORADIOBUTTON,100,370,90,10
    CONTROL         "Current Node's Edges",IDC_AIN_SEL_NODE_EDGES,"Button",
                    BS_AUTORADIOBUTTON,100,380,85,10
    PUSHBUTTON      "Forward",IDC_AIN_FORWARD,10,415,25,10
    PUSHBUTTON      "Left",IDC_AIN_LEFT,10,430,25,10
    PUSHBUTTON      "Up",IDC_AIN_UP,40,415,25,10
    PUSHBUTTON      "Back",IDC_AIN_BACK,70,415,25,10
    PUSHBUTTON      "Down",IDC_AIN_DOWN,40,430,25,10
    PUSHBUTTON      "Right",IDC_AIN_RIGHT,70,430,25,10
    EDITTEXT        IDC_AIN_MOVE_AMOUNT,61,446,35,12,ES_AUTOHSCROLL
    PUSHBUTTON      "Insert node on Edge",IDC_AIN_INSERT_NODE_ON_EDGE,100,
                    400,85,10
    PUSHBUTTON      "Make Edge",IDC_AIN_MAKE_EDGE,100,415,85,10
    PUSHBUTTON      "Delete Edge",IDC_AIN_DELETE_EDGE,100,430,85,10
    PUSHBUTTON      "Create Node at Viewer",IDC_AIN_CREATE_NODE_AT_VIEWER,10,
                    475,85,20
    PUSHBUTTON      "Mono print Node Stats",IDC_AIN_MPNS,105,475,85,20
    PUSHBUTTON      "Reset AI Path Nodes for current room",
                    IDC_AIN_RESET_ROOM_NODES,105,500,85,20,BS_MULTILINE
    PUSHBUTTON      "Move left",IDC_PATHPAD_MOVE_LEFT,5,55,45,15,
                    BS_MULTILINE
    PUSHBUTTON      "Move Obj No Orient",IDC_PATH_MOVE_OBJ_NO_ORIENT,5,75,45,
                    15,BS_MULTILINE
    PUSHBUTTON      "Move down",IDC_PATHPAD_MOVE_DOWN,55,75,45,15,
                    BS_MULTILINE
    PUSHBUTTON      "Move back",IDC_PATHPAD_MOVE_BACKWARD,105,75,40,15,
                    BS_MULTILINE
    PUSHBUTTON      "Move forward",IDC_PATHPAD_MOVE_FORWARD,105,35,40,15,
                    BS_MULTILINE
    PUSHBUTTON      "Move up",IDC_PATHPAD_MOVE_UP,55,35,45,15,BS_MULTILINE
    PUSHBUTTON      "Insert node after current node",IDC_PATHPAD_INSERT_NODE,
                    150,35,40,35,BS_MULTILINE
    PUSHBUTTON      "Delete this node",IDC_PATHPAD_DELETE_NODE,150,75,40,35,
                    BS_MULTILINE
    PUSHBUTTON      "Orientate Viewer",IDC_PATH_REORIENT_VIEWER_BUTTON,150,
                    15,40,15,BS_MULTILINE
    PUSHBUTTON      "Move Viewer",IDC_PATHPAD_CENTER_VIEW,55,55,45,15,
                    BS_MULTILINE
    PUSHBUTTON      "Prev node",IDC_PATHPAD_PREV_NODE,5,35,45,15,
                    BS_MULTILINE
    PUSHBUTTON      "Next node",IDC_PATHPAD_NEXT_NODE,5,15,45,15,
                    BS_MULTILINE
    PUSHBUTTON      "Move current node to current object",
                    IDC_PATH_MOVE_NODE_TO_CURRENT_OBJECT,5,95,95,15,
                    BS_MULTILINE
    PUSHBUTTON      "Move Object",IDC_PATH_MOVE_OBJECT_BUTTON,105,95,40,15,
                    BS_MULTILINE
    PUSHBUTTON      "Move right",IDC_PATHPAD_MOVE_RIGHT,105,55,40,15,
                    BS_MULTILINE
    GROUPBOX        "Current Node Functions",IDC_TEXTURE_GROUP,0,0,200,120
    COMBOBOX        IDC_PATHPAD_PULLDOWN,10,145,100,60,CBS_DROPDOWN | 
                    CBS_SORT | WS_VSCROLL | WS_TABSTOP
    PUSHBUTTON      "Create new path",IDC_PATHPAD_ADD_PATH,115,145,70,15,
                    BS_MULTILINE
    LTEXT           "Current Path",IDC_STATIC,10,135,60,8
    LTEXT           "Number of nodes:",IDC_PATHPAD_NUM_NODES,115,135,75,10
    LTEXT           "Current Node",IDC_STATIC,10,165,45,10
    EDITTEXT        IDC_CURRENT_NODE_EDIT,70,165,40,15,ES_AUTOHSCROLL
    PUSHBUTTON      "Delete this path",IDC_DELETE_PATH,115,165,70,15
    CONTROL         "Show paths",IDC_SHOW_NODES_CHECK,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,10,225,60,10
    GROUPBOX        "Editor Options",IDC_STATIC,0,205,200,35
    GROUPBOX        "General Info",IDC_STATIC,0,125,200,75
    EDITTEXT        IDC_PATH_INC_TEXT,100,215,50,12,ES_AUTOHSCROLL
    LTEXT           "Movement increment",IDC_STATIC,10,215,85,10
    LTEXT           "Current Room:",IDC_PATHPAD_CUR_NODE_ROOM,10,185,90,8
    GROUPBOX        "AI Node System",IDC_STATIC,0,245,200,290
    GROUPBOX        "Node Display Options",IDC_STATIC,90,255,105,60
    GROUPBOX        "Current Node Fuctions",IDC_STATIC,5,320,190,140
    GROUPBOX        "Misc. Node Functions",IDC_STATIC,5,465,190,65
    LTEXT           "Move Amount",IDC_STATIC,10,446,45,10
    LTEXT           "Current Node",IDC_STATIC,10,330,80,8
    LTEXT           "Other Node",IDC_STATIC,100,330,80,8
    PUSHBUTTON      "Insert Edge node at viewer",IDC_IENAV_BUTTON,100,445,85,
                    10
    PUSHBUTTON      "Auto Edge Room/Region",IDC_AER_BUTTON,10,500,85,20
    PUSHBUTTON      "Move VIEWER to NODE",IDC_MVTN_BUTTON,10,400,85,10
    PUSHBUTTON      "Orient Node To Viewer",IDC_ONTVGP_BUTTON,55,15,90,15
END

IDD_EDITLINEDLG DIALOG DISCARDABLE  0, 0, 194, 57
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "Edit text"
FONT 8, "MS Sans Serif"
BEGIN
    EDITTEXT        IDC_EDIT,5,35,125,15,ES_AUTOHSCROLL | ES_WANTRETURN
    DEFPUSHBUTTON   "OK",IDOK,135,15,50,14
    PUSHBUTTON      "Cancel",IDCANCEL,135,35,50,14
    LTEXT           "Prompt",IDC_PROMPT,5,10,125,20
END

IDD_CUSTOMIZE_OBJECT DIALOG DISCARDABLE  0, 0, 247, 101
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "Dialog"
FONT 8, "MS Sans Serif"
BEGIN
    DEFPUSHBUTTON   "OK",IDOK,180,80,50,14
    LTEXT           "Object Name:",IDC_OBJECT_NAME,5,10,45,10
    LTEXT           "Object type:",IDC_OBJECT_TYPE,5,25,50,10
    LTEXT           "Size:",IDC_OBJECT_SIZE,5,50,20,10
    LTEXT           "Mass:",IDC_OBJECT_MASS,5,65,25,10
    LTEXT           "Drag:",IDC_OBJECT_DRAG,5,80,25,10
    EDITTEXT        IDC_OBJECT_SIZE_EDIT,35,50,50,15,ES_AUTOHSCROLL
    EDITTEXT        IDC_OBJECT_MASS_EDIT,35,65,50,15,ES_AUTOHSCROLL
    EDITTEXT        IDC_OBJECT_DRAG_EDIT,35,80,50,15,ES_AUTOHSCROLL
    EDITTEXT        IDC_OBJECT_NAME_EDIT,55,5,65,15,ES_AUTOHSCROLL
    EDITTEXT        IDC_OBJECT_TYPE_EDIT,55,20,65,15,ES_AUTOHSCROLL
END

IDD_ANIMSTATES DIALOG DISCARDABLE  0, 0, 500, 429
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "Animation States"
FONT 7, "Small Fonts"
BEGIN
    GROUPBOX        "Animation Frames",IDC_STATIC,20,25,335,400
    CONTROL         "Alert",IDC_ALERT_CHECK,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,26,60,55,10
    EDITTEXT        IDC_ALERT_FROM_EDIT,80,55,25,15,ES_AUTOHSCROLL
    EDITTEXT        IDC_ALERT_TO_EDIT,146,55,25,15,ES_AUTOHSCROLL
    LTEXT           " through",IDC_STATIC,110,60,30,10
    COMBOBOX        IDC_CURRENT_STATE_PULLDOWN,61,5,104,50,CBS_DROPDOWN | 
                    CBS_SORT | WS_VSCROLL | WS_TABSTOP
    LTEXT           "Current State",IDC_STATIC,5,5,47,15
    CONTROL         "Death",IDC_DEATH_CHECK,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,26,75,55,10
    EDITTEXT        IDC_DEATH_FROM_EDIT,80,70,25,15,ES_AUTOHSCROLL
    EDITTEXT        IDC_DEATH_TO_EDIT,146,70,25,15,ES_AUTOHSCROLL
    LTEXT           " through",IDC_STATIC,110,75,30,10
    CONTROL         "Birth",IDC_MISSILE1_CHECK,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,26,89,55,10
    EDITTEXT        IDC_MISSILE1_FROM_EDIT,80,85,25,15,ES_AUTOHSCROLL
    EDITTEXT        IDC_MISSILE1_TO_EDIT,146,85,25,15,ES_AUTOHSCROLL
    LTEXT           " through",IDC_STATIC,110,89,30,10
    CONTROL         "-----",IDC_MISSILE1_RECOIL_CHECK,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,26,105,55,10
    EDITTEXT        IDC_RECOIL1_FROM_EDIT,80,100,25,15,ES_AUTOHSCROLL
    EDITTEXT        IDC_RECOIL1_TO_EDIT,146,100,25,15,ES_AUTOHSCROLL
    LTEXT           " through",IDC_STATIC,110,105,30,10
    CONTROL         "-----",IDC_MISSILE2_CHECK,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,26,120,55,10
    EDITTEXT        IDC_MISSILE2_FROM_EDIT,80,115,25,15,ES_AUTOHSCROLL
    EDITTEXT        IDC_MISSILE2_TO_EDIT,146,115,25,15,ES_AUTOHSCROLL
    LTEXT           " through",IDC_STATIC,110,120,30,10
    CONTROL         "-----",IDC_MISSILE2_RECOIL_CHECK,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,26,136,55,10
    EDITTEXT        IDC_RECOIL2_FROM_EDIT,80,129,25,15,ES_AUTOHSCROLL
    EDITTEXT        IDC_RECOIL2_TO_EDIT,146,129,25,15,ES_AUTOHSCROLL
    LTEXT           " through",IDC_STATIC,110,136,30,10
    CONTROL         "Melee 1",IDC_MELEE1_CHECK,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,26,150,55,10
    EDITTEXT        IDC_MELEE1_FROM_EDIT,80,145,25,15,ES_AUTOHSCROLL
    EDITTEXT        IDC_MELEE1_TO_EDIT,146,145,25,15,ES_AUTOHSCROLL
    LTEXT           " through",IDC_STATIC,110,150,30,10
    CONTROL         "MRecoil 1",IDC_MELEE1_RECOIL_CHECK,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,26,165,55,10
    EDITTEXT        IDC_MRECOIL1_FROM_EDIT,80,160,25,15,ES_AUTOHSCROLL
    EDITTEXT        IDC_MRECOIL1_TO_EDIT,146,160,25,15,ES_AUTOHSCROLL
    LTEXT           " through",IDC_STATIC,110,165,30,10
    CONTROL         "Melee 2",IDC_MELEE2_CHECK,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,26,181,55,10
    EDITTEXT        IDC_MELEE2_FROM_EDIT,80,176,25,15,ES_AUTOHSCROLL
    EDITTEXT        IDC_MELEE2_TO_EDIT,146,176,25,15,ES_AUTOHSCROLL
    LTEXT           " through",IDC_STATIC,110,181,30,10
    CONTROL         "MRecoil 2",IDC_MELEE2_RECOIL_CHECK,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,26,195,55,10
    EDITTEXT        IDC_MRECOIL2_FROM_EDIT,80,190,25,15,ES_AUTOHSCROLL
    EDITTEXT        IDC_MRECOIL2_TO_EDIT,146,190,25,15,ES_AUTOHSCROLL
    LTEXT           " through",IDC_STATIC,110,195,30,10
    CONTROL         "Idle",IDC_IDLE_CHECK,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,26,209,55,10
    EDITTEXT        IDC_IDLE_FROM_EDIT,80,205,25,15,ES_AUTOHSCROLL
    EDITTEXT        IDC_IDLE_TO_EDIT,146,205,25,15,ES_AUTOHSCROLL
    LTEXT           " through",IDC_STATIC,110,209,30,10
    CONTROL         "Quirk",IDC_QUIRK_CHECK,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,26,225,55,10
    EDITTEXT        IDC_QUIRK_FROM_EDIT,80,221,25,15,ES_AUTOHSCROLL
    EDITTEXT        IDC_QUIRK_TO_EDIT,146,221,25,15,ES_AUTOHSCROLL
    LTEXT           " through",IDC_STATIC,110,225,30,10
    CONTROL         "Flinch",IDC_FLINCH_CHECK,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,26,240,55,10
    EDITTEXT        IDC_FLINCH_FROM_EDIT,80,235,25,15,ES_AUTOHSCROLL
    EDITTEXT        IDC_FLINCH_TO_EDIT,146,235,25,15,ES_AUTOHSCROLL
    LTEXT           " through",IDC_STATIC,110,240,30,10
    CONTROL         "Taunt",IDC_TAUNT_CHECK,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,26,256,55,10
    EDITTEXT        IDC_TAUNT_FROM_EDIT,80,249,25,15,ES_AUTOHSCROLL
    EDITTEXT        IDC_TAUNT_TO_EDIT,146,249,25,15,ES_AUTOHSCROLL
    LTEXT           " through",IDC_STATIC,110,256,30,10
    CONTROL         "To Idle Standing",IDC_SPECIAL1_CHECK,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,26,271,79,10
    EDITTEXT        IDC_SPECIAL1_FROM_EDIT,110,265,25,15,ES_AUTOHSCROLL
    EDITTEXT        IDC_SPECIAL1_TO_EDIT,146,265,25,15,ES_AUTOHSCROLL
    CONTROL         "To Idle Flying",IDC_SPECIAL2_CHECK,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,26,285,74,10
    CONTROL         "To Idle Rolling",IDC_SPECIAL3_CHECK,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,26,301,79,10
    CONTROL         "To Idle Jumping",IDC_SPECIAL5_CHECK,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,26,329,79,10
    CONTROL         "To Idle Walking",IDC_SPECIAL4_CHECK,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,26,315,74,10
    CONTROL         "To Alert Standing",IDC_GOTO_STANDING_CHECK,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,26,350,80,10
    EDITTEXT        IDC_STANDING_FROM_EDIT,110,345,25,15,ES_AUTOHSCROLL
    EDITTEXT        IDC_STANDING_TO_EDIT,146,345,25,15,ES_AUTOHSCROLL
    LTEXT           "to",IDC_STATIC,135,350,10,8
    CONTROL         "To Alert Flying",IDC_GOTO_FLYING_CHECK,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,26,365,80,10
    EDITTEXT        IDC_FLYING_FROM_EDIT,110,360,25,15,ES_AUTOHSCROLL
    EDITTEXT        IDC_FLYING_TO_EDIT,146,360,25,15,ES_AUTOHSCROLL
    LTEXT           "to",IDC_STATIC,135,365,10,8
    CONTROL         "To Alert Rolling",IDC_GOTO_ROLLING_CHECK,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,26,380,80,10
    EDITTEXT        IDC_ROLLING_FROM_EDIT,110,375,25,15,ES_AUTOHSCROLL
    EDITTEXT        IDC_ROLLING_TO_EDIT,146,375,25,15,ES_AUTOHSCROLL
    LTEXT           "to",IDC_STATIC,135,380,10,8
    CONTROL         "To Alert Walking",IDC_GOTO_WALKING_CHECK,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,26,395,80,10
    EDITTEXT        IDC_WALKING_FROM_EDIT,110,390,25,15,ES_AUTOHSCROLL
    EDITTEXT        IDC_WALKING_TO_EDIT,146,390,25,15,ES_AUTOHSCROLL
    LTEXT           "to",IDC_STATIC,135,395,10,8
    CONTROL         "To Alert Jumping",IDC_GOTO_JUMPING_CHECK,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,26,409,80,10
    EDITTEXT        IDC_JUMPING_FROM_EDIT,110,405,25,15,ES_AUTOHSCROLL
    EDITTEXT        IDC_JUMPING_TO_EDIT,146,405,25,15,ES_AUTOHSCROLL
    LTEXT           "to",IDC_STATIC,135,409,10,8
    CONTROL         "",IDC_ANIMSTATE_VIEW,"Static",SS_ETCHEDFRAME,360,35,135,
                    125
    COMBOBOX        IDC_CURRENT_ANIMATION_PULLDOWN,400,160,95,107,
                    CBS_DROPDOWN | CBS_SORT | WS_VSCROLL | WS_TABSTOP
    LTEXT           "Current Animation",IDC_STATIC,360,160,40,15
    EDITTEXT        IDC_SPECIAL2_FROM_EDIT,110,280,25,15,ES_AUTOHSCROLL
    EDITTEXT        IDC_SPECIAL2_TO_EDIT,146,280,25,15,ES_AUTOHSCROLL
    EDITTEXT        IDC_SPECIAL3_FROM_EDIT,110,295,25,15,ES_AUTOHSCROLL
    EDITTEXT        IDC_SPECIAL3_TO_EDIT,146,295,25,15,ES_AUTOHSCROLL
    EDITTEXT        IDC_SPECIAL4_FROM_EDIT,110,310,25,15,ES_AUTOHSCROLL
    EDITTEXT        IDC_SPECIAL4_TO_EDIT,146,310,25,15,ES_AUTOHSCROLL
    EDITTEXT        IDC_SPECIAL5_FROM_EDIT,110,325,25,15,ES_AUTOHSCROLL
    EDITTEXT        IDC_SPECIAL5_TO_EDIT,146,325,25,15,ES_AUTOHSCROLL
    PUSHBUTTON      "Reset anim states to zero",IDC_RESET_ANIM_ZERO,185,5,95,
                    20
    PUSHBUTTON      "Rotate Left",IDC_ROT_LEFT,360,180,55,20
    PUSHBUTTON      "Rotate Right",IDC_ROT_RIGHT,435,180,60,20
    PUSHBUTTON      "OK",IDOK,285,5,60,20
    PUSHBUTTON      "Cancel",IDCANCEL,350,5,60,20
    EDITTEXT        IDC_SEC_PER_CYCLE_EDIT1,200,55,35,15,ES_AUTOHSCROLL
    EDITTEXT        IDC_SEC_PER_CYCLE_EDIT2,200,70,35,15,ES_AUTOHSCROLL
    EDITTEXT        IDC_SEC_PER_CYCLE_EDIT3,200,85,35,15,ES_AUTOHSCROLL
    EDITTEXT        IDC_SEC_PER_CYCLE_EDIT4,200,100,35,15,ES_AUTOHSCROLL
    EDITTEXT        IDC_SEC_PER_CYCLE_EDIT5,200,115,35,15,ES_AUTOHSCROLL
    EDITTEXT        IDC_SEC_PER_CYCLE_EDIT6,200,129,35,15,ES_AUTOHSCROLL
    EDITTEXT        IDC_SEC_PER_CYCLE_EDIT7,200,145,35,15,ES_AUTOHSCROLL
    EDITTEXT        IDC_SEC_PER_CYCLE_EDIT8,200,160,35,15,ES_AUTOHSCROLL
    EDITTEXT        IDC_SEC_PER_CYCLE_EDIT9,200,175,35,15,ES_AUTOHSCROLL
    EDITTEXT        IDC_SEC_PER_CYCLE_EDIT10,200,190,35,15,ES_AUTOHSCROLL
    EDITTEXT        IDC_SEC_PER_CYCLE_EDIT11,200,205,35,15,ES_AUTOHSCROLL
    EDITTEXT        IDC_SEC_PER_CYCLE_EDIT12,200,220,35,15,ES_AUTOHSCROLL
    EDITTEXT        IDC_SEC_PER_CYCLE_EDIT13,200,235,35,15,ES_AUTOHSCROLL
    EDITTEXT        IDC_SEC_PER_CYCLE_EDIT14,200,249,35,15,ES_AUTOHSCROLL
    EDITTEXT        IDC_SEC_PER_CYCLE_EDIT15,200,265,35,15,ES_AUTOHSCROLL
    EDITTEXT        IDC_SEC_PER_CYCLE_EDIT16,200,280,35,15,ES_AUTOHSCROLL
    EDITTEXT        IDC_SEC_PER_CYCLE_EDIT17,200,295,35,15,ES_AUTOHSCROLL
    EDITTEXT        IDC_SEC_PER_CYCLE_EDIT18,200,310,35,15,ES_AUTOHSCROLL
    EDITTEXT        IDC_SEC_PER_CYCLE_EDIT19,200,325,35,15,ES_AUTOHSCROLL
    EDITTEXT        IDC_SEC_PER_CYCLE_EDIT20,200,345,35,15,ES_AUTOHSCROLL
    EDITTEXT        IDC_SEC_PER_CYCLE_EDIT21,200,360,35,15,ES_AUTOHSCROLL
    EDITTEXT        IDC_SEC_PER_CYCLE_EDIT22,200,375,35,15,ES_AUTOHSCROLL
    EDITTEXT        IDC_SEC_PER_CYCLE_EDIT23,200,390,35,15,ES_AUTOHSCROLL
    EDITTEXT        IDC_SEC_PER_CYCLE_EDIT24,200,405,35,15,ES_AUTOHSCROLL
    LTEXT           "Anim State",IDC_STATIC,25,38,40,10
    LTEXT           "Keyframes",IDC_STATIC,108,38,40,10
    LTEXT           "Anim Cycle Time",IDC_STATIC,189,38,65,10
    COMBOBOX        IDC_ANIM_SOUND0_PULLDOWN,245,55,104,50,CBS_DROPDOWN | 
                    CBS_SORT | WS_VSCROLL | WS_TABSTOP
    COMBOBOX        IDC_ANIM_SOUND1_PULLDOWN,245,70,104,50,CBS_DROPDOWN | 
                    CBS_SORT | WS_VSCROLL | WS_TABSTOP
    COMBOBOX        IDC_ANIM_SOUND2_PULLDOWN,245,85,104,50,CBS_DROPDOWN | 
                    CBS_SORT | WS_VSCROLL | WS_TABSTOP
    COMBOBOX        IDC_ANIM_SOUND3_PULLDOWN,245,100,104,50,CBS_DROPDOWN | 
                    CBS_SORT | WS_VSCROLL | WS_TABSTOP
    COMBOBOX        IDC_ANIM_SOUND4_PULLDOWN,245,115,104,50,CBS_DROPDOWN | 
                    CBS_SORT | WS_VSCROLL | WS_TABSTOP
    COMBOBOX        IDC_ANIM_SOUND5_PULLDOWN,245,129,104,50,CBS_DROPDOWN | 
                    CBS_SORT | WS_VSCROLL | WS_TABSTOP
    COMBOBOX        IDC_ANIM_SOUND6_PULLDOWN,245,145,104,50,CBS_DROPDOWN | 
                    CBS_SORT | WS_VSCROLL | WS_TABSTOP
    COMBOBOX        IDC_ANIM_SOUND7_PULLDOWN,245,160,104,50,CBS_DROPDOWN | 
                    CBS_SORT | WS_VSCROLL | WS_TABSTOP
    COMBOBOX        IDC_ANIM_SOUND8_PULLDOWN,245,175,104,50,CBS_DROPDOWN | 
                    CBS_SORT | WS_VSCROLL | WS_TABSTOP
    COMBOBOX        IDC_ANIM_SOUND9_PULLDOWN,245,190,104,50,CBS_DROPDOWN | 
                    CBS_SORT | WS_VSCROLL | WS_TABSTOP
    COMBOBOX        IDC_ANIM_SOUND10_PULLDOWN,245,205,104,50,CBS_DROPDOWN | 
                    CBS_SORT | WS_VSCROLL | WS_TABSTOP
    COMBOBOX        IDC_ANIM_SOUND11_PULLDOWN,245,220,104,50,CBS_DROPDOWN | 
                    CBS_SORT | WS_VSCROLL | WS_TABSTOP
    COMBOBOX        IDC_ANIM_SOUND12_PULLDOWN,245,235,104,50,CBS_DROPDOWN | 
                    CBS_SORT | WS_VSCROLL | WS_TABSTOP
    COMBOBOX        IDC_ANIM_SOUND13_PULLDOWN,245,249,104,50,CBS_DROPDOWN | 
                    CBS_SORT | WS_VSCROLL | WS_TABSTOP
    COMBOBOX        IDC_ANIM_SOUND14_PULLDOWN,245,265,104,50,CBS_DROPDOWN | 
                    CBS_SORT | WS_VSCROLL | WS_TABSTOP
    COMBOBOX        IDC_ANIM_SOUND15_PULLDOWN,245,280,104,50,CBS_DROPDOWN | 
                    CBS_SORT | WS_VSCROLL | WS_TABSTOP
    COMBOBOX        IDC_ANIM_SOUND16_PULLDOWN,245,295,104,50,CBS_DROPDOWN | 
                    CBS_SORT | WS_VSCROLL | WS_TABSTOP
    COMBOBOX        IDC_ANIM_SOUND17_PULLDOWN,245,310,104,50,CBS_DROPDOWN | 
                    CBS_SORT | WS_VSCROLL | WS_TABSTOP
    COMBOBOX        IDC_ANIM_SOUND18_PULLDOWN,245,325,104,50,CBS_DROPDOWN | 
                    CBS_SORT | WS_VSCROLL | WS_TABSTOP
    COMBOBOX        IDC_ANIM_SOUND19_PULLDOWN,245,345,104,50,CBS_DROPDOWN | 
                    CBS_SORT | WS_VSCROLL | WS_TABSTOP
    COMBOBOX        IDC_ANIM_SOUND20_PULLDOWN,245,359,104,50,CBS_DROPDOWN | 
                    CBS_SORT | WS_VSCROLL | WS_TABSTOP
    COMBOBOX        IDC_ANIM_SOUND21_PULLDOWN,245,375,104,50,CBS_DROPDOWN | 
                    CBS_SORT | WS_VSCROLL | WS_TABSTOP
    COMBOBOX        IDC_ANIM_SOUND22_PULLDOWN,245,390,104,50,CBS_DROPDOWN | 
                    CBS_SORT | WS_VSCROLL | WS_TABSTOP
    COMBOBOX        IDC_ANIM_SOUND23_PULLDOWN,245,405,104,50,CBS_DROPDOWN | 
                    CBS_SORT | WS_VSCROLL | WS_TABSTOP
    LTEXT           "Anim Sound",IDC_STATIC,272,38,65,10
    LTEXT           "to",IDC_STATIC,135,270,10,8
    LTEXT           "to",IDC_STATIC,135,285,10,8
    LTEXT           "to",IDC_STATIC,135,300,10,8
    LTEXT           "to",IDC_STATIC,135,315,10,8
    LTEXT           "to",IDC_STATIC,135,329,10,8
END

IDD_FILEPAGE DIALOG DISCARDABLE  0, 0, 363, 305
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "Miscellaneous Files"
FONT 8, "MS Sans Serif"
BEGIN
    PUSHBUTTON      "Add new file",IDC_ADD_FILE,14,39,45,15
    PUSHBUTTON      "Lock",IDC_LOCK_FILE,62,39,35,15
    PUSHBUTTON      "Check in",IDC_CHECKIN_FILE,286,39,40,15
    PUSHBUTTON      "Delete",IDC_DELETE_FILE,225,39,50,15
    PUSHBUTTON      "Pages Checked out",IDC_FILES_OUT,282,7,70,15
    PUSHBUTTON      "Override to unlocked",IDC_OVERRIDE,100,39,85,15
    LISTBOX         IDC_FILELIST,9,58,183,237,LBS_SORT | LBS_MULTIPLESEL | 
                    LBS_OWNERDRAWFIXED | LBS_HASSTRINGS | 
                    LBS_NOINTEGRALHEIGHT | WS_VSCROLL | WS_TABSTOP
    LISTBOX         IDC_CHECKEDOUT_LIST,199,58,153,237,LBS_SORT | 
                    LBS_MULTIPLESEL | LBS_OWNERDRAWFIXED | LBS_HASSTRINGS | 
                    LBS_NOINTEGRALHEIGHT | WS_VSCROLL | WS_TABSTOP
    GROUPBOX        "Available Files",IDC_STATIC,7,26,190,272
    GROUPBOX        "Files Checked Out",IDC_STATIC,195,26,161,272
    COMBOBOX        IDC_DIRECTORY_FILTER,91,7,90,119,CBS_DROPDOWN | CBS_SORT | 
                    WS_VSCROLL | WS_TABSTOP
    LTEXT           "Directory Filter",IDC_STATIC,43,9,46,8
END

IDD_MEGACELL DIALOG DISCARDABLE  0, 0, 537, 316
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "Megacell Dialog"
FONT 8, "MS Sans Serif"
BEGIN
    PUSHBUTTON      "New Megacell",IDC_NEW_MEGACELL,5,5,78,14
    PUSHBUTTON      "Delete Megacell",IDC_DELETE_MEGACELL,86,5,64,14
    PUSHBUTTON      "Lock",IDC_LOCK_MEGACELL,155,5,21,14
    PUSHBUTTON      "Check In",IDC_CHECKIN_MEGACELL,180,5,34,14
    PUSHBUTTON      "<< Prev",IDC_PREVIOUS_MEGACELL,220,5,42,14
    PUSHBUTTON      "Next >>",IDC_NEXT_MEGACELL,265,5,30,14
    PUSHBUTTON      "Megacells Checked Out",IDC_MEGACELLS_OUT,300,5,85,14
    EDITTEXT        IDC_MEGACELL_NAME_EDIT,30,25,80,12,ES_AUTOHSCROLL
    EDITTEXT        IDC_MEGA_WIDTH_EDIT,135,25,25,12,ES_AUTOHSCROLL
    EDITTEXT        IDC_MEGA_HEIGHT_EDIT,195,25,25,12,ES_AUTOHSCROLL
    COMBOBOX        IDC_MEGACELL_PULLDOWN,285,25,100,15,CBS_DROPDOWN | 
                    CBS_SORT | WS_VSCROLL | WS_TABSTOP
    LTEXT           "Name",IDC_STATIC,5,25,25,10
    LTEXT           "Megacell ID",IDC_STATIC,240,25,45,10
    LTEXT           "Width",IDC_STATIC,115,25,20,10
    LTEXT           "Height",IDC_STATIC,170,25,25,10
    CONTROL         "",IDC_TEXTURE_LISTBOX,"Static",SS_BLACKFRAME,385,55,145,
                    245
    SCROLLBAR       IDC_TEXLIST_VERT,375,55,10,245,SBS_VERT
    LTEXT           "Textures",IDC_STATIC,385,45,55,10
    CONTROL         "",IDC_MEGACELL_VIEW,"Static",SS_BLACKFRAME,5,55,360,245
    LTEXT           "Megacell View",IDC_STATIC,5,45,50,10
    PUSHBUTTON      "Remap Megacell",IDC_REMAP_MEGACELL,465,33,60,10
    PUSHBUTTON      "Copy",IDC_COPY_MEGACELL,400,20,25,10
    PUSHBUTTON      "Paste",IDC_PASTE_MEGACELL,430,20,25,10
    PUSHBUTTON      "Rotate",IDC_ROTATE_MEGACELL,465,45,60,10
    PUSHBUTTON      "Import Megacell",IDC_IMPORT_TINY,400,0,60,10
    PUSHBUTTON      "Import Sky Dome",IDC_IMPORT_SKY,465,0,60,10
    PUSHBUTTON      "Import Sky Band",IDC_IMPORT_SKY_BAND,465,11,60,10
END

IDD_MEGAKEYPAD DIALOG DISCARDABLE  0, 0, 210, 461
STYLE DS_CONTROL | WS_CHILD
FONT 7, "Small Fonts"
BEGIN
    CONTROL         "",IDC_MEGACELL_SET_VIEW,"Static",SS_BLACKFRAME,5,175,
                    200,245
    PUSHBUTTON      "Next >>",IDC_NEXT_MEGA_SET,170,155,35,15,BS_FLAT
    PUSHBUTTON      "<< Prev",IDC_PREV_MEGA_SET,130,155,35,15,BS_FLAT
    LTEXT           "Megacell Name",IDC_MEGACELL_NAME_STATIC,5,120,200,10
    LTEXT           "Width",IDC_MEGA_WIDTH_STATIC,5,130,45,10
    LTEXT           "Height",IDC_MEGA_HEIGHT_STATIC,60,130,60,10
    CONTROL         "Randomize when placing",IDC_RANDOMIZE_MEGACELL_CHECK,
                    "Button",BS_AUTOCHECKBOX | BS_MULTILINE | WS_TABSTOP,10,
                    65,85,15
    PUSHBUTTON      "Fill selected min/max area with megacell",
                    IDC_FILL_WITH_MEGACELL,10,15,80,20,BS_MULTILINE | 
                    BS_FLAT
    PUSHBUTTON      "Fill only selected cells with megacell (Unify)",
                    IDC_UNIFY_MEGACELL,105,15,80,20,BS_MULTILINE | BS_FLAT
    PUSHBUTTON      "Select all cells that have this megacell",
                    IDC_SELECT_FROM_MEGACELL,10,40,80,20,BS_MULTILINE | 
                    BS_FLAT
    LTEXT           "X Granularity",IDC_STATIC,10,85,45,10
    EDITTEXT        IDC_X_GRANULAR_EDIT,65,85,25,12,ES_AUTOHSCROLL
    LTEXT           "Y Granularity",IDC_STATIC,10,100,45,10
    EDITTEXT        IDC_Y_GRANULAR_EDIT,65,100,25,12,ES_AUTOHSCROLL
    PUSHBUTTON      "Highlight megacell borders",IDC_HIGHLIGHT_BORDERS,105,
                    40,80,20,BS_MULTILINE | BS_FLAT
    PUSHBUTTON      "Use this megacell as a sky dome",IDC_MEGACELL_SKYDOME,
                    105,65,80,20,BS_MULTILINE | BS_FLAT
    PUSHBUTTON      "Use this megacell as as sky band",IDC_MEGACELL_SKYBAND,
                    105,90,80,20,BS_MULTILINE | BS_FLAT
END

IDD_ROOMKEYPAD DIALOG DISCARDABLE  0, 0, 210, 603
STYLE DS_CONTROL | WS_CHILD
FONT 7, "Small Fonts"
BEGIN
    PUSHBUTTON      "Load room",IDC_LOAD_ROOM,147,383,45,15,BS_FLAT
    PUSHBUTTON      "Save room",IDC_SAVE_ROOM_LOCALLY,110,30,85,15,BS_FLAT
    COMBOBOX        IDC_ROOM_PULLDOWN,50,385,91,155,CBS_DROPDOWN | CBS_SORT | 
                    WS_VSCROLL | WS_TABSTOP
    LTEXT           "Rooms",IDC_STATIC,14,385,30,10
    LTEXT           "Vertices:",IDC_VERTEX_COUNT,15,25,60,10
    LTEXT           "Faces:",IDC_FACE_COUNT,15,36,60,10
    GROUPBOX        "Paletted Rooms",IDC_STATIC,9,370,200,47
    GROUPBOX        "Current Room",IDC_STATIC,5,0,198,280
    PUSHBUTTON      "Mark Room",IDC_MARK_ROOM,110,10,85,15,BS_FLAT
    PUSHBUTTON      "Contract Edge",IDC_ROOMPAD_CONTRACT_EDGE,158,93,40,20,
                    BS_MULTILINE | BS_FLAT
    PUSHBUTTON      "Contract room",IDC_ROOMPAD_CONTRACT_ROOM,158,143,40,20,
                    BS_MULTILINE | BS_FLAT
    PUSHBUTTON      "Rotate Left",IDC_ROOMPAD_ROTATE_LEFT,19,225,35,20,
                    BS_MULTILINE | BS_FLAT
    PUSHBUTTON      "Rotate Right",IDC_ROOMPAD_ROTATE_RIGHT,59,225,35,20,
                    BS_MULTILINE | BS_FLAT
    PUSHBUTTON      "Next Vertex",IDC_ROOMPAD_NEXT_VERTEX,58,131,40,20,
                    BS_MULTILINE | BS_FLAT
    PUSHBUTTON      "Expand edge",IDC_ROOMPAD_EXPAND_EDGE,114,93,40,20,
                    BS_MULTILINE | BS_FLAT
    PUSHBUTTON      "Expand room",IDC_ROOMPAD_EXPAND_ROOM,114,143,40,20,
                    BS_MULTILINE | BS_FLAT
    PUSHBUTTON      "Expand face",IDC_ROOMPAD_EXPAND_FACE,114,119,40,20,
                    BS_MULTILINE | BS_FLAT
    PUSHBUTTON      "Pitch Back",IDC_ROOMPAD_PITCH_BACKWARD,19,200,35,20,
                    BS_MULTILINE | BS_FLAT
    PUSHBUTTON      "Next Edge",IDC_ROOMPAD_NEXT_EDGE,14,131,40,20,
                    BS_MULTILINE | BS_FLAT
    PUSHBUTTON      "Bank Right",IDC_ROOMPAD_BANK_RIGHT,59,250,35,19,
                    BS_MULTILINE | BS_FLAT
    PUSHBUTTON      "Bank Left",IDC_ROOMPAD_BANK_LEFT,19,250,35,19,
                    BS_MULTILINE | BS_FLAT
    PUSHBUTTON      "Pitch Forward",IDC_ROOMPAD_PITCH_FORWARD,59,200,35,20,
                    BS_MULTILINE | BS_FLAT
    PUSHBUTTON      "Contract face",IDC_ROOMPAD_CONTRACT_FACE,158,119,40,20,
                    BS_MULTILINE | BS_FLAT
    GROUPBOX        "Portals",IDC_STATIC,14,190,85,85
    GROUPBOX        "Scaling",IDC_STATIC,108,83,95,85
    GROUPBOX        "Movement",IDC_STATIC,109,190,95,85
    PUSHBUTTON      "Move backward",IDC_ROOMPAD_MOVE_BACKWARD,159,201,40,20,
                    BS_MULTILINE | BS_FLAT
    PUSHBUTTON      "Move down",IDC_ROOMPAD_MOVE_DOWN,159,251,40,20,
                    BS_MULTILINE | BS_FLAT
    PUSHBUTTON      "Move forward",IDC_ROOMPAD_MOVE_FORWARD,114,201,40,20,
                    BS_MULTILINE | BS_FLAT
    PUSHBUTTON      "Move up",IDC_ROOMPAD_MOVE_UP,114,251,40,20,BS_MULTILINE | 
                    BS_FLAT
    PUSHBUTTON      "Move right",IDC_ROOMPAD_MOVE_RIGHT,159,225,40,20,
                    BS_MULTILINE | BS_FLAT
    PUSHBUTTON      "Move left",IDC_ROOMPAD_MOVE_LEFT,114,225,40,20,
                    BS_MULTILINE | BS_FLAT
    PUSHBUTTON      "Combine Faces",IDC_ROOMPAD_COMBINE_FACES,10,425,40,20,
                    BS_MULTILINE | BS_FLAT
    CONTROL         "Alphaed surfaces as wireframe",IDC_SHOW_ALPHAED_CHECK,
                    "Button",BS_AUTOCHECKBOX | WS_TABSTOP,12,449,110,10
    PUSHBUTTON      "Next Portal",IDC_ROOMPAD_NEXT_PORTAL,14,340,40,20,
                    BS_FLAT
    LTEXT           "Vertices:",IDC_PALETTE_VERTEX_COUNT,17,402,70,10
    LTEXT           "Faces:",IDC_PALETTE_FACE_COUNT,96,401,70,10
    LTEXT           "Portals:",IDC_PORTAL_COUNT,15,47,60,10
    GROUPBOX        "Portals",IDC_STATIC,9,320,200,45
    LTEXT           "Current Portal:",IDC_CURRENT_PORTAL,15,330,70,8
    CONTROL         "Render portal faces",IDC_RENDER_PORTAL_FACES,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,115,325,80,10
    CONTROL         "Enable fly-through of rendered faces",
                    IDC_ENABLE_PORTAL_FLYTHROUGH,"Button",BS_AUTOCHECKBOX | 
                    BS_MULTILINE | WS_TABSTOP,65,340,140,10
    CONTROL         "External Rm",IDC_EXTERNAL_ROOM,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,15,57,55,10
    CONTROL         "Render all external rooms",IDC_RENDER_EXTERNAL_ROOMS,
                    "Button",BS_AUTOCHECKBOX | WS_TABSTOP,12,463,95,10
    PUSHBUTTON      "Next Face",IDC_NEXT_FACE,169,290,35,20,BS_MULTILINE | 
                    BS_FLAT
    GROUPBOX        "Faces",IDC_STATIC,9,280,200,40
    LTEXT           "Current face:",IDC_CURRENT_FACE,14,290,70,10
    CONTROL         "Touches terrain",IDC_TOUCHES_OUTSIDE,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,110,45,85,8
    CONTROL         "Refueling",IDC_ROOMPAD_REFUELING_CENTER,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,15,68,50,10
    CONTROL         "Red Goal",IDC_ROOMPAD_GOAL1,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,15,78,50,10
    CONTROL         "Blue Goal",IDC_ROOMPAD_GOAL2,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,15,87,50,10
    CONTROL         "Green Goal",IDC_ROOMPAD_GOAL3,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,15,97,55,10
    CONTROL         "Yellow Goal",IDC_ROOMPAD_GOAL4,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,15,108,55,10
    CONTROL         "Goal face",IDC_GOALFACE_CHECK,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,15,304,65,8
    CONTROL         "S1",IDC_SPECIAL_1,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,
                    69,57,25,8
    CONTROL         "S2",IDC_SPECIAL_2,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,
                    69,68,25,8
    CONTROL         "S3",IDC_SPECIAL_3,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,
                    69,78,25,8
    CONTROL         "S4",IDC_SPECIAL_4,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,
                    69,87,25,8
    CONTROL         "S5",IDC_SPECIAL_5,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,
                    69,97,25,8
    CONTROL         "S6",IDC_SPECIAL_6,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,
                    69,108,25,8
    PUSHBUTTON      "Compute BOA visibility",IDC_BOA_VIS_BUTTON,146,425,55,
                    20,BS_MULTILINE | BS_FLAT
    COMBOBOX        IDC_ROOM_AMBIENT,82,485,106,70,CBS_DROPDOWN | CBS_SORT | 
                    WS_VSCROLL | WS_TABSTOP
    LTEXT           "Ambient sound",IDC_STATIC,27,484,50,10
    GROUPBOX        "Mirrors",IDC_STATIC,9,529,200,50
    LTEXT           "Current mirror face:",IDC_MIRROR_FACE_TEXT,15,539,125,
                    10
    PUSHBUTTON      "Set current face as mirror",IDC_MIRROR_BUTTON,20,553,70,
                    20,BS_MULTILINE | BS_FLAT
    PUSHBUTTON      "Set no mirror for this room",IDC_NO_MIRROR_BUTTON,115,
                    553,75,20,BS_MULTILINE | BS_FLAT
    PUSHBUTTON      "Flip Face",IDC_ROOMPAD_FLIP_FACE,55,425,33,20,BS_FLAT
    GROUPBOX        "Audio",IDC_STATIC,7,474,200,52
    CONTROL         "Draw one room only",IDC_ROOMPAD_ONE_ROOM_ONLY,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,122,449,80,10
    CONTROL         "Don't draw terrain",IDC_ROOMPAD_NO_DRAW_TERRAIN,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,122,464,85,10
    CONTROL         "Triangulate",IDC_TRIANGULATE,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,15,118,55,10
    LTEXT           "Name: ",IDC_ROOM_NAME,15,15,82,10
    PUSHBUTTON      "Edit Name",IDC_EDIT_NAME,80,30,25,25,BS_MULTILINE
    LTEXT           "Environment",IDC_STATIC,34,505,46,8
    COMBOBOX        IDC_ROOM_ENVAUDIO,81,503,107,44,CBS_DROPDOWNLIST | 
                    CBS_SORT | WS_VSCROLL | WS_TABSTOP
    EDITTEXT        IDC_ROOMPAD_DAMAGE,67,155,37,12,ES_AUTOHSCROLL
    LTEXT           "Damage Amount",IDC_STATIC,7,157,59,10
    COMBOBOX        IDC_ROOMPAD_DAMAGETYPE,67,171,67,66,CBS_DROPDOWN | 
                    WS_VSCROLL | WS_TABSTOP
    LTEXT           "Damage Type",IDC_STATIC,7,172,55,11
    CONTROL         "Render Shell",IDC_RENDER_SHELL,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,95,291,65,7
    CONTROL         "Render Non-Shell",IDC_RENDER_NON_SHELL,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,94,302,65,7
    CONTROL         "Secret",IDC_SECRET_CHECK,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,110,55,80,8
    CONTROL         "Blocked",IDC_PORTAL_BLOCKED_CHECK,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,65,355,45,8
    CONTROL         "Block can go away",IDC_PORTAL_BLOCKED_CAN_GO_AWAY_CHECK,
                    "Button",BS_AUTOCHECKBOX | WS_TABSTOP,115,355,80,8
    PUSHBUTTON      "Find Bad Center Pnts",IDC_BAD_CENTERPOINTS,90,425,50,20,
                    BS_MULTILINE
    CONTROL         "Do not light this room",IDC_ROOM_SKIP_LIGHTING,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,110,65,88,10
END

IDD_OBJECTPROPERTIES DIALOG DISCARDABLE  0, 0, 229, 284
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "Object Properties"
FONT 8, "MS Sans Serif"
BEGIN
    DEFPUSHBUTTON   "OK",IDOK,64,263,50,14
    PUSHBUTTON      "Cancel",IDCANCEL,115,263,50,14
    CONTROL         "Tab1",IDC_PROPERTYTAB,"SysTabControl32",0x0,7,7,215,254
END

IDD_PROPPHYSICS DIALOG DISCARDABLE  0, 0, 200, 212
STYLE WS_CHILD
FONT 8, "MS Sans Serif"
BEGIN
    CONTROL         "Uses Turnroll",IDC_PTURNROLL,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,10,16,65,10
    CONTROL         "Auto Levelling",IDC_PLEVELLING,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,10,24,64,10
    CONTROL         "Bounces",IDC_PBOUNCE,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,10,32,57,10
    CONTROL         "Wiggles",IDC_PWIGGLE,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,10,40,54,10
    CONTROL         "Sticks",IDC_PSTICKS,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,10,48,52,10
    CONTROL         "Persisten",IDC_PPERSISTENT,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,10,56,71,10
    CONTROL         "Uses Thrust",IDC_PUSESTHRUST,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,10,65,79,10
    CONTROL         "Affected by Gravity",IDC_PGRAVITY,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,10,73,80,10
    CONTROL         "Affected by Magnetism",IDC_PMAGNETISM,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,10,81,87,10
    CONTROL         "Affected by Wind",IDC_PWIND,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,10,89,86,10
    GROUPBOX        "Flags",IDC_STATIC,7,7,96,94
    EDITTEXT        IDC_PVELOCITY_X,21,114,26,12,ES_AUTOHSCROLL
    EDITTEXT        IDC_PVELOCITY_Y,21,127,26,12,ES_AUTOHSCROLL
    EDITTEXT        IDC_PVELOCITY_Z,21,139,26,12,ES_AUTOHSCROLL
    LTEXT           "X:",IDC_STATIC,14,116,8,10
    LTEXT           "Y:",IDC_STATIC,14,128,8,10
    LTEXT           "Z:",IDC_STATIC,14,140,8,10
    GROUPBOX        "Velocity",IDC_STATIC,7,101,42,52
    EDITTEXT        IDC_PTHRUST_X,21,166,26,12,ES_AUTOHSCROLL
    EDITTEXT        IDC_PTHRUST_Y,21,179,26,12,ES_AUTOHSCROLL
    EDITTEXT        IDC_PTHRUST_Z,21,191,26,12,ES_AUTOHSCROLL
    LTEXT           "X:",IDC_STATIC,14,168,8,10
    LTEXT           "Y:",IDC_STATIC,14,180,8,10
    LTEXT           "Z:",IDC_STATIC,14,192,8,10
    GROUPBOX        "Thrust",IDC_STATIC,7,153,42,52
    EDITTEXT        IDC_PROTVEL_X,62,114,26,12,ES_AUTOHSCROLL
    EDITTEXT        IDC_PROTVEL_Y,62,127,26,12,ES_AUTOHSCROLL
    EDITTEXT        IDC_PROTVEL_Z,62,139,26,12,ES_AUTOHSCROLL
    LTEXT           "X:",IDC_STATIC,55,116,8,10
    LTEXT           "Y:",IDC_STATIC,55,128,8,10
    LTEXT           "Z:",IDC_STATIC,55,140,8,10
    GROUPBOX        "Rot Velocity",IDC_STATIC,48,101,55,52
    EDITTEXT        IDC_PROTTHRUST_X,62,166,26,12,ES_AUTOHSCROLL
    EDITTEXT        IDC_PROTTHRUST_Y,62,179,26,12,ES_AUTOHSCROLL
    EDITTEXT        IDC_PROTTHRUST_Z,62,191,26,12,ES_AUTOHSCROLL
    LTEXT           "X:",IDC_STATIC,55,168,8,10
    LTEXT           "Y:",IDC_STATIC,55,180,8,10
    LTEXT           "Z:",IDC_STATIC,55,192,8,10
    GROUPBOX        "Rot Thrust",IDC_STATIC,48,153,55,52
    EDITTEXT        IDC_PTURNROLLANGLE,167,7,26,12,ES_AUTOHSCROLL
    RTEXT           "Turnroll Angle",IDC_STATIC,119,8,47,8
    EDITTEXT        IDC_PNUM_BOUNCES,167,85,26,12,ES_AUTOHSCROLL
    RTEXT           "Num. of Bounces",IDC_STATIC,111,88,56,8
    EDITTEXT        IDC_PMASS,167,20,26,12,ES_AUTOHSCROLL
    RTEXT           "Mass",IDC_STATIC,149,22,18,8
    EDITTEXT        IDC_PDRAG,167,98,26,12,ES_AUTOHSCROLL
    RTEXT           "Drag",IDC_STATIC,151,100,16,8
    EDITTEXT        IDC_PFULL_THRUST,167,33,26,12,ES_AUTOHSCROLL
    RTEXT           "Max. Thrust",IDC_STATIC,129,35,38,8
    EDITTEXT        IDC_PROTDRAG,167,111,26,12,ES_AUTOHSCROLL
    RTEXT           "Rotational Drag",IDC_STATIC,115,113,51,8
    EDITTEXT        IDC_PFULL_ROTTHRUST,167,46,26,12,ES_AUTOHSCROLL
    RTEXT           "Max. Rot. Thrust",IDC_STATIC,112,48,55,8
    EDITTEXT        IDC_PMAX_TURNROLL_RATE,167,124,26,12,ES_AUTOHSCROLL
    RTEXT           "Max. TurnRoll Rate",IDC_STATIC,104,126,63,8
    EDITTEXT        IDC_PTURNROLL_RATIO,167,59,26,12,ES_AUTOHSCROLL
    RTEXT           "TurnRoll Ratio",IDC_STATIC,119,61,47,8
    EDITTEXT        IDC_PWIGGLE_AMPLITUDE,167,137,26,12,ES_AUTOHSCROLL
    RTEXT           "Wiggle Amplitude",IDC_STATIC,111,140,56,8
    EDITTEXT        IDC_PWIGGLES_PER_SECOND,167,150,26,12,ES_AUTOHSCROLL
    RTEXT           "Wiggles Per Sec.",IDC_STATIC,111,152,56,8
    CONTROL         "Unlimited Bounce",IDC_PUNLIMBOUNCE,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,122,73,71,10
END

IDD_PROPAI DIALOG DISCARDABLE  0, 0, 209, 205
STYLE WS_CHILD
FONT 8, "MS Sans Serif"
BEGIN
END

IDD_PROPSCRIPT DIALOG DISCARDABLE  0, 0, 212, 227
STYLE WS_CHILD
FONT 8, "MS Sans Serif"
BEGIN
    LISTBOX         IDC_SCRIPTSEL_LISTBOX,15,25,115,115,LBS_SORT | 
                    LBS_NOINTEGRALHEIGHT | WS_VSCROLL | WS_TABSTOP
    LTEXT           "Scripts:",IDC_STATIC,15,15,115,10
    PUSHBUTTON      "Script Wizard",IDC_SCRIPTSEL_SCRIPT_WIZ,150,25,50,15
    PUSHBUTTON      "Parameters",IDC_PARAMETERS,150,45,50,14,WS_DISABLED
END

IDD_STATUSDLG DIALOGEX 0, 0, 275, 100
STYLE DS_MODALFRAME | DS_3DLOOK | DS_CENTER | WS_CHILD
EXSTYLE WS_EX_CLIENTEDGE | WS_EX_STATICEDGE
FONT 8, "MS Sans Serif", 0, 0, 0x1
BEGIN
    CONTROL         "Progress1",IDC_STATUSPROGRESS,"msctls_progress32",
                    WS_BORDER,7,49,261,26
    LTEXT           "Please Wait...",IDC_STATUSTEXT,17,26,251,13
END

IDD_SCRIPTEDITOR DIALOG DISCARDABLE  0, 0, 437, 297
STYLE DS_MODALFRAME | DS_3DLOOK | WS_MINIMIZEBOX | WS_POPUP | WS_CAPTION | 
    WS_SYSMENU
CAPTION "OSIRIS Editor"
MENU 181
FONT 8, "MS Sans Serif"
BEGIN
    DEFPUSHBUTTON   "Save and Exit",IDOK,360,7,70,14,NOT WS_TABSTOP
    PUSHBUTTON      "Cancel and Exit",IDCANCEL,360,24,70,14,NOT WS_TABSTOP
    CTEXT           "Currently Set To:",IDC_STATIC,360,127,70,8,NOT WS_GROUP
    CTEXT           "None",IDC_SCRIPTEXTERNAL,360,135,70,8
    PUSHBUTTON      "Set External Editor",IDC_SET_EXTERNAL_VIEWER,360,144,70,
                    14,NOT WS_TABSTOP
    PUSHBUTTON      "Use External Editor",IDC_USE_EXTERNAL,360,98,70,24,NOT 
                    WS_TABSTOP
    PUSHBUTTON      "Export script to file",ID_SCRIPTEDITOR_EXPORT,360,188,
                    70,14,NOT WS_TABSTOP
    PUSHBUTTON      "Import script from file",ID_SCRIPTEDITOR_IMPORT,360,204,
                    70,14,NOT WS_TABSTOP
    PUSHBUTTON      "Compile Script",IDC_SCRIPT_COMPILE,359,60,71,22,NOT 
                    WS_TABSTOP
    CONTROL         "",IDC_SCRMOD_BOX,"Static",SS_BLACKFRAME,5,10,345,280
    CONTROL         "",IDC_SCRIPVIEW,"RICHEDIT",ES_MULTILINE | ES_WANTRETURN,
                    5,10,345,280
END

IDD_OSIRIS_STATUS DIALOGEX 0, 0, 212, 140
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION
CAPTION "Compiling Status..."
FONT 8, "MS Sans Serif", 0, 0, 0x1
BEGIN
    PUSHBUTTON      "Done",IDC_OSIRIS_DONE,81,108,50,14
    GROUPBOX        "Progress",IDC_STATIC,7,7,198,44
    CTEXT           "Initializing...",IDC_OSIRIS_PROGRESS,13,18,186,25,
                    SS_CENTERIMAGE
    EDITTEXT        IDC_OSIRIS_ERROR,7,53,198,54,ES_MULTILINE | 
                    ES_AUTOVSCROLL | ES_AUTOHSCROLL | ES_OEMCONVERT | 
                    ES_READONLY | WS_VSCROLL | WS_HSCROLL,WS_EX_STATICEDGE
END

IDD_SELECTRANGE_DIALOG DIALOG DISCARDABLE  0, 0, 186, 81
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "Select Range"
FONT 8, "MS Sans Serif"
BEGIN
    DEFPUSHBUTTON   "OK",IDOK,129,7,50,14
    LTEXT           "Lower bound:",IDC_STATIC,5,5,45,10
    EDITTEXT        IDC_LOWER_BOUND_EDIT,50,5,55,12,ES_AUTOHSCROLL
    LTEXT           "Upper bound:",IDC_STATIC,5,20,45,10
    EDITTEXT        IDC_UPPER_BOUND_EDIT,50,20,55,12,ES_AUTOHSCROLL
    CONTROL         "Select by height change",IDC_SELECT_SLOPE_CHECK,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,5,40,93,10
    LTEXT           "Change",IDC_STATIC,5,55,40,10
    EDITTEXT        IDC_SLOPE_EDIT,45,55,30,12,ES_AUTOHSCROLL | WS_DISABLED
    CONTROL         "Invert selection",IDC_INVERSE_CHECK,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,100,40,65,10
    CONTROL         "Select by texture grid",IDC_TEXTURE_GRID_CHECK,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,100,50,80,8
    CONTROL         "Random",IDC_RANDOM_CHECK,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,100,60,60,10
    CONTROL         "Select only cur texture",IDC_SELECT_CUR_TEXTURE_CHECK,
                    "Button",BS_AUTOCHECKBOX | WS_TABSTOP,100,70,85,8
END

IDD_PHYSICS DIALOG DISCARDABLE  0, 0, 447, 227
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "Physics Settings"
FONT 8, "MS Sans Serif"
BEGIN
    EDITTEXT        IDC_PHYSICS_MASS_EDIT,105,5,40,12,ES_AUTOHSCROLL | 
                    ES_WANTRETURN
    EDITTEXT        IDC_PHYSICS_DRAG_EDIT,105,20,40,12,ES_AUTOHSCROLL | 
                    ES_WANTRETURN
    EDITTEXT        IDC_PHYSICS_ROTDRAG_EDIT,105,35,40,12,ES_AUTOHSCROLL | 
                    ES_WANTRETURN
    EDITTEXT        IDC_PHYSICS_MAXTHRUST_EDIT,390,100,40,12,ES_AUTOHSCROLL | 
                    ES_WANTRETURN
    EDITTEXT        IDC_PHYSICS_ROTTHRUST_EDIT,390,114,40,12,ES_AUTOHSCROLL | 
                    ES_WANTRETURN
    EDITTEXT        IDC_PHYSICS_MAXTURNROLLRATE_EDIT,390,40,40,12,
                    ES_AUTOHSCROLL | ES_WANTRETURN
    EDITTEXT        IDC_PHYSICS_TURNROLLRATIO_EDIT,390,54,40,12,
                    ES_AUTOHSCROLL | ES_WANTRETURN
    CONTROL         "Wind",IDC_PHYSICS_WIND_CHECK,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,10,170,40,10
    CONTROL         "Ignore Own Conc Forces",IDC_PHYSICS_MAGNETISM_CHECK,
                    "Button",BS_AUTOCHECKBOX | WS_TABSTOP,10,180,95,10
    CONTROL         "Wiggle Enabled",IDC_PHYSICS_WIGGLE_CHECK,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,170,20,75,10
    EDITTEXT        IDC_PHYSICS_WIGGLESIZE_EDIT,250,40,40,12,ES_AUTOHSCROLL | 
                    ES_WANTRETURN
    EDITTEXT        IDC_PHYSICS_WIGGLEFREQ_EDIT,250,55,40,12,ES_AUTOHSCROLL | 
                    ES_WANTRETURN
    DEFPUSHBUTTON   "OK",IDOK,390,140,50,14
    PUSHBUTTON      "Cancel",IDCANCEL,390,165,50,14
    LTEXT           "Mass",IDC_STATIC,10,5,18,8
    LTEXT           "Wiggle Amplitude",IDC_PHYSICS_WIGGLESIZE_TEXT,170,40,61,
                    8
    LTEXT           "Max.Thrust",IDC_PHYSICS_MAXTHRUST_TEXT,310,100,36,8
    LTEXT           "Max. Rot.Thrust",IDC_PHYSICS_ROTTHRUST_TEXT,310,116,52,
                    8
    LTEXT           "Drag",IDC_PHYSICS_DRAG_TEXT,10,20,16,8
    LTEXT           "Rot. Drag",IDC_PHYSICS_ROTDRAG_TEXT,10,35,40,10
    LTEXT           "Max. Turn/Roll Rate",IDC_PHYSICS_MAXTURNROLLRATE_TEXT,
                    310,40,70,10
    LTEXT           "Turn/Roll Ratio",IDC_PHYSICS_TURNROLLRATIO_TEXT,310,55,
                    70,10
    LTEXT           "Wiggles Per Second",IDC_PHYSICS_WIGGLES_PER_SEC_TEXT,
                    170,55,70,8
    GROUPBOX        "Wiggle",IDC_STATIC,165,5,135,65
    LTEXT           "Initial Forward Velocity",IDC_STATIC,10,50,75,10
    EDITTEXT        IDC_PHYSICS_INIT_VELOCITY_EDIT,105,50,40,12,
                    ES_AUTOHSCROLL | ES_WANTRETURN
    CONTROL         "Thrusts",IDC_PHYSICS_THRUSTS_CHECK,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,310,85,37,9
    CONTROL         "Bouncy",IDC_PHYSICS_BOUNCY_CHECK,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,175,85,47,9
    LTEXT           "Initial Rot. Velocity X",IDC_STATIC,10,65,80,10
    EDITTEXT        IDC_PHYSICS_INIT_ROT_VELOCITY_X_EDIT,105,65,40,12,
                    ES_AUTOHSCROLL | ES_WANTRETURN
    CONTROL         "Fixed Rate Rotate",IDC_PHYSICS_FIXED_ROTATE_CHECK,
                    "Button",BS_AUTOCHECKBOX | WS_TABSTOP,10,150,71,9
    EDITTEXT        IDC_PHYSICS_PERCENT_LOSS_EDIT,250,114,40,12,
                    ES_AUTOHSCROLL | ES_WANTRETURN
    LTEXT           "Percent Loss",IDC_PERCENT_LOSS_TEXT,175,115,42,9
    CONTROL         "Fixed Velocity",IDC_PHYSICS_FIXED_VELOCITY_CHECK,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,10,140,59,9
    EDITTEXT        IDC_PHYSICS_INIT_ROT_VELOCITY_Y_EDIT,105,80,40,12,
                    ES_AUTOHSCROLL | ES_WANTRETURN
    EDITTEXT        IDC_PHYSICS_INIT_ROT_VELOCITY_Z_EDIT,105,95,40,12,
                    ES_AUTOHSCROLL | ES_WANTRETURN
    LTEXT           "Initial Rot. Velocity Y",IDC_STATIC,10,81,80,12
    LTEXT           "Initial Rot. Velocity Z",IDC_STATIC,10,95,80,14
    GROUPBOX        "Bounce",IDC_STATIC,165,70,135,60
    CONTROL         "Gravity",IDC_PHYSICS_GRAVITY_RADIO,"Button",
                    BS_AUTORADIOBUTTON,115,130,50,10
    CONTROL         "No gravity",IDC_PHYSICS_NO_GRAVITY_RADIO,"Button",
                    BS_AUTORADIOBUTTON,115,140,50,10
    CONTROL         "Lighter than air",IDC_PHYSICS_REVERSE_GRAVITY_RADIO,
                    "Button",BS_AUTORADIOBUTTON,115,150,65,10
    CONTROL         "Sticky",IDC_PHYSICS_STICKY_CHECK,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,235,85,47,9
    LTEXT           "Max Bounces",IDC_MAX_BOUNCES_TEXT,175,100,60,9
    EDITTEXT        IDC_PHYSICS_MAX_BOUNCES_EDIT,250,100,40,12,
                    ES_AUTOHSCROLL | ES_WANTRETURN
    CONTROL         "Turn/Roll Enabled",IDC_TURN_ROLL_CHECK,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,310,20,75,10
    GROUPBOX        "Turn Roll",IDC_STATIC,305,5,135,65
    GROUPBOX        "Thrust",IDC_STATIC,305,70,135,60
    CONTROL         "Auto Leveling",IDC_PHYSICS_AUTO_LEVELING_CHECK,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,10,130,65,10
    CONTROL         "No Collide And No Relink -- ask Chris about use -- ",
                    IDC_NO_COLLIDE_CHECK,"Button",BS_AUTOCHECKBOX | 
                    BS_MULTILINE | WS_TABSTOP,115,185,95,15
    CONTROL         "Zero Radius Wall Collisions",
                    IDC_PHYSICS_POINT_WALL_CHECK,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,10,160,100,10
    CONTROL         "Ignore Robot Collisions",
                    IDC_PHYSICS_IGNORE_ROBOTS_CHECK,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,115,165,90,10
    LTEXT           "Terminal Velocity (drag, mass, and thrust):",
                    IDC_PHYSICS_TERMINAL_VELOCITY_STATIC,219,196,200,10
    LTEXT           "Terminal Rot. Vel. (rot. drag, mass, and rot. thrust):",
                    IDC_PHYSICS_TERMINAL_ROT_VELOCITY_STATIC,220,210,195,10
    CONTROL         "Persistent",IDC_PHYSICS_PERSISTENT_CHECK,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,10,190,55,10
    CONTROL         "Ignore Concussive Forces",
                    IDC_PHYSICS_IGNORE_CONCUSSIVE_FORCES,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,115,175,100,10
    EDITTEXT        IDC_PHYSICS_HIT_DIE_ANGLE,105,110,40,12,ES_AUTOHSCROLL | 
                    ES_WANTRETURN
    LTEXT           "Hit Death Angle (0 - 90)",IDC_PHYSICS_MAX_HIT_ANGLE,10,
                    110,75,10
    PUSHBUTTON      "Copy Physics Info",IDC_PHYSICS_COPY,220,165,60,15
    PUSHBUTTON      "Paste Physics Info",IDC_PHYSICS_PASTE,300,165,60,15
    GROUPBOX        "Axis Locked (Local Axes)",IDC_STATIC,220,135,160,25
    CONTROL         "X",IDC_L_X_CHECK,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,
                    225,145,20,10
    CONTROL         "Y",IDC_L_Y_CHECK,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,
                    250,145,20,10
    CONTROL         "Z",IDC_L_Z_CHECK,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,
                    275,145,20,10
    CONTROL         "P",IDC_L_P_CHECK,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,
                    310,145,20,10
    CONTROL         "H",IDC_L_H_CHECK,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,
                    335,145,20,10
    CONTROL         "B",IDC_L_B_CHECK,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,
                    360,145,20,10
    CONTROL         "Dont Do Big Multi Sphere",IDC_NO_BIG_SPHERE_CHECK,
                    "Button",BS_AUTOCHECKBOX | WS_TABSTOP,10,200,95,8
    CONTROL         "Ignore Same Type/ID Collisions",
                    IDC_PHYSICS_IGNORE_SAME_CHECK,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,10,210,130,8
    CONTROL         "Ignore Door Collisions",IDC_PHYSICS_IGNORE_DOOR_CHECK,
                    "Button",BS_AUTOCHECKBOX | WS_TABSTOP,115,201,90,10
END

IDD_AI DIALOG DISCARDABLE  0, 0, 557, 339
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "AI Settings"
FONT 8, "MS Sans Serif"
BEGIN
    CONTROL         "Target by Distance - not FOV",
                    IDC_AI_TARGET_BY_DIST_CHECK,"Button",BS_AUTOCHECKBOX | 
                    BS_MULTILINE | WS_TABSTOP,455,55,80,20
    CONTROL         "Fires weapon 2",IDC_AI_FIRE_WEAPON_2_CHECK,"Button",
                    BS_AUTOCHECKBOX | WS_DISABLED | WS_TABSTOP,455,75,75,8
    CONTROL         "Use Melee 1",IDC_AI_MELEE_1_CHECK,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,305,25,55,8
    CONTROL         "Use Melee 2",IDC_AI_MELEE_2_CHECK,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,305,40,55,8
    DEFPUSHBUTTON   "OK",IDOK,225,318,50,14
    PUSHBUTTON      "Cancel",IDCANCEL,285,318,50,14
    LTEXT           "Field of View",IDC_STATIC,305,95,47,8
    LTEXT           "Max Velocity",IDC_STATIC,150,55,50,8
    EDITTEXT        IDC_AI_MAX_VELOCITY_EDIT,225,55,40,12,ES_AUTOHSCROLL
    EDITTEXT        IDC_AI_MAX_VEL_TIME_EDIT,225,71,40,12,ES_AUTOHSCROLL
    EDITTEXT        IDC_AI_FOV_EDIT,405,95,35,12,ES_AUTOHSCROLL
    LTEXT           "Max Velocity Time",IDC_STATIC,150,71,65,8
    GROUPBOX        "Movement Information",IDC_STATIC,145,5,150,230
    LTEXT           "Movement Type",IDC_STATIC,150,20,55,10
    LTEXT           "Movement Subtype",IDC_STATIC,150,35,70,10
    LTEXT           "Max Turn Rate",IDC_STATIC,150,85,50,8
    EDITTEXT        IDC_AI_MAX_TURN_RATE_EDIT,225,85,40,12,ES_AUTOHSCROLL
    COMBOBOX        IDC_AI_MOVEMENT_TYPE_COMBO,225,20,65,50,CBS_DROPDOWN | 
                    CBS_SORT | WS_VSCROLL | WS_TABSTOP
    COMBOBOX        IDC_AI_MOVEMENT_SUBTYPE_COMBO,225,35,65,55,CBS_DROPDOWN | 
                    CBS_SORT | WS_VSCROLL | WS_TABSTOP
    CONTROL         "Use Physics for movement (Thrusts)",
                    IDC_AI_USES_THRUST_CHECK,"Button",BS_AUTOCHECKBOX | 
                    WS_DISABLED | WS_TABSTOP,150,195,130,10
    GROUPBOX        "AI Information",IDC_STATIC,5,5,135,75
    LTEXT           "AI Class",IDC_STATIC,11,20,55,10
    LTEXT           "AI Type",IDC_STATIC,11,35,55,10
    LTEXT           "AI Initial Action",IDC_STATIC,11,50,55,10,WS_DISABLED
    GROUPBOX        "Other Information",IDC_STATIC,300,5,250,125
    CONTROL         "Hears noise",IDC_HEARS_NOISE_CHECK,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,305,200,55,10
    COMBOBOX        IDC_AI_CLASS_COMBO,71,20,65,50,CBS_DROPDOWN | CBS_SORT | 
                    WS_VSCROLL | WS_TABSTOP
    COMBOBOX        IDC_AI_INITIAL_ACTION_COMBO,71,50,65,45,CBS_DROPDOWN | 
                    CBS_SORT | WS_DISABLED | WS_VSCROLL | WS_TABSTOP
    COMBOBOX        IDC_AI_TYPE_COMBO,71,35,65,45,CBS_DROPDOWN | CBS_SORT | 
                    WS_VSCROLL | WS_TABSTOP
    CONTROL         "Stays In/Out ",IDC_AI_STAY_IN_OUT_CHECK,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,410,160,100,10
    LTEXT           "Attack Sound",IDC_STATIC,7,304,44,8
    LTEXT           "See Sound",IDC_STATIC,7,320,36,8
    COMBOBOX        IDC_AI_ATTACK_SOUND_COMBO,75,305,104,30,CBS_DROPDOWN | 
                    CBS_SORT | WS_VSCROLL | WS_TABSTOP
    COMBOBOX        IDC_AI_SEE_SOUND_COMBO,75,320,104,30,CBS_DROPDOWN | 
                    CBS_SORT | WS_VSCROLL | WS_TABSTOP
    LTEXT           "Flee Sound",IDC_STATIC,383,304,37,8
    LTEXT           "Angry Sound",IDC_STATIC,383,320,42,8
    COMBOBOX        IDC_AI_FLEE_SOUND_COMBO,447,304,104,30,CBS_DROPDOWN | 
                    CBS_SORT | WS_VSCROLL | WS_TABSTOP
    COMBOBOX        IDC_AI_ANGRY_SOUND_COMBO,447,320,104,30,CBS_DROPDOWN | 
                    CBS_SORT | WS_VSCROLL | WS_TABSTOP
    GROUPBOX        "Default Notifications",IDC_STATIC,300,135,100,100
    CONTROL         "See Target",IDC_AI_NOTIFY_SEEPLAYER_CHECK,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,305,150,85,10
    CONTROL         "Near Wall",IDC_AI_NOTIFY_NEARWALL_CHECK,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,305,160,85,10
    CONTROL         "Hit by weapon",IDC_AI_NOTIFY_HITBYWEAPON_CHECK,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,305,170,85,10
    CONTROL         "Player Fired",IDC_AI_NOTIFY_OBJFIRED_CHECK,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,305,180,85,10
    CONTROL         "Weapon Hit Player",IDC_AI_NOTIFY_WEAPONHITPLAYER_CHECK,
                    "Button",BS_AUTOCHECKBOX | WS_TABSTOP,305,190,85,10
    GROUPBOX        "Default Actions",IDC_STATIC,5,135,135,100
    CONTROL         "Auto Dodge Weapons",IDC_AI_DODGE_CHECK,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,10,150,90,10
    CONTROL         "Auto Avoid Walls",IDC_AI_AVOIDWALLS_CHECK,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,10,170,95,10
    CONTROL         "Auto Aquire Targets",IDC_AI_RETARGET_CHECK,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,10,160,105,10
    CONTROL         "Auto Turret Control and Aiming",
                    IDC_AUTO_TURRETCONTROL_CHECK,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,10,180,115,10
    CONTROL         "Auto Firing Control",IDC_AI_FIREATTARGET_CHECK,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,10,190,105,10
    GROUPBOX        "Behavioral options",IDC_STATIC,405,135,145,100
    CONTROL         "Taunts Only At Target Death",IDC_TAUNTATDEATH_CHECK,
                    "Button",BS_AUTOCHECKBOX | WS_TABSTOP,410,150,105,10
    CONTROL         "Flinches when hurt (needs anim)",IDC_AI_FLINCH_CHECK,
                    "Button",BS_AUTOCHECKBOX | WS_TABSTOP,410,200,130,10
    CONTROL         "Persistant",IDC_PERSISTANT_CHECK,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,410,170,80,10
    CONTROL         "Fluctuate Speed Properties +/- 5%",
                    IDC_AI_FLUCTUATE_PROPERTIES,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,150,185,135,10
    LTEXT           "Max Turn Rate Time",IDC_STATIC,150,100,75,8
    EDITTEXT        IDC_AI_MAX_TURN_RATE_TIME_EDIT,225,100,40,12,
                    ES_AUTOHSCROLL
    LTEXT           "Dodge Velocity %",IDC_STATIC,150,115,65,10
    LTEXT           "Dodge Percentage",IDC_STATIC,305,105,65,10
    LTEXT           "Attack Velocity %",IDC_STATIC,150,130,70,10
    LTEXT           "Flee Velocity %",IDC_STATIC,150,145,70,10
    EDITTEXT        IDC_AI_DODGE_VEL_PER_EDIT,225,115,40,12,ES_AUTOHSCROLL
    EDITTEXT        IDC_AI_ATTACK_VEL_PER_EDIT,225,130,40,12,ES_AUTOHSCROLL
    EDITTEXT        IDC_AI_FLEE_VEL_PER_EDIT,225,145,40,12,ES_AUTOHSCROLL
    EDITTEXT        IDC_AI_DODGE_PER_EDIT,405,105,35,12,ES_AUTOHSCROLL
    LTEXT           "Circle Distance",IDC_STATIC,150,165,65,10
    EDITTEXT        IDC_AI_CIRCLE_DISTANCE_EDIT,225,160,40,15,ES_AUTOHSCROLL
    EDITTEXT        IDC_AI_MELEE1_DIST_EDIT,365,25,35,12,ES_AUTOHSCROLL
    EDITTEXT        IDC_AI_MELEE2_DIST_EDIT,365,40,35,12,ES_AUTOHSCROLL
    LTEXT           "Damage",IDC_STATIC,365,15,35,10
    EDITTEXT        IDC_AI_MELEE1_LATENCY_EDIT,405,25,35,12,ES_AUTOHSCROLL
    EDITTEXT        IDC_AI_MELEE2_LATENCY_EDIT,405,40,35,12,ES_AUTOHSCROLL
    LTEXT           "Latency",IDC_STATIC,410,15,30,10
    CONTROL         "PTMC Aligned",IDC_AI_PTMC_RADIO,"Button",
                    BS_AUTORADIOBUTTON,455,85,75,10
    CONTROL         "Rebel Aligned",IDC_AI_REBEL_RADIO,"Button",
                    BS_AUTORADIOBUTTON,455,95,75,10
    CONTROL         "Hostile",IDC_AI_HOSTILE_RADIO,"Button",
                    BS_AUTORADIOBUTTON,455,105,75,10
    CONTROL         "Neutral",IDC_AI_NEUTRAL_RADIO,"Button",
                    BS_AUTORADIOBUTTON,455,115,75,10
    DEFPUSHBUTTON   "Default Flyer",IDC_AI_FLYER_BUTTON,10,95,55,14
    DEFPUSHBUTTON   "Defualt Walker",IDC_AI_WALKER_BUTTON,75,110,55,14
    DEFPUSHBUTTON   "Default Turret",IDC_AI_TURRET_BUTTON,75,95,55,14
    DEFPUSHBUTTON   "Default Melee",IDC_AI_MELEE_BUTTON,10,110,55,14
    GROUPBOX        "Default Types",IDC_STATIC,5,85,135,45
    CONTROL         "Ordered WB Firing",IDC_AI_ORDERED_WB_FIRING_CHECK,
                    "Button",BS_AUTOCHECKBOX | WS_TABSTOP,410,180,75,10
    CONTROL         "Always Orientate to Velocity",IDC_AI_ORIENT_VEL_CHECK,
                    "Button",BS_AUTOCHECKBOX | WS_TABSTOP,410,190,120,10
    CONTROL         "Ignore Height diff. for distances",
                    IDC_AI_XZ_TARGET_DIST,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,305,115,130,10
    LTEXT           "Night Vision",IDC_STATIC,5,240,45,10
    LTEXT           "Fog Vision",IDC_STATIC,5,250,45,10
    LTEXT           "Infighting with team",IDC_STATIC,130,240,65,8
    LTEXT           "Infighting with same type",IDC_STATIC,130,250,65,8
    CONTROL         "Slider2",IDC_AI_NV_SLIDER,"msctls_trackbar32",TBS_BOTH | 
                    TBS_NOTICKS | WS_TABSTOP,50,240,75,10
    CONTROL         "Slider2",IDC_AI_FV_SLIDER,"msctls_trackbar32",TBS_BOTH | 
                    TBS_NOTICKS | WS_TABSTOP,50,250,75,10
    LTEXT           "Target Leading Accuracy",IDC_STATIC,275,240,85,8
    LTEXT           "Target Leading Varience",IDC_STATIC,275,250,80,8
    CONTROL         "Slider2",IDC_AI_FIGHT_TEAM_SLIDER,"msctls_trackbar32",
                    TBS_BOTH | TBS_NOTICKS | WS_TABSTOP,195,240,75,10
    CONTROL         "Slider2",IDC_AI_ROAMING_SLIDER,"msctls_trackbar32",
                    TBS_BOTH | TBS_NOTICKS | WS_TABSTOP,475,260,75,10
    CONTROL         "Slider2",IDC_AI_TLA_SLIDER,"msctls_trackbar32",TBS_BOTH | 
                    TBS_NOTICKS | WS_TABSTOP,360,240,75,10
    CONTROL         "Slider2",IDC_AI_HEARING_SLIDER,"msctls_trackbar32",
                    TBS_BOTH | TBS_NOTICKS | WS_TABSTOP,475,240,75,10
    LTEXT           "Curiousity",IDC_STATIC,5,260,35,10
    LTEXT           "Agression",IDC_STATIC,130,260,40,10
    LTEXT           "Life Perservation",IDC_STATIC,420,270,55,10
    CONTROL         "Slider2",IDC_AI_CURIOUSITY_SLIDER,"msctls_trackbar32",
                    TBS_BOTH | TBS_NOTICKS | WS_TABSTOP,50,260,75,10
    CONTROL         "Slider2",IDC_AI_AGRESSION_SLIDER,"msctls_trackbar32",
                    TBS_BOTH | TBS_NOTICKS | WS_TABSTOP,195,260,75,10
    CONTROL         "Slider2",IDC_AI_FIRE_SPREAD_SLIDER,"msctls_trackbar32",
                    TBS_BOTH | TBS_NOTICKS | WS_TABSTOP,360,260,75,10
    LTEXT           "Hearing",IDC_STATIC,440,240,30,10
    CONTROL         "Slider2",IDC_AI_TLV_SLIDER,"msctls_trackbar32",TBS_BOTH | 
                    TBS_NOTICKS | WS_TABSTOP,360,250,75,10
    LTEXT           "Frustration",IDC_STATIC,440,250,35,10
    CONTROL         "Slider2",IDC_AI_FRUSTRATION_SLIDER,"msctls_trackbar32",
                    TBS_BOTH | TBS_NOTICKS | WS_TABSTOP,475,250,75,10
    CONTROL         "Slider2",IDC_AI_FIGHT_SAME_SLIDER,"msctls_trackbar32",
                    TBS_BOTH | TBS_NOTICKS | WS_TABSTOP,195,250,75,10
    LTEXT           "Roaming",IDC_STATIC,440,260,30,10
    LTEXT           "Fire Spread Variance",IDC_STATIC,275,260,70,10
    CONTROL         "Slider2",IDC_AI_LIFE_PRESERVATION_SLIDER,
                    "msctls_trackbar32",TBS_BOTH | TBS_NOTICKS | WS_TABSTOP,
                    475,270,75,10
    LTEXT           "Turret Sound",IDC_STATIC,193,303,42,8
    COMBOBOX        IDC_AI_TURRET_SOUND_COMBO,260,300,104,78,CBS_DROPDOWN | 
                    CBS_SORT | WS_VSCROLL | WS_TABSTOP
    LTEXT           "Coop Same",IDC_STATIC,5,274,40,10
    CONTROL         "Slider2",IDC_AI_CURIOUSITY_SLIDER2,"msctls_trackbar32",
                    TBS_BOTH | TBS_NOTICKS | WS_TABSTOP,50,274,75,10
    LTEXT           "Leadership",IDC_STATIC,130,274,35,10
    CONTROL         "Slider2",IDC_AI_CURIOUSITY_SLIDER3,"msctls_trackbar32",
                    TBS_BOTH | TBS_NOTICKS | WS_TABSTOP,195,274,75,10
    CONTROL         "Auto Avoid Friends",IDC_AI_AVOID_FRIENDS_CHECK,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,10,200,77,10
    LTEXT           "Coop Team",IDC_STATIC,5,285,40,10
    CONTROL         "Slider2",IDC_AI_CURIOUSITY_SLIDER4,"msctls_trackbar32",
                    TBS_BOTH | TBS_NOTICKS | WS_TABSTOP,50,285,75,10
    EDITTEXT        IDC_AVOID_FRIENDS_DIST_EDIT,100,200,35,10,ES_AUTOHSCROLL
    CONTROL         "Biased Flight Height",IDC_BFH_CHECK,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,455,15,80,10
    LTEXT           "Min Height",IDC_STATIC,455,25,45,10
    LTEXT           "Max Height",IDC_STATIC,455,35,50,10
    CONTROL         "Slider1",IDC_BFHI_SLIDER,"msctls_trackbar32",TBS_BOTH | 
                    TBS_NOTICKS | WS_TABSTOP,495,45,45,10
    LTEXT           "Importance:",IDC_STATIC,455,45,40,10
    EDITTEXT        IDC_BFHMIN_EDIT,505,25,30,10,ES_AUTOHSCROLL
    EDITTEXT        IDC_BFHMAX_EDIT,505,35,30,10,ES_AUTOHSCROLL
    CONTROL         "FOV from Upward Vector (not Foward Vector)  -- FOR TURRETS",
                    IDC_AI_AIM_WITH_UVEC,"Button",BS_AUTOCHECKBOX | 
                    BS_MULTILINE | WS_TABSTOP,305,55,130,15
    CONTROL         "FOV from first weapon battery's aiming point",
                    IDC_AIM_FOV_CHECK,"Button",BS_AUTOCHECKBOX | 
                    BS_MULTILINE | WS_TABSTOP,305,75,125,15
END

IDD_WORLDOBJECTSGENERIC DIALOG DISCARDABLE  0, 0, 547, 281
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "World objects: robots, powerups, buildings, etc."
FONT 8, "MS Sans Serif"
BEGIN
    DEFPUSHBUTTON   "Done",IDOK,495,265,40,9
    LTEXT           "Explosion Sound",IDC_STATIC,209,129,54,8
    LTEXT           "Name",IDC_STATIC,10,35,20,8
    COMBOBOX        IDC_GENERIC_EXPLOSION_SOUND_COMBO,275,126,104,68,
                    CBS_DROPDOWN | CBS_SORT | WS_VSCROLL | WS_TABSTOP
    LTEXT           "Ambient Sound",IDC_STATIC,209,142,49,8
    COMBOBOX        IDC_GENERIC_AMBIENT_SOUND_COMBO,275,137,104,50,
                    CBS_DROPDOWN | CBS_SORT | WS_VSCROLL | WS_TABSTOP
    COMBOBOX        IDC_NAME_PULLDOWN,35,35,65,99,CBS_DROPDOWNLIST | 
                    CBS_SORT | WS_VSCROLL | WS_TABSTOP
    EDITTEXT        IDC_GENERIC_MODEL_NAME_EDIT,35,100,95,12,ES_AUTOHSCROLL | 
                    ES_READONLY
    PUSHBUTTON      "Change Model",IDC_GENERIC_CHANGE_MODEL,145,95,50,10
    PUSHBUTTON      "Add New",IDC_GENERIC_ADD_NEW,150,5,50,10
    PUSHBUTTON      "Delete",IDC_GENERIC_DELETE,150,20,50,10
    PUSHBUTTON      "Lock",IDC_GENERIC_LOCK,385,5,21,15
    PUSHBUTTON      "Check In",IDC_GENERIC_CHECKIN,410,5,35,15
    PUSHBUTTON      "<< Previous",IDC_GENERIC_PREV,35,55,42,15
    PUSHBUTTON      "Next >>",IDC_GENERIC_NEXT,80,55,30,15
    PUSHBUTTON      "Show Locked",IDC_GENERIC_CHECKED_OUT,495,5,35,20,
                    BS_MULTILINE
    LTEXT           "Size",IDC_STATIC,210,7,16,8
    EDITTEXT        IDC_GENERIC_SIZE_EDIT,243,5,40,12,ES_AUTOHSCROLL | 
                    ES_NUMBER
    PUSHBUTTON      "Edit Physics",IDC_GENERIC_EDIT_PHYSICS,465,165,65,15
    CONTROL         "Uses physics",IDC_GENERIC_USES_PHYSICS,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,465,153,65,10
    CONTROL         "Uses AI",IDC_GENERIC_USES_AI,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,465,180,65,10
    PUSHBUTTON      "Edit AI Settings",IDC_GENERIC_EDIT_AI,465,190,65,15
    PUSHBUTTON      "Copy",IDC_GENERIC_COPY,385,265,40,9
    PUSHBUTTON      "Paste",IDC_GENERIC_PASTE,430,265,40,9
    CONTROL         "",IDC_GENERIC_OBJECTVIEW,"Static",SS_ETCHEDFRAME,385,35,
                    150,105
    PUSHBUTTON      "Define Animation States",IDC_GENERIC_DEFINE_ANIMSTATES,
                    390,158,65,20,BS_MULTILINE
    LTEXT           "Type",IDC_STATIC,10,20,20,10
    EDITTEXT        IDC_GENERIC_TYPE_NAME,35,20,40,12,ES_AUTOHSCROLL | 
                    ES_READONLY
    EDITTEXT        IDC_GENERIC_ID_EDIT,95,20,35,12,ES_AUTOHSCROLL | 
                    ES_READONLY
    LTEXT           "ID",IDC_STATIC,85,20,8,8
    PUSHBUTTON      "Change Name",IDC_GENERIC_CHANGE_NAME,150,35,50,10
    LTEXT           "Model",IDC_STATIC,10,100,20,10
    GROUPBOX        "Current object",IDC_STATIC,5,5,130,70
    PUSHBUTTON      "Undo Lock",IDC_GENERIC_UNDO_LOCK,450,5,40,15
    CONTROL         "Destroyable",IDC_GENERIC_DESTROYABLE,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,209,70,55,10
    EDITTEXT        IDC_GENERIC_HITPOINT_EDIT,244,92,32,12,ES_AUTOHSCROLL
    LTEXT           "Hit Points",IDC_STATIC,209,92,30,10
    PUSHBUTTON      "Edit Weapon Info",IDC_GENERIC_WEAPON_INFO_BUTTON,390,
                    180,65,10
    GROUPBOX        "Additional Info",IDC_STATIC,385,145,150,120
    PUSHBUTTON      "Edit Light Info",IDC_GENERIC_LIGHT_BUTTON,390,195,65,10
    PUSHBUTTON      "Compute Default Size",IDC_DEFAULT_RADIUS_BUTTON,295,5,
                    85,15
    EDITTEXT        IDC_SCRIPTNAME,207,28,94,12,ES_AUTOHSCROLL
    PUSHBUTTON      "Browse",IDC_SELSCRIPT,305,27,50,15
    LTEXT           "Num. Polys",IDC_NUM_POLYS,284,113,65,10
    LTEXT           "Impact Size",IDC_STATIC,283,72,43,10
    LTEXT           "Impact Damage",IDC_STATIC,283,85,53,10
    EDITTEXT        IDC_GENERIC_IMPACT_SIZE_EDIT,341,72,39,12,ES_AUTOHSCROLL
    EDITTEXT        IDC_GENERIC_IMPACT_DAMAGE_EDIT,341,85,39,12,
                    ES_AUTOHSCROLL
    GROUPBOX        "Models",IDC_STATIC,5,75,195,65
    CONTROL         "Hi-res",IDC_HIRES_RADIO,"Button",BS_AUTORADIOBUTTON,10,
                    85,35,10
    CONTROL         "Med-res",IDC_MEDRES_RADIO,"Button",BS_AUTORADIOBUTTON,
                    50,85,40,10
    CONTROL         "Lo-res",IDC_LORES_RADIO,"Button",BS_AUTORADIOBUTTON,95,
                    85,35,10
    PUSHBUTTON      "Set none",IDC_NOLOD,145,105,50,10
    LTEXT           "Impact Time",IDC_STATIC,283,97,55,10
    EDITTEXT        IDC_GENERIC_IMPACT_TIME_EDIT,341,97,39,12,ES_AUTOHSCROLL
    GROUPBOX        "Inventory Information",IDC_STATIC,5,197,368,80
    EDITTEXT        IDC_GENERIC_INVEN_DESCRIPTION,10,241,255,31,ES_MULTILINE | 
                    ES_AUTOHSCROLL | ES_WANTRETURN
    LTEXT           "Item Description",IDC_STATIC,11,232,52,8
    EDITTEXT        IDC_GENERIC_INVEN_ICONNAME,280,257,87,14,ES_AUTOHSCROLL
    LTEXT           "Icon Name",IDC_STATIC,270,247,36,8
    PUSHBUTTON      "Override to unlocked",IDC_OVERRIDE,150,50,50,20,
                    BS_MULTILINE
    CONTROL         "Can be selected on HUD",IDC_GENERIC_INVEN_SELECTABLE,
                    "Button",BS_AUTOCHECKBOX | WS_TABSTOP,90,215,98,10
    CONTROL         "Can NOT be used from HUD",IDC_GENERIC_INVEN_NONUSEABLE,
                    "Button",BS_AUTOCHECKBOX | WS_TABSTOP,90,205,105,10
    EDITTEXT        IDC_LOD_DISTANCE_EDIT,170,82,25,12,ES_AUTOHSCROLL
    LTEXT           "LOD Dist",IDC_STATIC,137,81,30,10
    CONTROL         "Mission Item",IDC_INVTYPE_MISSION,"Button",
                    BS_AUTORADIOBUTTON | WS_GROUP,207,213,55,10
    CONTROL         "Game Item",IDC_INVTYPE_GAME,"Button",BS_AUTORADIOBUTTON | 
                    WS_GROUP,207,223,50,10
    GROUPBOX        "Item Type",IDC_STATIC,203,205,62,31,WS_GROUP
    CONTROL         "Don't remove when used",IDC_INVEN_NOREMOVE,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,90,225,93,10
    CONTROL         "Visible when used",IDC_INVEN_VISWHENUSED,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,10,212,73,10
    GROUPBOX        "Death Spew",IDC_STATIC,7,150,294,47
    EDITTEXT        IDC_DEATH_POWERUP1_NUM_EDIT,88,169,25,12,ES_AUTOHSCROLL
    LTEXT           "Generic ID 1",IDC_STATIC,13,159,65,10
    EDITTEXT        IDC_DEATH_POWERUP1_PERCENT_EDIT,118,169,25,12,
                    ES_AUTOHSCROLL
    LTEXT           "Number",IDC_STATIC,88,159,25,8
    LTEXT           "Percent",IDC_STATIC,118,159,25,8
    COMBOBOX        IDC_DEATH_POWERUP2_PULLDOWN,158,169,65,99,
                    CBS_DROPDOWNLIST | CBS_SORT | WS_VSCROLL | WS_TABSTOP
    EDITTEXT        IDC_DEATH_POWERUP2_NUM_EDIT,233,169,25,12,ES_AUTOHSCROLL
    LTEXT           "Generic ID 2",IDC_STATIC,158,159,65,10
    EDITTEXT        IDC_DEATH_POWERUP2_PERCENT_EDIT,263,169,25,12,
                    ES_AUTOHSCROLL
    LTEXT           "Number",IDC_STATIC,233,159,25,8
    LTEXT           "Percent",IDC_STATIC,263,159,25,8
    CONTROL         "If player has GID1, don't spew it",
                    IDC_DEATH_POWERUP_USE2_IF_HAVE1_CHECK,"Button",
                    BS_AUTOCHECKBOX | BS_MULTILINE | WS_TABSTOP,13,184,125,8
    COMBOBOX        IDC_DEATH_POWERUP1_PULLDOWN,13,169,65,99,
                    CBS_DROPDOWNLIST | CBS_SORT | WS_VSCROLL | WS_TABSTOP
    CONTROL         "Only spew GID2 if zero GID1s are spewed",
                    IDC_GENERIC_DEATH_SPEW_2_IF_ZERO_1,"Button",
                    BS_AUTOCHECKBOX | BS_MULTILINE | WS_TABSTOP,157,185,140,
                    8
    LTEXT           "Texture RAM:",IDC_TEXTURE_RAM_TEXT,10,120,85,10
    GROUPBOX        "Default Script Module",IDC_STATIC,203,19,180,51
    PUSHBUTTON      "Compile Module",IDC_COMPILEMODULE,305,52,72,14
    EDITTEXT        IDC_SCRIPT_OVERRIDE,208,53,93,14,ES_AUTOHSCROLL
    LTEXT           "Script Name Override",IDC_STATIC,208,44,71,8
    CONTROL         "AI Scripted Death",IDC_GENERIC_AI_SCRIPTED_DEATH,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,209,82,70,9
    LTEXT           "Respawn scalar:",IDC_STATIC,390,207,55,10
    EDITTEXT        IDC_RESPAWN_SCALAR_EDIT,445,207,35,12,ES_AUTOHSCROLL
    PUSHBUTTON      "Edit Deaths",IDC_GENERIC_DEATHS,488,205,42,15
    EDITTEXT        IDC_GENERIC_SCORE_EDIT,244,108,32,12,ES_AUTOHSCROLL
    LTEXT           "Score",IDC_STATIC,209,110,25,8
    EDITTEXT        IDC_GENERIC_AMMO_EDIT,333,168,32,12,ES_AUTOHSCROLL
    LTEXT           "Ammo Count",IDC_GENERIC_AMMO_TEXT,306,165,28,18
    CONTROL         "Do ceiling check (not for powerups)",
                    IDC_OBJ_CEILING_CHECK,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,390,241,140,10
    CONTROL         "Fly through glass and forcefields",
                    IDC_OBJECT_FLY_THROUGH_RENDERED_PORTALS,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,390,250,120,10
    CONTROL         "Don't scale damage taken by difficulty",IDC_NSC_BUTTON,
                    "Button",BS_AUTOCHECKBOX | WS_TABSTOP,390,230,141,10
    CONTROL         "Don't scale movement properties by diff",
                    IDC_DSMPBD_CHECK,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,
                    390,220,140,10
    CONTROL         "Ambient Object",IDC_GENERIC_AMBIENT,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,75,144,71,8
END

IDD_SCRIPTWIZ DIALOG DISCARDABLE  0, 0, 372, 181
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "ScriptWizard!"
FONT 8, "MS Sans Serif"
BEGIN
    PUSHBUTTON      "Add Script",IDC_ADDSCRIPT,310,5,55,15
    LTEXT           "Modules",IDC_STATIC,15,10,115,10
    LISTBOX         IDC_SCR_LISTBOX,15,50,115,80,LBS_SORT | 
                    LBS_NOINTEGRALHEIGHT | WS_VSCROLL | WS_TABSTOP
    LTEXT           "Scripts",IDC_STATIC,15,40,115,10
    PUSHBUTTON      "Lock",IDC_LOCK,310,75,55,15
    PUSHBUTTON      "Unlock",IDC_UNLOCK,310,95,55,15
    PUSHBUTTON      "Undo Lock",IDC_UNDOLOCK,310,115,55,15
    PUSHBUTTON      "Add Event",IDC_ADDEVENT,310,25,55,15
    COMBOBOX        IDC_SCRMOD_BOX,15,20,120,55,CBS_DROPDOWNLIST | CBS_SORT | 
                    WS_VSCROLL | WS_TABSTOP
    LISTBOX         IDC_SCREVT_LISTBOX,170,50,115,80,LBS_SORT | 
                    LBS_NOINTEGRALHEIGHT | WS_VSCROLL | WS_TABSTOP
    LTEXT           "Events",IDC_STATIC,170,40,115,10
    PUSHBUTTON      "Edit...",IDC_EDITSCRIPT,310,45,55,15
    PUSHBUTTON      "Cancel",IDCANCEL,310,160,55,15
    PUSHBUTTON      "Ok",IDOK,245,160,55,15
END

IDD_SCRIPTSTUDIO DIALOG DISCARDABLE  0, 0, 462, 301
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION
CAPTION "Script Studio"
MENU 181
FONT 8, "MS Sans Serif"
BEGIN
    CONTROL         "",IDC_SCRMOD_BOX,"Static",SS_BLACKFRAME,5,5,450,280
    LTEXT           "Line:",IDC_STATIC,5,290,15,10
    LTEXT           "Static",IDC_LINENUM,25,290,50,10
END

IDD_ROBOT_WEAPON_DIALOG DIALOGEX 0, 0, 579, 352
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "Weapon Battery Information"
FONT 8, "MS Sans Serif", 0, 0, 0x1
BEGIN
    DEFPUSHBUTTON   "OK",IDOK,245,330,50,14
    PUSHBUTTON      "Cancel",IDCANCEL,301,330,50,14
    COMBOBOX        IDC_BARREL_0_COMBO,106,55,70,135,CBS_DROPDOWNLIST | 
                    CBS_SORT | WS_VSCROLL | WS_TABSTOP
    COMBOBOX        IDC_BARREL_1_COMBO,106,69,70,126,CBS_DROPDOWNLIST | 
                    CBS_SORT | WS_VSCROLL | WS_TABSTOP
    COMBOBOX        IDC_BARREL_2_COMBO,106,84,70,111,CBS_DROPDOWNLIST | 
                    CBS_SORT | WS_VSCROLL | WS_TABSTOP
    COMBOBOX        IDC_BARREL_3_COMBO,106,99,70,96,CBS_DROPDOWNLIST | 
                    CBS_SORT | WS_VSCROLL | WS_TABSTOP
    COMBOBOX        IDC_BARREL_4_COMBO,106,114,70,81,CBS_DROPDOWNLIST | 
                    CBS_SORT | WS_VSCROLL | WS_TABSTOP
    COMBOBOX        IDC_BARREL_5_COMBO,106,129,70,66,CBS_DROPDOWNLIST | 
                    CBS_SORT | WS_VSCROLL | WS_TABSTOP
    COMBOBOX        IDC_BARREL_6_COMBO,106,144,70,51,CBS_DROPDOWNLIST | 
                    CBS_SORT | WS_VSCROLL | WS_TABSTOP
    COMBOBOX        IDC_BARREL_7_COMBO,106,160,70,35,CBS_DROPDOWNLIST | 
                    CBS_SORT | WS_VSCROLL | WS_TABSTOP
    CONTROL         "",IDC_GP_MASK_0_0,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,
                    255,70,10,10
    CONTROL         "",IDC_GP_MASK_0_1,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,
                    265,70,10,10
    CONTROL         "",IDC_GP_MASK_0_2,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,
                    275,70,10,10
    CONTROL         "",IDC_GP_MASK_0_3,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,
                    285,70,10,10
    CONTROL         "",IDC_GP_MASK_0_4,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,
                    295,70,10,10
    CONTROL         "",IDC_GP_MASK_0_5,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,
                    305,70,10,10
    CONTROL         "",IDC_GP_MASK_0_6,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,
                    315,70,10,10
    CONTROL         "",IDC_GP_MASK_0_7,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,
                    325,70,10,10
    CONTROL         "",IDC_GP_MASK_1_0,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,
                    255,85,10,10
    CONTROL         "",IDC_GP_MASK_1_1,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,
                    265,85,10,10
    CONTROL         "",IDC_GP_MASK_1_2,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,
                    275,85,10,10
    CONTROL         "",IDC_GP_MASK_1_3,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,
                    285,85,10,10
    CONTROL         "",IDC_GP_MASK_1_4,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,
                    295,85,10,10
    CONTROL         "",IDC_GP_MASK_1_5,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,
                    305,85,10,10
    CONTROL         "",IDC_GP_MASK_1_6,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,
                    315,85,10,10
    CONTROL         "",IDC_GP_MASK_1_7,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,
                    325,85,10,10
    CONTROL         "",IDC_GP_MASK_2_0,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,
                    255,100,10,10
    CONTROL         "",IDC_GP_MASK_2_1,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,
                    265,100,10,10
    CONTROL         "",IDC_GP_MASK_2_2,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,
                    275,100,10,10
    CONTROL         "",IDC_GP_MASK_2_3,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,
                    285,100,10,10
    CONTROL         "",IDC_GP_MASK_2_4,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,
                    295,100,10,10
    CONTROL         "",IDC_GP_MASK_2_5,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,
                    305,100,10,10
    CONTROL         "",IDC_GP_MASK_2_6,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,
                    315,100,10,10
    CONTROL         "",IDC_GP_MASK_2_7,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,
                    325,100,10,10
    CONTROL         "",IDC_GP_MASK_3_0,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,
                    255,114,10,10
    CONTROL         "",IDC_GP_MASK_3_1,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,
                    265,114,10,10
    CONTROL         "",IDC_GP_MASK_3_2,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,
                    275,114,10,10
    CONTROL         "",IDC_GP_MASK_3_3,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,
                    285,114,10,10
    CONTROL         "",IDC_GP_MASK_3_4,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,
                    295,114,10,10
    CONTROL         "",IDC_GP_MASK_3_5,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,
                    305,114,10,10
    CONTROL         "",IDC_GP_MASK_3_6,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,
                    315,114,10,10
    CONTROL         "",IDC_GP_MASK_3_7,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,
                    325,114,10,10
    CONTROL         "",IDC_GP_MASK_4_0,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,
                    255,130,10,10
    CONTROL         "",IDC_GP_MASK_4_1,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,
                    265,130,10,10
    CONTROL         "",IDC_GP_MASK_4_2,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,
                    275,130,10,10
    CONTROL         "",IDC_GP_MASK_4_3,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,
                    285,130,10,10
    CONTROL         "",IDC_GP_MASK_4_4,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,
                    295,130,10,10
    CONTROL         "",IDC_GP_MASK_4_5,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,
                    305,130,10,10
    CONTROL         "",IDC_GP_MASK_4_6,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,
                    315,130,10,10
    CONTROL         "",IDC_GP_MASK_4_7,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,
                    325,130,10,10
    CONTROL         "",IDC_GP_MASK_5_0,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,
                    255,145,10,10
    CONTROL         "",IDC_GP_MASK_5_1,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,
                    265,145,10,10
    CONTROL         "",IDC_GP_MASK_5_2,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,
                    275,145,10,10
    CONTROL         "",IDC_GP_MASK_5_3,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,
                    285,145,10,10
    CONTROL         "",IDC_GP_MASK_5_4,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,
                    295,145,10,10
    CONTROL         "",IDC_GP_MASK_5_5,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,
                    305,145,10,10
    CONTROL         "",IDC_GP_MASK_5_6,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,
                    315,145,10,10
    CONTROL         "",IDC_GP_MASK_5_7,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,
                    325,145,10,10
    CONTROL         "",IDC_GP_MASK_6_0,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,
                    255,160,10,10
    CONTROL         "",IDC_GP_MASK_6_1,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,
                    265,160,10,10
    CONTROL         "",IDC_GP_MASK_6_2,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,
                    275,160,10,10
    CONTROL         "",IDC_GP_MASK_6_3,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,
                    285,160,10,10
    CONTROL         "",IDC_GP_MASK_6_4,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,
                    295,160,10,10
    CONTROL         "",IDC_GP_MASK_6_5,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,
                    305,160,10,10
    CONTROL         "",IDC_GP_MASK_6_6,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,
                    315,160,10,10
    CONTROL         "",IDC_GP_MASK_6_7,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,
                    325,160,10,10
    CONTROL         "",IDC_GP_MASK_7_0,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,
                    255,175,10,10
    CONTROL         "",IDC_GP_MASK_7_1,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,
                    265,175,10,10
    CONTROL         "",IDC_GP_MASK_7_2,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,
                    275,175,10,10
    CONTROL         "",IDC_GP_MASK_7_3,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,
                    285,175,10,10
    CONTROL         "",IDC_GP_MASK_7_4,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,
                    295,175,10,10
    CONTROL         "",IDC_GP_MASK_7_5,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,
                    305,175,10,10
    CONTROL         "",IDC_GP_MASK_7_6,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,
                    315,175,10,10
    CONTROL         "",IDC_GP_MASK_7_7,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,
                    325,175,10,10
    LTEXT           "Num Masks",IDC_STATIC,205,45,40,10
    EDITTEXT        IDC_GB_NUM_MASKS_EDIT,245,45,40,12,ES_AUTOHSCROLL
    LTEXT           "Num Turrets For Battery: 0",IDC_NUM_GB_TURRETS_STATIC,
                    27,220,95,10
    LTEXT           "Delay till Fire",IDC_STATIC,205,60,40,8
    EDITTEXT        IDC_GB_MASK_LATENCY_0_EDIT,205,70,35,12,ES_AUTOHSCROLL
    EDITTEXT        IDC_GB_MASK_LATENCY_1_EDIT,205,85,35,12,ES_AUTOHSCROLL
    EDITTEXT        IDC_GB_MASK_LATENCY_2_EDIT,205,100,35,12,ES_AUTOHSCROLL
    EDITTEXT        IDC_GB_MASK_LATENCY_3_EDIT,205,115,35,12,ES_AUTOHSCROLL
    EDITTEXT        IDC_GB_MASK_LATENCY_4_EDIT,205,130,35,12,ES_AUTOHSCROLL
    EDITTEXT        IDC_GB_MASK_LATENCY_5_EDIT,205,145,35,12,ES_AUTOHSCROLL
    EDITTEXT        IDC_GB_MASK_LATENCY_6_EDIT,205,160,35,12,ES_AUTOHSCROLL
    EDITTEXT        IDC_GB_MASK_LATENCY_7_EDIT,205,175,35,12,ES_AUTOHSCROLL
    LTEXT           "Turret SPR",IDC_STATIC,26,287,50,10
    GROUPBOX        "Turret Information",IDC_STATIC,20,207,105,112
    GROUPBOX        "Firing Mask Information",IDC_STATIC,195,30,370,165
    CONTROL         "NA",IDC_BARREL_0_RADIO,"Button",BS_AUTORADIOBUTTON,40,
                    55,60,10
    CONTROL         "NA",IDC_BARREL_1_RADIO,"Button",BS_AUTORADIOBUTTON,40,
                    70,60,10
    CONTROL         "NA",IDC_BARREL_2_RADIO,"Button",BS_AUTORADIOBUTTON,40,
                    85,60,10
    CONTROL         "NA",IDC_BARREL_3_RADIO,"Button",BS_AUTORADIOBUTTON,40,
                    100,60,10
    CONTROL         "NA",IDC_BARREL_4_RADIO,"Button",BS_AUTORADIOBUTTON,40,
                    115,60,10
    CONTROL         "NA",IDC_BARREL_5_RADIO,"Button",BS_AUTORADIOBUTTON,40,
                    130,60,10
    CONTROL         "NA",IDC_BARREL_6_RADIO,"Button",BS_AUTORADIOBUTTON,40,
                    145,60,10
    CONTROL         "NA",IDC_BARREL_7_RADIO,"Button",BS_AUTORADIOBUTTON,40,
                    160,60,10
    LTEXT           "Current Gun Battery",IDC_STATIC,5,5,70,10
    COMBOBOX        IDC_CURRENT_GB_COMBO,85,5,55,37,CBS_DROPDOWN | CBS_SORT | 
                    WS_VSCROLL | WS_TABSTOP
    GROUPBOX        "Gun Battery Information",IDC_STATIC,7,20,568,305
    LTEXT           "Total Gun Points 0",IDC_TOTAL_GP_STATIC,320,222,70,10
    LTEXT           "Total Gun Batteries 0",IDC_TOTAL_GB_STATIC,235,221,80,
                    10
    LTEXT           "Reaction Time",IDC_STATIC,26,301,50,10
    LTEXT           "Field Of View",IDC_STATIC,26,271,50,10
    LTEXT           "30.0",IDC_TURRET_FOV_STATIC,91,271,32,10
    LTEXT           "0.5",IDC_TURRET_SPR_STATIC,91,287,31,10
    LTEXT           "0.25",IDC_TURRET_REACTION_TIME_STATIC,91,300,31,10
    COMBOBOX        IDC_CURRET_TURRET_COMBO,78,234,35,85,CBS_DROPDOWN | 
                    CBS_SORT | WS_VSCROLL | WS_TABSTOP
    LTEXT           "Current Turret",IDC_STATIC,26,234,45,10
    LTEXT           "OOF Subobject",IDC_STATIC,26,255,55,10
    LTEXT           "0",IDC_TURRET_SOBJ_STATIC,91,255,25,10
    GROUPBOX        "Aiming Rules",IDC_STATIC,138,205,80,112
    CONTROL         "View Cone Angle",IDC_VIEW_CONE_ANGLE_CHECK,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,141,220,70,10
    CONTROL         "Max Distance",IDC_MAX_DISTANCE_CHECK,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,141,250,70,10
    CONTROL         "XZ Plane Angle",IDC_XZ_PLANE_ANGLE_CHECK,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,141,279,70,10
    EDITTEXT        IDC_VIEW_CONE_ANGLE_EDIT,141,234,35,12,ES_AUTOHSCROLL
    EDITTEXT        IDC_MAX_DISTANCE_EDIT,141,265,35,12,ES_AUTOHSCROLL
    EDITTEXT        IDC_XZ_PLANE_ANGLE_EDIT,141,295,35,12,ES_AUTOHSCROLL
    LTEXT           "0",IDC_STATIC,257,60,10,10
    LTEXT           "7",IDC_STATIC,329,60,10,10
    LTEXT           "Gun Slot",IDC_STATIC,281,60,35,10
    LTEXT           "0",IDC_STATIC,30,55,10,10
    LTEXT           "1",IDC_STATIC,30,70,10,10
    LTEXT           "2",IDC_STATIC,30,85,10,10
    LTEXT           "3",IDC_STATIC,30,100,10,10
    LTEXT           "4",IDC_STATIC,30,115,10,10
    LTEXT           "5",IDC_STATIC,30,130,10,10
    LTEXT           "6",IDC_STATIC,30,145,10,10
    LTEXT           "7",IDC_STATIC,30,160,10,10
    GROUPBOX        "Gun Slots (Toggled slot is for aiming)",IDC_STATIC,20,
                    45,165,150
    LTEXT           "Energy Use",IDC_STATIC,320,237,43,12
    EDITTEXT        IDC_ENERGY_USAGE_EDIT,365,237,21,12,ES_AUTOHSCROLL | 
                    ES_WANTRETURN
    LTEXT           "Ammo Use",IDC_STATIC,235,237,40,8
    EDITTEXT        IDC_AMMO_USAGE_EDIT,280,237,21,12,ES_AUTOHSCROLL | 
                    ES_WANTRETURN
    COMBOBOX        IDC_WB_FIRE0_SOUND_PULLDOWN,340,70,87,60,CBS_DROPDOWN | 
                    CBS_SORT | WS_VSCROLL | WS_TABSTOP
    COMBOBOX        IDC_WB_FIRE1_SOUND_PULLDOWN,340,86,87,60,CBS_DROPDOWN | 
                    CBS_SORT | WS_VSCROLL | WS_TABSTOP
    COMBOBOX        IDC_WB_FIRE2_SOUND_PULLDOWN,340,100,87,60,CBS_DROPDOWN | 
                    CBS_SORT | WS_VSCROLL | WS_TABSTOP
    COMBOBOX        IDC_WB_FIRE3_SOUND_PULLDOWN,340,115,87,60,CBS_DROPDOWN | 
                    CBS_SORT | WS_VSCROLL | WS_TABSTOP
    COMBOBOX        IDC_WB_FIRE4_SOUND_PULLDOWN,340,130,87,60,CBS_DROPDOWN | 
                    CBS_SORT | WS_VSCROLL | WS_TABSTOP
    COMBOBOX        IDC_WB_FIRE5_SOUND_PULLDOWN,340,145,87,60,CBS_DROPDOWN | 
                    CBS_SORT | WS_VSCROLL | WS_TABSTOP
    COMBOBOX        IDC_WB_FIRE6_SOUND_PULLDOWN,340,160,87,60,CBS_DROPDOWN | 
                    CBS_SORT | WS_VSCROLL | WS_TABSTOP
    COMBOBOX        IDC_WB_FIRE7_SOUND_PULLDOWN,340,175,87,60,CBS_DROPDOWN | 
                    CBS_SORT | WS_VSCROLL | WS_TABSTOP
    GROUPBOX        "Misc. Information",IDC_STATIC,230,205,165,112,0,0,
                    HIDC_STATIC
    LTEXT           "Firing Sound",IDC_STATIC,360,60,45,10
    LTEXT           "Start",IDC_STATIC,444,60,20,10
    LTEXT           "Fire",IDC_STATIC,475,60,20,10
    EDITTEXT        IDC_WB_FIRE0_START_TICK,440,70,25,12,ES_AUTOHSCROLL
    EDITTEXT        IDC_WB_FIRE0_FIRE_TICK,470,70,25,12,ES_AUTOHSCROLL
    EDITTEXT        IDC_WB_FIRE1_START_TICK,440,85,25,12,ES_AUTOHSCROLL
    EDITTEXT        IDC_WB_FIRE1_FIRE_TICK,470,85,25,12,ES_AUTOHSCROLL
    EDITTEXT        IDC_WB_FIRE2_START_TICK,440,100,25,12,ES_AUTOHSCROLL
    EDITTEXT        IDC_WB_FIRE2_FIRE_TICK,470,100,25,12,ES_AUTOHSCROLL
    EDITTEXT        IDC_WB_FIRE3_START_TICK,440,115,25,12,ES_AUTOHSCROLL
    EDITTEXT        IDC_WB_FIRE3_FIRE_TICK,470,115,25,12,ES_AUTOHSCROLL
    EDITTEXT        IDC_WB_FIRE4_START_TICK,440,130,25,12,ES_AUTOHSCROLL
    EDITTEXT        IDC_WB_FIRE4_FIRE_TICK,470,130,25,12,ES_AUTOHSCROLL
    EDITTEXT        IDC_WB_FIRE5_START_TICK,440,145,25,12,ES_AUTOHSCROLL
    EDITTEXT        IDC_WB_FIRE5_FIRE_TICK,470,145,25,12,ES_AUTOHSCROLL
    EDITTEXT        IDC_WB_FIRE6_START_TICK,440,160,25,12,ES_AUTOHSCROLL
    EDITTEXT        IDC_WB_FIRE6_FIRE_TICK,470,160,25,12,ES_AUTOHSCROLL
    EDITTEXT        IDC_WB_FIRE7_START_TICK,440,175,25,12,ES_AUTOHSCROLL
    EDITTEXT        IDC_WB_FIRE7_FIRE_TICK,470,175,25,12,ES_AUTOHSCROLL
    LTEXT           "Local WB Anim",IDC_STATIC,475,50,60,10
    LTEXT           "End",IDC_STATIC,505,60,20,10
    EDITTEXT        IDC_WB_FIRE0_END_TICK,500,70,25,12,ES_AUTOHSCROLL
    EDITTEXT        IDC_WB_FIRE1_END_TICK,500,85,25,12,ES_AUTOHSCROLL
    EDITTEXT        IDC_WB_FIRE2_END_TICK,500,100,25,12,ES_AUTOHSCROLL
    EDITTEXT        IDC_WB_FIRE3_END_TICK,500,115,25,12,ES_AUTOHSCROLL
    EDITTEXT        IDC_WB_FIRE4_END_TICK,500,130,25,12,ES_AUTOHSCROLL
    EDITTEXT        IDC_WB_FIRE5_END_TICK,500,145,25,12,ES_AUTOHSCROLL
    EDITTEXT        IDC_WB_FIRE6_END_TICK,500,160,25,12,ES_AUTOHSCROLL
    EDITTEXT        IDC_WB_FIRE7_END_TICK,500,175,25,12,ES_AUTOHSCROLL
    LTEXT           "Cycle Time",IDC_STATIC,525,60,40,10
    EDITTEXT        IDC_WB_FIRE0_CYCLE_TIME,530,70,25,12,ES_AUTOHSCROLL
    EDITTEXT        IDC_WB_FIRE1_CYCLE_TIME,530,85,25,12,ES_AUTOHSCROLL
    EDITTEXT        IDC_WB_FIRE2_CYCLE_TIME,530,100,25,12,ES_AUTOHSCROLL
    EDITTEXT        IDC_WB_FIRE3_CYCLE_TIME,530,115,25,12,ES_AUTOHSCROLL
    EDITTEXT        IDC_WB_FIRE4_CYCLE_TIME,530,130,25,12,ES_AUTOHSCROLL
    EDITTEXT        IDC_WB_FIRE5_CYCLE_TIME,530,145,25,12,ES_AUTOHSCROLL
    EDITTEXT        IDC_WB_FIRE6_CYCLE_TIME,530,160,25,12,ES_AUTOHSCROLL
    EDITTEXT        IDC_WB_FIRE7_CYCLE_TIME,530,175,25,12,ES_AUTOHSCROLL
    CONTROL         "Spray ",IDC_SPRAY_CHECK,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,237,257,70,10
    CONTROL         "Use for melee 1",IDC_USEFORMELEE1_CHECK,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,237,267,65,8
    CONTROL         "Use for melee 2",IDC_USEFORMELEE2_CHECK,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,237,277,65,8
    CONTROL         "No WB Anim",IDC_WB_NO_ANIM_RADIO,"Button",
                    BS_AUTORADIOBUTTON | WS_GROUP,320,257,65,10
    CONTROL         "Local WB Anim",IDC_WB_LOCAL_ANIM_RADIO,"Button",
                    BS_AUTORADIOBUTTON,320,267,65,10
    CONTROL         "Model WB Anim",IDC_WB_MODEL_ANIM_RADIO,"Button",
                    BS_AUTORADIOBUTTON,320,276,65,10
    CONTROL         "Random Firing Order",IDC_WB_RANDOM_CHECK,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,237,287,80,8
    CONTROL         "Guided",IDC_GUIDED_CHECK,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,237,297,75,10
    CONTROL         "",IDC_GP_MASK_Q_0,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,
                    405,228,10,10
    CONTROL         "",IDC_GP_MASK_Q_1,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,
                    415,228,10,10
    CONTROL         "",IDC_GP_MASK_Q_2,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,
                    425,228,10,10
    CONTROL         "",IDC_GP_MASK_Q_3,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,
                    435,228,10,10
    CONTROL         "",IDC_GP_MASK_Q_4,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,
                    445,228,10,10
    CONTROL         "",IDC_GP_MASK_Q_5,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,
                    455,228,10,10
    CONTROL         "",IDC_GP_MASK_Q_6,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,
                    465,228,10,10
    CONTROL         "",IDC_GP_MASK_Q_7,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,
                    475,228,10,10
    LTEXT           "0",IDC_STATIC,407,218,10,10
    LTEXT           "7",IDC_STATIC,475,218,10,10
    LTEXT           "Gun Slot",IDC_STATIC,431,218,35,10
    GROUPBOX        "Quad Powerup Settings",IDC_STATIC,400,205,88,37
    CONTROL         "On/Off Weapon",IDC_ONOFF_CHECK,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,237,307,69,8
    CONTROL         "User timeout",IDC_USER_TIMEOUT_CHECK,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,322,307,60,8
    CONTROL         "Force Fire Dir to Forward Vector",
                    IDC_FORCE_TO_FVEC_CHECK,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,425,270,120,8
    CONTROL         "Aim with Forward Vector",IDC_WBAIMFORWARD_CHECK,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,425,280,120,8
    CONTROL         "Force Fire Dir to Target Dir",IDC_FORCE_TO_TARGET_CHECK,
                    "Button",BS_AUTOCHECKBOX | WS_TABSTOP,425,260,120,8
END

IDD_GENERICLIGHT DIALOG DISCARDABLE  0, 0, 315, 135
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "Generic Light Properties"
FONT 8, "MS Sans Serif"
BEGIN
    LTEXT           "Light Distance",IDC_STATIC,5,10,50,10
    EDITTEXT        IDC_GENERIC_LIGHT_DISTANCE,55,10,30,12,ES_AUTOHSCROLL
    LTEXT           "Start red",IDC_STATIC,5,45,35,10
    EDITTEXT        IDC_GENERIC_RED_LIGHT,55,45,30,12,ES_AUTOHSCROLL
    LTEXT           "Start green",IDC_STATIC,5,55,45,10
    EDITTEXT        IDC_GENERIC_GREEN_LIGHT,55,55,30,12,ES_AUTOHSCROLL
    LTEXT           "Start blue",IDC_STATIC,5,65,45,10
    EDITTEXT        IDC_GENERIC_BLUE_LIGHT,55,65,30,12,ES_AUTOHSCROLL
    LTEXT           "Time Interval",IDC_STATIC,5,20,45,10
    EDITTEXT        IDC_GENERIC_PULSE_TIME,55,20,30,12,ES_AUTOHSCROLL
    CONTROL         "1",IDC_TIME_CHECK1,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,180,40,8,10
    CONTROL         "1",IDC_TIME_CHECK2,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,190,40,8,10
    CONTROL         "1",IDC_TIME_CHECK3,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,200,40,8,10
    CONTROL         "1",IDC_TIME_CHECK4,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,210,40,8,10
    CONTROL         "1",IDC_TIME_CHECK5,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,220,40,8,10
    CONTROL         "1",IDC_TIME_CHECK6,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,230,40,8,10
    CONTROL         "1",IDC_TIME_CHECK7,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,240,40,8,10
    CONTROL         "1",IDC_TIME_CHECK8,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,250,40,8,10
    CONTROL         "Flicker erraticly",IDC_FLICKER_RADIO,"Button",
                    BS_AUTORADIOBUTTON,115,30,60,10
    CONTROL         "Use Timer Bits",IDC_USE_TIMEBITS_RADIO,"Button",
                    BS_AUTORADIOBUTTON,115,40,60,10
    CONTROL         "Pulse",IDC_PULSE,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,
                    115,60,35,10
    CONTROL         "Always on",IDC_ALWAYS_ON_RADIO,"Button",
                    BS_AUTORADIOBUTTON,115,20,50,8
    LTEXT           "End red",IDC_STATIC,5,80,35,10
    EDITTEXT        IDC_GENERIC_RED_LIGHT2,55,80,30,12,ES_AUTOHSCROLL
    LTEXT           "End green",IDC_STATIC,5,90,40,10
    EDITTEXT        IDC_GENERIC_GREEN_LIGHT2,55,90,30,12,ES_AUTOHSCROLL
    LTEXT           "End Blue",IDC_STATIC,5,100,45,10
    EDITTEXT        IDC_GENERIC_BLUE_LIGHT2,55,100,30,12,ES_AUTOHSCROLL
    GROUPBOX        "Light cast",IDC_STATIC,110,7,175,78
    GROUPBOX        "Object render style",IDC_STATIC,110,90,80,40
    CONTROL         "Static light",IDC_RENDER_STATIC,"Button",
                    BS_AUTORADIOBUTTON,115,100,75,8
    CONTROL         "Gouraud shading",IDC_RENDER_GOURAUD,"Button",
                    BS_AUTORADIOBUTTON,115,110,75,8
    CONTROL         "Lightmaps",IDC_RENDER_LIGHTMAPS,"Button",
                    BS_AUTORADIOBUTTON,115,121,75,8
    CONTROL         "Alternate with end color",IDC_PULSE_TO_SECOND,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,180,60,90,10
    CONTROL         "Flicker slightly",IDC_FLICKER_SLIGHTLY_CHECK,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,115,70,60,10
    LTEXT           "Factor",IDC_STATIC,180,70,25,10
    EDITTEXT        IDC_FLICKER_SLIGHTLY_EDIT,205,70,25,10,ES_AUTOHSCROLL
    CONTROL         "Directional",IDC_DIRECTIONAL,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,195,15,50,8
    LTEXT           "FOV:",IDC_STATIC,195,25,20,10
    EDITTEXT        IDC_FOV_EDIT,215,25,20,10,ES_AUTOHSCROLL
    GROUPBOX        "Misc Properties",IDC_STATIC,200,90,85,40
    CONTROL         "No specularity",IDC_NO_SPECULAR_CHECK,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,205,100,60,10
END

IDD_LEVELPROPERTIES DIALOG DISCARDABLE  0, 0, 187, 226
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "Room Properties"
FONT 8, "MS Sans Serif"
BEGIN
    DEFPUSHBUTTON   "OK",IDOK,70,205,50,14,NOT WS_TABSTOP
    PUSHBUTTON      "Cancel",IDCANCEL,125,205,50,14
    GROUPBOX        "Properties",IDC_STATIC,5,10,170,184
    COMBOBOX        IDC_AMBIENT_SOUND,72,27,83,88,CBS_DROPDOWNLIST | 
                    CBS_SORT | WS_VSCROLL | WS_TABSTOP
    COMBOBOX        IDC_ENVAUD_LIST,72,46,83,77,CBS_DROPDOWNLIST | CBS_SORT | 
                    WS_VSCROLL | WS_TABSTOP
    LTEXT           "Ambient",IDC_STATIC,25,30,34,8
    LTEXT           "Environment",IDC_STATIC,25,49,45,8
END

IDD_SCRIPTSELECT DIALOG DISCARDABLE  0, 0, 189, 181
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "Select Script"
FONT 8, "MS Sans Serif"
BEGIN
    DEFPUSHBUTTON   "OK",IDOK,135,5,50,14
    PUSHBUTTON      "Cancel",IDCANCEL,135,25,50,14
    LISTBOX         IDC_SCRIPTSEL_LISTBOX,5,15,115,160,LBS_SORT | 
                    LBS_NOINTEGRALHEIGHT | WS_VSCROLL | WS_TABSTOP
    LTEXT           "Scripts:",IDC_STATIC,5,5,115,10
    PUSHBUTTON      "Add script",IDC_SCRIPTSEL_ADD,135,140,50,15
    PUSHBUTTON      "Parameters",IDC_PARAMETERS,135,110,50,14,WS_DISABLED
    PUSHBUTTON      "Edit script",IDC_EDITSCRIPT,135,161,50,14,WS_DISABLED
    PUSHBUTTON      "Default List",IDC_SCRTYPE,135,70,50,14
END

IDD_POWPROP_DIALOG DIALOG DISCARDABLE  0, 0, 327, 206
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "Powerup properties"
FONT 8, "MS Sans Serif"
BEGIN
    DEFPUSHBUTTON   "OK",IDOK,265,10,50,14
    PUSHBUTTON      "Cancel",IDCANCEL,265,30,50,14
    GROUPBOX        "Scripts",IDC_STATIC,5,5,165,195
    EDITTEXT        IDC_POWERUP_SCRIPT_EDIT,15,25,145,160,ES_MULTILINE | 
                    ES_AUTOVSCROLL | ES_AUTOHSCROLL | ES_WANTRETURN | 
                    WS_VSCROLL | WS_HSCROLL
END

IDD_ADDSCRIPT DIALOG DISCARDABLE  0, 0, 222, 61
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "Add script"
FONT 8, "MS Sans Serif"
BEGIN
    DEFPUSHBUTTON   "OK",IDOK,165,5,50,14
    PUSHBUTTON      "Cancel",IDCANCEL,165,25,50,14
    EDITTEXT        IDC_EDITNAME,35,5,100,15,ES_AUTOHSCROLL
    LTEXT           "Name",IDC_STATIC,5,5,30,15
    COMBOBOX        IDC_TYPESEL,35,25,100,45,CBS_DROPDOWNLIST | CBS_SORT | 
                    WS_VSCROLL | WS_TABSTOP
    LTEXT           "Type",IDC_STATIC,5,25,30,15
END

IDD_PLAYER_WEAPONS DIALOG DISCARDABLE  0, 0, 154, 272
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "Player Weapons Dialog"
FONT 8, "MS Sans Serif"
BEGIN
    DEFPUSHBUTTON   "OK",IDOK,50,250,50,14
    GROUPBOX        "Additional Weapon Battery Properties",IDC_STATIC,10,55,
                    135,185
    COMBOBOX        IDC_CURRENT_WEAPON_BATTERY_COMBO,70,5,65,195,
                    CBS_DROPDOWN | WS_VSCROLL | WS_TABSTOP
    LTEXT           "Current Weapon",IDC_STATIC,10,5,55,10
    PUSHBUTTON      "Edit Weapon Battery Info",IDC_EDIT_WB_BUTTON,20,30,115,
                    15
    CONTROL         "Has continuous firing sound",
                    IDC_CONTINUOUS_FIRING_SOUND,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,25,105,95,10
    CONTROL         "Has firing release sound",IDC_FIRING_RELEASE_SOUND,
                    "Button",BS_AUTOCHECKBOX | WS_TABSTOP,25,135,90,10
    COMBOBOX        IDC_FIRING_SOUND_PULLDOWN,25,115,85,130,CBS_DROPDOWN | 
                    CBS_SORT | WS_VSCROLL | WS_TABSTOP
    COMBOBOX        IDC_RELEASE_SOUND_PULLDOWN,25,145,85,100,CBS_DROPDOWN | 
                    CBS_SORT | WS_VSCROLL | WS_TABSTOP
    COMBOBOX        IDC_SPEW_POWERUP_PULLDOWN,25,175,85,55,CBS_DROPDOWN | 
                    CBS_SORT | WS_VSCROLL | WS_TABSTOP
    LTEXT           "Spew powerup",IDC_STATIC,25,165,46,8
    EDITTEXT        IDC_MAX_AMMO,25,205,85,12,ES_AUTOHSCROLL
    LTEXT           "Max Ammo",IDC_STATIC,25,195,85,10
    CONTROL         "Show as tenths",IDC_SHOW_TENTHS,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,26,223,86,8
    CONTROL         "Zoom weapon",IDC_ZOOM,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,25,90,71,6
    CONTROL         "On/Off weapon",IDC_ONOFF,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,25,80,73,7
    CONTROL         "Fires like Fusion",IDC_FIRES_FUSION,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,25,70,75,5
END

IDD_SCRIPTPARMS DIALOG DISCARDABLE  0, 0, 184, 142
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "Parameters"
FONT 8, "MS Sans Serif"
BEGIN
    DEFPUSHBUTTON   "OK",IDOK,140,10,35,15
    PUSHBUTTON      "Cancel",IDCANCEL,140,30,35,14
    LTEXT           "None",IDC_PARM1,10,15,45,10
    EDITTEXT        IDC_PARMEDIT1,55,15,45,10,ES_AUTOHSCROLL | WS_DISABLED
    LTEXT           "None",IDC_PARM2,10,30,45,10
    EDITTEXT        IDC_PARMEDIT2,55,30,45,10,ES_AUTOHSCROLL | WS_DISABLED
    LTEXT           "None",IDC_PARM3,10,45,45,10
    EDITTEXT        IDC_PARMEDIT3,55,45,45,10,ES_AUTOHSCROLL | WS_DISABLED
    LTEXT           "None",IDC_PARM4,10,60,45,10
    EDITTEXT        IDC_PARMEDIT4,55,60,45,10,ES_AUTOHSCROLL | WS_DISABLED
    LTEXT           "None",IDC_PARM5,10,75,45,10
    EDITTEXT        IDC_PARMEDIT5,55,75,45,10,ES_AUTOHSCROLL | WS_DISABLED
    LTEXT           "None",IDC_PARM6,10,90,45,10
    EDITTEXT        IDC_PARMEDIT6,55,90,45,10,ES_AUTOHSCROLL | WS_DISABLED
    LTEXT           "None",IDC_PARM7,10,105,45,10
    EDITTEXT        IDC_PARMEDIT7,55,105,45,10,ES_AUTOHSCROLL | WS_DISABLED
    LTEXT           "None",IDC_PARM8,10,120,45,10
    EDITTEXT        IDC_PARMEDIT8,55,120,45,10,ES_AUTOHSCROLL | WS_DISABLED
    PUSHBUTTON      "<",IDC_PARAMLIST,105,15,15,10,WS_DISABLED
    PUSHBUTTON      "<",IDC_PARAMLIST2,105,30,15,10,WS_DISABLED
    PUSHBUTTON      "<",IDC_PARAMLIST3,105,45,15,10,WS_DISABLED
    PUSHBUTTON      "<",IDC_PARAMLIST4,105,60,15,10,WS_DISABLED
    PUSHBUTTON      "<",IDC_PARAMLIST5,105,75,15,10,WS_DISABLED
    PUSHBUTTON      "<",IDC_PARAMLIST6,105,90,15,10,WS_DISABLED
    PUSHBUTTON      "<",IDC_PARAMLIST7,105,105,15,10,WS_DISABLED
    PUSHBUTTON      "<",IDC_PARAMLIST8,105,120,15,10,WS_DISABLED
END

IDD_OBJTREEDLG DIALOG DISCARDABLE  0, 0, 239, 172
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "Objects"
FONT 8, "MS Sans Serif"
BEGIN
    DEFPUSHBUTTON   "OK",IDOK,190,10,45,14
    PUSHBUTTON      "Cancel",IDCANCEL,190,30,45,14
    CONTROL         "Tree1",IDC_OBJTREE,"SysTreeView32",TVS_HASBUTTONS | 
                    TVS_HASLINES | TVS_LINESATROOT | TVS_DISABLEDRAGDROP | 
                    WS_BORDER | WS_TABSTOP,5,10,175,155
    PUSHBUTTON      "Go to...",IDC_GOTO,190,130,45,15
    PUSHBUTTON      "Delete",IDC_DELETE,190,150,45,15
    PUSHBUTTON      "Clear All",IDC_CLEAR,190,75,45,14
END

IDD_LEVEL_INFO DIALOG DISCARDABLE  0, 0, 249, 172
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "Level information"
FONT 8, "MS Sans Serif"
BEGIN
    EDITTEXT        IDC_LEVEL_NAME,10,20,165,12,ES_AUTOHSCROLL
    EDITTEXT        IDC_DESIGNER,10,50,165,12,ES_AUTOHSCROLL
    EDITTEXT        IDC_COPYRIGHT,10,80,165,12,ES_AUTOHSCROLL
    EDITTEXT        IDC_NOTES,10,110,165,50,ES_MULTILINE | ES_AUTOHSCROLL | 
                    ES_WANTRETURN | WS_VSCROLL
    DEFPUSHBUTTON   "OK",IDOK,190,125,50,14
    PUSHBUTTON      "Cancel",IDCANCEL,190,145,50,14
    LTEXT           "Level name",IDC_STATIC,10,10,45,10
    LTEXT           "Designer(s)",IDC_STATIC,10,40,55,10
    LTEXT           "Notes",IDC_STATIC,10,100,40,10
    LTEXT           "Copyright",IDC_STATIC,10,70,55,10
END

IDD_PROCEDURAL_DIALOG DIALOG DISCARDABLE  0, 0, 457, 301
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "Procedurals"
FONT 8, "MS Sans Serif"
BEGIN
    PUSHBUTTON      "Update",IDC_PROCEDURAL_UPDATE,220,5,50,15
    CONTROL         "",IDC_PROCEDURAL_VIEW,"Static",SS_BLACKFRAME,5,5,210,
                    175
    PUSHBUTTON      "Clear",IDC_CLEAR_PROCEDURALS,220,165,50,15
    COMBOBOX        IDC_PROCEDURAL_PULLDOWN,220,45,100,105,CBS_DROPDOWN | 
                    WS_VSCROLL | WS_TABSTOP
    PUSHBUTTON      "Import PCX Palette",IDC_IMPORT_PALETTE,375,50,70,15
    CONTROL         "",IDC_PALETTE_VIEW,"Static",SS_BLACKFRAME,375,15,70,30
    GROUPBOX        "Palette",IDC_STATIC,370,5,80,95
    PUSHBUTTON      "Use Default Palette",IDC_DEFAULT_PALETTE,375,80,70,15
    LTEXT           "Heat",IDC_STATIC,220,25,25,10
    CONTROL         "Slider1",IDC_FP_HEAT_SLIDER,"msctls_trackbar32",
                    TBS_BOTH | TBS_NOTICKS | WS_TABSTOP,245,25,102,12
    LTEXT           "",IDC_HEAT_TEXT,350,25,20,10
    LTEXT           "Size",IDC_STATIC,220,140,25,10
    CONTROL         "Slider1",IDC_SIZE_SLIDER,"msctls_trackbar32",TBS_BOTH | 
                    TBS_NOTICKS | WS_TABSTOP,245,140,102,12
    LTEXT           "",IDC_SIZE_TEXT2,350,140,30,10
    LTEXT           "Freq",IDC_STATIC,220,125,25,10
    CONTROL         "Slider1",IDC_FREQ_SLIDER,"msctls_trackbar32",TBS_BOTH | 
                    TBS_NOTICKS | WS_TABSTOP,245,125,102,12
    LTEXT           "",IDC_FREQUENCY_TEXT,350,125,30,10
    LTEXT           "Speed",IDC_STATIC,220,110,25,10
    CONTROL         "Slider1",IDC_SPEED_SLIDER,"msctls_trackbar32",TBS_BOTH | 
                    TBS_NOTICKS | WS_TABSTOP,245,110,102,12
    LTEXT           "",IDC_SPEED_TEXT,350,110,30,10
    COMBOBOX        IDC_PROC_COPY_LIST,219,191,136,88,CBS_DROPDOWN | 
                    CBS_SORT | WS_VSCROLL | WS_TABSTOP
    PUSHBUTTON      "Copy Procedural",IDC_COPY_PROCEDURAL,363,190,67,14
END

IDD_TABLE_FILE_EDITOR DIALOGEX 0, 0, 315, 226
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU
EXSTYLE WS_EX_CLIENTEDGE
CAPTION "Table File Text Editor v1.0"
MENU IDR_TABLE_FILE_EDIT_MENU
FONT 8, "MS Sans Serif", 0, 0, 0x1
BEGIN
    GROUPBOX        "Current Generic Object",IDC_TABLEEDIT_MAIN_GROUP,9,8,
                    296,61,BS_LEFT
    COMBOBOX        IDC_TABLEEDIT_NAME_LIST,65,25,229,180,CBS_DROPDOWNLIST | 
                    CBS_SORT | WS_VSCROLL | WS_TABSTOP
    LTEXT           "Page Name:",IDC_STATIC,18,27,42,8
    PUSHBUTTON      " <<  Previous",IDC_TABLEEDIT_BTN_PREVIOUS,65,45,112,14
    PUSHBUTTON      "Next  >> ",IDC_TABLEEDIT_BTN_NEXT,182,45,112,14
    EDITTEXT        IDC_TABLEEDIT_DESCRIPTION_EDIT,20,130,275,76,
                    ES_MULTILINE | ES_WANTRETURN | WS_VSCROLL | WS_HSCROLL
    LTEXT           "Description:",IDC_STATIC,20,119,38,8
    GROUPBOX        "Generic Object's Data",IDC_STATIC,10,75,296,140
    EDITTEXT        IDC_TABLEEDIT_SCREEN_NAME_EDIT,71,95,224,14,
                    ES_AUTOHSCROLL
    LTEXT           "Screen Name:",IDC_STATIC,19,98,46,8
END

IDD_WATER_PROCEDURAL DIALOG DISCARDABLE  0, 0, 472, 236
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "Water Procedural Dialog"
FONT 8, "MS Sans Serif"
BEGIN
    PUSHBUTTON      "Update",IDC_PROCEDURAL_UPDATE,220,5,50,15
    CONTROL         "",IDC_WATER_PROCEDURAL_VIEW,"Static",SS_BLACKFRAME,5,5,
                    210,175
    PUSHBUTTON      "Clear",IDC_CLEAR_PROCEDURALS,220,165,50,15
    COMBOBOX        IDC_PROCEDURAL_PULLDOWN,220,75,100,105,CBS_DROPDOWN | 
                    WS_VSCROLL | WS_TABSTOP
    LTEXT           "Size",IDC_STATIC,220,125,15,10
    LTEXT           "Freq",IDC_STATIC,220,110,15,10
    LTEXT           "Height",IDC_STATIC,220,95,25,10
    LTEXT           "Thickness",IDC_STATIC,220,25,35,10
    LTEXT           "Light",IDC_STATIC,220,41,20,10
    LTEXT           "Eval. Time",IDC_STATIC,290,5,35,10
    EDITTEXT        IDC_PROC_EVAL_TIME_EDIT,330,5,25,10,ES_AUTOHSCROLL
    CONTROL         "Slider1",IDC_WP_HEIGHT_SLIDER,"msctls_trackbar32",
                    TBS_BOTH | TBS_NOTICKS | WS_TABSTOP,245,95,102,12
    CONTROL         "Slider1",IDC_WP_FREQUENCY_SLIDER,"msctls_trackbar32",
                    TBS_BOTH | TBS_NOTICKS | WS_TABSTOP,245,110,102,12
    CONTROL         "Slider1",IDC_WP_SIZE_SLIDER,"msctls_trackbar32",
                    TBS_BOTH | TBS_NOTICKS | WS_TABSTOP,245,125,102,12
    LTEXT           "",IDC_HEIGHT_TEXT,350,95,30,10
    LTEXT           "",IDC_FREQ_TEXT,350,110,30,10
    LTEXT           "",IDC_SIZE_TEXT,350,125,30,10
    CONTROL         "Slider1",IDC_WP_THICKNESS_SLIDER,"msctls_trackbar32",
                    TBS_BOTH | TBS_NOTICKS | WS_TABSTOP,260,25,102,12
    LTEXT           "",IDC_THICKNESS_TEXT,365,25,30,10
    CONTROL         "Slider1",IDC_WP_LIGHT_SLIDER,"msctls_trackbar32",
                    TBS_BOTH | TBS_NOTICKS | WS_TABSTOP,260,41,102,12
    LTEXT           "",IDC_LIGHT_TEXT,365,41,30,10
    LTEXT           "Osclt. Time",IDC_STATIC,220,55,35,10
    EDITTEXT        IDC_PROC_OSC_TIME_EDIT,260,55,25,10,ES_AUTOHSCROLL
    LTEXT           "Oscillate to value:",IDC_STATIC,310,55,60,10
    EDITTEXT        IDC_PROC_OSC_VALUE_EDIT,375,55,20,10,ES_AUTOHSCROLL
    COMBOBOX        IDC_PROC_COPY_LIST,219,191,136,88,CBS_DROPDOWN | 
                    CBS_SORT | WS_VSCROLL | WS_TABSTOP
    PUSHBUTTON      "Copy Procedural",IDC_COPY_PROCEDURAL,363,190,67,14
END

IDD_AMBIENTSOUNDPATTERNS DIALOG DISCARDABLE  0, 0, 214, 262
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "Ambient Sound Patterns"
FONT 8, "MS Sans Serif"
BEGIN
    DEFPUSHBUTTON   "OK",IDOK,65,245,35,14
    COMBOBOX        IDC_ASP_LIST,10,45,95,75,CBS_DROPDOWN | CBS_SORT | 
                    WS_VSCROLL | WS_TABSTOP
    LISTBOX         IDC_ASP_ELEMENT_LIST,5,145,170,90,LBS_NOINTEGRALHEIGHT | 
                    WS_VSCROLL | WS_TABSTOP
    LTEXT           "Ambient Sound Pattern",IDC_STATIC,10,35,90,10
    LTEXT           "Sound Elements",IDC_STATIC,5,135,115,10
    PUSHBUTTON      "Edit",IDC_ASP_EDIT_ELEMENT,180,165,25,15
    PUSHBUTTON      "New",IDC_ASP_NEW_ELEMENT,180,145,25,15
    PUSHBUTTON      "New Pattern",IDC_ASP_NEW,150,55,55,15
    PUSHBUTTON      "Delete Pattern",IDC_ASP_DELETE,150,75,55,15
    PUSHBUTTON      "Rename Pattern",IDC_ASP_RENAME,150,95,55,15
    PUSHBUTTON      "Lock",IDC_ASP_LOCK,40,5,40,15
    PUSHBUTTON      "Check in",IDC_ASP_CHECK_IN,85,5,40,15
    PUSHBUTTON      "Delete",IDC_ASP_DELETE_ELEMENT,180,185,25,15
    DEFPUSHBUTTON   "Cancel",IDCANCEL,110,245,35,14
    PUSHBUTTON      "Undo Lock",IDC_ASP_UNDO_LOCK,130,5,40,15
    EDITTEXT        IDC_ASP_MIN_DELAY,10,75,50,12,ES_AUTOHSCROLL
    LTEXT           "Max Delay",IDC_STATIC,10,95,40,10
    LTEXT           "Min Delay",IDC_STATIC,10,65,60,10
    EDITTEXT        IDC_ASP_MAX_DELAY,10,105,50,12,ES_AUTOHSCROLL
END

IDD_AMBIENTSOUNDELEMENT DIALOG DISCARDABLE  0, 0, 164, 112
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "Ambient Sound Element"
FONT 8, "MS Sans Serif"
BEGIN
    DEFPUSHBUTTON   "OK",IDOK,105,5,50,14
    PUSHBUTTON      "Cancel",IDCANCEL,105,22,50,14
    EDITTEXT        IDC_ASE_PROBABILITY,5,90,60,12,ES_AUTOHSCROLL
    EDITTEXT        IDC_ASE_MAX_VOLUME,50,60,35,12,ES_AUTOHSCROLL
    EDITTEXT        IDC_MIN_VOLUME,5,60,35,12,ES_AUTOHSCROLL
    COMBOBOX        IDC_ASE_SOUND_LIST,5,25,80,85,CBS_DROPDOWN | CBS_SORT | 
                    WS_VSCROLL | WS_TABSTOP
    LTEXT           "Sample",IDC_STATIC,5,15,65,10
    LTEXT           "Min volume",IDC_STATIC,5,50,40,10
    LTEXT           "Max volume",IDC_STATIC,50,50,40,10
    LTEXT           "Probability",IDC_STATIC,5,80,55,10
END

IDD_MATCENKEYPAD DIALOG DISCARDABLE  0, 0, 205, 611
STYLE DS_CONTROL | WS_CHILD
FONT 7, "Small Fonts"
BEGIN
    GROUPBOX        "Matcen Information",IDC_STATIC,5,0,195,605
    PUSHBUTTON      "Prev Matcen",IDC_MAT_PREV_BUTTON,15,10,55,15
    PUSHBUTTON      "Delete Matcen",IDC_MAT_DELETE_BUTTON,75,30,55,15
    PUSHBUTTON      "New Matcen",IDC_MAT_NEW_BUTTON,75,10,55,15
    PUSHBUTTON      "Next Matcen",IDC_MAT_NEXT_BUTTON,15,30,55,15
    LTEXT           "Number of Matcens: 1",IDC_MAT_NUM_STATIC,105,75,85,10
    LTEXT           "Matcen Name",IDC_STATIC,10,85,152,8
    EDITTEXT        IDC_MAT_NAME_EDIT,10,95,180,12,ES_AUTOHSCROLL
    PUSHBUTTON      "Copy Matcen",IDC_MAT_COPY_BUTTON,135,10,55,15
    PUSHBUTTON      "Paste Matcen",IDC_MAT_PASTE_BUTTON,135,30,55,15
    LTEXT           "Current Matcen: 1",IDC_MAT_CUR_STATIC,10,75,70,10
    CONTROL         "Enabled",IDC_MAT_ENABLED_CHECK,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,125,390,50,11
    CONTROL         "Room and faces",IDC_MAT_ROOM_RADIO,"Button",
                    BS_AUTORADIOBUTTON,15,155,90,10
    CONTROL         "Object and gunpoints",IDC_MAT_OBJECT_RADIO,"Button",
                    BS_AUTORADIOBUTTON,15,165,86,10
    EDITTEXT        IDC_MAT_NUM_SPAWN_EDIT,83,129,30,12,ES_AUTOHSCROLL
    LTEXT           "Num Spawn Points",IDC_STATIC,10,130,67,10
    EDITTEXT        IDC_MAT_SPAWN0_EDIT,155,144,30,12,ES_AUTOHSCROLL
    EDITTEXT        IDC_MAT_SPAWN2_EDIT,155,174,30,12,ES_AUTOHSCROLL
    EDITTEXT        IDC_MAT_SPAWN1_EDIT,155,159,30,12,ES_AUTOHSCROLL
    EDITTEXT        IDC_MAT_SPAWN3_EDIT,155,189,30,12,ES_AUTOHSCROLL
    GROUPBOX        "Spawn Pts",IDC_STATIC,149,129,42,80
    COMBOBOX        IDC_MAT_PROD_0_COMBO,10,230,70,55,CBS_DROPDOWN | 
                    CBS_SORT | WS_VSCROLL | WS_TABSTOP
    COMBOBOX        IDC_MAT_PROD_2_COMBO,10,261,70,54,CBS_DROPDOWN | 
                    CBS_SORT | WS_VSCROLL | WS_TABSTOP
    COMBOBOX        IDC_MAT_PROD_1_COMBO,10,246,70,54,CBS_DROPDOWN | 
                    CBS_SORT | WS_VSCROLL | WS_TABSTOP
    COMBOBOX        IDC_MAT_PROD_3_COMBO,10,277,70,48,CBS_DROPDOWN | 
                    CBS_SORT | WS_VSCROLL | WS_TABSTOP
    COMBOBOX        IDC_MAT_PROD_4_COMBO,10,290,70,50,CBS_DROPDOWN | 
                    CBS_SORT | WS_VSCROLL | WS_TABSTOP
    COMBOBOX        IDC_MAT_PROD_5_COMBO,10,306,70,54,CBS_DROPDOWN | 
                    CBS_SORT | WS_VSCROLL | WS_TABSTOP
    COMBOBOX        IDC_MAT_PROD_6_COMBO,10,320,70,50,CBS_DROPDOWN | 
                    CBS_SORT | WS_VSCROLL | WS_TABSTOP
    COMBOBOX        IDC_MAT_PROD_7_COMBO,10,335,70,55,CBS_DROPDOWN | 
                    CBS_SORT | WS_VSCROLL | WS_TABSTOP
    LTEXT           "Production List",IDC_STATIC,10,220,65,8
    EDITTEXT        IDC_MAT_PROD_0_QUATITY_EDIT,90,231,30,12,ES_AUTOHSCROLL
    EDITTEXT        IDC_MAT_PROD_1_QUATITY_EDIT,90,245,30,12,ES_AUTOHSCROLL
    EDITTEXT        IDC_MAT_PROD_2_QUATITY_EDIT,90,260,30,12,ES_AUTOHSCROLL
    EDITTEXT        IDC_MAT_PROD_3_QUATITY_EDIT,90,276,30,12,ES_AUTOHSCROLL
    EDITTEXT        IDC_MAT_PROD_4_QUATITY_EDIT,90,290,30,12,ES_AUTOHSCROLL
    EDITTEXT        IDC_MAT_PROD_5_QUATITY_EDIT,90,304,30,12,ES_AUTOHSCROLL
    EDITTEXT        IDC_MAT_PROD_6_QUATITY_EDIT,90,320,30,12,ES_AUTOHSCROLL
    EDITTEXT        IDC_MAT_PROD_7_QUATITY_EDIT,90,335,30,12,ES_AUTOHSCROLL
    EDITTEXT        IDC_MAT_PROD_0_SEC_EDIT,125,231,30,12,ES_AUTOHSCROLL
    EDITTEXT        IDC_MAT_PROD_1_SEC_EDIT,125,245,30,12,ES_AUTOHSCROLL
    EDITTEXT        IDC_MAT_PROD_2_SEC_EDIT,125,260,30,12,ES_AUTOHSCROLL
    EDITTEXT        IDC_MAT_PROD_3_SEC_EDIT,125,276,30,12,ES_AUTOHSCROLL
    EDITTEXT        IDC_MAT_PROD_4_SEC_EDIT,125,290,30,12,ES_AUTOHSCROLL
    EDITTEXT        IDC_MAT_PROD_5_SEC_EDIT,125,305,30,12,ES_AUTOHSCROLL
    EDITTEXT        IDC_MAT_PROD_6_SEC_EDIT,125,320,30,12,ES_AUTOHSCROLL
    EDITTEXT        IDC_MAT_PROD_7_SEC_EDIT,125,335,30,12,ES_AUTOHSCROLL
    EDITTEXT        IDC_MAT_PROD_0_PRIORITY_EDIT,160,231,30,12,
                    ES_AUTOHSCROLL
    EDITTEXT        IDC_MAT_PROD_1_PRIORITY_EDIT,160,245,30,12,
                    ES_AUTOHSCROLL
    EDITTEXT        IDC_MAT_PROD_2_PRIORITY_EDIT,160,260,30,12,
                    ES_AUTOHSCROLL
    EDITTEXT        IDC_MAT_PROD_3_PRIORITY_EDIT,160,276,30,12,
                    ES_AUTOHSCROLL
    EDITTEXT        IDC_MAT_PROD_4_PRIORITY_EDIT,160,290,30,12,
                    ES_AUTOHSCROLL
    EDITTEXT        IDC_MAT_PROD_5_PRIORITY_EDIT,160,304,30,12,
                    ES_AUTOHSCROLL
    EDITTEXT        IDC_MAT_PROD_6_PRIORITY_EDIT,160,320,30,12,
                    ES_AUTOHSCROLL
    EDITTEXT        IDC_MAT_PROD_7_PRIORITY_EDIT,160,335,30,12,
                    ES_AUTOHSCROLL
    EDITTEXT        IDC_MAX_OBJ_PROD_EDIT,125,356,30,12,ES_AUTOHSCROLL
    LTEXT           "Maximum objects produced",IDC_STATIC,10,356,100,10
    EDITTEXT        IDC_PROD_RATE_MULTI_EDIT,125,365,30,12,ES_AUTOHSCROLL
    LTEXT           "Production rate multiplier",IDC_STATIC,10,365,90,10
    LTEXT           "Quantity",IDC_STATIC,90,220,30,8
    LTEXT           "Seconds",IDC_STATIC,125,220,30,8
    LTEXT           "Priority",IDC_STATIC,160,220,30,8
    CONTROL         "Ordered Production",IDC_MAT_ORDERED_RADIO,"Button",
                    BS_AUTORADIOBUTTON,10,401,80,10
    CONTROL         "Random Production",IDC_MAT_RANDOM_RADIO,"Button",
                    BS_AUTORADIOBUTTON,10,391,90,10
    LTEXT           "Number of object types:",IDC_STATIC,10,200,80,10
    EDITTEXT        IDC_MAT_NUM_TYPES_EDIT,90,200,30,12,ES_AUTOHSCROLL
    PUSHBUTTON      "Move Viewer",IDC_MAT_MOVE_VIEWER_BUTTON,75,50,55,15
    EDITTEXT        IDC_MAX_ALIVE_CHILDREN_EDIT,125,375,30,12,ES_AUTOHSCROLL
    LTEXT           "Maximum alive children",IDC_STATIC,10,375,90,10
    COMBOBOX        IDC_MAT_ENABLED_SOUND_COMBO,90,470,105,55,CBS_DROPDOWN | 
                    CBS_SORT | WS_VSCROLL | WS_TABSTOP
    COMBOBOX        IDC_MAT_PROD_SOUND_COMBO,90,500,105,55,CBS_DROPDOWN | 
                    CBS_SORT | WS_VSCROLL | WS_TABSTOP
    COMBOBOX        IDC_MAT_DISABLED_SOUND_COMBO,90,485,105,55,CBS_DROPDOWN | 
                    CBS_SORT | WS_VSCROLL | WS_TABSTOP
    LTEXT           "Active Sound",IDC_STATIC,10,470,75,10
    LTEXT           "? sound - what we need",IDC_STATIC,10,485,75,10
    LTEXT           "Production Sound",IDC_STATIC,10,500,75,10
    CONTROL         "Script Controlled",IDC_MAT_SC_RADIO,"Button",
                    BS_AUTORADIOBUTTON,20,445,70,10
    CONTROL         "While player near",IDC_MAT_WPN_RADIO,"Button",
                    BS_AUTORADIOBUTTON,20,425,70,10
    CONTROL         "After player visible",IDC_MAT_APV_RADIO,"Button",
                    BS_AUTORADIOBUTTON,105,435,85,10
    CONTROL         "While player visible",IDC_MAT_WPV_RADIO,"Button",
                    BS_AUTORADIOBUTTON,20,435,85,10
    CONTROL         "After player near",IDC_MAT_APN_RADIO,"Button",
                    BS_AUTORADIOBUTTON,105,425,85,10
    GROUPBOX        "When to produce",IDC_STATIC,10,415,185,45
    COMBOBOX        IDC_MAT_PROD_EFFECT_COMBO,90,525,105,45,CBS_DROPDOWN | 
                    WS_VSCROLL | WS_TABSTOP
    LTEXT           "Production Effect",IDC_STATIC,10,525,75,10
    CONTROL         "Unassigned",IDC_MAT_UNASSIGNED_RADIO,"Button",
                    BS_AUTORADIOBUTTON,15,175,78,8
    PUSHBUTTON      "Paste Objref",IDC_MAT_PASTE_OBJREF_BUTTON,15,50,55,15
    LTEXT           "Attach Room/Objref",IDC_STATIC,10,115,71,10
    GROUPBOX        "Attach Type",IDC_STATIC,10,145,100,45
    LTEXT           "Pre-Production Time",IDC_STATIC,10,565,70,10
    LTEXT           "Post-Production Time",IDC_STATIC,10,580,80,10
    EDITTEXT        IDC_MAT_EFFECT_PREPROD_TIME_EDIT,90,565,30,12,
                    ES_AUTOHSCROLL
    EDITTEXT        IDC_MAT_EFFECT_POSTPROD_TIME_EDIT,90,580,30,12,
                    ES_AUTOHSCROLL
    COMBOBOX        IDC_MAT_PROD_TEXTURE_COMBO,90,540,105,45,CBS_DROPDOWN | 
                    CBS_SORT | WS_VSCROLL | WS_TABSTOP
    LTEXT           "Production Texture",IDC_STATIC,10,540,75,10
    CONTROL         "No Hurt Player",IDC_MAT_NHP_CHECK,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,125,400,70,11
    COMBOBOX        IDC_MAT_ROOM_OBJECT_COMBO,83,114,108,82,CBS_DROPDOWN | 
                    CBS_SORT | WS_VSCROLL | WS_TABSTOP
END

IDD_LEVELKEYPAD DIALOGEX 0, 0, 200, 908
STYLE DS_CONTROL | WS_CHILD
FONT 7, "Small Fonts", 0, 0, 0x1
BEGIN
    PUSHBUTTON      "New Goal",IDC_LEVEL_NEW_GOAL_BUTTON,10,195,45,15
    PUSHBUTTON      "Next Goal",IDC_LEVEL_NEXT_GOAL_BUTTON,80,195,45,15
    PUSHBUTTON      "Delete Goal",IDC_LEVEL_DELETE_GOAL_BUTTON,150,195,45,15
    PUSHBUTTON      "Prev Goal",IDC_LEVEL_PREV_GOAL_BUTTON,80,210,45,15
    EDITTEXT        IDC_LEVEL_GOAL_NAME_EDIT,10,255,185,12,ES_AUTOHSCROLL
    EDITTEXT        IDC_LEVEL_GOAL_LOC_NAME_EDIT,10,279,185,12,
                    ES_AUTOHSCROLL
    EDITTEXT        IDC_LEVEL_GOAL_DESC_EDIT,10,315,185,25,ES_MULTILINE | 
                    ES_AUTOVSCROLL | WS_VSCROLL
    EDITTEXT        IDC_LEVEL_GOAL_PRIORITY_EDIT1,145,380,50,12,
                    ES_AUTOHSCROLL
    EDITTEXT        IDC_EDIT57,145,390,50,14,ES_AUTOHSCROLL
    CONTROL         "Room",IDC_LEVEL_GOAL_ROOM_ITEM_RADIO,"Button",
                    BS_AUTORADIOBUTTON | WS_GROUP,15,466,38,10
    CONTROL         "Object Name",IDC_LEVEL_GOAL_OBJ_ITEM_RADIO,"Button",
                    BS_AUTORADIOBUTTON,15,481,59,10
    CONTROL         "Trigger Name",IDC_TRIGGER_RADIO,"Button",
                    BS_AUTORADIOBUTTON,15,495,70,10
    CONTROL         "Anywhere on the terrain (Only this item)",
                    IDC_LEVEL_GOAL_CELL_ITEM_RADIO,"Button",
                    BS_AUTORADIOBUTTON,15,511,143,10
    CONTROL         "Anywhere in a mine (Only this item)",
                    IDC_LEVEL_GOAL_MINE_RADIO,"Button",BS_AUTORADIOBUTTON,15,
                    521,129,10
    COMBOBOX        IDC_LEVELGOAL_ROOM_COMBO,90,465,100,53,CBS_DROPDOWN | 
                    CBS_SORT | WS_VSCROLL | WS_TABSTOP
    COMBOBOX        IDC_LEVELGOALOBJECT_COMBO,90,481,100,64,CBS_DROPDOWN | 
                    CBS_SORT | WS_VSCROLL | WS_TABSTOP
    COMBOBOX        IDC_LEVELGOALTRIGGER_COMBO,90,495,100,67,CBS_DROPDOWN | 
                    CBS_SORT | WS_VSCROLL | WS_TABSTOP
    CONTROL         "On Activate (triggers)",IDC_LG_ACTIVATE_RADIO,"Button",
                    BS_AUTORADIOBUTTON | WS_GROUP,15,551,85,10
    CONTROL         "On Enter(rooms/terrain)",IDC_LG_ENTER_RADIO,"Button",
                    BS_AUTORADIOBUTTON,15,561,95,10
    CONTROL         "On Pickup/Destroyed",IDC_LG_DESTROYED_RADIO,"Button",
                    BS_AUTORADIOBUTTON,15,570,90,10
    CONTROL         "Player or Player Weapon Collide",
                    IDC_LG_PPW_COLLIDE_RADIO,"Button",BS_AUTORADIOBUTTON,15,
                    580,135,10
    CONTROL         "Player Collide",IDC_LG_LC_RADIO,"Button",
                    BS_AUTORADIOBUTTON,110,551,85,10
    CONTROL         "DALLAS Controlled",IDC_LG_DALLAS_RADIO,"Button",
                    BS_AUTORADIOBUTTON,110,561,80,10
    CONTROL         "Secondary Goal",IDC_LEVEL_GOAL_SEC_GOAL_CHECK,"Button",
                    BS_AUTOCHECKBOX | WS_GROUP | WS_TABSTOP,10,601,70,8,
                    WS_EX_TRANSPARENT
    CONTROL         "Enabled",IDC_LEVEL_GOAL_ENABLED_CHECK,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,10,610,45,8
    CONTROL         "Completed",IDC_LEVEL_GOAL_COMPLETED_CHECK,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,10,620,55,8,
                    WS_EX_TRANSPARENT
    LTEXT           "Objective and Goal List",IDC_STATIC,10,631,115,8
    CONTROL         "This goal is an Objective",IDC_LEVEL_GOAL_TLG_CHECK,
                    "Button",BS_AUTOCHECKBOX | WS_TABSTOP,90,601,100,11
    CONTROL         "GB Doesn't Know Location",IDC_LEVEL_GOAL_GDKL_CHECK,
                    "Button",BS_AUTOCHECKBOX | WS_TABSTOP,90,610,105,10,
                    WS_EX_TRANSPARENT
    CONTROL         "Not Location Based",IDC_LEVEL_GOAL_NLB_CHECK,"Button",
                    BS_AUTOCHECKBOX | BS_MULTILINE | WS_TABSTOP,90,620,95,10,
                    WS_EX_TRANSPARENT
    GROUPBOX        "Completion Rule",IDC_STATIC,10,540,185,55
    LTEXT           "Gravity (32.2 is Earth)",IDC_STATIC,10,10,78,8
    EDITTEXT        IDC_LEVEL_GRAVITY_EDIT,90,10,38,12,ES_AUTOHSCROLL
    GROUPBOX        "Goal Information",IDC_STATIC,5,181,195,669
    LTEXT           "Priority (starts at zero)",IDC_STATIC,10,380,80,8
    LTEXT           "Location/Object name (I.e. Reactor Core)",IDC_STATIC,10,
                    270,137,8
    LTEXT           "Goal Description (I.e.  You must destroy the reactor to disable the matcens and open the excape hatch.)",
                    IDC_STATIC,10,295,185,15
    LTEXT           "Number of goals: 1",IDC_LEVEL_GOAL_NUM_GOALS_STATIC,10,
                    230,70,10
    LTEXT           "Goal name (I.e.  Destroy the reactor)",IDC_STATIC,10,
                    244,152,8
    GROUPBOX        "World Properties",IDC_STATIC,7,0,195,60
    LTEXT           "Current Goal: 1",IDC_LEVEL_GOAL_CUR_GOAL_STATIC,85,230,
                    70,10
    PUSHBUTTON      "Prev Item",IDC_LEVEL_GOAL_PREV_ITEM_BUTTON,80,431,50,10
    PUSHBUTTON      "Delete Item",IDC_LEVEL_GOAL_DELETE_ITEM_BUTTON,10,431,
                    50,10
    PUSHBUTTON      "New Item",IDC_LEVEL_GOAL_NEW_ITEM_BUTTON,10,415,50,10
    PUSHBUTTON      "Next Item",IDC_LEVEL_GOAL_NEXT_ITEM_BUTTON,80,415,50,10
    LTEXT           "Number of items: 1",IDC_LEVEL_GOAL_NUM_ITEMS_STATIC,15,
                    445,70,10
    LTEXT           "Current Items: 1",IDC_LEVEL_GOAL_CUR_ITEM_STATIC,105,
                    445,70,10
    GROUPBOX        "Current Goal Item (Must be same type)",IDC_STATIC,10,
                    455,185,80
    PUSHBUTTON      "Move Viewer",IDC_LEVEL_GOAL_GOTO_ITEM_BUTTON,145,431,50,
                    10
    GROUPBOX        "Ambient creatures",IDC_STATIC,7,851,190,50
    COMBOBOX        IDC_AMB_TYPE1,13,875,65,60,CBS_DROPDOWN | CBS_SORT | 
                    WS_VSCROLL | WS_TABSTOP
    EDITTEXT        IDC_AMB_N_GROUPS1,82,875,25,15,ES_AUTOHSCROLL
    EDITTEXT        IDC_AMB_MIN1,147,871,25,15,ES_AUTOHSCROLL
    CONTROL         "",IDC_AMB_OUTSIDE1,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,182,876,15,10
    COMBOBOX        IDC_AMB_TYPE2,10,876,65,60,CBS_DROPDOWN | CBS_SORT | 
                    WS_VSCROLL | WS_TABSTOP
    EDITTEXT        IDC_AMB_N_GROUPS2,82,871,25,15,ES_AUTOHSCROLL
    EDITTEXT        IDC_AMB_MIN2,147,871,25,15,ES_AUTOHSCROLL
    CONTROL         "",IDC_AMB_OUTSIDE2,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,182,876,15,10
    COMBOBOX        IDC_AMB_TYPE3,10,878,65,65,CBS_DROPDOWN | CBS_SORT | 
                    WS_VSCROLL | WS_TABSTOP
    EDITTEXT        IDC_AMB_N_GROUPS3,82,871,25,15,ES_AUTOHSCROLL
    EDITTEXT        IDC_AMB_MIN3,147,871,25,15,ES_AUTOHSCROLL
    CONTROL         "",IDC_AMB_OUTSIDE3,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,182,876,15,10
    COMBOBOX        IDC_AMB_TYPE4,10,878,65,60,CBS_DROPDOWN | CBS_SORT | 
                    WS_VSCROLL | WS_TABSTOP
    EDITTEXT        IDC_AMB_N_GROUPS4,82,871,25,15,ES_AUTOHSCROLL
    EDITTEXT        IDC_AMB_MIN4,147,871,25,15,ES_AUTOHSCROLL
    CONTROL         "",IDC_AMB_OUTSIDE4,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,182,876,15,10
    COMBOBOX        IDC_AMB_TYPE5,10,878,65,60,CBS_DROPDOWN | CBS_SORT | 
                    WS_VSCROLL | WS_TABSTOP
    EDITTEXT        IDC_AMB_N_GROUPS5,82,871,25,15,ES_AUTOHSCROLL
    EDITTEXT        IDC_AMB_MIN5,147,871,25,15,ES_AUTOHSCROLL
    CONTROL         "",IDC_AMB_OUTSIDE5,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,182,876,15,10
    COMBOBOX        IDC_AMB_TYPE6,10,875,65,65,CBS_DROPDOWN | CBS_SORT | 
                    WS_VSCROLL | WS_TABSTOP
    EDITTEXT        IDC_AMB_N_GROUPS6,82,871,25,15,ES_AUTOHSCROLL
    EDITTEXT        IDC_AMB_MIN6,147,871,25,15,ES_AUTOHSCROLL
    CONTROL         "",IDC_AMB_OUTSIDE6,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,182,876,15,10
    LTEXT           "Creature Type",IDC_STATIC,18,865,55,8
    LTEXT           "Alive",IDC_STATIC,82,865,25,8
    LTEXT           "Min",IDC_STATIC,147,865,25,8
    LTEXT           "Inside",IDC_STATIC,178,865,25,8
    LTEXT           "For each group",IDC_STATIC,118,855,50,8
    LTEXT           "Max",IDC_STATIC,82,855,20,8
    EDITTEXT        IDC_AMB_MAX1,118,871,25,15,ES_AUTOHSCROLL
    EDITTEXT        IDC_AMB_MAX2,118,871,25,15,ES_AUTOHSCROLL
    EDITTEXT        IDC_AMB_MAX3,118,871,25,15,ES_AUTOHSCROLL
    EDITTEXT        IDC_AMB_MAX4,118,871,25,15,ES_AUTOHSCROLL
    EDITTEXT        IDC_AMB_MAX5,118,871,25,15,ES_AUTOHSCROLL
    EDITTEXT        IDC_AMB_MAX6,118,871,25,15,ES_AUTOHSCROLL
    LTEXT           "Max",IDC_STATIC,118,865,25,8
    EDITTEXT        IDC_GOAL_TABLE_EDIT,10,642,185,203,ES_MULTILINE | 
                    ES_AUTOVSCROLL | ES_READONLY | ES_WANTRETURN | 
                    WS_VSCROLL
    LTEXT           "Completion HUD Message (I.e. Reactor destroyed!  Now get out of the mine!)",
                    IDC_STATIC,10,345,185,15,WS_DISABLED
    EDITTEXT        IDC_LEVEL_GOAL_COMPLETION_EDIT,10,365,185,12,
                    ES_AUTOHSCROLL | WS_DISABLED
    LTEXT           "Primary Objective List (0 - 3)",IDC_STATIC,10,395,100,8
    CONTROL         "Check ceiling collisions inside rooms and outside",
                    IDC_LEVEL_CEILING_EVERYWHERE,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,15,35,180,10
    EDITTEXT        IDC_LEVEL_CEILING_EDIT,90,22,38,12,ES_AUTOHSCROLL
    LTEXT           "Ceiling height",IDC_STATIC,10,22,65,8
    LTEXT           "AMBIENT CREATURES - NO USED -SOMEDAY?",IDC_STATIC,13,
                    891,180,10
    GROUPBOX        "Sound Overrides",IDC_STATIC,5,135,185,45
    COMBOBOX        IDC_FFOS_COMBO,100,160,90,98,CBS_DROPDOWN | CBS_SORT | 
                    WS_VSCROLL | WS_TABSTOP
    COMBOBOX        IDC_BGOS_COMBO,10,160,85,89,CBS_DROPDOWN | CBS_SORT | 
                    WS_VSCROLL | WS_TABSTOP
    LTEXT           "Forcefield Sound",IDC_STATIC,100,150,65,10
    LTEXT           "Glass Breaking Sound",IDC_STATIC,10,150,80,10
    COMBOBOX        IDC_FFT1_COMBO,15,75,95,80,CBS_DROPDOWN | CBS_SORT | 
                    WS_VSCROLL | WS_TABSTOP
    COMBOBOX        IDC_FFT2_COMBO,15,90,95,95,CBS_DROPDOWN | CBS_SORT | 
                    WS_VSCROLL | WS_TABSTOP
    GROUPBOX        "Special Forcefields and bounce factors",IDC_STATIC,10,
                    65,185,65
    EDITTEXT        IDC_FFT1_EDIT,103,879,50,15,ES_AUTOHSCROLL
    EDITTEXT        IDC_FFT2_EDIT,120,90,50,15,ES_AUTOHSCROLL
    EDITTEXT        IDC_FIXED_FFT1_EDIT,120,75,50,15,ES_AUTOHSCROLL
    CONTROL         "Powerups on this level should ignore wind",
                    IDC_LEVEL_PIW,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,15,
                    45,180,10
    COMBOBOX        IDC_FFT3_COMBO,15,105,95,95,CBS_DROPDOWN | CBS_SORT | 
                    WS_VSCROLL | WS_TABSTOP
    EDITTEXT        IDC_FFT3_EDIT,120,105,50,15,ES_AUTOHSCROLL
END

IDD_BRIEF_MAIN DIALOG DISCARDABLE  0, 0, 572, 335
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "Briefing Editor"
FONT 8, "MS Sans Serif"
BEGIN
    EDITTEXT        IDC_BRIEF_TITLE,458,23,84,14,ES_AUTOHSCROLL
    EDITTEXT        IDC_BRIEF_STATIC,459,47,20,14,ES_AUTOHSCROLL
    EDITTEXT        IDC_BRIEF_GLITCH,485,47,20,14,ES_AUTOHSCROLL
    COMBOBOX        IDC_BRIEF_SCREEN_LIST,464,78,85,98,CBS_DROPDOWN | 
                    WS_VSCROLL | WS_TABSTOP
    PUSHBUTTON      "Delete",IDC_BRIEF_SCREEN_DEL,465,93,41,14
    PUSHBUTTON      "Edit",IDC_BRIEF_SCREEN_EDIT,507,93,41,14
    PUSHBUTTON      "Add Screen",IDC_BRIEF_SCREEN_ADD,465,112,82,14
    COMBOBOX        IDC_BRIEF_EFFECT_LIST,465,144,83,107,CBS_DROPDOWN | 
                    WS_VSCROLL | WS_TABSTOP
    PUSHBUTTON      "Delete",IDC_BRIEF_EFFECT_DEL,466,159,41,14
    PUSHBUTTON      "Edit",IDC_BRIEF_EFFECT_EDIT,508,159,41,14
    PUSHBUTTON      "Add Text Effect",IDC_BRIEF_EFFECT_TEXT,465,178,82,14
    PUSHBUTTON      "Add Bitmap Effect",IDC_BRIEF_EFFECT_BITMAP,465,194,82,
                    14
    PUSHBUTTON      "Add Polymodel Effect",IDC_BRIEF_EFFECT_POLYMODEL,465,
                    210,82,14
    PUSHBUTTON      "Add Movie Effect",IDC_BRIEF_EFFECT_MOVIE,465,226,82,14
    PUSHBUTTON      "Add Sound Effect",IDC_BRIEF_EFFECT_SOUND,465,242,82,14
    PUSHBUTTON      "Add Button Effect",IDC_BRIEF_EFFECT_BUTTON,465,258,82,
                    14
    PUSHBUTTON      "Add Voice Effect",IDC_BRIEF_EFFECT_VOICE,465,274,82,14
    PUSHBUTTON      "Save",IDC_BRIEF_SAVE,451,296,36,14
    DEFPUSHBUTTON   "Quit",IDOK,529,296,36,14
    GROUPBOX        "Screens",IDC_STATIC,454,69,111,65
    GROUPBOX        "Effects on Current Screen",IDC_STATIC,454,135,111,159
    CONTROL         "",IDC_BRIEF_PICTURE,"Static",SS_BLACKFRAME,7,7,441,321
    GROUPBOX        "Global",IDC_STATIC,453,7,112,60
    LTEXT           "Title:",IDC_STATIC,458,15,16,8
    LTEXT           "Static:",IDC_STATIC,458,39,21,8
    LTEXT           "Glitch:",IDC_STATIC,484,39,21,8
    DEFPUSHBUTTON   "Load",IDC_LOAD,489,296,37,14
    PUSHBUTTON      "Manage Status",IDC_MANAGE,451,314,56,14
END

IDD_BRIEF_ADDSCREEN DIALOG DISCARDABLE  0, 0, 330, 159
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "Briefing Screen"
FONT 8, "MS Sans Serif"
BEGIN
    COMBOBOX        IDC_BRIEF_ADDS_LAYOUT_LIST,7,16,85,87,CBS_DROPDOWN | 
                    WS_VSCROLL | WS_TABSTOP
    DEFPUSHBUTTON   "OK",IDOK,7,58,50,14
    PUSHBUTTON      "Cancel",IDCANCEL,7,74,50,14
    LTEXT           "Layout Screen To Use",IDC_STATIC,7,7,72,8
    EDITTEXT        IDC_BRIEF_ADDS_DESC,7,102,94,14,ES_AUTOHSCROLL
    LTEXT           "Description",IDC_STATIC,7,92,63,8
    CONTROL         "",IDC_BRIEF_ADDS_PICTURE,"Static",SS_BLACKFRAME,130,7,
                    193,145
    PUSHBUTTON      "Mission Flags",IDC_MISSIONFLAGS,7,138,50,14
END

IDD_BRIEF_ADDTEXT DIALOG DISCARDABLE  0, 0, 302, 295
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "Text Effect"
FONT 8, "MS Sans Serif"
BEGIN
    EDITTEXT        IDC_BRIEF_T_ID,7,166,32,14,ES_AUTOHSCROLL | ES_NUMBER
    EDITTEXT        IDC_BRIEF_T_SPEED,7,192,32,14,ES_AUTOHSCROLL
    EDITTEXT        IDC_BRIEF_T_STARTTIME,7,217,32,14,ES_AUTOHSCROLL
    CONTROL         "Static",IDC_BRIEF_T_STATIC,"Button",BS_AUTORADIOBUTTON | 
                    WS_GROUP,89,166,34,10
    CONTROL         "Typing Out",IDC_BRIEF_T_TYPE,"Button",
                    BS_AUTORADIOBUTTON,89,176,51,10
    CONTROL         "Fade In",IDC_BRIEF_T_FADEIN,"Button",BS_AUTORADIOBUTTON,
                    89,186,40,10
    CONTROL         "Fade Out",IDC_BRIEF_T_FADEOUT,"Button",
                    BS_AUTORADIOBUTTON,89,196,45,10
    CONTROL         "Scroll Left to Right",IDC_BRIEF_T_SL2R,"Button",
                    BS_AUTORADIOBUTTON,89,206,74,10
    CONTROL         "Scroll Right To Left",IDC_BRIEF_T_SR2L,"Button",
                    BS_AUTORADIOBUTTON,89,216,77,10
    CONTROL         "Scroll Top To Bottom",IDC_BRIEF_T_ST2B,"Button",
                    BS_AUTORADIOBUTTON,89,226,80,10
    CONTROL         "Scroll Bottom To Top",IDC_BRIEF_T_SB2T,"Button",
                    BS_AUTORADIOBUTTON,89,236,81,10
    EDITTEXT        IDC_BRIEF_T_UL_X,243,169,20,13,ES_AUTOHSCROLL | 
                    ES_NUMBER
    EDITTEXT        IDC_BRIEF_T_UL_Y,269,169,20,13,ES_AUTOHSCROLL | 
                    ES_NUMBER
    EDITTEXT        IDC_BRIEF_T_LR_X,243,184,20,13,ES_AUTOHSCROLL | 
                    ES_NUMBER
    EDITTEXT        IDC_BRIEF_T_LR_Y,269,184,20,13,ES_AUTOHSCROLL | 
                    ES_NUMBER
    COMBOBOX        IDC_BRIEF_T_PREDEF,175,208,86,45,CBS_DROPDOWN | 
                    WS_VSCROLL | WS_GROUP | WS_TABSTOP
    DEFPUSHBUTTON   "OK",IDOK,245,258,50,14
    PUSHBUTTON      "Cancel",IDCANCEL,245,274,50,14
    LTEXT           "Text Information",IDC_STATIC,7,7,52,8
    LTEXT           "ID:",IDC_STATIC,7,157,10,8
    GROUPBOX        "Effect Type",IDC_STATIC,84,159,89,90,WS_GROUP
    LTEXT           "Speed:",IDC_STATIC,7,183,24,8
    LTEXT           "StartTime",IDC_STATIC,7,209,31,8
    LTEXT           "Upper Left Corner:",IDC_STATIC,177,171,61,8
    LTEXT           "Lower Right Corner:",IDC_STATIC,175,187,64,8
    LTEXT           "X Value",IDC_STATIC,241,159,26,8
    LTEXT           "Y Value",IDC_STATIC,269,159,26,8
    LTEXT           "Predef Coordinates",IDC_STATIC,176,200,63,8
    EDITTEXT        IDC_BRIEF_T_DESC,7,274,112,14,ES_AUTOHSCROLL
    LTEXT           "Effect Description",IDC_STATIC,7,265,63,8
    EDITTEXT        IDC_BRIEF_COLOR_R,133,274,21,14,ES_AUTOHSCROLL
    EDITTEXT        IDC_BRIEF_COLOR_G,164,274,21,14,ES_AUTOHSCROLL
    EDITTEXT        IDC_BRIEF_COLOR_B,195,274,22,14,ES_AUTOHSCROLL
    PUSHBUTTON      "Pal",IDC_BRIEF_COLOR_PAL,219,274,16,14
    LTEXT           "G",IDC_STATIC,155,278,8,8
    LTEXT           "B",IDC_STATIC,187,278,8,8
    LTEXT           "R",IDC_STATIC,126,278,8,8
    COMBOBOX        IDC_BRIEF_T_FONT,175,231,61,45,CBS_DROPDOWN | WS_VSCROLL | 
                    WS_GROUP | WS_TABSTOP
    LTEXT           "Font",IDC_STATIC,176,223,28,8
    CONTROL         "",IDC_RICHFRAME,"Static",SS_BLACKFRAME,7,15,288,138
    CONTROL         "Tab Stop (Can Scroll with Keyboard)",IDC_TABSTOP,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,7,252,130,10
    PUSHBUTTON      "Mission Flags",IDC_MISSIONFLAGS,245,224,50,14
END

IDD_BRIEF_ADDBITMAP DIALOG DISCARDABLE  0, 0, 195, 162
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "Bitmap Effect"
FONT 8, "MS Sans Serif"
BEGIN
    EDITTEXT        IDC_BRIEF_B_SPEED,7,15,27,14,ES_AUTOHSCROLL
    EDITTEXT        IDC_BRIEF_B_STARTTIME,43,15,27,14,ES_AUTOHSCROLL
    EDITTEXT        IDC_BRIEF_B_X,15,41,19,14,ES_AUTOHSCROLL | ES_NUMBER
    EDITTEXT        IDC_BRIEF_B_Y,46,41,19,14,ES_AUTOHSCROLL | ES_NUMBER
    EDITTEXT        IDC_BRIEF_B_FILENAME,7,102,110,14,ES_AUTOHSCROLL
    PUSHBUTTON      "Choose",IDC_BRIEF_B_CHOOSE,78,116,39,14
    CONTROL         "Static",IDC_BRIEF_B_STATIC,"Button",BS_AUTORADIOBUTTON | 
                    WS_GROUP,127,15,34,10
    CONTROL         "Blur In",IDC_BRIEF_B_BLURIN,"Button",BS_AUTORADIOBUTTON,
                    127,25,36,10
    CONTROL         "Blur Out",IDC_BRIEF_B_BLUROUT,"Button",
                    BS_AUTORADIOBUTTON,127,35,41,10
    CONTROL         "Scanline In",IDC_BRIEF_B_SCANIN,"Button",
                    BS_AUTORADIOBUTTON,127,45,51,10
    CONTROL         "Scanline Out",IDC_BRIEF_B_SCANOUT,"Button",
                    BS_AUTORADIOBUTTON,127,55,57,10
    CONTROL         "Inverse In",IDC_BRIEF_B_INVIN,"Button",
                    BS_AUTORADIOBUTTON,127,65,47,10
    CONTROL         "Inverse Out",IDC_BRIEF_B_INVOUT,"Button",
                    BS_AUTORADIOBUTTON,127,75,53,10
    CONTROL         "Stretch In",IDC_BRIEF_B_STRETCHIN,"Button",
                    BS_AUTORADIOBUTTON,127,85,47,10
    CONTROL         "Stretch Out",IDC_BRIEF_B_STRETCHOUT,"Button",
                    BS_AUTORADIOBUTTON,127,95,52,10
    DEFPUSHBUTTON   "OK",IDOK,7,126,50,14,WS_GROUP
    PUSHBUTTON      "Cancel",IDCANCEL,7,141,50,14
    GROUPBOX        "Effect Types",IDC_STATIC,122,7,64,102,WS_GROUP
    LTEXT           "Speed:",IDC_STATIC,7,7,24,8
    LTEXT           "StartTime:",IDC_STATIC,43,7,33,8
    LTEXT           "Upper Left Corner Position:",IDC_STATIC,7,31,86,8
    LTEXT           "X:",IDC_STATIC,7,43,8,8
    LTEXT           "Y:",IDC_STATIC,37,43,8,8
    LTEXT           "Filename:",IDC_STATIC,7,93,31,8
    EDITTEXT        IDC_BRIEF_B_DESC,76,141,112,14,ES_AUTOHSCROLL
    LTEXT           "Effect Description",IDC_STATIC,75,132,63,8
    COMBOBOX        IDC_BRIEF_B_PREDEF,7,80,68,53,CBS_DROPDOWN | WS_VSCROLL | 
                    WS_TABSTOP
    LTEXT           "Predef Coordinates",IDC_STATIC,8,71,61,8
    PUSHBUTTON      "Mission Flags",IDC_MISSIONFLAGS,142,113,46,14
    CONTROL         "Don't render until starttime",IDC_BRIEF_B_NORENDER,
                    "Button",BS_AUTOCHECKBOX | WS_TABSTOP,7,59,97,10
END

IDD_BRIEF_ADDMOVIE DIALOG DISCARDABLE  0, 0, 218, 105
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "Movie Effect"
FONT 8, "MS Sans Serif"
BEGIN
    EDITTEXT        IDC_BRIEF_M_X,21,17,19,14,ES_AUTOHSCROLL | ES_NUMBER
    EDITTEXT        IDC_BRIEF_M_Y,52,17,19,14,ES_AUTOHSCROLL | ES_NUMBER
    EDITTEXT        IDC_BRIEF_M_FILENAME,101,17,110,14,ES_AUTOHSCROLL
    EDITTEXT        IDC_BRIEF_M_FPS,13,41,22,14,ES_AUTOHSCROLL
    EDITTEXT        IDC_BRIEF_M_STARTTIME,39,41,27,14,ES_AUTOHSCROLL
    CONTROL         "Looping",IDC_BRIEF_M_LOOPING,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,102,34,41,10
    PUSHBUTTON      "Choose",IDC_BRIEF_M_CHOOSE,172,31,39,14
    DEFPUSHBUTTON   "OK",IDOK,57,84,50,14
    PUSHBUTTON      "Cancel",IDCANCEL,109,84,50,14
    LTEXT           "Upper Left Corner Position:",IDC_STATIC,7,7,86,8
    LTEXT           "X:",IDC_STATIC,13,19,8,8
    LTEXT           "Y:",IDC_STATIC,43,19,8,8
    LTEXT           "FPS:",IDC_STATIC,13,33,16,8
    LTEXT           "StartTime:",IDC_STATIC,39,33,33,8
    LTEXT           "Filename:",IDC_STATIC,102,7,31,8
    EDITTEXT        IDC_BRIEF_M_DESC,14,68,112,14,ES_AUTOHSCROLL
    LTEXT           "Effect Description",IDC_STATIC,14,58,63,8
    PUSHBUTTON      "Mission Flags",IDC_MISSIONFLAGS,161,60,50,14
END

IDD_BRIEF_ADDSOUND DIALOG DISCARDABLE  0, 0, 186, 90
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "Sound Effect"
FONT 8, "MS Sans Serif"
BEGIN
    EDITTEXT        IDC_BRIEF_S_STARTTIME,7,15,28,14,ES_AUTOHSCROLL
    EDITTEXT        IDC_BRIEF_S_FILENAME,59,16,110,14,ES_AUTOHSCROLL
    PUSHBUTTON      "Choose",IDC_BRIEF_S_CHOOSE,130,31,39,14
    CONTROL         "Play Once",IDC_BRIEF_S_PLAYONCE,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,59,32,49,10
    DEFPUSHBUTTON   "OK",IDOK,7,69,50,14
    PUSHBUTTON      "Cancel",IDCANCEL,58,69,50,14
    LTEXT           "StartTime:",IDC_STATIC,7,7,33,8
    LTEXT           "Filename:",IDC_STATIC,59,7,31,8
    EDITTEXT        IDC_BRIEF_S_DESC,7,53,112,14,ES_AUTOHSCROLL
    LTEXT           "Effect Description",IDC_STATIC,7,43,63,8
    PUSHBUTTON      "Mission Flags",IDC_MISSIONFLAGS,129,69,50,14
END

IDD_BRIEF_ADDBUTTON DIALOG DISCARDABLE  0, 0, 214, 279
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "OnScreen Button"
FONT 8, "MS Sans Serif"
BEGIN
    CONTROL         "Flasher",IDC_BRIEF_BT_FLASHER,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,13,16,39,10
    EDITTEXT        IDC_BRIEF_BT_FLASHSTART,65,26,29,14,ES_AUTOHSCROLL
    EDITTEXT        IDC_BRIEF_BT_FLASH_FILENAME,13,51,110,14,ES_AUTOHSCROLL
    PUSHBUTTON      "Choose",IDC_BRIEF_BT_FLASH_CHOOSE,84,66,39,14
    EDITTEXT        IDC_BRIEF_BT_ID,161,24,29,14,ES_AUTOHSCROLL | ES_NUMBER
    EDITTEXT        IDC_BRIEF_BT_PID,161,50,29,14,ES_AUTOHSCROLL | ES_NUMBER
    EDITTEXT        IDC_BRIEF_BT_SID,161,75,29,14,ES_AUTOHSCROLL | ES_NUMBER
    CONTROL         "Text Scroll Up",IDC_BRIEF_BT_SCROLLUP,"Button",
                    BS_AUTORADIOBUTTON | WS_GROUP,14,137,61,10
    CONTROL         "Text Scroll Down",IDC_BRIEF_BT_SCROLLDOWN,"Button",
                    BS_AUTORADIOBUTTON,14,147,70,10
    CONTROL         "Next Screen",IDC_BRIEF_BT_NEXTSCREEN,"Button",
                    BS_AUTORADIOBUTTON,14,157,55,10
    CONTROL         "Prev Screen",IDC_BRIEF_BT_PREVSCREEN,"Button",
                    BS_AUTORADIOBUTTON,14,167,55,10
    CONTROL         "Quit To MainMenu",IDC_BRIEF_BT_QUIT,"Button",
                    BS_AUTORADIOBUTTON,14,177,73,10
    CONTROL         "Jump to Screen",IDC_BRIEF_BT_JUMP,"Button",
                    BS_AUTORADIOBUTTON,14,187,65,10
    EDITTEXT        IDC_BRIEF_BT_JUMP_SCREEN,27,196,19,12,ES_AUTOHSCROLL
    CONTROL         "On Mouse Up",IDC_BRIEF_BT_MOUSEUP,"Button",
                    BS_AUTORADIOBUTTON | WS_GROUP,123,140,60,10
    CONTROL         "On Mouse Down",IDC_BRIEF_BT_MOUSEDOWN,"Button",
                    BS_AUTORADIOBUTTON,123,151,69,10
    CONTROL         "While Mouse Down",IDC_BRIEF_BT_MOUSEHOLD,"Button",
                    BS_AUTORADIOBUTTON,123,161,78,10
    EDITTEXT        IDC_BRIEF_BT_FILENAME,97,185,110,14,ES_AUTOHSCROLL
    PUSHBUTTON      "Choose",IDC_BRIEF_BT_CHOOSE,168,199,39,14
    DEFPUSHBUTTON   "OK",IDOK,53,258,50,14
    PUSHBUTTON      "Cancel",IDCANCEL,109,258,50,14
    LTEXT           "Flash StartTime:",IDC_STATIC,13,27,52,8
    LTEXT           "Flash Filename:",IDC_STATIC,13,42,50,8
    GROUPBOX        "Flash Information",IDC_STATIC,7,7,122,122
    LTEXT           "ID:",IDC_STATIC,161,15,10,8
    LTEXT           "Parent ID:",IDC_STATIC,161,41,33,8
    LTEXT           "Sibling ID:",IDC_STATIC,161,66,33,8
    GROUPBOX        "ID Information",IDC_STATIC,149,7,58,89
    CONTROL         "Glow When Mouse Over",IDC_BRIEF_BT_GLOW,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,13,116,94,10
    GROUPBOX        "Click Types",IDC_STATIC,119,132,88,40,WS_GROUP
    GROUPBOX        "Button Types",IDC_STATIC,7,129,85,81
    LTEXT           "Filename:",IDC_STATIC,97,175,31,8
    EDITTEXT        IDC_BRIEF_BT_DESC,7,242,112,14,ES_AUTOHSCROLL
    LTEXT           "Effect Description",IDC_STATIC,7,233,63,8
    EDITTEXT        IDC_BRIEF_BT_FLASH_FILENAME_FOCUS,13,86,110,14,
                    ES_AUTOHSCROLL
    PUSHBUTTON      "Choose",IDC_BRIEF_BT_FLASH_CHOOSE_FOCUS,84,100,39,14
    LTEXT           "Flash-Focus Filename:",IDC_STATIC,13,76,71,8
    EDITTEXT        IDC_BRIEF_BT_FILENAME_FOCUS,97,216,110,14,ES_AUTOHSCROLL
    PUSHBUTTON      "Choose",IDC_BRIEF_BT_CHOOSE_FOCUS,168,230,39,14
    LTEXT           "Focus Filename:",IDC_STATIC,97,206,52,8
    EDITTEXT        IDC_BRIEF_BT_X,15,220,19,14,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_TABSTOP
    EDITTEXT        IDC_BRIEF_BT_Y,46,220,19,14,ES_AUTOHSCROLL | ES_NUMBER
    LTEXT           "Upper Left Corner Position:",IDC_STATIC,7,210,86,8
    LTEXT           "X:",IDC_STATIC,7,222,8,8
    LTEXT           "Y:",IDC_STATIC,37,222,8,8
    PUSHBUTTON      "Mission Flags",IDC_MISSIONFLAGS,157,109,50,14
END

IDD_LOCALIZESCRIPT DIALOG DISCARDABLE  0, 0, 302, 250
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "Script String Editor"
FONT 8, "MS Sans Serif"
BEGIN
    DEFPUSHBUTTON   "Quit",IDOK,245,25,50,14
    PUSHBUTTON      "Save",IDC_SAVE,245,5,50,14,WS_DISABLED
    PUSHBUTTON      "Load script",IDC_LOADSCRIPT,5,5,60,14
    LISTBOX         IDC_STRINGLIST,10,60,280,115,LBS_NOINTEGRALHEIGHT | 
                    WS_DISABLED | WS_VSCROLL | WS_TABSTOP
    EDITTEXT        IDC_STRINGEDIT,10,185,280,45,ES_MULTILINE | 
                    ES_AUTOHSCROLL | ES_WANTRETURN | WS_DISABLED | 
                    WS_VSCROLL
END

IDD_TABLE_FILE_FILTER DIALOGEX 0, 0, 293, 279
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU
EXSTYLE WS_EX_CLIENTEDGE
CAPTION "Table File Filter v1.0"
MENU IDR_TABLE_FILE_FILTER_MENU
FONT 8, "MS Sans Serif", 0, 0, 0x1
BEGIN
    LISTBOX         IDC_TABLEFILTER_LEVEL_LIST,7,49,184,39,LBS_SORT | 
                    LBS_NOINTEGRALHEIGHT | WS_VSCROLL | WS_TABSTOP
    LTEXT           "Level filters:",IDC_STATIC,7,40,96,8
    PUSHBUTTON      "Add level(s)...",IDC_TABLEFILTER_BTN_ADDLEVEL,202,49,84,
                    16
    PUSHBUTTON      "Remove level(s)",IDC_TABLEFILTER_BTN_REMOVELEVEL,202,73,
                    84,16
    CONTROL         "List1",IDC_TABLEFILTER_DATA_LIST,"SysListView32",
                    LVS_REPORT | LVS_SORTASCENDING | WS_BORDER | WS_TABSTOP,
                    7,107,279,129
    LTEXT           "Pages to include:",IDC_STATIC,7,95,56,8
    PUSHBUTTON      "Add Page...",IDC_TABLEFILTER_BTN_ADDPAGE,7,243,64,14
    PUSHBUTTON      "Remove Page",IDC_TABLEFILTER_BTN_REMOVEPAGE,76,243,64,
                    14
    PUSHBUTTON      "Create Filtered Table File...",
                    IDC_TABLEFILTER_BTN_CREATENEWTABLEFILE,178,243,108,14
    LTEXT           "Status:",IDC_STATIC,7,264,24,8
    LTEXT           "Waiting for user...",IDC_TABLEFILTER_STATUSTEXT,35,264,
                    251,8
    LTEXT           "0",IDC_TABLEFILTER_NUMPAGESTEXT,67,95,37,8
    LTEXT           "Location of TableParse Output File:",IDC_STATIC,7,7,138,
                    8
    EDITTEXT        IDC_OUTPUTFILE_EDIT,7,16,184,14,ES_AUTOHSCROLL
END

IDD_TABLEFILEFILTER_ADDPAGE_DLG DIALOG DISCARDABLE  0, 0, 186, 125
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "Add Page"
FONT 8, "MS Sans Serif"
BEGIN
    EDITTEXT        IDC_TABLEFILEFILTER_ADDNAME_EDIT,7,18,172,14,
                    ES_AUTOHSCROLL
    COMBOBOX        IDC_TABLEFILEFILTER_ADDTYPE_COMBO,7,52,172,66,
                    CBS_DROPDOWNLIST | CBS_SORT | WS_VSCROLL | WS_TABSTOP
    DEFPUSHBUTTON   "Add",IDOK,38,104,50,14
    PUSHBUTTON      "Cancel",IDCANCEL,96,104,50,14
    LTEXT           "Page Name:",IDC_STATIC,7,7,41,8
    LTEXT           "Page Type:",IDC_STATIC,7,41,48,8
END

IDD_MULTICHECK DIALOG DISCARDABLE  0, 0, 275, 182
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "Manage dialog"
FONT 8, "MS Sans Serif"
BEGIN
    DEFPUSHBUTTON   "OK",IDOK,218,7,50,14
    PUSHBUTTON      "Cancel",IDCANCEL,218,24,50,14
    LISTBOX         IDC_MULTICHECKIN_LISTBOX,10,25,195,140,LBS_SORT | 
                    LBS_NOINTEGRALHEIGHT | LBS_EXTENDEDSEL | WS_VSCROLL | 
                    WS_TABSTOP
    LTEXT           "Choose the files you wish to check in",
                    IDC_DIRECTIONS_TEXT,5,5,190,15
END

IDD_SCRIPTINTERFACE DIALOGEX 0, 0, 495, 327
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "D3 Level And Script Interface"
FONT 8, "MS Sans Serif", 0, 0, 0x1
BEGIN
    LISTBOX         IDC_LIST_AVAILABLE,14,39,147,89,LBS_SORT | 
                    LBS_MULTIPLESEL | LBS_OWNERDRAWFIXED | LBS_HASSTRINGS | 
                    LBS_NOINTEGRALHEIGHT | WS_VSCROLL | WS_TABSTOP
    LISTBOX         IDC_LIST_CHECKEDOUT,180,39,149,89,LBS_SORT | 
                    LBS_MULTIPLESEL | LBS_OWNERDRAWFIXED | LBS_HASSTRINGS | 
                    LBS_NOINTEGRALHEIGHT | WS_VSCROLL | WS_TABSTOP
    CONTROL         "Levels",IDC_RADIO_LEVELS,"Button",BS_AUTORADIOBUTTON | 
                    WS_GROUP,358,76,37,10
    CONTROL         "Scripts",IDC_RADIO_SCRIPTS,"Button",BS_AUTORADIOBUTTON,
                    358,89,37,10
    PUSHBUTTON      "Add New",IDC_ADDNEW,29,132,50,14,WS_GROUP
    PUSHBUTTON      "Check Out",IDC_CHECKOUT,99,132,50,14
    PUSHBUTTON      "Check In",IDC_CHECKIN,181,132,50,14
    PUSHBUTTON      "Undo Check Out",IDC_UNDOCHECKOUT,234,132,64,14,
                    WS_DISABLED
    PUSHBUTTON      "Delete",IDC_DELETE,300,132,29,14
    PUSHBUTTON      "Create New Script File",IDC_CREATESCRIPT,346,138,82,14
    DEFPUSHBUTTON   "Exit",IDOK,438,138,50,14
    COMBOBOX        IDC_SELECTED_SCRIPT,13,184,104,72,CBS_DROPDOWN | 
                    CBS_SORT | WS_VSCROLL | WS_TABSTOP
    PUSHBUTTON      "Edit Script",IDC_EDITSCRIPT,122,166,50,14
    PUSHBUTTON      "Compile",IDC_COMPILE,121,183,50,14
    CONTROL         "Game",IDC_SCRIPTGAME,"Button",BS_AUTORADIOBUTTON | 
                    WS_GROUP,195,173,35,10
    CONTROL         "Level",IDC_SCRIPTLEVEL,"Button",BS_AUTORADIOBUTTON,195,
                    184,33,10
    PUSHBUTTON      "Configure Compiler",IDC_CONFIGCOMPILER,383,163,98,14,
                    WS_GROUP
    EDITTEXT        IDC_COMPILER_OUTPUT,11,199,473,118,ES_MULTILINE | 
                    ES_AUTOVSCROLL | ES_AUTOHSCROLL | ES_READONLY | 
                    WS_VSCROLL | WS_HSCROLL,WS_EX_DLGMODALFRAME
    GROUPBOX        "Available",IDC_AVAILABLE_GRP,7,29,161,122
    GROUPBOX        "Checked Out",IDC_CHECKEDOUT_GRP,174,28,161,123
    GROUPBOX        "View",IDC_STATIC,345,65,65,39
    GROUPBOX        "Script Compiler",IDC_STATIC,7,155,481,165
    LTEXT           "Script To Compile",IDC_STATIC,15,173,56,8
    GROUPBOX        "Current Status:",IDC_STATIC,7,7,327,20
    LTEXT           "",IDC_STATUS,9,15,321,8
    GROUPBOX        "Script Type",IDC_STATIC,191,162,47,36
    PUSHBUTTON      "Mass Compile",IDC_MASSCOMPILE,327,163,50,14
    CONTROL         "Display non-checked out scripts",
                    IDC_LIST_NONCHECKED_OUT,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,327,178,123,10
    CONTROL         "Auto Check In/Out Scripts w/ Levels",
                    IDC_AUTOCHECKSCRIPTS,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,327,188,133,10
END

IDD_DALLAS_MAIN_DIALOG DIALOGEX 0, 0, 449, 434
STYLE DS_MODALFRAME | DS_CENTER | WS_MINIMIZEBOX | WS_POPUP | WS_CAPTION | 
    WS_SYSMENU
EXSTYLE WS_EX_CLIENTEDGE
CAPTION "Dallas Graphical Script Editor v1.0"
FONT 8, "MS Sans Serif", 0, 0, 0x1
BEGIN
    CONTROL         "Tree1",IDC_EVENT_TREE,"SysTreeView32",TVS_HASBUTTONS | 
                    TVS_HASLINES | TVS_LINESATROOT | TVS_SHOWSELALWAYS | 
                    TVS_NOTOOLTIPS | WS_TABSTOP,7,15,269,346,
                    WS_EX_CLIENTEDGE
    PUSHBUTTON      "New Script",IDC_NEW_EVENT_BUTTON,285,18,50,12
    PUSHBUTTON      "Insert Script",IDC_INSERT_EVENT_BUTTON,285,33,50,12
    PUSHBUTTON      "Delete Script",IDC_DELETE_EVENT_BUTTON,285,48,50,12
    PUSHBUTTON      "Expand All",IDC_EXPAND_ALL_BUTTON,285,84,50,12
    PUSHBUTTON      "Collapse All",IDC_COLLAPSE_ALL_BUTTON,285,99,50,12
    PUSHBUTTON      "Copy ",IDC_COPY_BUTTON,285,131,50,12
    PUSHBUTTON      "Paste",IDC_PASTE_BUTTON,285,146,50,12
    PUSHBUTTON      "Import",IDC_IMPORT_BUTTON,285,182,50,12
    PUSHBUTTON      "Export",IDC_EXPORT_BUTTON,285,198,50,12
    LISTBOX         IDC_MESSAGE_LIST,349,25,85,86,LBS_SORT | 
                    LBS_NOINTEGRALHEIGHT | WS_VSCROLL | WS_TABSTOP
    PUSHBUTTON      "Change Name ID",IDC_CHANGE_MESSAGE_NAME_BUTTON,349,115,
                    85,12
    EDITTEXT        IDC_MESSAGE_EDIT,349,131,85,54,ES_MULTILINE | 
                    ES_AUTOVSCROLL
    PUSHBUTTON      "New",IDC_NEW_MESSAGE_BUTTON,349,191,41,14
    PUSHBUTTON      "Delete",IDC_DELETE_MESSAGE_BUTTON,393,191,41,14
    PUSHBUTTON      "Open the User Types Workshop",IDC_UTYPES_BUTTON,286,221,
                    156,14
    CONTROL         "None",IDC_NONE_RADIO,"Button",BS_AUTORADIOBUTTON | 
                    WS_GROUP | WS_TABSTOP,295,257,33,10
    CONTROL         "Specific",IDC_SPECIFIC_RADIO,"Button",
                    BS_AUTORADIOBUTTON | WS_TABSTOP,295,270,41,10
    CONTROL         "By Event",IDC_EVENT_RADIO,"Button",BS_AUTORADIOBUTTON | 
                    WS_TABSTOP,295,283,45,10
    COMBOBOX        IDC_EVENT_COMBO,293,295,141,132,CBS_DROPDOWNLIST | 
                    WS_VSCROLL | WS_TABSTOP
    CONTROL         "All Level Scripts",IDC_LEVEL_RADIO,"Button",
                    BS_AUTORADIOBUTTON,355,257,66,10
    CONTROL         "All Object Scripts",IDC_ALLOBJ_RADIO,"Button",
                    BS_AUTORADIOBUTTON | WS_TABSTOP,355,270,69,10
    CONTROL         "All Trigger Scripts",IDC_ALLTRIG_RADIO,"Button",
                    BS_AUTORADIOBUTTON | WS_TABSTOP,355,283,71,10
    PUSHBUTTON      "Highlight Selected Owner Scripts",IDC_HIGHLIGHT_BUTTON,
                    293,314,141,13
    PUSHBUTTON      "Save",IDC_SAVE_BUTTON,286,347,75,14
    PUSHBUTTON      "Exit",IDCANCEL,367,347,75,14
    EDITTEXT        IDC_HELP_EDIT,7,367,435,60,ES_MULTILINE | ES_READONLY | 
                    ES_WANTRETURN | WS_VSCROLL
    LTEXT           "Scripts:",IDC_STATIC,8,4,66,8
    GROUPBOX        "Message List",IDC_STATIC,342,14,100,196
    GROUPBOX        "Script Highlighting",IDC_STATIC,286,244,156,90
END

IDD_CREATESCRIPT DIALOG DISCARDABLE  0, 0, 161, 85
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "Create New Script File"
FONT 8, "MS Sans Serif"
BEGIN
    EDITTEXT        IDC_FILENAME,7,23,76,14,ES_AUTOHSCROLL
    CONTROL         "Level",IDC_LEVEL,"Button",BS_AUTORADIOBUTTON | WS_GROUP,
                    100,19,33,10
    CONTROL         "Game",IDC_GAME,"Button",BS_AUTORADIOBUTTON,100,32,35,10
    DEFPUSHBUTTON   "OK",IDOK,28,64,50,14,WS_GROUP
    PUSHBUTTON      "Cancel",IDCANCEL,83,64,50,14
    LTEXT           "Filename:",IDC_STATIC,7,13,42,8
    GROUPBOX        "Script File Type",IDC_STATIC,92,7,62,45
END

IDD_CONFIGCOMPILER DIALOG DISCARDABLE  0, 0, 242, 135
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "Configure Compiler"
FONT 8, "MS Sans Serif"
BEGIN
    CONTROL         "Optimize",IDC_OPTIMIZE,"Button",BS_AUTORADIOBUTTON | 
                    WS_GROUP,13,18,49,10
    CONTROL         "COFF Info",IDC_COFF,"Button",BS_AUTORADIOBUTTON,13,30,
                    48,10
    CONTROL         "C7 Info",IDC_C7,"Button",BS_AUTORADIOBUTTON,13,42,39,10
    CONTROL         "None",IDC_WARN_NONE,"Button",BS_AUTORADIOBUTTON | 
                    WS_GROUP,15,72,33,10
    CONTROL         "Low",IDC_WARN_LOW,"Button",BS_AUTORADIOBUTTON,15,82,29,
                    10
    CONTROL         "Medium",IDC_WARN_MED,"Button",BS_AUTORADIOBUTTON,15,92,
                    41,10
    CONTROL         "High",IDC_WARN_HIGH,"Button",BS_AUTORADIOBUTTON,15,102,
                    31,10
    CONTROL         "Highest",IDC_RADIO8,"Button",BS_AUTORADIOBUTTON,15,112,
                    40,10
    EDITTEXT        IDC_COMPILER,83,17,146,12,ES_AUTOHSCROLL | WS_GROUP
    PUSHBUTTON      "Browse",IDC_BROWSE,131,39,50,14
    DEFPUSHBUTTON   "Done",IDOK,95,113,50,14
    GROUPBOX        "Debugging Info",IDC_STATIC,7,7,66,49
    GROUPBOX        "Virtual Compiler",IDC_STATIC,79,7,156,49
    GROUPBOX        "Warning Level",IDC_STATIC,7,61,66,66
END

IDD_DALLAS_GENERIC_PROMPT_DIALOG DIALOG DISCARDABLE  0, 0, 186, 63
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "Generic Prompt Dialog"
FONT 8, "MS Sans Serif"
BEGIN
    EDITTEXT        IDC_GENERIC_DATA_EDIT,7,17,172,14,ES_AUTOHSCROLL
    DEFPUSHBUTTON   "OK",IDOK,40,42,50,14
    PUSHBUTTON      "Cancel",IDCANCEL,95,42,50,14
    LTEXT           "Enter the generic data:",IDC_GENERIC_PROMPT_STATIC,7,7,
                    172,8
END

IDD_DALLAS_VECTOR_DIALOG DIALOG DISCARDABLE  0, 0, 166, 87
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "Vector Value Prompt"
FONT 8, "MS Sans Serif"
BEGIN
    EDITTEXT        IDC_DATA1_EDIT,79,7,80,14,ES_AUTOHSCROLL
    EDITTEXT        IDC_DATA2_EDIT,79,24,80,14,ES_AUTOHSCROLL
    EDITTEXT        IDC_DATA3_EDIT,79,41,80,14,ES_AUTOHSCROLL
    DEFPUSHBUTTON   "OK",IDOK,32,66,50,14
    PUSHBUTTON      "Cancel",IDCANCEL,87,66,50,14
    LTEXT           "X Vector Coordinate:",IDC_STATIC,7,11,69,8
    LTEXT           "Z Vector Coordinate:",IDC_STATIC,7,46,69,8
    LTEXT           "Y Vector Coordinate:",IDC_STATIC,7,28,69,8
END

IDD_COMPILEALL DIALOG DISCARDABLE  0, 0, 442, 375
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "Mass Compile"
FONT 8, "MS Sans Serif"
BEGIN
    DEFPUSHBUTTON   "OK",IDOK,385,23,50,14
    PUSHBUTTON      "Compile All",IDC_BUILD,385,46,50,14
    LISTBOX         IDC_LIST,15,26,119,183,LBS_SORT | LBS_MULTIPLESEL | 
                    LBS_OWNERDRAWFIXED | LBS_HASSTRINGS | 
                    LBS_NOINTEGRALHEIGHT | WS_VSCROLL | WS_TABSTOP
    GROUPBOX        "Status",IDC_STATIC,145,20,187,189
    LTEXT           "Current File:",IDC_STATIC,153,38,38,8
    LTEXT           "",IDC_FILENAME,153,48,176,9
    LTEXT           "Step 1: Check out file",IDC_STATIC,153,65,72,8
    LTEXT           "",IDC_STEP1,153,74,173,11
    LTEXT           "Step 2: Determine Script Type",IDC_STATIC,153,87,99,8
    LTEXT           "",IDC_STEP2,153,98,176,12
    LTEXT           "Step 3: Compile file",IDC_STATIC,153,114,64,8
    LTEXT           "",IDC_STEP3,153,123,176,11
    LTEXT           "Step 4: Check In",IDC_STATIC,153,139,99,8
    LTEXT           "",IDC_STEP4,153,148,176,11
    GROUPBOX        "Display",IDC_STATIC,340,80,95,48
    CONTROL         "Checked Out Only",IDC_LOCKED,"Button",
                    BS_AUTORADIOBUTTON | WS_GROUP,348,91,74,10
    CONTROL         "Non-Checked Out Only",IDC_NOTLOCKED,"Button",
                    BS_AUTORADIOBUTTON,348,102,89,10
    CONTROL         "Both",IDC_BOTH,"Button",BS_AUTORADIOBUTTON,348,112,31,
                    10
    CONTROL         "Auto Check-In",IDC_AUTOCHECKIN,"Button",BS_AUTOCHECKBOX | 
                    WS_GROUP | WS_TABSTOP,356,196,61,10
    EDITTEXT        IDC_OUTPUT,7,226,428,142,ES_MULTILINE | WS_VSCROLL | 
                    WS_HSCROLL
    PUSHBUTTON      "Select All",IDC_SELECTALL,334,133,101,14
    PUSHBUTTON      "Select only out of date scripts",IDC_OODSCRIPTS,334,149,
                    101,14
    PUSHBUTTON      "Select only out of sync scripts",IDC_OOSSCRIPTS,334,165,
                    101,14
    CONTROL         "Don't Checkout files",IDC_NOCHECKOUT,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,356,185,79,10
END

IDD_QUICKCOMPILE DIALOG DISCARDABLE  0, 0, 326, 167
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "Quick Compile"
FONT 8, "MS Sans Serif"
BEGIN
    DEFPUSHBUTTON   "OK",IDOK,138,146,50,14
    EDITTEXT        IDC_TEXT,7,7,312,137,ES_MULTILINE | ES_AUTOVSCROLL | 
                    ES_AUTOHSCROLL | ES_READONLY | WS_VSCROLL | WS_HSCROLL
END

IDD_LIGHTINGSTATUS DIALOG DISCARDABLE  0, 0, 186, 76
STYLE DS_MODALFRAME | DS_SETFOREGROUND | DS_CENTER | WS_POPUP | WS_VISIBLE | 
    WS_CAPTION
CAPTION "Lighting status"
FONT 8, "MS Sans Serif"
BEGIN
    LTEXT           "Percentage complete:",IDC_LIGHTING_COMPLETE,35,15,125,
                    15
    PUSHBUTTON      "Stop Lighting",IDC_STOP_LIGHTING,45,45,90,20
END

IDD_BRIEFMISSIONFLAGS DIALOG DISCARDABLE  0, 0, 218, 234
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "Mission Flag Settings"
FONT 8, "MS Sans Serif"
BEGIN
    CONTROL         "Mission Flag 01",IDC_D_FLAG1,"Button",
                    BS_AUTORADIOBUTTON | BS_LEFTTEXT | WS_GROUP,7,73,65,10
    CONTROL         "",IDC_S_FLAG1,"Button",BS_AUTORADIOBUTTON,74,74,12,8
    CONTROL         "",IDC_N_FLAG1,"Button",BS_AUTORADIOBUTTON,84,74,12,8
    CONTROL         "Mission Flag 02",IDC_D_FLAG2,"Button",
                    BS_AUTORADIOBUTTON | BS_LEFTTEXT | WS_GROUP,7,81,65,10
    CONTROL         "",IDC_S_FLAG2,"Button",BS_AUTORADIOBUTTON,74,82,12,8
    CONTROL         "",IDC_N_FLAG2,"Button",BS_AUTORADIOBUTTON,84,82,12,8
    CONTROL         "Mission Flag 03",IDC_D_FLAG3,"Button",
                    BS_AUTORADIOBUTTON | BS_LEFTTEXT | WS_GROUP,7,89,65,10
    CONTROL         "",IDC_S_FLAG3,"Button",BS_AUTORADIOBUTTON,74,90,12,8
    CONTROL         "",IDC_N_FLAG3,"Button",BS_AUTORADIOBUTTON,84,90,12,8
    CONTROL         "Mission Flag 04",IDC_D_FLAG4,"Button",
                    BS_AUTORADIOBUTTON | BS_LEFTTEXT | WS_GROUP,7,97,65,10
    CONTROL         "",IDC_S_FLAG4,"Button",BS_AUTORADIOBUTTON,74,98,12,8
    CONTROL         "",IDC_N_FLAG4,"Button",BS_AUTORADIOBUTTON,84,98,12,8
    CONTROL         "Mission Flag 05",IDC_D_FLAG5,"Button",
                    BS_AUTORADIOBUTTON | BS_LEFTTEXT | WS_GROUP,7,105,65,10
    CONTROL         "",IDC_S_FLAG5,"Button",BS_AUTORADIOBUTTON,74,106,12,8
    CONTROL         "",IDC_N_FLAG5,"Button",BS_AUTORADIOBUTTON,84,106,12,8
    CONTROL         "Mission Flag 06",IDC_D_FLAG6,"Button",
                    BS_AUTORADIOBUTTON | BS_LEFTTEXT | WS_GROUP,7,113,65,10
    CONTROL         "",IDC_S_FLAG6,"Button",BS_AUTORADIOBUTTON,74,114,12,8
    CONTROL         "",IDC_N_FLAG6,"Button",BS_AUTORADIOBUTTON,84,114,12,8
    CONTROL         "Mission Flag 07",IDC_D_FLAG7,"Button",
                    BS_AUTORADIOBUTTON | BS_LEFTTEXT | WS_GROUP,7,121,65,10
    CONTROL         "",IDC_S_FLAG7,"Button",BS_AUTORADIOBUTTON,74,122,12,8
    CONTROL         "",IDC_N_FLAG7,"Button",BS_AUTORADIOBUTTON,84,122,12,8
    CONTROL         "Mission Flag 08",IDC_D_FLAG8,"Button",
                    BS_AUTORADIOBUTTON | BS_LEFTTEXT | WS_GROUP,7,129,65,10
    CONTROL         "",IDC_S_FLAG8,"Button",BS_AUTORADIOBUTTON,74,130,12,8
    CONTROL         "",IDC_N_FLAG8,"Button",BS_AUTORADIOBUTTON,84,130,12,8
    CONTROL         "Mission Flag 09",IDC_D_FLAG9,"Button",
                    BS_AUTORADIOBUTTON | BS_LEFTTEXT | WS_GROUP,7,137,65,10
    CONTROL         "",IDC_S_FLAG9,"Button",BS_AUTORADIOBUTTON,74,138,12,8
    CONTROL         "",IDC_N_FLAG9,"Button",BS_AUTORADIOBUTTON,84,138,12,8
    CONTROL         "Mission Flag 10",IDC_D_FLAG10,"Button",
                    BS_AUTORADIOBUTTON | BS_LEFTTEXT | WS_GROUP,7,146,65,10
    CONTROL         "",IDC_S_FLAG10,"Button",BS_AUTORADIOBUTTON,74,146,12,8
    CONTROL         "",IDC_N_FLAG10,"Button",BS_AUTORADIOBUTTON,84,146,12,8
    CONTROL         "Mission Flag 11",IDC_D_FLAG11,"Button",
                    BS_AUTORADIOBUTTON | BS_LEFTTEXT | WS_GROUP,7,155,65,10
    CONTROL         "",IDC_S_FLAG11,"Button",BS_AUTORADIOBUTTON,74,156,12,8
    CONTROL         "",IDC_N_FLAG11,"Button",BS_AUTORADIOBUTTON,84,156,12,8
    CONTROL         "Mission Flag 12",IDC_D_FLAG12,"Button",
                    BS_AUTORADIOBUTTON | BS_LEFTTEXT | WS_GROUP,7,163,65,10
    CONTROL         "",IDC_S_FLAG12,"Button",BS_AUTORADIOBUTTON,74,164,12,8
    CONTROL         "",IDC_N_FLAG12,"Button",BS_AUTORADIOBUTTON,84,164,12,8
    CONTROL         "Mission Flag 13",IDC_D_FLAG13,"Button",
                    BS_AUTORADIOBUTTON | BS_LEFTTEXT | WS_GROUP,7,172,65,10
    CONTROL         "",IDC_S_FLAG13,"Button",BS_AUTORADIOBUTTON,74,173,12,8
    CONTROL         "",IDC_N_FLAG13,"Button",BS_AUTORADIOBUTTON,84,173,12,8
    CONTROL         "Mission Flag 14",IDC_D_FLAG14,"Button",
                    BS_AUTORADIOBUTTON | BS_LEFTTEXT | WS_GROUP,7,181,65,10
    CONTROL         "",IDC_S_FLAG14,"Button",BS_AUTORADIOBUTTON,74,182,12,8
    CONTROL         "",IDC_N_FLAG14,"Button",BS_AUTORADIOBUTTON,84,182,12,8
    CONTROL         "Mission Flag 15",IDC_D_FLAG15,"Button",
                    BS_AUTORADIOBUTTON | BS_LEFTTEXT | WS_GROUP,7,189,65,10
    CONTROL         "",IDC_S_FLAG15,"Button",BS_AUTORADIOBUTTON,74,190,12,8
    CONTROL         "",IDC_N_FLAG15,"Button",BS_AUTORADIOBUTTON,84,190,12,8
    CONTROL         "Mission Flag 16",IDC_D_FLAG16,"Button",
                    BS_AUTORADIOBUTTON | BS_LEFTTEXT | WS_GROUP,7,197,65,10
    CONTROL         "",IDC_S_FLAG16,"Button",BS_AUTORADIOBUTTON,74,198,12,8
    CONTROL         "",IDC_N_FLAG16,"Button",BS_AUTORADIOBUTTON,84,198,12,8
    CONTROL         "Mission Flag 17",IDC_D_FLAG17,"Button",
                    BS_AUTORADIOBUTTON | BS_LEFTTEXT | WS_GROUP,117,73,65,10
    CONTROL         "",IDC_S_FLAG17,"Button",BS_AUTORADIOBUTTON,184,74,12,8
    CONTROL         "",IDC_N_FLAG17,"Button",BS_AUTORADIOBUTTON,194,74,12,8
    CONTROL         "Mission Flag 18",IDC_D_FLAG18,"Button",
                    BS_AUTORADIOBUTTON | BS_LEFTTEXT | WS_GROUP,117,81,65,10
    CONTROL         "",IDC_S_FLAG18,"Button",BS_AUTORADIOBUTTON,184,82,12,8
    CONTROL         "",IDC_N_FLAG18,"Button",BS_AUTORADIOBUTTON,194,82,12,8
    CONTROL         "Mission Flag 19",IDC_D_FLAG19,"Button",
                    BS_AUTORADIOBUTTON | BS_LEFTTEXT | WS_GROUP,117,89,65,10
    CONTROL         "",IDC_S_FLAG19,"Button",BS_AUTORADIOBUTTON,184,90,12,8
    CONTROL         "",IDC_N_FLAG19,"Button",BS_AUTORADIOBUTTON,194,90,12,8
    CONTROL         "Mission Flag 20",IDC_D_FLAG20,"Button",
                    BS_AUTORADIOBUTTON | BS_LEFTTEXT | WS_GROUP,117,97,65,10
    CONTROL         "",IDC_S_FLAG20,"Button",BS_AUTORADIOBUTTON,184,98,12,8
    CONTROL         "",IDC_N_FLAG20,"Button",BS_AUTORADIOBUTTON,194,98,12,8
    CONTROL         "Mission Flag 21",IDC_D_FLAG21,"Button",
                    BS_AUTORADIOBUTTON | BS_LEFTTEXT | WS_GROUP,117,105,65,
                    10
    CONTROL         "",IDC_S_FLAG21,"Button",BS_AUTORADIOBUTTON,184,106,12,8
    CONTROL         "",IDC_N_FLAG21,"Button",BS_AUTORADIOBUTTON,194,106,12,8
    CONTROL         "Mission Flag 22",IDC_D_FLAG22,"Button",
                    BS_AUTORADIOBUTTON | BS_LEFTTEXT | WS_GROUP,117,113,65,
                    10
    CONTROL         "",IDC_S_FLAG22,"Button",BS_AUTORADIOBUTTON,184,114,12,8
    CONTROL         "",IDC_N_FLAG22,"Button",BS_AUTORADIOBUTTON,194,114,12,8
    CONTROL         "Mission Flag 23",IDC_D_FLAG23,"Button",
                    BS_AUTORADIOBUTTON | BS_LEFTTEXT | WS_GROUP,117,121,65,
                    10
    CONTROL         "",IDC_S_FLAG23,"Button",BS_AUTORADIOBUTTON,184,122,12,8
    CONTROL         "",IDC_N_FLAG23,"Button",BS_AUTORADIOBUTTON,194,122,12,8
    CONTROL         "Mission Flag 24",IDC_D_FLAG24,"Button",
                    BS_AUTORADIOBUTTON | BS_LEFTTEXT | WS_GROUP,117,129,65,
                    10
    CONTROL         "",IDC_S_FLAG24,"Button",BS_AUTORADIOBUTTON,184,130,12,8
    CONTROL         "",IDC_N_FLAG24,"Button",BS_AUTORADIOBUTTON,194,130,12,8
    CONTROL         "Mission Flag 25",IDC_D_FLAG25,"Button",
                    BS_AUTORADIOBUTTON | BS_LEFTTEXT | WS_GROUP,117,137,65,
                    10
    CONTROL         "",IDC_S_FLAG25,"Button",BS_AUTORADIOBUTTON,184,138,12,8
    CONTROL         "",IDC_N_FLAG25,"Button",BS_AUTORADIOBUTTON,194,138,12,8
    CONTROL         "Mission Flag 26",IDC_D_FLAG26,"Button",
                    BS_AUTORADIOBUTTON | BS_LEFTTEXT | WS_GROUP,117,146,65,
                    10
    CONTROL         "",IDC_S_FLAG26,"Button",BS_AUTORADIOBUTTON,184,146,12,8
    CONTROL         "",IDC_N_FLAG26,"Button",BS_AUTORADIOBUTTON,194,146,12,8
    CONTROL         "Mission Flag 27",IDC_D_FLAG27,"Button",
                    BS_AUTORADIOBUTTON | BS_LEFTTEXT | WS_GROUP,117,155,65,
                    10
    CONTROL         "",IDC_S_FLAG27,"Button",BS_AUTORADIOBUTTON,184,156,12,8
    CONTROL         "",IDC_N_FLAG27,"Button",BS_AUTORADIOBUTTON,194,156,12,8
    CONTROL         "Mission Flag 28",IDC_D_FLAG28,"Button",
                    BS_AUTORADIOBUTTON | BS_LEFTTEXT | WS_GROUP,117,163,65,
                    10
    CONTROL         "",IDC_S_FLAG28,"Button",BS_AUTORADIOBUTTON,184,164,12,8
    CONTROL         "",IDC_N_FLAG28,"Button",BS_AUTORADIOBUTTON,194,164,12,8
    CONTROL         "Mission Flag 29",IDC_D_FLAG29,"Button",
                    BS_AUTORADIOBUTTON | BS_LEFTTEXT | WS_GROUP,117,172,65,
                    10
    CONTROL         "",IDC_S_FLAG29,"Button",BS_AUTORADIOBUTTON,184,173,12,8
    CONTROL         "",IDC_N_FLAG29,"Button",BS_AUTORADIOBUTTON,194,173,12,8
    CONTROL         "Mission Flag 30",IDC_D_FLAG30,"Button",
                    BS_AUTORADIOBUTTON | BS_LEFTTEXT | WS_GROUP,117,181,65,
                    10
    CONTROL         "",IDC_S_FLAG30,"Button",BS_AUTORADIOBUTTON,184,182,12,8
    CONTROL         "",IDC_N_FLAG30,"Button",BS_AUTORADIOBUTTON,194,182,12,8
    CONTROL         "Mission Flag 31",IDC_D_FLAG31,"Button",
                    BS_AUTORADIOBUTTON | BS_LEFTTEXT | WS_GROUP,117,189,65,
                    10
    CONTROL         "",IDC_S_FLAG31,"Button",BS_AUTORADIOBUTTON,184,190,12,8
    CONTROL         "",IDC_N_FLAG31,"Button",BS_AUTORADIOBUTTON,194,190,12,8
    CONTROL         "Mission Flag 32",IDC_D_FLAG32,"Button",
                    BS_AUTORADIOBUTTON | BS_LEFTTEXT | WS_GROUP,117,197,65,
                    10
    CONTROL         "",IDC_S_FLAG32,"Button",BS_AUTORADIOBUTTON,184,198,12,8
    CONTROL         "",IDC_N_FLAG32,"Button",BS_AUTORADIOBUTTON,194,198,12,8
    DEFPUSHBUTTON   "OK",IDOK,54,213,50,14,WS_GROUP
    PUSHBUTTON      "Cancel",IDCANCEL,113,213,50,14
    LTEXT           "N\no\nt\n \nU\ns\ne\nd",IDC_STATIC,66,7,8,65
    LTEXT           "B\nE\n \nS\ne\nt",IDC_STATIC,76,25,8,47
    LTEXT           "N\nO\nT\n \nS\ne\nt\n",IDC_STATIC,85,15,8,57
    LTEXT           "N\no\nt\n \nU\ns\ne\nd",IDC_STATIC,176,7,8,65
    LTEXT           "B\nE\n \nS\ne\nt",IDC_STATIC,186,25,8,47
    LTEXT           "N\nO\nT\n \nS\ne\nt\n",IDC_STATIC,195,15,8,57
END

IDD_DEATH_DIALOG DIALOG DISCARDABLE  0, 0, 269, 267
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "Object Death Settings"
FONT 8, "MS Sans Serif"
BEGIN
    EDITTEXT        IDC_DEATH_DELAY_MIN,51,24,35,12,ES_AUTOHSCROLL | 
                    WS_GROUP
    EDITTEXT        IDC_DEATH_DELAY_MAX,51,40,35,12,ES_AUTOHSCROLL
    CONTROL         "From Death Anim",IDC_DEATH_DELAY_FROM_ANIM,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,12,59,68,10
    CONTROL         "Fireball",IDC_DEATH_FIREBALL,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,12,102,55,8
    CONTROL         "Breaks Apart",IDC_DEATH_BREAKS_APART,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,12,112,55,8
    CONTROL         "Blast Ring",IDC_DEATH_BLAST_RING,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,12,122,55,8
    CONTROL         "Remains",IDC_DEATH_REMAINS,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,12,132,55,8
    CONTROL         "Loses Anti-Gravity",IDC_DEATH_LOSES_ANTIGRAV,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,12,142,70,8
    CONTROL         "Small",IDC_DEATH_EXPLOSION_SMALL,"Button",
                    BS_AUTORADIOBUTTON | WS_GROUP,110,225,40,8
    CONTROL         "Medium",IDC_DEATH_EXPLOSION_MEDIUM,"Button",
                    BS_AUTORADIOBUTTON,110,235,40,8
    CONTROL         "Large",IDC_DEATH_EXPLOSION_LARGE,"Button",
                    BS_AUTORADIOBUTTON,110,245,40,8
    CONTROL         "Sparks",IDC_DEATH_DELAY_SPARKS,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,110,25,75,10
    CONTROL         "Fireballs",IDC_DEATH_DELAY_FIREBALLS,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,110,35,75,10
    CONTROL         "Smoke",IDC_DEATH_DELAY_SMOKE,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,110,45,75,10
    CONTROL         "Flies in Air",IDC_DEATH_DELAY_FLYING,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,110,55,75,10
    CONTROL         "Loses Anti-Gravity",IDC_DEATH_DELAY_LOSES_ANTIGRAV,
                    "Button",BS_AUTOCHECKBOX | WS_TABSTOP,110,75,75,10
    CONTROL         "Fade Away",IDC_DEATH_DELAY_FADE_AWAY,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,110,85,75,10
    CONTROL         "Fireball",IDC_DEATH_CONTACT_FIREBALL,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,110,145,55,8
    CONTROL         "Breaks Apart",IDC_DEATH_CONTACT_BREAKS_APART,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,110,155,55,8
    CONTROL         "Blast Ring",IDC_DEATH_CONTACT_BLAST_RING,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,110,165,55,8
    CONTROL         "Remains",IDC_DEATH_CONTACT_REMAINS,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,110,175,55,8
    CONTROL         "Debris Smokes",IDC_DEATH_DEBRIS_SMOKES,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,12,186,70,10
    CONTROL         "Fireball",IDC_DEATH_DEBRIS_CONTACT_FIREBALL,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,12,225,55,8
    CONTROL         "Blast Ring",IDC_DEATH_DEBRIS_CONTACT_BLAST_RING,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,12,235,55,8
    CONTROL         "Remains",IDC_DEATH_DEBRIS_CONTACT_REMAINS,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,12,245,55,8
    PUSHBUTTON      "Copy",IDC_DEATH_COPY,220,135,35,13
    PUSHBUTTON      "Paste",IDC_DEATH_PASTE,220,150,35,13
    PUSHBUTTON      "Descent",IDC_DEATH_COMMON_DESCENT,215,25,40,13
    PUSHBUTTON      "Quick Expl.",IDC_DEATH_COMMON_QUICK_EXPLOSION,215,40,40,
                    14
    PUSHBUTTON      "Shoot Up",IDC_DEATH_COMMON_SHOOT_UP,215,55,40,14
    PUSHBUTTON      "R2D2",IDC_DEATH_COMMON_R2D2,215,70,40,12
    PUSHBUTTON      "Animal",IDC_DEATH_COMMON_ANIMAL,215,84,40,13
    DEFPUSHBUTTON   "OK",IDOK,215,195,50,14
    PUSHBUTTON      "Cancel",IDCANCEL,215,212,50,14
    GROUPBOX        "Delay Before Death",IDC_STATIC,7,10,85,67
    GROUPBOX        "Delay Options",IDC_DEATH_DELAY_GROUP,105,10,85,102
    LTEXT           "Min Delay",IDC_DEATH_DELAY_MIN_LABEL,12,27,36,8
    LTEXT           "Max Delay",IDC_DEATH_DELAY_MAX_LABEL,12,42,36,8
    GROUPBOX        "Upon Death",IDC_DEATH_STATIC,7,87,85,77
    GROUPBOX        "Common Deaths",IDC_STATIC,205,10,60,95
    GROUPBOX        "Explosion Size",IDC_DEATH_EXPLOSION_OPTIONS_GROUP,105,
                    210,85,50
    GROUPBOX        "Upon Contact",IDC_DEATH_CONTACT_GROUP,105,130,85,60
    GROUPBOX        "Debris Contact",IDC_DEATH_DEBRIS_CONTACT_GROUP,7,210,85,
                    50
    GROUPBOX        "Debris Options",IDC_DEATH_DEBRIS_OPTIONS_GROUP,7,171,85,
                    30
    CONTROL         "Fades Away",IDC_DEATH_FADES_AWAY,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,12,152,75,10
    CONTROL         "Plays Sound",IDC_DEATH_DELAY_SOUND,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,110,95,75,10
    CONTROL         "No Tumble When Fly",IDC_DEATH_DELAY_NO_TUMBLE,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,110,65,75,10
END

IDD_DALLAS_SOUND_DIALOG DIALOG DISCARDABLE  0, 0, 193, 250
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION
CAPTION "Sound Selection Prompt"
FONT 8, "MS Sans Serif"
BEGIN
    DEFPUSHBUTTON   "OK",IDOK,35,229,59,14
    PUSHBUTTON      "Cancel",IDCANCEL,100,229,59,14
    LISTBOX         IDC_SOUND_LIST,18,23,157,164,LBS_SORT | 
                    LBS_NOINTEGRALHEIGHT | WS_VSCROLL | WS_TABSTOP
    GROUPBOX        "Available Sounds:",IDC_STATIC,7,7,179,204
    PUSHBUTTON      "Play Sound",IDC_PLAY_SOUND_BUTTON,18,191,77,12
    PUSHBUTTON      "Stop All Sounds",IDC_STOP_SOUNDS_BUTTON,98,191,77,12
END

IDD_DALLAS_FLAG_DIALOG DIALOG DISCARDABLE  0, 0, 173, 197
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION
CAPTION "Flag Selection Prompt"
FONT 8, "MS Sans Serif"
BEGIN
    DEFPUSHBUTTON   "OK",IDOK,31,176,54,14
    PUSHBUTTON      "Cancel",IDCANCEL,89,176,54,14
    LISTBOX         IDC_FLAG_LIST,7,17,159,151,LBS_SORT | LBS_OWNERDRAWFIXED | 
                    LBS_HASSTRINGS | LBS_NOINTEGRALHEIGHT | WS_VSCROLL | 
                    WS_TABSTOP
    LTEXT           "Select the flags you would like to set:",
                    IDC_FLAG_PROMPT_STATIC,7,7,159,8
END

IDD_CUSTDEFSCRIPT DIALOG DISCARDABLE  0, 0, 186, 87
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "Custom Default Script Setup"
FONT 8, "MS Sans Serif"
BEGIN
    DEFPUSHBUTTON   "OK",IDOK,35,66,50,14
    PUSHBUTTON      "Cancel",IDCANCEL,101,66,50,14
    LTEXT           "Module Name",IDC_STATIC,7,7,48,8
    EDITTEXT        IDC_MODULE,7,15,104,14,ES_AUTOHSCROLL
    LTEXT           "Script Name",IDC_STATIC,7,33,40,8
    EDITTEXT        IDC_NAME,7,42,104,14,ES_AUTOHSCROLL
    PUSHBUTTON      "Browse",IDC_BROWSE,115,14,50,14
END

IDD_DALLAS_TEXTURE_DIALOG DIALOG DISCARDABLE  0, 0, 157, 221
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION
CAPTION "Texture Selection Prompt"
FONT 8, "MS Sans Serif"
BEGIN
    DEFPUSHBUTTON   "OK",IDOK,18,200,58,14
    PUSHBUTTON      "Cancel",IDCANCEL,80,200,58,14
    LISTBOX         IDC_TEXTURE_LIST,6,18,144,168,LBS_SORT | 
                    LBS_NOINTEGRALHEIGHT | WS_VSCROLL | WS_TABSTOP
    LTEXT           "Select a texture from the list below:",IDC_STATIC,6,7,
                    144,8
END

IDD_DALLAS_USERTYPES_DIALOG DIALOG DISCARDABLE  0, 0, 178, 279
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION
CAPTION "Dallas User Types Workshop"
FONT 8, "MS Sans Serif"
BEGIN
    DEFPUSHBUTTON   "Return to Dallas",IDOK,7,254,164,18
    COMBOBOX        IDC_UTYPE_COMBO,7,17,164,114,CBS_DROPDOWNLIST | CBS_SORT | 
                    WS_VSCROLL | WS_TABSTOP
    LISTBOX         IDC_VALUES_LIST,18,54,142,131,LBS_NOINTEGRALHEIGHT | 
                    WS_VSCROLL | WS_TABSTOP
    PUSHBUTTON      "Change Value Name",IDC_CHANGE_BUTTON,18,191,142,13,
                    BS_FLAT
    PUSHBUTTON      "Add New Value",IDC_ADD_BUTTON,18,208,69,19,BS_FLAT
    PUSHBUTTON      "Delete Value",IDC_DELETE_BUTTON,91,208,69,19,BS_FLAT
    LTEXT           "Select the User Type to edit:",IDC_STATIC,7,7,164,8
    GROUPBOX        "User Defined Values",IDC_STATIC,7,40,164,195
END

IDD_DALLAS_IMPORT_DIALOG DIALOG DISCARDABLE  0, 0, 373, 202
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "Import Script Prompt"
FONT 8, "MS Sans Serif"
BEGIN
    DEFPUSHBUTTON   "Import",IDOK,64,183,119,14
    PUSHBUTTON      "Cancel",IDCANCEL,189,183,119,14
    LISTBOX         IDC_SCRIPT_LIST,6,18,360,156,LBS_NOINTEGRALHEIGHT | 
                    WS_VSCROLL | WS_TABSTOP
    LTEXT           "Select the script to import from the list below:",
                    IDC_STATIC,6,7,144,8
END

IDD_BRIEF_MANAGE DIALOG DISCARDABLE  0, 0, 256, 178
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "Manage System Status"
FONT 8, "MS Sans Serif"
BEGIN
    DEFPUSHBUTTON   "OK",IDOK,199,157,50,14
    LISTBOX         IDC_LIST,7,44,117,127,LBS_SORT | LBS_NOINTEGRALHEIGHT | 
                    WS_VSCROLL | WS_TABSTOP
    LTEXT           "The following files are not in the\nmanage system.  You need to add\nthese to the manage system.  Use\nthe button to the right.",
                    IDC_STATIC,7,7,115,35
    PUSHBUTTON      "Add File",IDC_ADD,127,43,50,14
END

IDD_FILEPAGEADD DIALOG DISCARDABLE  0, 0, 274, 186
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "Add File(s)"
FONT 8, "MS Sans Serif"
BEGIN
    DEFPUSHBUTTON   "OK",IDOK,163,165,50,14
    PUSHBUTTON      "Cancel",IDCANCEL,217,165,50,14
    COMBOBOX        IDC_DIRS,7,167,90,83,CBS_DROPDOWN | CBS_SORT | 
                    WS_VSCROLL | WS_TABSTOP
    LTEXT           "Destination Directory:",IDC_STATIC,7,158,68,8
    COMBOBOX        IDC_LIST_DRIVES,7,135,130,71,CBS_DROPDOWN | CBS_SORT | 
                    WS_VSCROLL | WS_TABSTOP
    LISTBOX         IDC_LIST_FOLDERS,7,23,130,104,LBS_SORT | 
                    LBS_NOINTEGRALHEIGHT | WS_VSCROLL | WS_TABSTOP
    LISTBOX         IDC_LIST_FILES,143,23,124,125,LBS_SORT | LBS_MULTIPLESEL | 
                    LBS_NOINTEGRALHEIGHT | WS_VSCROLL | WS_TABSTOP
    LTEXT           "You can select multiple files",IDC_STATIC,143,13,113,8
    LTEXT           "",IDC_CURRPATH,7,13,128,8
END

IDD_DALLAS_STRM_AUDIO_DIALOG DIALOG DISCARDABLE  0, 0, 193, 250
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION
CAPTION "Streaming Audio File Prompt"
FONT 8, "MS Sans Serif"
BEGIN
    DEFPUSHBUTTON   "OK",IDOK,35,229,59,14
    PUSHBUTTON      "Cancel",IDCANCEL,99,229,59,14
    LISTBOX         IDC_STRM_AUDIO_LIST,17,23,157,164,LBS_SORT | 
                    LBS_NOINTEGRALHEIGHT | WS_VSCROLL | WS_TABSTOP
    GROUPBOX        "Available Streaming Audio Files:",IDC_STATIC,7,7,179,
                    204
    PUSHBUTTON      "Play Streaming Audio File",IDC_PLAY_STRM_AUDIO_BUTTON,
                    17,191,157,12
END

IDD_TERRAIN_SOUND_DIALOG DIALOG DISCARDABLE  0, 0, 214, 102
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "Terrain Ambient Sound"
FONT 8, "MS Sans Serif"
BEGIN
    COMBOBOX        IDC_TERRAIN_SOUND_COMBO,10,25,85,70,CBS_DROPDOWN | 
                    CBS_SORT | WS_VSCROLL | WS_TABSTOP
    EDITTEXT        IDC_TERRAIN_SOUND_LOW_ALT,135,15,35,12,ES_AUTOHSCROLL
    EDITTEXT        IDC_TERRAIN_SOUND_HIGH_ALT,175,15,35,12,ES_AUTOHSCROLL
    EDITTEXT        IDC_TERRAIN_SOUND_LOW_VOLUME,135,35,35,12,ES_AUTOHSCROLL
    EDITTEXT        IDC_TERRAIN_SOUND_HIGH_VOLUME,175,35,35,12,
                    ES_AUTOHSCROLL
    DEFPUSHBUTTON   "OK",IDOK,160,65,50,14
    PUSHBUTTON      "&Next",IDC_TERRAIN_SOUND_NEXT,10,80,40,15
    PUSHBUTTON      "&Previous",IDC_TERRAIN_SOUND_PREV,55,80,40,15
    PUSHBUTTON      "Cancel",IDCANCEL,160,82,50,14
    LTEXT           "Band 0:",IDC_TERRAIN_SOUND_BAND_TEXT,10,10,34,9
    LTEXT           "Low",IDC_STATIC,145,5,20,10
    LTEXT           "Volume",IDC_STATIC,110,36,25,10
    LTEXT           "Elevation",IDC_STATIC,104,17,30,10
    LTEXT           "High",IDC_STATIC,185,5,20,10
END

IDD_SOUNDSOURCE_DIALOG DIALOG DISCARDABLE  0, 0, 157, 90
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "SoundSource Properties"
FONT 8, "MS Sans Serif"
BEGIN
    DEFPUSHBUTTON   "OK",IDOK,100,52,50,14
    PUSHBUTTON      "Cancel",IDCANCEL,100,69,50,14
    COMBOBOX        IDC_SOUND_COMBO,15,20,85,75,CBS_DROPDOWN | CBS_SORT | 
                    WS_VSCROLL | WS_TABSTOP
    LTEXT           "Sound",IDC_STATIC,15,10,40,8
    EDITTEXT        IDC_VOLUME,15,50,45,12,ES_AUTOHSCROLL
    LTEXT           "Volume",IDC_STATIC,15,40,60,10
    PUSHBUTTON      "Select",IDC_SELECT,105,20,44,13
END

IDD_SYNCSCRIPTS DIALOG DISCARDABLE  0, 0, 188, 226
STYLE DS_MODALFRAME | DS_3DLOOK | WS_POPUP | WS_CAPTION
CAPTION "Syncing Scripts"
FONT 8, "MS Sans Serif"
BEGIN
    CONTROL         "List1",IDC_LIST,"SysListView32",LVS_REPORT | 
                    LVS_SINGLESEL | LVS_NOSORTHEADER | WS_BORDER | 
                    WS_TABSTOP,7,7,174,212
END

IDD_GENERIC_DEATHS DIALOG DISCARDABLE  0, 0, 254, 107
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "Generic Object Deaths"
FONT 8, "MS Sans Serif"
BEGIN
    DEFPUSHBUTTON   "OK",IDOK,195,65,50,14
    PUSHBUTTON      "Cancel",IDCANCEL,195,82,50,14
    PUSHBUTTON      "Edit Death",IDC_GENERIC_DEATH_EDIT1,131,20,40,15
    PUSHBUTTON      "Edit Death",IDC_GENERIC_DEATH_EDIT2,131,40,40,15
    PUSHBUTTON      "Edit Death",IDC_GENERIC_DEATH_EDIT3,131,60,40,15
    PUSHBUTTON      "Edit Death",IDC_GENERIC_DEATH_EDIT4,131,80,40,15
    EDITTEXT        IDC_GENEREIC_DEATH_PROB1,81,20,45,15,ES_AUTOHSCROLL
    LTEXT           "Death 1 Probability",IDC_STATIC,15,22,65,10
    LTEXT           "Death 3 Probability",IDC_STATIC,15,62,65,10
    LTEXT           "Death 2 Probability",IDC_STATIC,15,42,65,10
    LTEXT           "Death 4 Probability",IDC_STATIC,15,82,65,10
    EDITTEXT        IDC_GENEREIC_DEATH_PROB2,81,40,45,15,ES_AUTOHSCROLL
    EDITTEXT        IDC_GENEREIC_DEATH_PROB3,81,60,45,15,ES_AUTOHSCROLL
    EDITTEXT        IDC_GENEREIC_DEATH_PROB4,81,80,45,15,ES_AUTOHSCROLL
END

IDD_ORPHAN DIALOG DISCARDABLE  0, 0, 246, 207
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "Orphan Hunter"
FONT 8, "MS Sans Serif"
BEGIN
    DEFPUSHBUTTON   "Kill 'Em",IDOK,189,86,50,14
    PUSHBUTTON      "Quit",IDCANCEL,189,186,50,14
    LTEXT           "This dialog will go through the selected directory and remove any orphan files that are in that directory.  An orphaned file is a file that exists physically on your harddrive but is not referenced in the manage system.",
                    IDC_STATIC,7,7,232,29
    COMBOBOX        IDC_DIRECTORY,7,55,93,103,CBS_DROPDOWN | CBS_SORT | 
                    WS_VSCROLL | WS_TABSTOP
    LTEXT           "Selected Directory",IDC_STATIC,7,46,59,8
    LISTBOX         IDC_LIST,7,86,174,114,LBS_SORT | LBS_MULTIPLESEL | 
                    LBS_OWNERDRAWFIXED | LBS_HASSTRINGS | 
                    LBS_NOINTEGRALHEIGHT | WS_VSCROLL | WS_TABSTOP
    LTEXT           "Orphaned Files In Selected Directory",IDC_STATIC,7,77,
                    116,8
    LTEXT           "",IDC_STATS,105,51,100,26
    CONTROL         280,IDC_STATIC,"Static",SS_BITMAP,207,35,32,30
    LTEXT           "Status:",IDC_STATIC,7,35,23,8
    LTEXT           "",IDC_STATUS,31,35,168,8
END

IDD_VIEWER_DIALOG DIALOGEX 0, 0, 145, 153
STYLE DS_MODALFRAME | DS_CENTER | WS_MINIMIZEBOX | WS_POPUP | WS_CAPTION | 
    WS_SYSMENU
EXSTYLE WS_EX_CLIENTEDGE
CAPTION "Viewer Properties"
FONT 7, "Small Fonts", 0, 0, 0x1
BEGIN
    GROUPBOX        "Position",IDC_STATIC,4,4,137,72
    GROUPBOX        "Orientation",IDC_STATIC,4,79,137,71
    EDITTEXT        IDC_XPOS_EDIT,21,14,40,12,ES_AUTOHSCROLL
    LTEXT           "X:",IDC_STATIC,11,18,8,8
    CONTROL         "X Movement On",IDC_XMOVE_CHECK,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,68,15,71,10
    EDITTEXT        IDC_YPOS_EDIT,21,28,40,12,ES_AUTOHSCROLL
    LTEXT           "Y:",IDC_STATIC,11,31,8,8
    CONTROL         "Y Movement On",IDC_YMOVE_CHECK,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,68,31,70,10
    EDITTEXT        IDC_ZPOS_EDIT,21,44,40,12,ES_AUTOHSCROLL
    LTEXT           "Z:",IDC_STATIC,11,47,8,8
    CONTROL         "Z Movement On",IDC_ZMOVE_CHECK,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,68,46,68,10
    PUSHBUTTON      "Commit",IDC_POS_COMMIT_BUTTON,21,60,40,12,BS_FLAT
    EDITTEXT        IDC_HEADING_EDIT,21,89,40,12,ES_AUTOHSCROLL
    LTEXT           "H:",IDC_STATIC,11,92,8,8
    EDITTEXT        IDC_PITCH_EDIT,21,103,40,12,ES_AUTOHSCROLL
    LTEXT           "P:",IDC_STATIC,11,108,8,8
    EDITTEXT        IDC_BANK_EDIT,21,119,40,12,ES_AUTOHSCROLL
    LTEXT           "B:",IDC_STATIC,11,122,8,8
    PUSHBUTTON      "Commit",IDC_ORIENT_COMMIT_BUTTON,21,135,40,12,BS_FLAT
    LTEXT           "Align X:",IDC_STATIC,67,92,25,8
    PUSHBUTTON      "+",IDC_ALIGN_XPOS_BUTTON,94,89,19,12,BS_FLAT
    PUSHBUTTON      "-",IDC_ALIGN_XNEG_BUTTON,115,89,19,12,BS_FLAT
    LTEXT           "Align Y:",IDC_STATIC,67,106,25,8
    PUSHBUTTON      "+",IDC_ALIGN_YPOS_BUTTON,94,103,19,12,BS_FLAT
    PUSHBUTTON      "-",IDC_ALIGN_YNEG_BUTTON,115,103,19,12,BS_FLAT
    LTEXT           "Align Z:",IDC_STATIC,67,122,25,8
    PUSHBUTTON      "+",IDC_ALIGN_ZPOS_BUTTON,94,119,19,12,BS_FLAT
    PUSHBUTTON      "-",IDC_ALIGN_ZNEG_BUTTON,115,119,19,12,BS_FLAT
    PUSHBUTTON      "Align Up with Y+",IDC_ALIGN_UPYPOS_BUTTON,66,135,68,12,
                    BS_FLAT
END


#ifndef _MAC
/////////////////////////////////////////////////////////////////////////////
//
// Version
//

VS_VERSION_INFO VERSIONINFO
 FILEVERSION 1,0,0,1
 PRODUCTVERSION 1,0,0,1
 FILEFLAGSMASK 0x3fL
#ifdef _DEBUG
 FILEFLAGS 0x1L
#else
 FILEFLAGS 0x0L
#endif
 FILEOS 0x4L
 FILETYPE 0x1L
 FILESUBTYPE 0x0L
BEGIN
    BLOCK "StringFileInfo"
    BEGIN
        BLOCK "040904b0"
        BEGIN
            VALUE "CompanyName", "\0"
            VALUE "FileDescription", "EDITOR MFC Application\0"
            VALUE "FileVersion", "1, 0, 0, 1\0"
            VALUE "InternalName", "EDITOR\0"
            VALUE "LegalCopyright", "Copyright © 1996\0"
            VALUE "OriginalFilename", "EDITOR.EXE\0"
            VALUE "ProductName", "EDITOR Application\0"
            VALUE "ProductVersion", "1, 0, 0, 1\0"
        END
    END
    BLOCK "VarFileInfo"
    BEGIN
        VALUE "Translation", 0x409, 1200
    END
END

#endif    // !_MAC


/////////////////////////////////////////////////////////////////////////////
//
// DESIGNINFO
//

#ifdef APSTUDIO_INVOKED
GUIDELINES DESIGNINFO DISCARDABLE 
BEGIN
    IDD_ABOUTBOX, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 191
        TOPMARGIN, 7
        BOTTOMMARGIN, 95
    END

    IDD_KEYPAD_DLGBAR, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 214
        TOPMARGIN, 7
        BOTTOMMARGIN, 511
    END

    IDD_TEXTUREKEYPAD, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 203
        TOPMARGIN, 7
        BOTTOMMARGIN, 525
    END

    IDD_OBJECTKEYPAD, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 203
        TOPMARGIN, 7
        BOTTOMMARGIN, 546
    END

    IDD_LIGHTINGKEYPAD, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 203
        TOPMARGIN, 7
        BOTTOMMARGIN, 474
    END

    IDD_TRIGGERKEYPAD, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 205
        TOPMARGIN, 7
        BOTTOMMARGIN, 454
    END

    IDD_STATUS, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 288
        TOPMARGIN, 7
        BOTTOMMARGIN, 175
    END

    IDD_WORLDOBJECTSDOOR, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 312
        TOPMARGIN, 7
        BOTTOMMARGIN, 255
    END

    IDD_WORLDOBJECTSLIGHT, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 327
        TOPMARGIN, 7
        BOTTOMMARGIN, 290
    END

    IDD_WORLDOBJECTSPLAYER, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 395
        TOPMARGIN, 7
        BOTTOMMARGIN, 224
    END

    IDD_WORLDTEXTURES, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 610
        TOPMARGIN, 7
        BOTTOMMARGIN, 384
    END

    IDD_WORLDWEAPONS, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 540
        TOPMARGIN, 7
    END

    IDD_FONT_DIALOG, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 445
        TOPMARGIN, 7
        BOTTOMMARGIN, 309
    END

    IDD_WORLDSOUNDS, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 420
        TOPMARGIN, 7
        BOTTOMMARGIN, 224
    END

    IDD_FLOATING_KEYPAD, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 251
        TOPMARGIN, 7
        BOTTOMMARGIN, 207
    END

    IDD_PREFERENCES, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 300
        TOPMARGIN, 7
        BOTTOMMARGIN, 178
    END

    IDD_HOGDIALOG, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 450
        TOPMARGIN, 7
        BOTTOMMARGIN, 264
    END

    IDD_DOORWAYKEYPAD, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 198
        TOPMARGIN, 7
        BOTTOMMARGIN, 454
    END

    IDD_TERRAINKEYPAD, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 203
        TOPMARGIN, 7
        BOTTOMMARGIN, 507
    END

    IDD_D3SPLASH, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 208
        TOPMARGIN, 7
        BOTTOMMARGIN, 142
    END

    IDD_PATHKEYPAD, DIALOG
    BEGIN
        RIGHTMARGIN, 200
        TOPMARGIN, 7
        BOTTOMMARGIN, 531
    END

    IDD_EDITLINEDLG, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 187
        TOPMARGIN, 7
        BOTTOMMARGIN, 50
    END

    IDD_CUSTOMIZE_OBJECT, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 240
        TOPMARGIN, 7
        BOTTOMMARGIN, 94
    END

    IDD_ANIMSTATES, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 493
        TOPMARGIN, 7
        BOTTOMMARGIN, 422
    END

    IDD_FILEPAGE, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 356
        TOPMARGIN, 7
        BOTTOMMARGIN, 298
    END

    IDD_MEGACELL, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 530
        TOPMARGIN, 7
        BOTTOMMARGIN, 309
    END

    IDD_MEGAKEYPAD, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 203
        TOPMARGIN, 7
        BOTTOMMARGIN, 454
    END

    IDD_ROOMKEYPAD, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 203
        TOPMARGIN, 7
        BOTTOMMARGIN, 596
    END

    IDD_OBJECTPROPERTIES, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 222
        TOPMARGIN, 7
        BOTTOMMARGIN, 277
    END

    IDD_PROPPHYSICS, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 193
        TOPMARGIN, 7
        BOTTOMMARGIN, 205
    END

    IDD_PROPAI, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 202
        TOPMARGIN, 10
        BOTTOMMARGIN, 198
    END

    IDD_PROPSCRIPT, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 205
        TOPMARGIN, 7
        BOTTOMMARGIN, 220
    END

    IDD_STATUSDLG, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 268
        TOPMARGIN, 7
        BOTTOMMARGIN, 93
    END

    IDD_SCRIPTEDITOR, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 430
        TOPMARGIN, 7
        BOTTOMMARGIN, 290
    END

    IDD_OSIRIS_STATUS, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 205
        TOPMARGIN, 7
        BOTTOMMARGIN, 124
    END

    IDD_SELECTRANGE_DIALOG, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 179
        TOPMARGIN, 7
        BOTTOMMARGIN, 74
    END

    IDD_PHYSICS, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 440
        TOPMARGIN, 7
        BOTTOMMARGIN, 220
    END

    IDD_AI, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 550
        TOPMARGIN, 7
        BOTTOMMARGIN, 332
    END

    IDD_WORLDOBJECTSGENERIC, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 540
        TOPMARGIN, 7
        BOTTOMMARGIN, 274
        HORZGUIDE, 158
    END

    IDD_SCRIPTWIZ, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 365
        TOPMARGIN, 7
        BOTTOMMARGIN, 174
    END

    IDD_SCRIPTSTUDIO, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 455
        TOPMARGIN, 7
        BOTTOMMARGIN, 294
    END

    IDD_ROBOT_WEAPON_DIALOG, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 572
        TOPMARGIN, 7
        BOTTOMMARGIN, 345
    END

    IDD_GENERICLIGHT, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 308
        TOPMARGIN, 7
        BOTTOMMARGIN, 128
    END

    IDD_LEVELPROPERTIES, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 180
        TOPMARGIN, 7
        BOTTOMMARGIN, 219
    END

    IDD_SCRIPTSELECT, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 182
        TOPMARGIN, 7
        BOTTOMMARGIN, 174
    END

    IDD_POWPROP_DIALOG, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 320
        TOPMARGIN, 7
        BOTTOMMARGIN, 199
    END

    IDD_ADDSCRIPT, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 215
        TOPMARGIN, 7
        BOTTOMMARGIN, 54
    END

    IDD_PLAYER_WEAPONS, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 147
        TOPMARGIN, 7
        BOTTOMMARGIN, 265
    END

    IDD_SCRIPTPARMS, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 177
        TOPMARGIN, 7
        BOTTOMMARGIN, 135
    END

    IDD_OBJTREEDLG, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 232
        TOPMARGIN, 7
        BOTTOMMARGIN, 165
    END

    IDD_LEVEL_INFO, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 242
        TOPMARGIN, 7
        BOTTOMMARGIN, 165
    END

    IDD_PROCEDURAL_DIALOG, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 450
        TOPMARGIN, 7
        BOTTOMMARGIN, 294
    END

    IDD_TABLE_FILE_EDITOR, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 308
        TOPMARGIN, 7
        BOTTOMMARGIN, 219
    END

    IDD_WATER_PROCEDURAL, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 465
        TOPMARGIN, 7
        BOTTOMMARGIN, 229
    END

    IDD_AMBIENTSOUNDPATTERNS, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 207
        TOPMARGIN, 7
        BOTTOMMARGIN, 255
    END

    IDD_AMBIENTSOUNDELEMENT, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 157
        TOPMARGIN, 7
        BOTTOMMARGIN, 105
    END

    IDD_MATCENKEYPAD, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 198
        TOPMARGIN, 7
        BOTTOMMARGIN, 604
    END

    IDD_LEVELKEYPAD, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 193
        TOPMARGIN, 7
        BOTTOMMARGIN, 901
    END

    IDD_BRIEF_MAIN, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 565
        TOPMARGIN, 7
        BOTTOMMARGIN, 328
    END

    IDD_BRIEF_ADDSCREEN, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 323
        TOPMARGIN, 7
        BOTTOMMARGIN, 152
    END

    IDD_BRIEF_ADDTEXT, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 295
        TOPMARGIN, 7
        BOTTOMMARGIN, 288
    END

    IDD_BRIEF_ADDBITMAP, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 188
        TOPMARGIN, 7
        BOTTOMMARGIN, 155
    END

    IDD_BRIEF_ADDMOVIE, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 211
        TOPMARGIN, 7
        BOTTOMMARGIN, 98
    END

    IDD_BRIEF_ADDSOUND, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 179
        TOPMARGIN, 7
        BOTTOMMARGIN, 83
    END

    IDD_BRIEF_ADDBUTTON, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 207
        TOPMARGIN, 7
        BOTTOMMARGIN, 272
    END

    IDD_LOCALIZESCRIPT, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 295
        TOPMARGIN, 7
        BOTTOMMARGIN, 243
    END

    IDD_TABLE_FILE_FILTER, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 286
        TOPMARGIN, 7
        BOTTOMMARGIN, 272
    END

    IDD_TABLEFILEFILTER_ADDPAGE_DLG, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 179
        TOPMARGIN, 7
        BOTTOMMARGIN, 118
    END

    IDD_MULTICHECK, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 268
        TOPMARGIN, 7
        BOTTOMMARGIN, 175
    END

    IDD_SCRIPTINTERFACE, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 488
        TOPMARGIN, 7
        BOTTOMMARGIN, 320
    END

    IDD_DALLAS_MAIN_DIALOG, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 442
        TOPMARGIN, 7
        BOTTOMMARGIN, 427
    END

    IDD_CREATESCRIPT, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 154
        TOPMARGIN, 7
        BOTTOMMARGIN, 78
    END

    IDD_CONFIGCOMPILER, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 235
        TOPMARGIN, 7
        BOTTOMMARGIN, 127
    END

    IDD_DALLAS_GENERIC_PROMPT_DIALOG, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 179
        TOPMARGIN, 7
        BOTTOMMARGIN, 56
    END

    IDD_DALLAS_VECTOR_DIALOG, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 159
        TOPMARGIN, 7
        BOTTOMMARGIN, 80
    END

    IDD_COMPILEALL, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 435
        TOPMARGIN, 7
        BOTTOMMARGIN, 368
    END

    IDD_QUICKCOMPILE, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 319
        TOPMARGIN, 7
        BOTTOMMARGIN, 160
    END

    IDD_LIGHTINGSTATUS, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 179
        TOPMARGIN, 7
        BOTTOMMARGIN, 69
    END

    IDD_BRIEFMISSIONFLAGS, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 211
        TOPMARGIN, 7
        BOTTOMMARGIN, 227
    END

    IDD_DEATH_DIALOG, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 262
        TOPMARGIN, 7
        BOTTOMMARGIN, 260
    END

    IDD_DALLAS_SOUND_DIALOG, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 186
        TOPMARGIN, 7
        BOTTOMMARGIN, 243
    END

    IDD_DALLAS_FLAG_DIALOG, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 166
        TOPMARGIN, 7
        BOTTOMMARGIN, 190
    END

    IDD_CUSTDEFSCRIPT, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 179
        TOPMARGIN, 7
        BOTTOMMARGIN, 80
    END

    IDD_DALLAS_TEXTURE_DIALOG, DIALOG
    BEGIN
        LEFTMARGIN, 6
        RIGHTMARGIN, 150
        TOPMARGIN, 7
        BOTTOMMARGIN, 214
    END

    IDD_DALLAS_USERTYPES_DIALOG, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 171
        TOPMARGIN, 7
        BOTTOMMARGIN, 272
    END

    IDD_DALLAS_IMPORT_DIALOG, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 366
        TOPMARGIN, 7
        BOTTOMMARGIN, 195
    END

    IDD_BRIEF_MANAGE, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 249
        TOPMARGIN, 7
        BOTTOMMARGIN, 171
    END

    IDD_FILEPAGEADD, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 267
        TOPMARGIN, 7
        BOTTOMMARGIN, 179
    END

    IDD_DALLAS_STRM_AUDIO_DIALOG, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 186
        TOPMARGIN, 7
        BOTTOMMARGIN, 243
    END

    IDD_TERRAIN_SOUND_DIALOG, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 207
        TOPMARGIN, 7
        BOTTOMMARGIN, 95
    END

    IDD_SOUNDSOURCE_DIALOG, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 150
        TOPMARGIN, 7
        BOTTOMMARGIN, 83
    END

    IDD_SYNCSCRIPTS, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 181
        TOPMARGIN, 7
        BOTTOMMARGIN, 219
    END

    IDD_GENERIC_DEATHS, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 247
        TOPMARGIN, 7
        BOTTOMMARGIN, 100
    END

    IDD_ORPHAN, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 239
        TOPMARGIN, 7
        BOTTOMMARGIN, 200
    END

    IDD_VIEWER_DIALOG, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 138
        TOPMARGIN, 7
        BOTTOMMARGIN, 146
    END
END
#endif    // APSTUDIO_INVOKED


/////////////////////////////////////////////////////////////////////////////
//
// Dialog Info
//

IDD_OBJECTKEYPAD DLGINIT
BEGIN
    IDC_COORDSYS_SELECT, 0x403, 6, 0
0x6f4c, 0x6163, 0x006c, 
    IDC_COORDSYS_SELECT, 0x403, 7, 0
0x6956, 0x7765, 0x7265, "\000" 
    IDC_COORDSYS_SELECT, 0x403, 6, 0
0x6f57, 0x6c72, 0x0064, 
    0
END


/////////////////////////////////////////////////////////////////////////////
//
// String Table
//

STRINGTABLE DISCARDABLE 
BEGIN
    IDP_OLE_INIT_FAILED     "OLE initialization failed.  Make sure that the OLE libraries are the correct version."
    ID_INDICATOR_CURSEG     "No curroom"
    ID_INDICATOR_MARKSEG    "No marked"
    ID_INDICATOR_PLAYERPOS  "X:+0000.00    Y:+0000.00  Z:+0000.00"
    ID_INDICATOR_NET        "Network"
    ID_INDICATOR_BMMEM      "Bitmap memory: 9999999 K"
END

STRINGTABLE PRELOAD DISCARDABLE 
BEGIN
    IDR_MAINFRAME           "Descent III Editor\n\nDescent III level\nD3 level Files(*.d3l)\n.d3l\nD3Levelt\nDescent III level"
END

STRINGTABLE PRELOAD DISCARDABLE 
BEGIN
    AFX_IDS_APP_TITLE       "Descent III Editor"
    AFX_IDS_IDLEMESSAGE     "Ready"
END

STRINGTABLE DISCARDABLE 
BEGIN
    ID_INDICATOR_EXT        "EXT"
    ID_INDICATOR_CAPS       "CAP"
    ID_INDICATOR_NUM        "NUM"
    ID_INDICATOR_SCRL       "SCRL"
    ID_INDICATOR_OVR        "OVR"
    ID_INDICATOR_REC        "REC"
END

STRINGTABLE DISCARDABLE 
BEGIN
    ID_FILE_NEW             "Create a new document\nNew"
    ID_FILE_OPEN            "Open an existing document\nOpen"
    ID_FILE_CLOSE           "Close the active document\nClose"
    ID_FILE_SAVE            "Save the active document\nSave"
    ID_FILE_SAVE_AS         "Save the active document with a new name\nSave As"
    ID_FILE_PAGE_SETUP      "Change the printing options\nPage Setup"
    ID_FILE_PRINT_SETUP     "Change the printer and printing options\nPrint Setup"
    ID_FILE_PRINT           "Print the active document\nPrint"
    ID_FILE_PRINT_PREVIEW   "Display full pages\nPrint Preview"
    ID_FILE_UPDATE          "Update the container to show any changes\nUpdate"
    ID_FILE_SAVE_COPY_AS    "Save a copy of the active document with a new name\nSave Copy"
END

STRINGTABLE DISCARDABLE 
BEGIN
    ID_APP_ABOUT            "Display program information, version number and copyright\nAbout"
    ID_APP_EXIT             "Quit the application; prompts to save documents\nExit"
END

STRINGTABLE DISCARDABLE 
BEGIN
    ID_FILE_MRU_FILE1       "Open this document"
    ID_FILE_MRU_FILE2       "Open this document"
    ID_FILE_MRU_FILE3       "Open this document"
    ID_FILE_MRU_FILE4       "Open this document"
    ID_FILE_MRU_FILE5       "Open this document"
    ID_FILE_MRU_FILE6       "Open this document"
    ID_FILE_MRU_FILE7       "Open this document"
    ID_FILE_MRU_FILE8       "Open this document"
    ID_FILE_MRU_FILE9       "Open this document"
    ID_FILE_MRU_FILE10      "Open this document"
    ID_FILE_MRU_FILE11      "Open this document"
    ID_FILE_MRU_FILE12      "Open this document"
    ID_FILE_MRU_FILE13      "Open this document"
    ID_FILE_MRU_FILE14      "Open this document"
    ID_FILE_MRU_FILE15      "Open this document"
    ID_FILE_MRU_FILE16      "Open this document"
END

STRINGTABLE DISCARDABLE 
BEGIN
    ID_NEXT_PANE            "Switch to the next window pane\nNext Pane"
    ID_PREV_PANE            "Switch back to the previous window pane\nPrevious Pane"
END

STRINGTABLE DISCARDABLE 
BEGIN
    ID_WINDOW_CASCADE       "Views are overlapping windows."
    ID_WINDOW_SPLIT         "Split the active window into panes\nSplit"
END

STRINGTABLE DISCARDABLE 
BEGIN
    ID_EDIT_CLEAR           "Erase the selection\nErase"
    ID_EDIT_CLEAR_ALL       "Erase everything\nErase All"
    ID_EDIT_COPY            "Copy the selection and put it on the Clipboard\nCopy"
    ID_EDIT_CUT             "Cut the selection and put it on the Clipboard\nCut"
    ID_EDIT_FIND            "Find the specified text\nFind"
    ID_EDIT_PASTE           "Insert Clipboard contents\nPaste"
    ID_EDIT_REPEAT          "Repeat the last action\nRepeat"
    ID_EDIT_REPLACE         "Replace specific text with different text\nReplace"
    ID_EDIT_SELECT_ALL      "Select the entire document\nSelect All"
    ID_EDIT_UNDO            "Undo the last action\nUndo"
    ID_EDIT_REDO            "Redo the previously undone action\nRedo"
END

STRINGTABLE DISCARDABLE 
BEGIN
    ID_VIEW_TOOLBAR         "Show or hide the toolbar\nToggle ToolBar"
    ID_VIEW_STATUS_BAR      "Show or hide the status bar\nToggle StatusBar"
END

STRINGTABLE DISCARDABLE 
BEGIN
    AFX_IDS_SCSIZE          "Change the window size"
    AFX_IDS_SCMOVE          "Change the window position"
    AFX_IDS_SCMINIMIZE      "Reduce the window to an icon"
    AFX_IDS_SCMAXIMIZE      "Enlarge the window to full size"
    AFX_IDS_SCNEXTWINDOW    "Switch to the next document window"
    AFX_IDS_SCPREVWINDOW    "Switch to the previous document window"
    AFX_IDS_SCCLOSE         "Close the active window and prompts to save the documents"
END

STRINGTABLE DISCARDABLE 
BEGIN
    AFX_IDS_SCRESTORE       "Restore the window to normal size"
    AFX_IDS_SCTASKLIST      "Activate Task List"
    ID_INDICATOR_ENGINE     "Terrain View"
    ID_INDICATOR_VIEWER     "Viewer 111"
END

STRINGTABLE DISCARDABLE 
BEGIN
    ID_BUTTON_OUTLINE       "Toggle Outline mode in the Texture-mapped view\nToggle Outline"
    ID_VIEW_KEYPAD_TOGGLE   "Show or hide keypad toolbar."
    ID_WINDOW_TILE          "Places views on screen without overlapping."
    ID_FILE_LEAVE_EDITOR    "Go to Descent III menu"
END

STRINGTABLE DISCARDABLE 
BEGIN
    ID_FILE_PLAYIN640X480   "Play level in fullscreen."
END

STRINGTABLE DISCARDABLE 
BEGIN
    ID_BUTTON_WINDOWSELECTION 
                            "Toggle box selection between IN WINDOW and ACROSS EDGE\nToggle box selection"
    ID_EDIT_DELETE          "Delete the selected segments\nDelete"
    ID_BUTTON_OBJMOVE       "Toggle object move mode between Object & Viewer reference\nToggle Object Move Mode "
END

STRINGTABLE DISCARDABLE 
BEGIN
    ID_BUTTON_LIGHTING      "Toggle lighting\nToggle lighting"
END

STRINGTABLE DISCARDABLE 
BEGIN
    ID_OBJBUTTON            "Display object list."
END

STRINGTABLE DISCARDABLE 
BEGIN
    ID_SUBEDITORS_TABLEFILEFILTER "Starts up the Table File Filter program."
END

STRINGTABLE DISCARDABLE 
BEGIN
    IDD_ORPHANHUNTER        "Starts up the Orphan Hunter Dialog"
END

#endif    // English (U.S.) resources
/////////////////////////////////////////////////////////////////////////////



#ifndef APSTUDIO_INVOKED
/////////////////////////////////////////////////////////////////////////////
//
// Generated from the TEXTINCLUDE 3 resource.
//
#define _AFX_NO_SPLITTER_RESOURCES
#define _AFX_NO_OLE_RESOURCES
#define _AFX_NO_TRACKER_RESOURCES
#define _AFX_NO_PROPERTY_RESOURCES

#if !defined(AFX_RESOURCE_DLL) || defined(AFX_TARG_ENU)
#ifdef _WIN32
LANGUAGE 9, 1
#pragma code_page(1252)
#endif
#include "res\editor.rc2"  // non-Microsoft Visual C++ edited resources
#include "afxres.rc"         // Standard components
#include "afxprint.rc"      // printing/print preview resources
#include "afxolesv.rc"      // OLE server resources
#endif
/////////////////////////////////////////////////////////////////////////////
#endif    // not APSTUDIO_INVOKED

